unit ChessPrint;

interface
                                        
uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  quickrpt, Qrctrls, SchachProcs, comctrls, StdCtrls, Richedit, Notation;

const trimChars = [#10, #13, ' '];

type
  TChessText = (ctHeadLine, ctDiagram, ctResult, ctHVar, ct1Var, ct2Var, ctComment);

  TFormatInfo = record
    fia, fil : longint;
    fiFont : TFont;
    fiText : TChessText;
  end;
  PFormatInfo = ^TFormatInfo;

  TQRChessRichText = class(TQRRichText)
  constructor Create(AOwner : TComponent); override;
  destructor Destroy; override;
  private
    { Private-Deklarationen }
    FFen, FPGNGames, sPrint : string;
    FFirstPosition : Boolean;
    re : TRichedit;
    Notation : TNotation;
    lFormatInfo : TList;
    AChar, CharDone : longint;
    b : TBitmap;
    FFontHeadLine, FFontHVar, FFont1Var, FFont2Var, FFontDia, FFontResult, FFontComment : TFont;
    FFontCharFile : String;
    procedure IniEdit;
    procedure ClearFormatInfo;
    procedure DoFormat;
    function  GetLastLine(il : longint) : longint;
    function  GetLastVisibleChar : longint;
    procedure SetFEN(sFEN : string);
    procedure SetFFontHeadLine(f : TFont);
    procedure SetFFontHVar(f : TFont);
    procedure SetFFont1Var(f : TFont);
    procedure SetFFont2Var(f : TFont);
    procedure SetFFontDia(f : TFont);
    procedure SetFFontResult(f : TFont);
    procedure SetFFontComment(f : TFont);
  protected
    { Protected-Deklarationen }
  public
    (*
    If this file is defined, the font for diagrams is read from this file
    (property "FontDia" is replaced)
    Please see the function "readFontFile" in the unit "SchachProcs"
    in which way the characters for the pieces must be written (it's a "iniFile")

    As an alternative, you can code some chess fonts in "schachfont.pas" and
    then assign one of them to "FontDia"
    *)
    property FontCharFile : String read FFontCharFile write FFontCharFile;
    (*
    Must be called in "OnNeedData" of TQuickReport.
    Returns false if no data available
    *)
    function ProcessPGN : Boolean;
  published
    (*
    true = the starting position of each game is printed as diagram
    *)
    property FirstPosition : Boolean read FFirstPosition write FFirstPosition;
    (*
    If Fen is set instead of "PGNGames", only a diagram is printed
    *)
    property FEN : string read FFEN write SetFEN;
    (*
    The games to be printed.
    Diagrams can be inserted as comment ("noDiagrammZeichen" of the unit "SchachProcs.pas")
    *)
    property PGNGames : string read FPGNGames write FPGNGames;
    (*
    With the following properties the printout can be determined
    *)
    property FontHeadLine : TFont read FFontHeadLine write SetFFontHeadLine;
    property FontHVar     : TFont read FFontHVar     write SetFFontHVar;
    property Font1Var     : TFont read FFont1Var     write SetFFont1Var;
    property Font2Var     : TFont read FFont2Var     write SetFFont2Var;
    (*
    The characters of the pieces for this font must be defined in "schachfonts.pas"
    *)
    property FontDia      : TFont read FFontDia      write SetFFontDia;
    property FontResult   : TFont read FFontResult   write SetFFontResult;
    property FontComment  : TFont read FFontComment  write SetFFontComment;
  end;

implementation

constructor TQRChessRichText.Create(AOwner : TComponent);

  procedure SetFontProp(f : TFont; s : string; i : integer; c : TColor; fs : TFontStyles);
  begin
    f.Name := s;
    f.Size := i;
    f.Color := c;
    f.Style := fs;
  end;

begin
  inherited Create(AOwner);
  b := TBitmap.Create;
  lFormatInfo := TList.Create;
  FFontHeadLine   := TFont.Create;
  SetFontProp(FFontHeadLine, 'Arial', 12, clMaroon, [fsBold]);
  FFontHVar       := TFont.Create;
  SetFontProp(FFontHVar, 'Courier New', 10, clMaroon, [fsBold]);
  FFont1Var       := TFont.Create;
  SetFontProp(FFont1Var, 'Courier New', 10, clMaroon, []);
  FFont2Var       := TFont.Create;
  SetFontProp(FFont2Var, 'Courier New', 8, clBlack, []);
  FFontDia        := TFont.Create;
  SetFontProp(FFontDia, '1Echecs', 20, clBlack, []);
  FFontResult     := TFont.Create;
  SetFontProp(FFontResult, 'Arial', 12, clMaroon, [fsBold]);
  FFontComment    := TFont.Create;
  SetFontProp(FFontComment, 'Arial', 10, clMaroon, []);
  if csDesigning in ComponentState then exit;
  re := TRichedit.Create(AOwner);
  re.Visible := false;
  re.Parent := self;
  Notation := TNotation.Create(AOwner);
  Notation.Visible := false;
  Notation.Parent := self;
end;

destructor TQRChessRichText.Destroy;
begin
  ClearFormatInfo;
  lFormatInfo.Free;
  FFontHeadLine.Free;
  FFontHVar.Free;
  FFont1Var.Free;
  FFont2Var.Free;
  FFontDia.Free;
  FFontResult.Free;
  FFontComment.Free;
  b.Free;
  if re <> Nil then re.free;
  if Notation <> Nil then Notation.Free;
  inherited destroy;
end;

procedure TQRChessRichText.ClearFormatInfo;
Var i : integer;
begin
  For i := 0 to lFormatInfo.Count - 1 do Dispose(PFormatInfo(lFormatInfo[i]));
  lFormatInfo.Clear;
  AChar := 0;
  CharDone := 0;
  sPrint := '';
end;

procedure TQRChessRichText.IniEdit;
Var rc : TRect;
begin
  re.Width := Width;
  re.Height := Height;
  re.Font.Assign(Font);
  Notation.Width := 800;
  Notation.Height := 600;
  SendMessage(Handle, em_getrect, 0, longint(@(rc)));
  SendMessage(re.Handle, em_setrect, 0, longint(@(rc)));
  if ParentRichEdit = Nil then ParentRichEdit := re;
end;

procedure TQRChessRichText.SetFEN(sFEN : string);
begin
  if re = Nil then exit;
  if sFEN = FFEN then exit;
  if csDesigning in ComponentState then exit;
  IniEdit;
  FFEN := sFEN;
  re.Font.Assign(FFontDia);
  re.Text := FENtoSFDia(FFEN, re.Font.Name, FFontCharFile);
  re.SelStart := 1;  
end;

function TQRChessRichText.GetLastLine(il : longint) : longint;
Var i : integer;
begin
  Result := lFormatInfo.Count;
  try
    for i := 0 to lFormatInfo.Count - 1 do
    with PFormatInfo(lFormatInfo[i])^ do begin
      if (il >= fia) and (il <= fia + fil) then begin
        Result := i;
        exit;
      end;
    end;
  except
  end;
end;

function TQRChessRichText.GetLastVisibleChar : longint;
var Range: TFormatRange;
    OldMap: Integer;
    dc:hdc;
begin
  Result := length(sPrint);
  try
    DC := GetDC(re.Handle);
    try
      FillChar(Range, SizeOf(TFormatRange), 0);
      with Range do begin
        hdc := dc;
        hdcTarget := hdc;
        SendMessage(re.Handle, em_getrect, 0, longint(@(rc)));
        rc.left := rc.Left * 1440 div Screen.PixelsPerInch;
        rc.top := rc.Top * 1440 div Screen.PixelsPerInch;
        rc.right := rc.Right * 1440 div Screen.PixelsPerInch;
        rc.bottom := rc.Bottom * 1440 div Screen.PixelsPerInch;
        rcPage := rc;
        chrg.cpMax := -1;
        OldMap := SetMapMode(hdc, MM_TEXT);
      end;
      SendMessage(re.Handle, EM_FORMATRANGE, 0, 0);
      Result:=SendMessage(re.Handle, EM_FORMATRANGE, 0, Longint(@Range));
      dec(Result, 2);
      SendMessage(re.Handle, EM_FORMATRANGE, 0, 0);
      SetMapMode(Range.hdc, OldMap);
    finally
      ReleaseDC(re.Handle, DC);
    end;
  except
  end;
end;

function TQRChessRichText.ProcessPGN : Boolean;
Var i, j, k, l, m, LastChar, rLastChar, rChar : integer;
    First : Boolean;
    s : string;

begin
  Result := false;
  try
    if re = Nil then exit;
    rChar := -1;
    if sPrint = '' then begin
      IniEdit;
      DoFormat;
    end;
    inc(AChar);
    if AChar > length(sPrint) then begin
      ClearFormatInfo;
      exit;
    end;                
    re.Clear;
    j := GetLastLine(AChar);
    First := true;
    k := -1;
    m := 0;
    for i := j to lFormatInfo.Count - 1 do
    with PFormatInfo(lFormatInfo[i])^ do begin
      re.SelAttributes.Assign(fiFont);
      if First then s := copy(sPrint, AChar, fil - (AChar - fia))
      else s := copy(sPrint, fia, fil);
      re.SelText := s;
      b.canvas.font.assign(fiFont);
      for l := 1 to length(s) do if s[l] = #13 then inc(m);
      l := b.canvas.TextHeight('Mg');
      if (k = -1) or (k < l) then k := l;
      if m * k > height then break;
      First := false;
    end;
    re.selstart := 1;
    LastChar := GetLastVisibleChar;
    i := GetLastLine(AChar + LastChar);
    rLastChar := LastChar;
    if (i >= 0) and (i < lFormatInfo.Count) then begin
      case PFormatInfo(lFormatInfo[i])^.fiText of
        ctHeadLine : if i - 1 >= 0 then with PFormatInfo(lFormatInfo[i - 1])^ do LastChar := fia + fil;
//        ctResult   :
        ctDiagram  : begin
                       if i - 1 >= 0 then begin
                         with PFormatInfo(lFormatInfo[i - 1])^ do LastChar := fia + fil;
                         if PFormatInfo(lFormatInfo[i - 1])^.fiText = ctHeadline then
                           if i - 2 >= 0 then with PFormatInfo(lFormatInfo[i - 2])^ do LastChar := fia + fil;
                       end;    
                     end;
      end;
    end;
    if rLastChar <> LastChar then begin
      if LastChar <> rChar then AChar := LastChar - 1
      else AChar := length(sPrint);
      dec(LastChar);
    end
    else begin
      if LastChar > 0 then AChar := AChar + LastChar + 1
      else AChar := length(sPrint);
      rChar := -1;
    end;
    re.selstart := AChar - CharDone;
    re.SelLength := length(re.Text);
    re.SelText := '';
    while (AChar > 0) and (AChar < length(sPrint)) and (sPrint[AChar + 1] in trimChars)
      do inc(AChar);
    CharDone := AChar;
    Result := true;
  except
  end;
end;

procedure TQRChessRichText.DoFormat;
Var i, j, rj, HVar  : longint;
    sPart, sNotat : String;
    Done : Boolean;
    Stllg : TSchachbrett;
    sl, slGames : TStringList;
    pd : TPartiedaten;

  procedure InsertReturn;
  begin
    if not ((sPrint <> '') and (sPrint[length(sPrint)] = #10)) then
      sPart := noReturn;
  end;

  procedure DotrimLeft(Var s : string);
  begin
    while (s <> '') and (s[1] = ' ') do delete(s, 1, 1);
  end;

  procedure eInsert(ChessText : TChessText);
  Var fi : PFormatinfo;
  begin
    if sPart = '' then exit;
    if (sPrint <> '') and (sPrint[length(sPrint)] = #10) then
      DotrimLeft(sPart);
    New(fi);
    FillChar(fi^, SizeOf(fi^), 0);
    Case ChessText of
      ctHeadLine : fi^.fiFont := fontHeadLine;
      ctHVar     : fi^.fiFont := ffontHVar;
      ct1Var     : fi^.fiFont := ffont1Var;
      ct2Var     : fi^.fiFont := ffont2Var;
      ctDiagram  : begin
                     fi^.fiFont := ffontDia;
                     if (length(sPrint) > 1) and (sPrint[length(sPrint)] = #10) then begin
                       Delete(sPrint, length(sPrint) - 2, 2);
                       if lFormatInfo.Count <> 0 then
                         dec(PFormatInfo(lFormatInfo[lFormatInfo.Count - 1])^.fil, 2);
                       InsertReturn;  
                     end;

                   end;
      ctResult   : fi^.fiFont := ffontResult;
      ctComment  : fi^.fiFont := ffontComment;
    end;
    fi^.fia := length(sPrint) + 1;
    fi^.fil := length(sPart);
    fi^.fiText := ChessText;
    lFormatInfo.Add(fi);
    sPrint := sPrint + sPart;
    sPart := '';
  end;

  procedure OrtEintragen;
  begin
    with Notation.Partiedaten.EventData do begin
      if (ort <> '') and (ort <> '?') then begin
        sPart := sPart + ort;
        if jahr <> 0 then sPart := sPart + noReturn;
      end;
      if jahr<>0 then sPart := sPart + inttostr(jahr) + noReturn;
    end;
  end;

begin
  sl := TStringList.Create;
  slGames := TStringList.Create;
  try
    sl.Text := FPGNGames;
    i := 0;
    while i < sl.Count do begin
      if Pos(uppercase(PGNPartieStart), trim(uppercase(sl[i]))) = 1 then begin
        j := i + 1;
        sPart := sl[i] + noReturn;
        while j < sl.Count do begin
          if Pos(uppercase(PGNPartieStart), trim(uppercase(sl[j]))) <> 1 then
            sPart := sPart + sl[j] + noReturn
          else break;
          inc(j);
        end;
        slGames.Add(sPart);
        i := j - 1;
      end;
      inc(i);
    end;
    for i := 0 to slGames.Count - 1 do begin
      Notation.SetGame(slGames[i]);
      if FirstPosition then begin
        if Notation.Partiedaten.EventData.FENString <> '' then begin
          sPart := trim(Notation.Partiedaten.Kommentar);
          if (sPart <> '') and (sPart[1] = '{') then
            insert(noDiagrammZeichen, sPart, 2)
          else sPart := '{' + noDiagrammZeichen + '}';
          pd := Notation.Partiedaten;
          pd.Kommentar := sPart;
          Notation.Partiedaten := pd;
          Notation.NotationRefresh;
        end;
      end;
      sNotat := trim(Notation.GetNotation);
      j := length(sNotat);
      while (j > 0) and (not (sNotat[j] in trimChars)) do dec(j);
      delete(sNotat, j + 1, length(sNotat));
      j := 0;
      HVar := 0;
      sPart := '';
      with Notation.Partiedaten.EventData do begin
        sPart := Weiss + ' - ' + Schwarz + noReturn;
        if (Quelle <> '') and (Quelle <> '?') then begin
          sPart := sPart + Quelle;
          if ((ort<>'') and (ort<>'?')) or (jahr <> 0) then begin
            sPart := sPart + ', ';
            OrtEintragen;
          end
          else sPart := sPart + noReturn;
        end
        else OrtEintragen;
        sPart := sPart + noReturn;
      end;
      eInsert(ctHeadLine);
      Done := false;
      while j < length(sNotat) do begin
          inc(j);
          case sNotat[j] of
            '[' : begin
                    eInsert(ctHVar);
                    inc(HVar);
                  end;
            ']' : begin
                    sPart := sPart + sNotat[j];
                    Done := true;
                    eInsert(ct1Var);
                    dec(HVar);
                  end;
            '(' : begin
                    eInsert(ct1Var);
                    inc(HVar);
                  end;
            ')' : begin
                    sPart := sPart + sNotat[j];
                    Done := true;
                    dec(HVar);
                    eInsert(ct2Var);
                  end;
            '{' : begin
                    case HVar of
                      0 : eInsert(ctHVar);
                      1 : eInsert(ct1Var);
                      2 : eInsert(ct2Var);
                    end;
                    rj := j;
                    inc(j);
                    while (sNotat[j] <> '}') and (j <= length(sNotat)) do begin
                      if (sNotat[j] <> '}') and (sNotat[j] <> noDiagrammzeichen)
                        then sPart := sPart + sNotat[j];
                      if sNotat[j] = noDiagrammzeichen then begin
                      eInsert(ctComment);
                      while (rj > 1) and ((sNotat[rj] = '{') or (sNotat[rj] = ' '))  {auf Zug setzen}
                        do dec(rj);
                      Stllg := Notation.GetPositionAt(rj); {Achtung - diese Funktion stellt auch Notation auf diese Stellung ein, wie auf Mausklick}
                      sPart := noReturn + PostoSFDia(stllg, FontDia.Name, FFontCharFile);
                      eInsert(ctDiagram);
                      sPart := noReturn;
                    end;
                    inc(j);
                  end;
                  eInsert(ctComment);
                  Done := true;
                end;
        end;
        if not Done then sPart := sPart + sNotat[j];
        Done := false;
      end;
      eInsert(ctHVar);
      with Notation.Partiedaten.EventData do
      if Ergebnis <> KeinErgebnis then begin
        InsertReturn;
        sPart := sPart + noReturn + Ergebnis;
        eInsert(ctResult);
      end;
      if i < slGames.Count - 1 then begin
        sPart := noReturn + noReturn + noReturn;;
        eInsert(ctResult);
      end;
    end;
  finally
    sl.Free;
    slGames.Free;
  end;
end;

procedure TQRChessRichText.SetFFontHeadLine(f : TFont);
begin
  FFontHeadLine.Assign(f);
end;

procedure TQRChessRichText.SetFFontHVar(f : TFont);
begin
  FFontHVar.Assign(f);
end;

procedure TQRChessRichText.SetFFont1Var(f : TFont);
begin
  FFont1Var.Assign(f);
end;

procedure TQRChessRichText.SetFFont2Var(f : TFont);
begin
  FFont2Var.Assign(f);
end;

procedure TQRChessRichText.SetFFontDia(f : TFont);
begin
  FFontDia.Assign(f);
end;

procedure TQRChessRichText.SetFFontResult(f : TFont);
begin
  FFontResult.Assign(f);
end;

procedure TQRChessRichText.SetFFontComment(f : TFont);
begin
  FFontComment.Assign(f);
end;


end.
