unit AThread;

(*
The process for the Winboard engine.
*)

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls, StdCtrls, Buttons, {SyncObjs,}  SchachProcs,  Menus, Main;

type

  TAnalyseThreadWB = class(TThread)
    constructor Create(f : TForm; aWBInfo : TWBInfo);
  private
    Ausgabetext : string;
    FinishLoop: Boolean;
    wbStop : Boolean;
    AParent : TForm;
    Closing : Boolean;
    Finished : Boolean;
    procedure WBANALYSEERGEBNIS (ns : PChar);
    function  WBGetMoveNumber : integer;
  protected
    procedure Execute; override;
  public
    { Public-Deklarationen }
    WBInfo : TWBInfo;
    procedure DoWriteOut;
    procedure WriteOut(ns : PChar);
    procedure RepeatWBEngine;
    procedure StopWBEngine;
    procedure TerminateWBEngine;
    function  WritePipe(s : string; h : THandle) : Boolean;
  end;

implementation

uses BrettWin;

function TAnalyseThreadWB.WritePipe(s : string; h : THandle) : Boolean;
Var writebuf : array[0..BufSize] of Char;
    BytesWrote: Dword;
    i : integer;
begin
  FillChar(writeBuf, SizeOf(writeBuf), 0);
  For i := 1 to length(s) do writebuf[i - 1] := s[i];
  Result := WriteFile(h, writebuf, length(s), BytesWrote, Nil);
end;

function TAnalyseThreadWB.WBGetMoveNumber : integer;
Var s : string;
begin
 Result := 1;
 try
    s := (AParent as TMDIBrett).Notation1.GetNextMoveNumberAsString;
    while Pos('.', s) <> 0 do delete(s, Pos('.', s), 1);
    Result := StrToInt(s);
  except
  end;
end;


constructor TAnalyseThreadWB.Create(f : TForm; aWBInfo : TWBInfo);
begin
  AParent := f as TMDIBrett;
  WBInfo := aWBInfo;
  WriteOut('Starting ...');
  inherited Create(false);
end;

procedure TAnalyseThreadWB.Execute;
Var wsc, wlc, bsc, blc : LongBool;
    epFeld, amZug, ZugNummer : Longint;
    i, j : integer;
    s, s1 : string;
    zug : TZug;
    stllg : TSchachbrett;
    ReadBuf: array[0..BufSize] of Char;
    BytesRead: Dword;

begin
  try
    if not WritePipe(WBInfo.WbIni, WBInfo.hTo)
      then exit;
    WritePipe('level 10 10000 0' + ret, WBInfo.hto);
    repeat
      FinishLoop := true;
      i := WBGetMoveNumber;
      stllg := (AParent as TMDIBrett).Brett1.GetBrStellung;
      zug :=(AParent as TMDIBrett).Brett1.GetBrZug;
      if uppercase(WBInfo.WbSetBoard) = 'SETBOARD' then begin
          s := '';
          s := PosToFen(Stllg, Zug, i, Nil);

          WritePipe(WBInfo.WbSetBoard + ' ' + s + ret +
            WBInfo.WbStartAnalyze, WBInfo.hto);
        end
        else begin  // alte edit modus
          WritePipe('new' + ret + 'force' + ret, WBInfo.hto);
          if Zug.amZug = SCHWARZ then
            WritePipe('a2a3' + ret, WBInfo.hto);
          WritePipe(WBInfo.WbSetBoard+ ret + '#' + ret, WBInfo.hto); // editmodus und clear
          for i := 0 to 63 do begin
            case Stllg[c(i)] of
              wk : WritePipe('K' + Feldnamen[i] + ret, WBInfo.hto);
              wd : WritePipe('Q' + Feldnamen[i] + ret, WBInfo.hto);
              wt : WritePipe('R' + Feldnamen[i] + ret, WBInfo.hto);
              wl : WritePipe('B' + Feldnamen[i] + ret, WBInfo.hto);
              ws : WritePipe('N' + Feldnamen[i] + ret, WBInfo.hto);
              wb : WritePipe('P' + Feldnamen[i] + ret, WBInfo.hto);
            end
          end;
          WritePipe('c' + ret, WBInfo.hto);  // Farbe wechseln
          for i := 0 to 63 do begin
            case Stllg[c(i)] of
              sk : WritePipe('K' + Feldnamen[i] + ret, WBInfo.hto);
              sd : WritePipe('Q' + Feldnamen[i] + ret, WBInfo.hto);
              st : WritePipe('R' + Feldnamen[i] + ret, WBInfo.hto);
              sl : WritePipe('B' + Feldnamen[i] + ret, WBInfo.hto);
              ss : WritePipe('N' + Feldnamen[i] + ret, WBInfo.hto);
              sb : WritePipe('P' + Feldnamen[i] + ret, WBInfo.hto);
            end
          end;

          WritePipe('.' + ret + WBInfo.WbStartAnalyze, WBInfo.hto);
        end;
        wbStop := false;
        repeat
          FillChar(ReadBuf, SizeOf(ReadBuf), 0);
          ReadFile(WBInfo.hfrom, ReadBuf, BufSize, BytesRead, nil);
          WBAnalyseErgebnis(ReadBuf);
        until WbStop;
      until FinishLoop;
  except
  end;
  Finished := true;
end;

procedure TAnalyseThreadWB.StopWBEngine;
begin
  if Suspended then Resume;
  wbStop := true;
  WritePipe(WBInfo.WbGetState, WBInfo.hto);
  WritePipe(WBInfo.WbStopAnalyze, WBInfo.hto);
end;

procedure TAnalyseThreadWB.RepeatWBEngine;
begin
  FinishLoop := false;
  WriteOut('- Initialize ...');
  StopWBEngine;
end;


procedure TAnalyseThreadWB.WriteOut(ns : PChar);
begin
  if Closing then exit;
  Ausgabetext := StrPas(ns);
  Synchronize(DoWriteOut);
end;

procedure TAnalyseThreadWB.DoWriteOut;
begin
  (AParent as TMDIBrett).Memo1.Text := Ausgabetext;
end;

procedure TAnalyseThreadWB.TerminateWBEngine;
var closeTime : TDateTime;
begin
  if Closing then exit;
  try
    Closing := true;
    WriteOut('Finishing ...');
    FinishLoop := true;
    StopWBEngine;
    WritePipe(WBInfo.WbQuit, WBInfo.hto);
    closeTime := Time;
    while (not Finished) and
    (Time - closeTime < 3{Sec} / 24 / 3600) do begin
    end;
    if not Finished then begin
      TerminateProcess(WBInfo.hProcess, 0);
      TerminateThread(Handle, 0);
    end
    else begin
      CloseHandle(WBInfo.hTo);
      CloseHandle(WBInfo.hFrom);
      CloseHandle(WBInfo.hProcess);
    end;
    DeleteFile(WBInfo.CraftyFile);
  except
  end;
end;

procedure TAnalyseThreadWB.WBANALYSEERGEBNIS (ns : PChar);
Var s, hs1, hs2, wert : string;
    i, j, k, nr : integer;
begin
  s := string(ns);
  if WBInfo.WBNoParseOutput then begin
    WriteOut(PChar(s));
    exit;
  end;

  i := Pos(#9, s);
  if i <> 0 then
    Delete(s, 1, i);
  s := trim(s);
  i := Pos(#13, s);
  if i <> 0 then Delete(s, i, length(s));

  i := Pos('.', s);
  if i = 0 then begin
    i := 1;
    while (i < length(s)) and (not (s[i] in ['A'..'Z', 'a'..'z']))
      do inc(i);
    if i >= length(s) then i := 0
    else dec(i);
  end;
  if i <> 0 then begin
    while (i > 0) and (s[i] <> ' ') do dec(i);
    hs1 := Copy(s, 1, i);
    delete(s, 1, i);
    i := 1;  // Tiefe (1. Zahl) und Wert (3. Zahl) lesen
    while (i <= length(hs1)) and (not (hs1[i] in ['0'..'9'])) do inc(i);
    hs2 := 'Depth: ';
    while (i <= length(hs1)) and (hs1[i] <> ' ') do begin
      hs2 := hs2 + hs1[i];
      inc(i);
    end;
    hs2 := hs2 + '  ' + 'Value: ';
    while (i <= length(hs1)) and (hs1[i] = ' ') do inc(i);
    Wert := '';
    while (i <= length(hs1)) and (hs1[i] <> ' ') do begin
      Wert := Wert + hs1[i];
      inc(i);
    end;
    try
      i := StrToInt(wert);
    except
      Wert := '';
    end;
    if wert <> '' then begin
      Str(i/100:3:2, wert);
    end;
    hs2 := hs2 + wert;
    i := 1;   // ersten Zug lesen
    hs1 := '';
    while (i <= length(s)) and (not (s[i] in ['A'..'z'])) do begin
      hs1 := hs1 + s[i];
      inc(i);
    end;
    while (i <= length(s)) and (not (s[i] in [' ', '!', '?'])) do begin
      hs1 := hs1 + s[i];
      inc(i);
    end;
    nr := WBGetMoveNumber;
    i := length(s);
    while i > 1 do begin
      if (s[i] = '.') and
      ((i > 1) and (s[i - 1] in ['0'..'9'])) then
      begin
        j := i;
        dec(i);
        while (i > 0) and (s[i] in ['0'..'9']) do dec(i);
        inc(i);
        hs1 := copy(s, i, j - i);
        k := 1;
        try
          k := StrToInt(hs1);
        except
          break;
        end;
        delete(s, i, j - i);
        hs1 := IntToStr(k + nr -1);
        insert(hs1, s, i);
      end;
      dec(i);
    end;

    s := hs2 + Ret + s;
  end;
  WriteOut(PChar(s));
end;

end.
