unit ExploreWin;

interface

uses Windows, Classes, Graphics, Forms, Controls,
  FileCtrl, StdCtrls, ExtCtrls, Buttons, Spin, ComCtrls, Dialogs;

type
  TMDIDatabase = class(TForm)
    pLeft: TPanel;
    DirectoryListBox1: TDirectoryListBox;
    FileEdit: TEdit;
    FileListBox1: TFileListBox;
    FilterComboBox1: TFilterComboBox;
    DriveComboBox1: TDriveComboBox;
    Panel1: TPanel;
    Label1: TLabel;
    lbDatabase: TListBox;
    bPGN: TButton;
    bCCF: TButton;
    procedure FileListBox1Click(Sender: TObject);
    procedure FileEditKeyPress(Sender: TObject; var Key: Char);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
    procedure FormCreate(Sender: TObject);
    procedure FormKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure FormShow(Sender: TObject);
    procedure lbDatabaseDblClick(Sender: TObject);
    procedure lbDatabaseDragDrop(Sender, Source: TObject; X, Y: Integer);
    procedure lbDatabaseDragOver(Sender, Source: TObject; X, Y: Integer;
      State: TDragState; var Accept: Boolean);
    procedure lbDatabaseKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure FileListBox1Change(Sender: TObject);
    procedure FileListBox1DblClick(Sender: TObject);
    procedure bPGNClick(Sender: TObject);
    procedure bCCFClick(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
  private
    procedure CreateNewIndexFile (Sender : TObject);
    procedure createNewDatabase(ext : String);
  end;

var
  MDIDatabase: TMDIDatabase;

implementation

uses SysUtils, Main, local, localconst, Chessdatabase, PGN,
     DialogsLng, SchachProcs;

{$R *.dfm}

procedure TMDIDatabase.CreateNewIndexFile (Sender : TObject);
begin     //ccf
  MessageDlg(LOGetMess(LOCreateNewIndexfile), mtError, [mbOk], 0);
end;

procedure TMDIDatabase.FileListBox1Click(Sender: TObject);
var
  FileExt: string;
begin
  FileExt := AnsiUpperCase(ExtractFileExt(FileListBox1.Filename));
  if (FileExt = '.BMP') or (FileExt = '.ICO') or (FileExt = '.WMF') or
    (FileExt = '.EMF') then
  begin
  end;
end;

procedure TMDIDatabase.FileEditKeyPress(Sender: TObject; var Key: Char);
begin
  if Key = #13 then 
  begin
    FileListBox1.ApplyFilePath(FileEdit.Text);
    Key := #0;
  end;
end;

procedure TMDIDatabase.FormClose(Sender: TObject;
  var Action: TCloseAction);
begin
  Action := caFree;
end;

procedure TMDIDatabase.FormCloseQuery(Sender: TObject;
  var CanClose: Boolean);
begin
  if MainForm.Working then CanClose := false
  else begin
    if WindowState = wsNormal then begin
      MainForm.DatabaseHeight := Height;
      MainForm.DatabaseWidth := Width;
    end;
  end;
  Mainform.DatabaseFolder := DirectoryListBox1.Directory;
end;

procedure TMDIDatabase.FormCreate(Sender: TObject);
begin
  Top := 0;
  Left := 0;
  if Mainform.DatabaseHeight > 200 then Height := Mainform.DatabaseHeight;
  if Mainform.DatabaseWidth > 200 then  Width :=  Mainform.DatabaseWidth;
  DirectoryListBox1.Directory := Mainform.DatabaseFolder;
  LOSaveDataBaseWinStrings(self);
end;

procedure TMDIDatabase.FormKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  if key = 27 then Close;  
end;

procedure TMDIDatabase.FormShow(Sender: TObject);
begin
  if LO_lng <> LO_German then LOLoadDataBaseWinStrings(LO_lng, self);
end;

procedure TMDIDatabase.lbDatabaseDblClick(Sender: TObject);
begin
  MainForm.CreateMDIPGNChild(ListboxNameToDatabaseName(
    lbDatabase.Items[lbDatabase.ItemIndex]));
end;

procedure TMDIDatabase.lbDatabaseDragDrop(Sender, Source: TObject; X,
  Y: Integer);
Var p : TPoint;
    Item, ID : longint;
    i : longint;
    s, FileName, sGames : string;
    ccfFile : TChessDatabase;
    idummy : Int64;
    pgn : TPGN;
begin
Mainform.taggleAnalysis(true);
try
  if (Sender is TListbox) and (Source is TPGN) then begin
    pgn := Source as TPGN;
    p.x := X;
    p.y := Y;
    Item := lbDatabase.ItemAtPos(p, true);
    if Item = -1 then exit;
    FileName := ListboxNameToDatabaseName(lbDatabase.Items[Item]);
    if FileName = pgn.PGNFile then exit;
    if pgn.SelCount = 0 then exit;
    for I := MainForm.MDIChildCount - 1 downto 0 do
    if MainForm.MDIChildren[I].Caption = FileName then begin
      MessageDlgLng(LOGetMess(LO_DieDatenbank) + FileName + LOGetMess(LO_istbereits),
      mtError, [mbOk], 0);
      exit;
    end;
    if MessageDlgLng(LOGetMess(LO_MarkiertePartien) +
    FileName + LOGetMess(LO_angehaengt), mtConfirmation, [mbYes, mbNo], 0) <> mrYes
      then exit;
    ccfFile := TChessDatabase.Create(self); //ccf
    if ccfFile.isChessDatabase(Filename) then begin
      ccfFile.OnCreateNewIndexFile := CreateNewIndexFile;
      // must be called!
      ccfFile.CheckAndLoadFile(Filename);
      if ccfFile.Filename = '' then begin
        FreeAndNil(ccfFile);
        MessageDlgLng(LOGetMess(LO_Speicherfehler), mtError, [mbOk], 0);
        exit;
      end;
    end
    else
      FreeAndNil(ccfFile);
    if (ccfFile = Nil) and (not checkPGN(Filename)) then begin
      MessageDlgLng(LOGetMess(LO_Speicherfehler), mtError, [mbOk], 0);
      exit;
    end;
    Screen.Cursor := crHourGlass;
    enabled := false;
    getParentForm(pgn).Enabled := false;
    if ccfFile <> Nil then begin
      if pgn.isccfFile then
        pgn.ccfRaw := true
      else
        ccfFile.Notation := pgn.ccfNotation;  
    end;
    if ccfFile <> Nil then
      ccfFile.FilesOpen := true;
    pgn.ccfFilesOpen := true;
    try
      MainForm.Abbrechen := false;
      i := 0;
      sGames := '';
      while i < pgn.Items.Count do begin
        if pgn.Selected[i] then begin
          s := pgn.GetGameAsString(pgn.ListeIndexToDatabaseIndex(i), ID);
          if s <> '' then begin
            if ccfFile <> Nil then begin
              if pgn.ccfRaw then
                ccfFile.addGameByIndex(idummy, s)
              else
                ccfFile.addGameAsPGNByIndex(idummy, s);
            end
            else
              sGames := sGames + #13#10#13#10 + s;
            Application.ProcessMessages;
            if MainForm.Abbrechen then Break;
          end;
        end;
        inc(i);
      end;
      if ccfFile = Nil then
        if not AppendGameToPGN(FileName, sGames) {a function to append games onto databases which were not opened by TPGN before}
          then MessageDlgLng(LOGetMess(LO_Speicherfehler), mtError, [mbOk], 0);
    finally
      pgn.ccfRaw := false;
      if ccfFile <> Nil then
        ccfFile.FilesOpen := false;
      pgn.ccfFilesOpen := false;
      getParentForm(pgn).Enabled := true;
      enabled := true;
      if ccfFile <> Nil then
        ccfFile.Free;
      Screen.Cursor := crDefault;
    end;
  end;
finally
  Mainform.taggleAnalysis(false);
end;    
end;


procedure TMDIDatabase.lbDatabaseDragOver(Sender, Source: TObject; X,
  Y: Integer; State: TDragState; var Accept: Boolean);
Var p : TPoint;
begin
  p.x := X;
  p.y := Y;
  lbDatabase.Itemindex := lbDatabase.ItemAtPos(p, true);
  Accept := (Source is TPGN) and (lbDatabase.Itemindex <> -1);
end;

procedure TMDIDatabase.lbDatabaseKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  if key = 13 then
    if lbDatabase.Itemindex <> -1 then lbDatabaseDblClick(self);
end;

procedure TMDIDatabase.FileListBox1Change(Sender: TObject);
var FileExt: string;
    i : Integer;
begin
  i := FileListBox1.Items.Count - 1;
  while i > -1 do begin
    FileExt := AnsiUpperCase(ExtractFileExt(FileListBox1.Items[i]));
    if (FileExt = '.CCFIDX') or (FileExt = '.CCFTMP') then
      FileListBox1.Items.Delete(i);
    dec(i);
  end;
end;

procedure TMDIDatabase.FileListBox1DblClick(Sender: TObject);
begin
  if FileListBox1.ItemIndex = -1 then exit;
  MainForm.CreateMDIPGNChild(FileListBox1.FileName);
end;

procedure TMDIDatabase.createNewDatabase(ext : String);
var s, s1 : String;
begin
  s := '';
  if InputQuery(LOGetMess(LODateinameCaption), LOGetMess(LODateinameHint), s) then
  begin
    s := IncludeTrailingBackSlash(DirectoryListBox1.Directory) + s;
    s1 := uppercase(ExtractFileExt(s));
    if (s1 <> '.PGN') and (s1 <> '.CCF')
      then s := s + ext;
    Mainform.CreateMDIPGNChild(s);
    FilterComboBox1.Filter := '*.pgn; *.ccf|*.pgn|*.*|*.*';
    FilterComboBox1.Filter := '*.pgn; *.ccf|*.pgn;*.ccf|*.*|*.*';
  end;
end;

procedure TMDIDatabase.bPGNClick(Sender: TObject);
begin
  createNewDatabase('.pgn');
end;

procedure TMDIDatabase.bCCFClick(Sender: TObject);
var s : String;
begin
  createNewDatabase('.ccf');
end;

procedure TMDIDatabase.FormDestroy(Sender: TObject);
begin
  MDIDatabase := Nil;
end;

end.
