unit PGN;

(*
Please note that this TPgn is also used as interface for files created by TChessDatabase.
If the file extension is ".pgn", the original methods are used,
otherwise it uses the methods starting with "ccf..." which use the functions of ChessDatabase.
*)

{$B-}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  stdctrls, SchachProcs, ChessDatabase, uCustomDatabase, Notation;

const

  pgnAddGame      = -1;  {invalid ID for SaveGame - then the game will be appended}

type
  TCancelMode = (cmCancelDenied, cmNoCancel, cmCancel);  {for the trace event in longer processes}

  TGameHeader = record
     ID, FilePositionA, FilePositionE : longint;
   end;
   PGameHeader = ^TGameHeader;

  TKriterium = (soNothing, soWeiss, soSchwarz, soquelle, soAnnotator, soort, soergebnis, soEco,
                soJahr, soWElo, soSElo, soPlyCount);

  TBeforeCreatePGNFileEvent = procedure (Sender: TObject; FileName : string; Var CreateNewPGNFile : Boolean) of object;
  TPGNTraceEvent = procedure (Sender: TObject; Percent : integer; Var CancelMode : TCancelMode) of object;
  TPGNDblClickEvent = procedure (Sender: TObject; PGNGame : string; GameID : longint) of object;
  TFilterRequestEvent = procedure (Sender: TObject; GameHeader : TEventData; Var ShowThisGame : Boolean) of object;
  TPGNDrawItemEvent = procedure (Control : TWinControl; Index: Integer; Rect: TRect; GameHeader : TEventData; GameIndex : longint) of object;
  TPi = ^longint;

  TPGN = class(TCustomListbox)
    constructor Create(AOwner : TComponent); override;
    destructor destroy; override;
  private
    { Private-Deklarationen }
    FPGNFile          : String;
    FOnPGNTrace       : TPGNTraceEvent;
    FOnFilterRequest  : TFilterRequestEvent;
    FOnPGNDblClick    : TPGNDblClickEvent;
    FOnPGNDrawItem    : TPGNDrawItemEvent;
    FBeforeCreatePGNFile: TBeforeCreatePGNFileEvent;
    FPGNSorted        : TKriterium;
    FLockRefresh      : Boolean;

    FOnlyPartsRead    : Boolean;
    filePGN    : File;
    LastID     : longint;     
  //  Abbruch    : TCancelMode;
    slTemp     : TStringList;
    lHeader    : TList;
    FileOpen, FileReading, Deleting  : Boolean;
    EvD        : TEventData;
    BackupFile : String;

    ccfpreparing : Boolean;
    ccfFile : TChessDatabase;
    ccfFilterList : TStringlist;
    FccfFilterReset : Boolean;
    ccfDatabaseCheckStep : Integer;
    FccfDatabaseCheckSteps : Integer;  // number of integrity checks of records while loading a file
    FccfNotation : TNotation;
    FonlyPGN : Boolean;
    FccfRaw : Boolean;  // if true, the games strings are not in pgn

    procedure SetPGNFile(s : string);
    procedure SetLockRefresh(b : Boolean);

    function  AddGame(Var ID : longint; sGame : String) : Boolean;
    procedure ClosePGNFile;
    procedure AddIndexObject(lHeaderIndex, GameID, PosA, posE : longint);
    procedure DeleteIndexObject(lHeaderindex : longint);
    function  AdjustGameString(lHeaderIndex : longint; sGame : string; Modus : integer) : string;
    function  GetFilePosIndex(ID : longint; Var FilePosindexA : longint) : longint;
    procedure UpdateFileIndexe(lHeaderIndex, summand : longint);
    procedure UpdateItems(ListeIndex, Summand : longint);
    function  DoGetGameAsString(lHeaderIndex : longint; var recordIsDeleted : Boolean) : string;
    procedure DoslIndexeClear;
    procedure RefreshEventData(lHeaderIndex : longint);
    function  DoSaveGame(ListeIndex : longint; Var ID : longint; sGame : String; Modus : integer) : Boolean;
    procedure AddPGNItem(ListeIndex, lHeaderIndex : longint);
    procedure DeleteItem(ListeIndex : longint);
    function  GetItemsVisibleCount : longint;
    function  GetItemsAllCount : longint;

    function  GetItemsObject(Listeindex : longint) : longint;
    function  DoDeleteGame(AIndex : longint; IsListeIndex : Boolean) : Boolean;
    function  GetListeIndexByID(AID : longint) : longint;
    function  Backup(idx : longint; von : Integer) : Boolean;

    function createChessDatabase(Filename : String) : Boolean;
    function  DataFind(Control: TWinControl; FindString: String): Integer;
    procedure Data(Control: TWinControl; Index: Integer; var Data: String);
    procedure CreateNewIndexFile (Sender : TObject);
    procedure DatabaseCompress (Sender : TObject; var Cancel : Boolean; percent : Integer);
    procedure DatabaseCheck (Sender : TObject; var checkstep : int64; var cancel : Boolean);
    procedure DatabaseError (Sender : TObject);
    procedure ccfIncCount(i : Int64);
    procedure setccfFilesOpen(b : Boolean);
  protected
    procedure DrawItem(Index: Integer; Rect: TRect; State: TOwnerDrawState);  override;
    procedure DblClick; override;
    procedure KeyPress(var Key: Char); override;

    function  ccfListeindex2Databaseindex(idx : longint) : longint;
  public
    (*
    Is set after "sortieren(criteria)".
    Then the list ist sorted by the determined criteria
    *)
    property  PGNSorted : TKriterium read FPGNSorted;
    (*
    Items which are displayed in the list
    *)
    property  ItemsVisibleCount : longint read GetItemsVisibleCount;
    (*
    all items = games in the pgn file
    *)
    property  ItemsAllCount : longint read GetItemsAllCount;
    (*
    true = loading was interrupted
    *)
    property  OnlyPartsRead : Boolean read FOnlyPartsRead;
    (*
    pgn file to load (empty string --> file will be closed)
    IMPORTANT: If the extension is NOT ".pgn",
    the component will use the functions of ChessDatabase,
    that means, it must be a file created with TChessDatabase.
    *)
    property  PGNFile : string read FPGNFile write SetPGNFile;
    (*
    true = the list is not repainted
    *)
    property  LockRefresh : Boolean read FLockRefresh write SetLockRefresh;

    { BEGIN: For the integration of TChessDatabase into the TPGN interface }

    (*
    If you want to reset a filtered list, you must repeatedly set
    "ccfFilterReset" to true and then use the method "Filtern".
    *)
    property  ccfFilterReset : Boolean write FccfFilterReset;
    (*
    Do not use it. It is set due to the file size.
    *)
    property  ccfDatabaseCheckSteps : Integer read FccfDatabaseCheckSteps write FccfDatabaseCheckSteps;
    (*
    If you want to compress a pgn string, you must set it,
    otherwise the game is saved as a pgn string into a TChessDatabase.
    *)
    property  ccfNotation : TNotation read FccfNotation write FccfNotation ;  // needed to compress the games (getRawGame)
    (*
    true = ALL files will be handled as pgn, even if they don't have the extension ".pgn"
    *)
    property  onlyPGN : Boolean read FonlyPGN write FonlyPGN;
    (*
    true = the file functions use those functions of TChessDatabase
    which don't await pgn as parameter.
    Use it if compressed game strings (instead of pgn strings) shall be copied from one database to another.
    *)
    property  ccfRaw : Boolean read FccfRaw write FccfRaw;
    (*
    Opens the files of TChessDatabase (see the comments there).
    Use it if many copy actions shall be processed.
    Finally you MUST set it to false!
    *)
    property  ccfFilesOpen : Boolean write setccfFilesOpen;

    { END: For the integration of TChessDatabase into the TPGN interface }

    (*
    Saves the game to a physical position ("ListeIndex")
    Remember the "ID" if you want to replace that game later
    *)
    function  SaveGame(ListeIndex : longint; Var ID : longint; sGame : String) : Boolean;
    (*
    Replaces the game (remember the id after "GetGameAsString" or "SaveGame")
    *)
    function  ReplaceGame(Var ID : longint; sGame : String) : Boolean;
    (*
    Deletes the specified game
    (the index of list can be changed after games were deleted previously!)
    *)
    function  DeleteGame(ListeIndex : longint) : Boolean;
    (*
    Deletes the game by ID
    *)
    function  DeleteGameByID(AID : longint) : Boolean;
    (*
    Returns the pgn string ("lHeaderIndex" is the database index from 0 to x)
    *)
    function  GetGameAsString(lHeaderIndex : longint; Var ID : longint{; var recordIsDeleted : Boolean}) : string;
    (*
    The list can be sorted and filtered -
    converts the index of list to the database index
    *)
    function  ListeIndexToDatabaseIndex(ListeIndex : longint) : longint;
    (*
    This function fires "OnFilterRequest" where you can determine
    whether the game shall be displayed
    *)
    procedure Filtern;
    (*
    Sorts the list
    *)
    procedure Sortieren(Kriterium : TKriterium);


    { BEGIN: For the integration of TChessDatabase into the TPGN interface }

    (*
    true = the loaded file is created with TChessDatabase (not pgn)
    *)
    function  isCCFFile : Boolean;
    (*
    You may use it after some games were marked as deleted.
    Unlike in pgn databases, the games are not irrecoverably removed after "DeleteGame".
    *)
    function ccfCompress : Boolean;
    (*
    Deletes the entire database (TChessDatabases consist of three files)
    *)
    function ccfDeleteDatabase : Boolean;

    { END: For the integration of TChessDatabase into the TPGN interface }

  published
    (*
    progress of loading
    *)
    property OnPGNTrace : TPGNTraceEvent read FOnPGNTrace write FOnPGNTrace;
    (*
    fired after double click or enter on the list
    *)
    property OnPGNDblClick : TPGNDblClickEvent read FOnPGNDblClick write FOnPGNDblClick;
    (*
    MUST be handled for displaying the games
    *)
    property OnPGNDrawItem : TPGNDrawItemEvent read FOnPGNDrawItem write FOnPGNDrawItem;
    (*
    you can determine whether the game shall be displayed
    *)
    property OnFilterRequest : TFilterRequestEvent read FOnFilterRequest write FOnFilterRequest;
    (*
    a specified file doesn't yet exists
    *)
    property BeforeCreatePGNFile : TBeforeCreatePGNFileEvent read FBeforeCreatePGNFile write FBeforeCreatePGNFile;

    property Align;
    property BorderStyle;
    property Color;
    property Columns;
    property Ctl3D;
    property DragCursor;
    property DragMode;
    property Enabled;
    property ExtendedSelect;
    property Font;
    property IntegralHeight;
    property ItemHeight;
    property MultiSelect;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property TabOrder;
    property Visible;

    property OnClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMeasureItem;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnStartDrag;
  end;

implementation

const

  BufferSize      = 16384;
  pgnRetString    = #$D#$A;
  pgnAdd          = 1;
  pgnInsert       = 2;
  pgnReplace      = 3;

procedure TPGN.setccfFilesOpen(b : Boolean);
begin
  if ccfFile = Nil then exit;
  ccfFile.FilesOpen := b;
end;

procedure TPGN.SetLockRefresh(b : Boolean);
begin
  Deleting := b;
  FLockRefresh := b;
  if not b then refresh;
end;

procedure TPGN.Sortieren(Kriterium : TKriterium);
Var i, j : longint;
    sl : TStringList;
    s : string;
    prcnt, Prcnt100, NewPrcnt, LastPrcnt : Integer;
    Abbruch : TCancelMode;

    ie, iGes : longint;

function DoTrace : Boolean;
begin
  Result := false;
  inc(prcnt);
  NewPrcnt := GetPercent(prcnt, prcnt100);
  if NewPrcnt <> LastPrcnt then begin
    if assigned(FOnPGNTrace) then FOnPGNTrace(self, NewPrcnt, Abbruch);
    if Abbruch = cmCancel then Result := true;
    LastPrcnt := NewPrcnt;
  end;
end;

procedure QuickSort(iLo, iHi: Integer);
  var
    Lo, Hi : Integer;
    Mid, T : string;
  begin
    Lo := iLo;
    Hi := iHi;
    Mid := sl[(Lo + Hi) div 2];
    repeat
      while sl[Lo] < Mid do Inc(Lo);
      while sl[Hi] > Mid do Dec(Hi);
      if Lo <= Hi then
      begin
        T := sl[Lo];
        sl[Lo] := sl[Hi];
        sl[Hi] := T;

        if ccfFile <> Nil then begin
          T := ccfFilterList[Lo];
          ccfFilterList[Lo] := ccfFilterList[Hi];
          ccfFilterList[Hi] := T;
        end
        else begin

        T := Items[Lo];
        Items[Lo] := Items[Hi];
        Items[Hi] := T;

        end;

        Inc(Lo);
        Dec(Hi);
        if Abbruch = cmCancel then exit;
      end;
    until Lo > Hi;
    DoTrace;
    if Hi > iLo then QuickSort(iLo, Hi);
    if Lo < iHi then QuickSort(Lo, iHi);
  end;

begin
  if (FileReading) or (not FileOpen) or (Items.Count = 0) then exit;
  LastPrcnt := -1;
  Prcnt := 0;
  FileReading := true;
  Deleting := true;  {prevents the game refresh (drawItem), just as it's done while deleting games}
  sl := TStringList.Create;
  Screen.Cursor := crHourGlass;
  Items.BeginUpdate;
  try
    if Kriterium <> FPGNSorted then begin
      Prcnt100 := Items.Count * 2;
      FPGNSorted := Kriterium;
      Abbruch := cmNoCancel;

      if ccfFile <> Nil then begin
        iGes := ccfFile.CountWithDeletedRecords;
        if Kriterium = soNothing then begin
          FreeAndNil(ccfFilterList);
          count := iGes;
          exit;
        end;
        if ccfFilterList = Nil then begin
          ccfFilterList := TStringlist.Create;
          for i := 0 to iGes - 1 do
            ccfFilterList.Add(IntToStr(i));
        end;
        ie := ccfFilterList.count - 1;
      end
      else
        ie := Items.Count - 1;

      if ccfFile <> Nil then
        ccfFile.FilesOpen := true;
      try
        for i := 0 to ie do begin
          if Kriterium = soNothing then begin
            Items[i] := '';  // wenn ccfFile Nil, wurde schon oben rausgespr.
          end
          else begin
            if ccfFile = Nil then
              Items[i] := IntToStr(i);
            j := GetItemsObject(I);
            RefreshEventData(j);
            case Kriterium of
            soWeiss    : sl.Add(EvD.Weiss);
            soSchwarz  : sl.Add(EvD.Schwarz);
            soquelle   : sl.Add(EvD.quelle);
            soAnnotator: sl.Add(EvD.Annotator);
            soOrt      : sl.Add(EvD.Ort);
            soergebnis : sl.Add(EvD.ergebnis);
            soEco      : sl.Add(EvD.Eco);
            soJahr     : sl.Add(IntToStr(EvD.Jahr));
            soWElo     : sl.Add(IntToStr(EvD.WElo));
            soSElo     : sl.Add(IntToStr(EvD.SElo));
            soPlyCount : sl.Add(IntToStr(EvD.PlyCount));
            end;
            if DoTrace
              then exit;
          end;
        end;
      finally
        if ccfFile <> Nil then
          ccfFile.FilesOpen := false;
      end;
      Abbruch := cmNoCancel;
      if Kriterium <> soNothing then QuickSort(0, sl.Count - 1);
    end
    else if Kriterium <> soNothing then begin
      Prcnt100 := Items.Count;
      Abbruch := cmNoCancel;
      
      if ccfFile <> Nil then begin
        if ccfFilterList <> Nil then
          j := ccfFilterList.Count - 1
        else
          j := -1;  // dann wird anschl. schleife nicht durchlaufen  
      end
      else
        j := Items.Count - 1;
      i := 0;
      while i < j do begin
        if ccfFile <> Nil then begin
          s := ccfFilterList[j];
          ccfFilterList[j] := ccfFilterList[i];
          ccfFilterList[i] := s;
          dec(j);
          inc(i);
        end
        else begin

          s := Items[j];
          if s = '' then dec(j)  {Game was appended to the sorted list - ignore}
          else begin
            Items[j] := Items[i];
            Items[i] := s;
            dec(j);
            inc(i);
          end;

        end;
        if DoTrace then exit;
      end;
    end;
  finally
    if assigned(FOnPGNTrace) then FOnPGNTrace(self, 0, Abbruch);
    Deleting := false;
    if not FLockRefresh then
      Refresh;
    sl.Free;
    Screen.Cursor := crDefault;
    FileReading := false;
    Items.EndUpdate;
  end;
end;

function TPGN.isCCFFile : Boolean;
begin
  Result := ccfFile <> Nil;
end;

function TPGN.DataFind(Control: TWinControl; FindString: String): Integer;
begin
  // still not used
end;

procedure TPGN.Data(Control: TWinControl; Index: Integer; var Data: String);
begin
  Data := IntToStr(Index);
end;

function TPGN.ccfListeindex2Databaseindex(idx : longint) : longint;
begin
  if ccfFilterlist = Nil then
    Result := idx
  else if (idx >= 0) and (idx < ccfFilterlist.Count) then
    Result := StrToIntDef(ccfFilterlist[idx], -1)
  else
    Result := -1;
end;

// wandelt den index einer (gefilterten/sortierten) liste in den Datenbankindex um
function TPGN.GetItemsObject(Listeindex : longint) : longint;
begin
  if ccfFile <> Nil then begin
    Result := ccfListeindex2Databaseindex(ListeIndex);
  end
  else begin

    if Items[ListeIndex] = '' then Result := longint(Items.Objects[ListeIndex])
    else begin
      Result := longint(Items.Objects[StrToInt(Items[ListeIndex])]);
    end;
  end;
end;

function TPGN.ListeIndexToDatabaseIndex(ListeIndex : longint) : longint;
begin
  if (ListeIndex >= 0) and (ListeIndex < Items.Count)
  then Result := GetItemsObject(ListeIndex)//longint(Items.Objects[ListeIndex])
  else Result := -1;
end;

procedure TPGN.DrawItem(Index: Integer; Rect: TRect; State: TOwnerDrawState);
Var i : longint;
begin
  if (ccfFile <> Nil) and (ccfFile.Working) then exit;
  if FLockRefresh or Deleting then exit;  {must not be executed while deleting}
  if assigned(FOnPGNDrawItem) then begin
    i := GetItemsObject(Index);//longint(items.Objects[Index]);
    RefreshEventData(i);
    FOnPGNDrawItem(self, Index, Rect, EvD, i);
  end
  else inherited DrawItem(Index, Rect, State);
end;

procedure TPGN.Filtern;
Var i : longint;
    prcnt, Prcnt100, NewPrcnt, LastPrcnt : Integer;
    Abbruch : TCancelMode;

    ccfCount : longint;
    ShowThisGame : Boolean;
begin
  if (FileReading) or (not FileOpen) then exit;
  LastPrcnt := -1;
  Prcnt100 := lHeader.Count;
  Prcnt := 0;
  Abbruch := cmNoCancel;
  FileReading := true;
  screen.cursor := crHourglass;
  Items.BeginUpdate;
  try
    FPGNSorted := soNothing;

    if ccfFile <> Nil then begin
      Count := 0;
      ccfCount := ccfFile.CountWithDeletedRecords;
      if (FccfFilterReset) or (not assigned(FOnFilterRequest)) then begin
        FccfFilterReset := false;
        FreeAndNil(ccfFilterList);
        Count := ccfCount;
      end
      else begin
        ccfFile.FilesOpen := true;
        try
          if ccfFilterList = Nil then
            ccfFilterList := TStringlist.Create;
          ccfFilterList.Clear;
          Prcnt100 := ccfCount;
          for i := 0 to ccfCount - 1 do begin
            RefreshEventData(i);
            ShowThisGame := true;
            FOnFilterRequest(self, EvD, ShowThisGame);
            if ShowThisGame then
              ccfIncCount(i);
            inc(prcnt);
            NewPrcnt := GetPercent(prcnt, prcnt100);
            if NewPrcnt <> LastPrcnt then begin
              if assigned(FOnPGNTrace) then FOnPGNTrace(self, NewPrcnt, Abbruch);
              if Abbruch = cmCancel then break;
              LastPrcnt := NewPrcnt;
            end;
          end;
        finally
          ccfFile.FilesOpen := false;
        end;
        Count := ccfFilterList.Count;
      end;
      if assigned(FOnPGNTrace) then FOnPGNTrace(self, 0, Abbruch);
      exit;
    end;

    Clear;
    for i := 0 to lHeader.Count - 1 do begin
      AddPGNItem(Items.Count, i);
      inc(prcnt);
      NewPrcnt := GetPercent(prcnt, prcnt100);
      if NewPrcnt <> LastPrcnt then begin
        if assigned(FOnPGNTrace) then FOnPGNTrace(self, NewPrcnt, Abbruch);
        if Abbruch = cmCancel then break;
        LastPrcnt := NewPrcnt;
      end;
    end;
    if assigned(FOnPGNTrace) then FOnPGNTrace(self, 0, Abbruch);
  finally
    Items.EndUpdate;
    FileReading := false;
    screen.cursor := crDefault;
  end;
end;

procedure TPGN.KeyPress(var Key: Char);
Var ID : longint;
    s : string;
 //   recordIsDeletedDummy : Boolean;
begin
  inherited Keypress(Key);
  if (key = #13) and (SelCount <> 0) then begin
    if (assigned(FOnPGNDblClick)) and (FileOpen) then begin
      s := GetGameAsString(GetItemsObject(ItemIndex), ID{, recordIsDeletedDummy});
      if (s <> '') and (ID <> -1) then FOnPGNDblClick(self, s, ID);
    end;
  end;
end;

procedure TPGN.DblClick;
Var ID : longint;
    s : string;
 //   recordIsDeletedDummy : Boolean;
begin
  inherited DblClick;
  if (assigned(FOnPGNDblClick)) and (FileOpen) then begin
    s := GetGameAsString(GetItemsObject(ItemIndex), ID{, recordIsDeletedDummy});
    if (s <> '') and (ID <> -1) then FOnPGNDblClick(self, s, ID);
  end;
end;

constructor TPGN.Create(AOwner : TComponent);
begin
  inherited Create(AOwner);
  FccfDatabaseCheckSteps := 200;
  slTemp := TStringlist.Create;
  EvD.slRestData := TStringlist.Create;
  lHeader := TList.Create;
  Style := lbOwnerDrawVariable;
  Sorted := false;
  FPGNSorted := soNothing;
end;

destructor TPGN.destroy;
begin
  try
    DoslIndexeClear;
    lHeader.Free;
    slTemp.free;
    EvD.slRestData.Free;
    ClosePGNFile;
    inherited destroy;
  except
  end;
end;

procedure TPGN.ClosePGNFile;
begin
  if FileOpen then

  if ccfFile <> Nil then begin
    FreeAndNil(ccfFile);
    FreeAndNil(ccfFilterList);
  end
  else

  try
    CloseFile(filePGN);
  except
  end;
  FileOpen := false;
end;

procedure TPGN.RefreshEventData(lHeaderIndex : longint);
Var j : longint;
    s : string;
    recordIsDeleted : Boolean;
begin
  slTemp.Text := DoGetGameAsString(lHeaderIndex, recordIsDeleted);
  ResetEventData(EvD);
  j := 0;
  while j < slTemp.Count do begin
    s := trim(slTemp[j]);
    if (s = '') or ((s <> '') and (s[1] <> '[')) then break;
    SetEventData(EvD, s);
    inc(j);
  end;
  inc(j);
  while j < slTemp.Count do begin
    EvD.tempData := EvD.tempData + slTemp[j] + #13#10;
    inc(j);
  end;
  EvD.Deleted := recordIsDeleted;
  if (slTemp.Count = 0) then begin
    EvD.Weiss := 'Error';
    EvD.Schwarz := '...';
  end;
end;

procedure TPGN.UpDateFileIndexe(lHeaderIndex, summand : longint);
Var i : longint;
begin
  for i := lHeaderIndex to lHeader.Count - 1 do begin
    PGameHeader(lHeader[i])^.FilePositionA :=
      PGameHeader(lHeader[i])^.FilePositionA + summand;
    PGameHeader(lHeader[i])^.FilePositionE :=
      PGameHeader(lHeader[i])^.FilePositionE + summand;
  end;
end;

procedure TPGN.UpdateItems(ListeIndex, Summand : longint);
Var i, j : longint;
begin
  j := ListeIndex;
  while j < Items.count do begin
    i := longint(Items.Objects[j]);
    i := i + Summand;
    Items.Objects[j] := Pointer(i);
    inc(j);
  end;
end;

procedure TPGN.DeleteIndexObject(lHeaderindex : longint);
Var gh : PGameHeader;
begin
  gh := PGameHeader(lHeader[lHeaderindex]);
  if assigned(gh) then Dispose(gh);
  lHeader.Delete(lHeaderindex);
end;

procedure TPGN.DeleteItem(ListeIndex : longint);
Var i, j, k, f : longint;
begin
  i := ListeIndex;
  if Items[i] <> '' then i := StrToInt(Items[i]);
  if FPGNSorted <> soNothing then begin
    if Listeindex > i then begin
      for j := Items.count - 1 downto 0 do begin
        if Items[j] <> '' then begin
          val(Items[j], k, f);
          if f = 0 then begin
            if k > i then k := k - 1;
            if (j >= i) and (j < Listeindex) then begin
              if j < Items.Count - 1 then Items[j + 1] := IntToStr(k);
            end
            else Items[j] := IntToStr(k);
          end;
        end;
      end;
    end
    else if Listeindex < i then begin
      for j := 0 to Items.count - 1 do begin
        if Items[j] <> '' then begin
          val(Items[j], k, f);
          if f = 0 then begin
            if k > i then k := k - 1;
            if (j > Listeindex) and (j <= i) then begin
              if j > 0 then Items[j - 1] := IntToStr(k);
            end
            else Items[j] := IntToStr(k);
          end;
        end;
      end;
    end
    else for j := 0 to Items.count - 1 do begin
      if Items[j] <> '' then begin
        val(Items[j], k, f);
        if f = 0 then begin
          if k > i then begin
            k := k - 1;
            Items[j] := IntToStr(k);
          end;
        end;
      end;
    end;
  end;
  Items.Delete(i);
  UpdateItems(i, -1);
end;

procedure TPGN.AddPGNItem(ListeIndex, lHeaderIndex : longint);
Var ShowThisGame : Boolean;
begin
  if assigned(FOnFilterRequest) then begin
    RefreshEventData(lHeaderIndex);
    ShowThisGame := true;
    FOnFilterRequest(self, EvD, ShowThisGame);
    if ShowThisGame then begin
      Items.InsertObject(ListeIndex, '', Pointer(lHeaderIndex));
      UpdateItems(ListeIndex + 1, 1);  {here Item is always ''}
    end;
  end
  else begin
    Items.InsertObject(ListeIndex, '', Pointer(lHeaderIndex));
    UpdateItems(ListeIndex + 1, 1);
  end;
end;

procedure TPGN.AddIndexObject(lHeaderIndex, GameID, posA, posE : longint);
Var gh : PGameHeader;
begin
  gh := New(PGameHeader);
  with gh^ do begin
    ID := GameID;
    FilePositionA := PosA;
    FilePositionE := PosE;
  end;
  lHeader.Insert(lHeaderIndex, Pointer(gh));
end;

procedure TPGN.DoslIndexeClear;
Var i : integer;
    gh : PGameHeader;
    c : TCursor;
begin
  c := screen.cursor;
  screen.cursor := crHourglass;
  try
    for i := 0 to lHeader.Count - 1 do begin
      gh := PGameHeader(lHeader[i]);
      if assigned(gh) then Dispose(gh);
    end;
    lHeader.Clear;
  finally
    screen.cursor := c;
  end;
end;

function TPGN.createChessDatabase(Filename : String) : Boolean;
var f : int64File;
    s : String;

    ccfTemp, ccfLast : TChessDatabase;
begin
  Result := false;
  ccfDatabaseCheckStep := 0;
  ccfpreparing := true;
  s := ChangeFileExt(Filename, ExtractFileExt(Filename) + defaultidxFileExt);
  if FileExists(s) then try
    ccfpreparing := false;
    AssignAndResetint64File(f, s);
    try
      ccfDatabaseCheckStep := trunc(FileSize(f) / FccfDatabaseCheckSteps);
    finally
     closeFile(f);
    end;
  except
  end;
  ccfLast := ccfFile;
  try
    ccfTemp := TChessDatabase.create(self);
    ccfFile := ccfTemp;  // the event handler check for ccfFile = Nil
    ccfTemp.OnDatabaseCheckTrace := DatabaseCheck;
    ccfTemp.OnDatabaseError := DatabaseError;
    ccfTemp.OnCreateNewIndexFile := CreateNewIndexFile;
    ccfTemp.OnDatabaseCompressTrace := DatabaseCompress;
    ccfTemp.OnCreateNewIndexfileTrace := DatabaseCompress;
    // if Notation is not set, the game will be saved uncompressed
    ccfTemp.Notation := ccfNotation;
    ccfTemp.CheckAndLoadFile(Filename);
    ccfDatabaseCheckStep := 0;
    if ccfTemp.Filename = '' then begin
      ccfTemp.Free;
      exit;
    end;
  finally
    ccfFile := ccfLast;
    ccfpreparing := false;
  end;
  ClosePGNFile;
  ccfFile := ccfTemp;
  FileOpen := true;
  Result := true;
  Style := lbVirtualOwnerDraw;
 // OnDataFind := DataFind;
  OnData := Data;
end;

procedure TPGN.SetPGNFile(s : string);
var
  NumRead, prcnt, Prcnt100, NewPrcnt, LastPrcnt : Integer;
  Buf : array[1..BufferSize + 1] of Char;
  hBuf: array[0..length(PGNPartieStart)] of Char;  { zum Vergleich des Reststrings (beachten: letztes Nullbyte u. PGNPartiestart ist eigentlich um 1 Zeichen (#$A) lnger, aber hier mu es eins weniger sein, damit der Vergleichstring nicht im normalen Block nochmal gefunden wird}
  cs, cs1, csRest : PChar;
  i, j, rFilePos, rFilePosA : longint;
  first, CreateNewPGNFile : Boolean;
  sRest : string;
  f : File;
  rFilterRq : TFilterRequestEvent;
  Abbruch    : TCancelMode;

procedure AddListItems(a, e : longint);
begin
  if rFilePosA <> -1 then begin
    AddIndexObject(lHeader.Count, lHeader.Count, a, e);
    AddPGNItem(Items.Count, lHeader.Count - 1);
  end;
  rFilePosA := e;
end;

begin
  if FileReading then exit;
  if uppercase(s) = uppercase(FPGNFile) then begin
    if not FOnlyPartsRead then exit;
  end
  else begin
    if ccfFile <> Nil then begin
      Count := 0;
      Style := lbOwnerDrawVariable;
     end;
    If not FileExists(s) then begin
      if s = '' then begin
        Items.Clear;
        DoslIndexeClear;
        ClosePGNFile;
        FPGNFile := '';
        exit;
      end
      else begin
        CreateNewPGNFile := true;
        if assigned(BeforeCreatePGNFile) then BeforeCreatePGNFile(self, s, CreateNewPGNFile);
        if CreateNewPGNFile then begin

          if (uppercase(ExtractFileExt(s)) = '.PGN') or (FonlyPGN) then begin
            ClosePGNFile;
            assignFile(f, s);
            rewrite(f);
            Close(f);
          end

          else
            if not createChessDatabase(s) then exit;

        end
        else exit;
      end;
    end
    else begin

      if (uppercase(ExtractFileExt(s)) = '.PGN') or (FonlyPGN) then begin
        if not CheckPGN(s) then exit;
        ClosePGNFile;
      end

      else
        if not createChessDatabase(s) then exit;

    end;
  end;
  screen.cursor := crHourglass;
  Abbruch := cmNoCancel;
  LastPrcnt := -1;
  rFilterRq := OnFilterRequest;
  OnFilterRequest := Nil;
  try
    if ccfFile <> Nil then begin
      Count := ccfFile.CountWithDeletedRecords;  // liefert alle datenstze inkl. gelschte
      FPGNFile := s;
      exit;
    end;

    Buf[BufferSize + 1] := #0;
    first := true;
    cs := @Buf;
    ClosePGNFile;
    sRest := '';
    AssignAndResetFile(filePGN, 1, s);	
    FPGNFile := s;
    FileOpen := true;
    FileReading := true;
    FOnlyPartsRead := true;
    rFilePosA := -1;
    Items.Clear;
    DoslIndexeClear;
    Prcnt100 := FileSize(filePGN) div BufferSize;
    Prcnt := 0;
    repeat
      rFilePos := FilePos(filePGN);
      BlockRead(filePGN, Buf, BufferSize, NumRead);
      buf[NumRead + 1] := #0;
      if first then cs1 := StrPos(cs, PGNPartieStart)
      else begin
        cs1 := StrPos(cs, #$A + PGNPartieStart);
        if cs1 <> Nil then begin
          StrLCopy(hBuf, cs, SizeOf(hBuf) - 1);
          csRest := @hBuf;
          sRest := sRest + string(csRest);
          i := Pos(#$A + PGNPartieStart, sRest);
          if i <> 0 then
            AddListItems(rFilePosA, rFilePos + (i - length(PGNPartieStart)));
        end;
      end;

      while cs1 <> Nil do begin
        if first then j := rFilePos + longint(cs1 - cs)
        else j := rFilePos + longint(cs1 - cs) + 1;  {without +1 it would point to #10}
        AddListItems(rFilePosA, j);
        inc(cs1);
        cs1 := StrPos(cs1, #$A + PGNPartieStart);
        First := false;
      end;
      if numRead > length(PGNPartieStart) - 1 then begin
        csRest := @buf[NumRead - (length(PGNPartieStart) - 1)];
        sRest := string(csRest);          {problematisch knnte werden: 1. erstes PGNPartiestart steht erst kurz vor Bufferende; 2. In Partie-Zgenotat. beginnt erste Zeile mit PGNPartiestart (in Kommentar oder so) - beides ist unwahrscheinlich}
      end;

      inc(prcnt);
      NewPrcnt := GetPercent(prcnt, prcnt100);
      if (lHeader.Count > 50) and (NewPrcnt <> LastPrcnt) then begin
        if assigned(FOnPGNTrace) then FOnPGNTrace(self, NewPrcnt, Abbruch);
        LastPrcnt := NewPrcnt;
      end;
    until (NumRead = 0) or (Abbruch = cmCancel);
    if Abbruch <> cmCancel then begin
      FOnlyPartsRead := false;
      AddListItems(rFilePosA, FileSize(filePGN));
    end;
  finally
    OnFilterRequest := rFilterRq;
    FileReading := false;
    LastID := lHeader.Count - 1;
    screen.cursor := crDefault;
    if assigned(FOnPGNTrace) then FOnPGNTrace(self, 0, Abbruch);
    if not FLockRefresh then
      Refresh;
  end;
end;

function TPGN.AdjustGameString(lHeaderIndex : longint; sGame : string; Modus : integer) : string;
begin
  sGame := trim(sGame);
  if lHeader.Count = 0 then
    Result := sGame
  else if Modus = pgnAdd then  {in this case ID is undefined}
    Result := pgnRetString + pgnRetString + sGame
  else if Modus = pgnInsert then Result := sGame + pgnRetString + pgnRetString
  else if lHeader.Count = 1 then Result := sGame
  else begin
    if (lHeaderIndex = lHeader.Count - 1) and (not onlyPartsRead) then Result := sGame
    else Result := sGame + pgnRetString + pgnRetString;
  end;
end;

function TPGN.GetFilePosIndex(ID : longint; Var FilePosindexA : longint) : longint;
Var i : longint;
begin
  Result := -1;
  i := 0;
  while i < lHeader.Count do begin
    if PGameHeader(lHeader[i])^.ID = ID then begin
      Result := i;
      FilePosIndexA := PGameHeader(lHeader[i])^.FilePositionA;
      break;
    end;
    inc(i);
  end;
end;

function TPGN.AddGame(Var ID : longint; sGame : String) : Boolean;
Var s : string;
    cswrite : Pointer;
    i : longint;

    i64 : Int64;
begin
  Result := false;
  if OnlyPartsRead then exit;

  if ccfFile <> Nil then begin
    i64 := ID;
    if ccfRaw then
      Result := ccfFile.addGameByIndex(i64, sGame)
    else
      Result := ccfFile.addGameAsPGNByIndex(i64, sGame);
    ID := i64;
    if Result then
      ccfIncCount(ID);
    exit;
  end;

  s := AdjustGameString(ID, sGame, pgnAdd);
  i := FileSize(filePGN);
  Seek(filePGN, i);
  GetMem(cswrite, length(s) + 1);
  screen.cursor := crHourglass;
  try
    StrCopy(csWrite, PChar(s));
    BlockWrite(filePGN, cswrite^, length(s));
    if lHeader.Count <> 0 then i := i + 4;  {if the file was empty, no empty line was inserted at the start}
    AddIndexObject(lHeader.Count, LastID + 1, i, FileSize(filePGN));
    AddPGNItem(Items.Count, lHeader.Count - 1);
    inc(LastID);
    ID := LastID;
    Result := true;
  finally
    screen.cursor := crDefault;
    FreeMem(cswrite, length(s) + 1);
  end;
end;

function TPGN.DeleteGameByID(AID : longint) : Boolean;
begin
  Result := DoDeleteGame(AID, false);
end;

function TPGN.DeleteGame(ListeIndex : longint) : Boolean;
begin
  Result := DoDeleteGame(ListeIndex, true);
end;

function TPGN.GetListeIndexByID(AID : longint) : longint;
Var i, j : integer;
begin
  Result := -1;
  for i := 0 to Items.Count - 1 do begin
    j := GetItemsObject(i);
    if (j >= 0) and (j < lHeader.Count) then
      if PGameHeader(lHeader[j])^.ID = AID then begin
        Result := i;
        exit;
      end;
  end;
end;

function TPGN.Backup(idx : longint; von : Integer) : Boolean;
var
  toF: file;
  NumRead, NumWritten: Integer;
  s1, s2, s3 : String;
  i : Integer;
  prcnt, prcnt100, NewPrcnt, LastPrcnt : integer;
  Buffer : Pointer;
  Abbruch    : TCancelMode;
begin
  Result := false;
  try
    BackupFile := '';
    Abbruch := cmNoCancel;
    LastPrcnt := -1;
    Prcnt100 := (FileSize(filePGN) - idx) div BufferSize;
    Prcnt := 0;
    s1 := ExtractFilePath(PGNFile) + 'backup' + FormatDateTime('dd.mm.', now);
    s2 := '_game' + IntToStr(von) + '-' + IntToStr(lHeader.Count) + '_' + ExtractFilename(PGNFile);
    i := 1;
    while (i < 26) and (FileExists(s1 + chr(i + 96) + s2)) do
      inc(i);
    BackupFile := s1 + chr(i + 96) + s2;
    GetMem(Buffer, BufferSize);
    AssignFile(ToF, BackupFile);
    Rewrite(ToF, 1);
    try
      seek(filePGN, idx);
      repeat
        BlockRead(filePGN, Buffer^, BufferSize, NumRead);
        BlockWrite(ToF, Buffer^, NumRead, NumWritten);
        inc(prcnt);
        NewPrcnt := GetPercent(prcnt, prcnt100);
        if NewPrcnt <> LastPrcnt then begin
          if assigned(FOnPGNTrace) then FOnPGNTrace(self, NewPrcnt, Abbruch);
          LastPrcnt := NewPrcnt;
        end;
        if Abbruch = cmCancel then exit;
      until (NumRead = 0) or (NumWritten <> NumRead);
      Result := true;
    finally
      CloseFile(ToF);
      FreeMem(Buffer, BufferSize);
      if assigned(FOnPGNTrace) then FOnPGNTrace(self, -1, Abbruch);
    end;
  except;
  end;
end;

function TPGN.DoDeleteGame(AIndex : longint; IsListeIndex : Boolean) : Boolean;
Var NumRead, NumWritten, readindex, writeindex, l, lHeaderNummer, i : longint;
    Buffer : Pointer;
    prcnt, prcnt100, NewPrcnt, LastPrcnt : integer;
    Abbruch    : TCancelMode;
begin
  Result := false;
  if OnlyPartsRead then exit;
  if (FileReading) or (not FileOpen) then exit;

  if ccfFile <> Nil then begin
    i := AIndex;
    if IsListeIndex then
    i := ccfListeindex2Databaseindex(i);
    Result := ccfFile.deleteGameByIndex(i);
(*  deleted games are only striked out
    if Result then
      Count := Count - 1;*)
    exit;
  end;

  if IsListeIndex then begin
    if (AIndex < 0) or (Aindex >= Items.Count) then exit;
    lHeaderNummer := GetItemsObject(AIndex);
  end
  else begin
    if AIndex < 0 then exit;
    lHeaderNummer := GetFilePosIndex(AIndex, i);
  end;
  if not ((lHeadernummer >= 0) and (lHeadernummer < lHeader.Count)) then exit;
  GetMem(Buffer, BufferSize);
  screen.cursor := crHourglass;
  FileReading := true;
  Deleting := true; {Prevents Draw-Event}
  try
    if lHeadernummer < lHeader.Count - 1 then
    readindex := PGameHeader(lHeader[lHeadernummer + 1]).FilePositionA
    else readindex := PGameHeader(lHeader[lHeadernummer]).FilePositionE;
    writeindex := PGameHeader(lHeader[lHeadernummer]).FilePositionA;
    l := writeindex - readindex;

    if not Backup(writeindex, lHeadernummer + 1) then begin
      DeleteFile(BackupFile);
      exit;
    end;
    Abbruch := cmCancelDenied;
    LastPrcnt := -1;
    Prcnt100 := FileSize(filePGN) div BufferSize;
    Prcnt := 0;
    repeat
      seek(filePGN, readindex);
      BlockRead(filePGN, Buffer^, BufferSize, NumRead);
      inc(readindex, BufferSize);
      seek(filePGN, writeindex);
      if NumRead <> 0 then BlockWrite(filePGN, Buffer^, NumRead, NumWritten);
      inc(writeindex, BufferSize);
      inc(prcnt);
      NewPrcnt := GetPercent(prcnt, prcnt100);
      if NewPrcnt <> LastPrcnt then begin
        if assigned(FOnPGNTrace) then FOnPGNTrace(self, NewPrcnt, Abbruch);
        LastPrcnt := NewPrcnt;
      end;
    until (NumRead = 0) or (NumRead <> NumWritten) or (readindex >= Filesize(filePGN));
    truncate(filePGN);
    if IsListeIndex then i := AIndex
    else i := GetListeIndexByID(AIndex);
    DeleteIndexObject(lHeadernummer);
    if lHeadernummer < lHeader.Count then
      UpDateFileIndexe(lHeadernummer, l);
    if i >= 0 then DeleteItem(i);  // -1 kann GetListeIndexByID zurckgeben
    DeleteFile(BackupFile);
    Result := true;
  finally
    screen.cursor := crDefault;
    FreeMem(Buffer, BufferSize);
    Deleting := false;
    if assigned(FOnPGNTrace) then FOnPGNTrace(self, 0, Abbruch);

    FileReading := false;
    if not FLockRefresh then
      Refresh;
  end;
end;

procedure TPGN.ccfIncCount(i : Int64);
var j : int64;
begin
  if ccfFilterList <> Nil then
    ccfFilterList.Add(IntToStr(i));
  j := Topindex;
  Count := Count + 1;
  Topindex := j;
end;

function TPGN.DoSaveGame(ListeIndex : longint; Var ID : longint; sGame : String; Modus : integer) : Boolean;
Var
  NumRead, NumWritten, NumWrite, BufSize, Nummer: Longint;
  csread, cswrite, cstemp : Pointer;
  l, lr, readindex, writeindex : longint;
  s : string;
  prcnt, prcnt100, NewPrcnt, LastPrcnt : integer;
  i64 : Int64;
  Abbruch    : TCancelMode;
begin
  Result := false;
  if OnlyPartsRead then exit;
  if (FileReading) or (not FileOpen) then exit;

  if ccfFile <> Nil then begin
    i64 := ID;
    if ListeIndex = pgnAddGame then begin
      if ccfRaw then
        Result := ccfFile.addGameByIndex(i64, sGame)
      else
        Result := ccfFile.addGameAsPGNByIndex(i64, sGame);
      ID := i64;
      if Result then
        ccfIncCount(ID);
    end
    else begin
      if ccfRaw then
        Result := ccfFile.replaceGameAsPGNByIndex(i64, sGame)
      else
        Result := ccfFile.replaceGameAsPGNByIndex(i64, sGame);
    end;
    exit;
  end;

  if (ListeIndex = pgnAddGame) or
  ((Modus = pgnInsert) and (FPGNSorted <> soNothing)){no inserting if the list is sorted}
  then begin
    Result := AddGame(ID, sGame);
    exit;
  end;
  nummer := GetFilePosIndex(ID, readindex);
  if nummer = -1 then begin
    Result := AddGame(ID, sGame);
    exit;
  end;
  s := AdjustGameString(ID, sGame, Modus);
  l := length(s);
  if l < Buffersize then BufSize := BufferSize
  else BufSize := l;
  GetMem(csread, BufSize + 1);
  GetMem(cswrite, BufSize + 1);
  StrCopy(cswrite, PChar(s));
  screen.cursor := crHourglass;
  FileReading := true;
  try
    writeindex := readindex;
    lr := 0;
    if modus = pgnReplace then begin
      readindex := PGameHeader(lHeader[nummer]).FilePositionE;
      lr := l - (readindex - writeindex);
    end;

    if not Backup(writeindex, nummer + 1) then begin
      DeleteFile(BackupFile);
      exit;
    end;

    Abbruch := cmCancelDenied;
    LastPrcnt := -1;
    seek(filePGN, readindex);
    BlockRead(filePGN, csread^, BufSize, NumRead);
    inc(readindex, BufSize);
    seek(filePGN, writeindex);
    BlockWrite(filePGN, cswrite^, l, NumWritten);  {write new game}
    inc(writeindex, l);
    StrCopy(cswrite, csread);
    NumWrite := NumRead;
    Prcnt100 := FileSize(filePGN) div BufSize;
    Prcnt := 0;
    repeat
      seek(filePGN, readindex);
      BlockRead(filePGN, csread^, BufSize, NumRead);
      inc(readindex, BufSize);
      seek(filePGN, writeindex);
      BlockWrite(filePGN, cswrite^, NumWrite, NumWritten);
      inc(writeindex, BufSize);
      NumWrite := NumRead;
      cstemp := cswrite;
      cswrite := csread;
      csread := cstemp;
      inc(prcnt);
      NewPrcnt := GetPercent(prcnt, prcnt100);
      if NewPrcnt <> LastPrcnt then begin
        if assigned(FOnPGNTrace) then FOnPGNTrace(self, NewPrcnt, Abbruch);
        LastPrcnt := NewPrcnt;
      end;
    until NumRead = 0;

    if modus = pgnReplace then begin
      truncate(filePGN);
      PGameHeader(lHeader[nummer])^.FilePositionE :=
        PGameHeader(lHeader[nummer])^.FilePositionE + lr;
      UpDateFileIndexe(nummer + 1, lr);
    end
    else begin
      AddIndexObject(nummer, LastID + 1, PGameHeader(lHeader[nummer])^.FilePositionA,
      PGameHeader(lHeader[nummer])^.FilePositionA + l);
      UpDateFileIndexe(nummer + 1, l);
      AddPGNItem(ListeIndex, nummer);
      inc(LastID);
      ID := LastID;
    end;
    DeleteFile(BackupFile);
    Result := true;
  finally
    screen.cursor := crDefault;
    FreeMem(csread, BufSize + 1);
    FreeMem(cswrite, BufSize + 1);
    if assigned(FOnPGNTrace) then FOnPGNTrace(self, 0, Abbruch);
    FileReading := false;
  end;
end;

function TPGN.SaveGame(ListeIndex : longint; Var ID : longint; sGame : String) : Boolean;
Var i : longint;
begin
  if (ListeIndex <= pgnAddGame) or (ListeIndex >= Items.Count)
  or (ccfFile <> Nil) then i := pgnAddGame
  else begin
    i := ListeIndex;
    ID := PGameHeader(lHeader[GetItemsObject(ListeIndex){longint(Items.Objects[ListeIndex])}])^.ID;
  end;
  Result := DoSaveGame(i, ID, sGame, pgnInsert);
end;

function TPGN.ReplaceGame(Var ID : longint; sGame : String) : Boolean;
begin
  Result := DoSaveGame(-2, ID, sGame, pgnReplace);  { -2 is only dummy - not used with pgnReplace}
end;

function TPGN.DoGetGameAsString(lHeaderIndex : longint; var recordIsDeleted : Boolean) : string;
Var eindex, aindex, diff, rFilePos : longint;
    csread : Pointer;

    handle : String;
begin
  Result := '';

  if ccfFile <> Nil then begin
    // handle wird hier nicht verwendet
    // das geht nur, weil reihenfolge sich nicht ndert nach lschen
    // es darf nur nicht zwischendurch komprimiert werden!!!
    if ccfRaw then begin
      if not ccfFile.getGameByIndex(lHeaderIndex, handle, Result, recordIsDeleted) then
        Result := ''
    end
    else if not ccfFile.getGameAsPGNByIndex(lHeaderIndex, handle, Result, recordIsDeleted) then
      Result := '';
  end
  else

  try
    aindex := PGameHeader(lHeader[lHeaderIndex])^.FilePositionA;
    eindex := PGameHeader(lHeader[lHeaderIndex])^.FilePositionE;
    diff := eIndex - aIndex;
    GetMem(csread, diff + 1);
    rFilePos := FilePos(filePGN);
    try
      FillChar(csread^, diff + 1, #0);
      seek(filePGN, aindex);
      BlockRead(filePGN, csread^, diff);
      Result := trim(String(PChar(csread)));
    finally
      FreeMem(csread, diff + 1);
      seek(filePGN, rFilePos);
    end;
  except
  end;
end;

function TPGN.GetGameAsString(lHeaderIndex : longint; Var ID : longint{; var recordIsDeleted : Boolean}) : string;
var recordIsDeleted : Boolean;
begin
  Result := '';
  if not FileOpen then exit;
  Screen.Cursor := crHourGlass;
  try
    ID := -1;

    if ccfFile <> Nil then begin
      Result := DoGetGameAsString(lHeaderIndex, recordIsDeleted);
      if Result <> '' then ID := lHeaderIndex;
    end
    else

    if (lHeaderIndex >= 0) and (lHeaderIndex < lHeader.Count) then begin
      Result := DoGetGameAsString(lHeaderIndex, recordIsDeleted);
      if Result <> '' then ID := PGameHeader(lHeader[lHeaderIndex])^.ID;
    end;
  finally
    Screen.Cursor := crDefault;
  end;
end;

function  TPGN.GetItemsVisibleCount : longint;
begin
  Result := Items.Count;
end;

function  TPGN.GetItemsAllCount : longint;
begin
  if ccfFile <> Nil then
    Result := ccfFile.Count
  else
    Result := lHeader.Count;
end;

procedure TPGN.DatabaseCheck (Sender : TObject; var checkstep : int64; var cancel : Boolean);
begin
  checkStep := ccfDatabaseCheckStep;
end;

procedure TPGN.DatabaseError (Sender : TObject);
begin
  if ccFFile = Nil then exit;
  MessageDlg('Serious error while reading file ' + ccFFile.Filename, mtError, [mbOk], 0);
//  ccfFile.OnDatabaseError := Nil;
end;

procedure TPGN.DatabaseCompress (Sender : TObject; var Cancel : Boolean; percent : Integer);
var c : TCancelMode;
    b : Boolean;
begin
  b := Deleting;
  Deleting := true;
  try
    c := cmNoCancel;
    if assigned(FOnPGNTrace) then FOnPGNTrace(self, percent, c);
    if c = cmCancel then
      Cancel := true;
  finally
    Deleting := b;
  end;
end;

procedure TPGN.CreateNewIndexFile(Sender : TObject);
begin
  if ccFFile = Nil then exit;
  if not ccfpreparing then  // dann war datei fehlerhaft, ansonsten war es wahrscheinlich beim anlegen
    MessageDlg('Error in database. Index file will be recreated - it could take a longer while.', mtError, [mbOk], 0);
end;

function TPGN.ccfDeleteDatabase : Boolean;
begin
  Result := false;
  if ccFFile = Nil then exit;
  Count := 0;
  Result := ccfFile.DeleteDatabase;
  if Result then
    PGNFile := '';
end;

function TPGN.ccfCompress : Boolean;
var c : TCancelMode;
begin
  Result := false;
  if ccFFile = Nil then exit;
  c := cmNoCancel;
  Result := ccfFile.compress;
  if assigned(FOnPGNTrace) then FOnPGNTrace(self, 0, c);
end;

end.

