unit SchachProcs;

(*
This unit must be written to "uses" of all forms using these components.
Some common chess functions are declared here.
Some of them are briefly explained, see below.

Especially see "AusgangsStellung" for the data type TSchachbrett and "TZug"
*)

{$B-}

interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
  Forms, Dialogs, ExtCtrls, StdCtrls, Buttons, Menus, clipbrd,
  Schachfonts, IniFiles;

Const

  {fr sucheDame, sucheTurm usw.}
  LZ : array[1..4] of integer = (-9,-11,9,11);  { Laeufer Richtungen }
  TZ : array[1..4] of integer = (-1,10,1,-10);  { Turm               }
  SZ : array[1..8] of integer = (19,21,12,-8,-19,-21,-12,8);   { Springer           }
  DZ : array[1..8] of integer = (-9,-11,9,11,-1,10,1,-10);
  KZ : array[1..8] of integer = (-9,-11,9,11,-1,10,1,-10);

  KEINWERT = 1000;  {Rckgabewert, falls Funktion aktuellerWert keinen findet (Fehler im Memotext)
  - 1000 ist kein mgl. Wert (geht blo bis 320}

  //crHandcursor = 5;

  nach_rechts = 0;   {fr Stellungsverschieben}
  nach_links  = 1;
  nach_oben   = 2;
  nach_unten  = 3;
  viertel_drehen = 4;
  spiegeln_an_y  = 5;

  KEINEROCHADE = 0;
  KURZEROCHADE = 1;
  LANGEROCHADE = 2;
  KURZEFISCHERROCHADE = 3;
  LANGEFISCHERROCHADE = 4;

  WK = 6; WD = 5; WT = 2; WL = 3; WS = 4; WB = 1;
  SK =-6; SD =-5; ST =-2; SL =-3; SS =-4; SB =-1;

  {**** Figuren. Schema: <Farbe>_<Figur> *}
  S_K = -6;        { Kennung fuer Schwarze Figuren }
  S_D = -5;
  S_S = -4;
  S_L = -3;
  S_T = -2;
  S_B = -1;
  W_B =  1;          { Kennung fuer weisse Figuren }
  W_T =  2;
  W_L =  3;
  W_S =  4;
  W_D =  5;
  W_K =  6;
  LEER=  0;          { Leeres Feld }
  RAND= 100;         { Umrandung des Schachbrettes }

  legaler_Zug   = 0;
  matt          = 1;
  patt          = 2;
  illegaler_zug = 3;
  mit_Matt      = 4;
  mit_Patt      = 5; {INFO dazu wre zustzlich Prfung nach ausgef. Zug erf. - tritt aber selten auf}
  fehler        = 6;
  mit_schach    = 7;

  nr = 0;         {fr Rochademglichkeiten-Set}
  wkr = 1;
  wlr = 2;
  skr = 3;
  slr = 4;

  a1=21;b1=22;c1=23;d1=24;e1=25;f1=26;g1=27;h1=28;
  a2=31;b2=32;c2=33;d2=34;e2=35;f2=36;g2=37;h2=38;
  a3=41;b3=42;c3=43;d3=44;e3=45;f3=46;g3=47;h3=48;
  a4=51;b4=52;c4=53;d4=54;e4=55;f4=56;g4=57;h4=58;
  a5=61;b5=62;c5=63;d5=64;e5=65;f5=66;g5=67;h5=68;
  a6=71;b6=72;c6=73;d6=74;e6=75;f6=76;g6=77;h6=78;
  a7=81;b7=82;c7=83;d7=84;e7=85;f7=86;g7=87;h7=88;
  a8=91;b8=92;c8=93;d8=94;e8=95;f8=96;g8=97;h8=98;

  {*
  *** Farbe, die am Zug ist
  *}
  WEISS  =  1;
  SCHWARZ= -1;

  PGNPartieStart  = '[Event ';

  Gewinn = '1-0';    {PGN - Ergebnis}
  Remis  = '1/2-1/2';
  Verlust= '0-1';
  keinErgebnis = '*';

  feldnamen : array[0..63] of string[2] =
 ('a1','b1','c1','d1','e1','f1','g1','h1',
  'a2','b2','c2','d2','e2','f2','g2','h2',
  'a3','b3','c3','d3','e3','f3','g3','h3',
  'a4','b4','c4','d4','e4','f4','g4','h4',
  'a5','b5','c5','d5','e5','f5','g5','h5',
  'a6','b6','c6','d6','e6','f6','g6','h6',
  'a7','b7','c7','d7','e7','f7','g7','h7',
  'a8','b8','c8','d8','e8','f8','g8','h8');

   FeldFarbe : array[0..63] of byte =  (1,0,1,0,1,0,1,0,
                                        0,1,0,1,0,1,0,1,
                                        1,0,1,0,1,0,1,0,
                                        0,1,0,1,0,1,0,1,
                                        1,0,1,0,1,0,1,0,
                                        0,1,0,1,0,1,0,1,
                                        1,0,1,0,1,0,1,0,
                                        0,1,0,1,0,1,0,1);

  FENFigurennamen : array[-6..6] of char = ('k','q','n','b','r','p',' ','P','R','B','N','Q','K');

  noDiagrammzeichen = #127;
  noReturn = #13#10;

  mess : array[1..12] of string = ('Kann Datei nicht lesen.','Cannot read file.',
      'Partie nicht gefunden.','Game not found.','Fehler im PGN-Format.','Error in PGN.',
      'Beim Einlesen der Zge ist ein Fehler aufgetreten.',
      'Error reading moves.','Datei nicht gefunden.','File not found.',
      'Partie konnte nicht gespeichert werden.', 'Could not save game.');

  FRCKennung = 'f1';  { diese (ungltige) ep-Feld im FEN wird als Kennung der Fischer-R.-Chess-Ausgangsstellung benutzt}

  minGame = '[Event ""]' + noReturn + '[White "N. N."]' + noReturn + '[Black "N. N."]' + noReturn + noReturn + '*';

  //Fernschach
  condmoves = 'Conditional moves';   // mglichst nicht ndern
  acceptedcondmoves = 'Accepted conditional moves';  // mglichst nicht ndern
  TimeLineStart = '  {';
  DateTimeFormat = 'DateTimeFormat';
  WhiteDifferenceToGMT = 'WhiteDifferenceToGMT';
  BlackDifferenceToGMT = 'BlackDifferenceToGMT';
  GameId = 'GameId';

type

  TFischerInfo = record
    enabled : Boolean;
    FischerLeftWhiteRook,
    FischerRightWhiteRook,
    FischerLeftBlackRook,
    FischerRightBlackRook,
    FischerWhiteKing,
    FischerBlackKing : integer;  {AusgangsFelder fr Fischerrochade merken}
  end;
  PFischerInfo = ^TFischerInfo;

  TSchachbrett = array[0..119] of integer;
  PSchachbrett = ^TSchachbrett;
  TRochademoeglichkeiten = set of 1..4;
  TZug = packed record Rochademoeglichkeiten :
         (*
         Info for POSITION

         castlings available
         *)
         TRochademoeglichkeiten;
         (*
         The square to where an e.p. capture would be possible in the next move
         *)
         epFeld,
         (*
         1 = White to move; -1 = black to move
         *)
         amZug,
         (*
         Info for MOVE

         from-square, destination-square
         *)
         von, nach,
         (*
         Only if the move was an e.p. capture - the square of the opponent's pawn
         *)
         epSchlag,
         (*
         Castling (KEINEROCHADE = no castling; KURZEROCHADE = castling to the king side; LANGEROCHADE = castling to the queen side)
         *)
         RochadeArt,
         (*
         Promotion piece
         *)
         UFigur,
         (*
         Captured piece, if it was a capture
         *)
         geschlFigur,
         (*
         Comments like "!", "!?", etc.
         *)                                    
         attr1,
         (*
         Additional info (check, mate, move marker)
         *)                    
         attr2 : Integer;//711 shortint; {attr1 fr !,? oder Schach}
         (*
         Symbol of Chess Informant (+-, +=, etc.)
         *)
         infattr, infAttrBefore : Byte;
         end;
  PZug = ^TZug;

  TEventData = record
    Weiss, Schwarz, quelle, Annotator, ort, ergebnis, Eco, WhitedifferenceToGMTValue, BlackdifferenceToGMTValue, DateTimeFormatValue, GameIdValue : string;
    Jahr, Monat, Tag, Runde, WElo, SElo, PlyCount, EventDateJahr, EventDateMonat, EventDateTag : word;//Integer;
    FENString : string;
    TimeControlGameSec, TimeControlMoveSec : word;//Integer;
    MetaDescription, MetaKeywords : string;
    WhiteClubID, BlackClubID : String;
    slRestData : TStringlist;
    clock, tempdata : String;
    Deleted : Boolean;
  end;
  PEventData = ^TEventData;

  TGameState = (gsRun, gsMate, gsPatt, gsDraw);

const
  AusgangsStellung : TSchachbrett =  (
  RAND,RAND,RAND,RAND,RAND,RAND,RAND,RAND,RAND,RAND,
  RAND,RAND,RAND,RAND,RAND,RAND,RAND,RAND,RAND,RAND,
  RAND,W_T, W_S ,W_L ,W_D ,W_K ,W_L ,W_S ,W_T, RAND,
  RAND,W_B, W_B ,W_B ,W_B ,W_B ,W_B ,W_B ,W_B, RAND,
  RAND,LEER,LEER,LEER,LEER,LEER,LEER,LEER,LEER,RAND,
  RAND,LEER,LEER,LEER,LEER,LEER,LEER,LEER,LEER,RAND,
  RAND,LEER,LEER,LEER,LEER,LEER,LEER,LEER,LEER,RAND,
  RAND,LEER,LEER,LEER,LEER,LEER,LEER,LEER,LEER,RAND,
  RAND,S_B ,S_B ,S_B ,S_B ,S_B ,S_B ,S_B ,S_B, RAND,
  RAND,S_T, S_S ,S_L ,S_D ,S_K ,S_L ,S_S ,S_T, RAND,
  RAND,RAND,RAND,RAND,RAND,RAND,RAND,RAND,RAND,RAND,
  RAND,RAND,RAND,RAND,RAND,RAND,RAND,RAND,RAND,RAND);

  leereStellung : TSchachbrett =(
  RAND,RAND,RAND,RAND,RAND,RAND,RAND,RAND,RAND,RAND,
  RAND,RAND,RAND,RAND,RAND,RAND,RAND,RAND,RAND,RAND,
  RAND,LEER,LEER,LEER,LEER,LEER,LEER,LEER,LEER, RAND,
  RAND,LEER,LEER,LEER,LEER,LEER,LEER,LEER,LEER, RAND,
  RAND,LEER,LEER,LEER,LEER,LEER,LEER,LEER,LEER,RAND,
  RAND,LEER,LEER,LEER,LEER,LEER,LEER,LEER,LEER,RAND,
  RAND,LEER,LEER,LEER,LEER,LEER,LEER,LEER,LEER,RAND,
  RAND,LEER,LEER,LEER,LEER,LEER,LEER,LEER,LEER,RAND,
  RAND,LEER,LEER,LEER,LEER,LEER,LEER,LEER,LEER, RAND,
  RAND,LEER,LEER,LEER,LEER,LEER,LEER,LEER,LEER, RAND,
  RAND,RAND,RAND,RAND,RAND,RAND,RAND,RAND,RAND,RAND,
  RAND,RAND,RAND,RAND,RAND,RAND,RAND,RAND,RAND,RAND);

  (*
  converts 0..63 to 21..98 - so you can access the squares on TSchachbrett with c(0) .. c(63)}
  *)
  function  c(feld : byte) : byte;
  (*
  converts 21..98 to 0..63 um
  *)
  function  cc(feld : byte) : byte;

  function  Coord2Number(feldname : String) : Integer;

  function  GetPercent(part, Hundred : integer) : integer;

  (*
  TSchachbrett to FEN; TZug must contain the information for the position:
  TRochademoeglichkeiten, epFeld, amZug. Fischerinfo should be Nil
  *)
  function  PosToFEN(Position : TSchachbrett; PosInfo : TZug; nr : integer; FischerInfo : PFischerinfo) : string;
  (*
  FEN to TSchachbrett
  *)
  function  FENToPos(epd : string; Var Stllg : TSchachBrett; Var StllgInfo : TZug; Var znr : word; FischerInfo : PFischerinfo) : Boolean;
  (*
  Checks a file (filename = s) whether it is pgn
  *)
  function  CheckPGN(s : String) : Boolean;
  (*
  Resets all data of a pgn header
  *)
  procedure ResetEventData(Var e : TEventData);
  (*
  Set one pgn header. s is an entire line: [Event "xxx"], etc.
  *)
  procedure SetEventData(Var e : TEventData; s : string);
  (*
  Sttlg should be the starting position.
  This function extracts the information for king and rook location
  *)
  procedure SetFischerInfo(FischerInfo : PFischerInfo; Stllg : TSchachBrett);
  (*
  Appends a pgn string to a text file
  *)
  function  AppendGameToPGN(FileName, Games : string) : Boolean;
  (*
  Creates a valid FRC position
  *)
  function  GetFischerRandomPosition : TSchachbrett;
  (*
  rarely used: returns a string "SaZ (4 + 3)". That means: Black to move, 4 white pieces, 3 black pieces
  *)
  function  GetPiecesNumber(Stellung : TSchachbrett; Zug : TZug) : string;
  (*
  returns
  
   +---+---+---+---+---+---+---+---+
8  | *T| *S| *L| *D| *K| *L| *S| *T|
   +---+---+---+---+---+---+---+---+
7  | *B| *B| *B| *B| *B| *B| *B| *B|
   +---+---+---+---+---+---+---+---+
6  |   |   |   |   |   |   |   |   |
   +---+---+---+---+---+---+---+---+
5  |   |   |   |   |   |   |   |   |
   +---+---+---+---+---+---+---+---+
4  |   |   |   |   |   |   |   |   |
   +---+---+---+---+---+---+---+---+
3  |   |   |   |   |   |   |   |   |
   +---+---+---+---+---+---+---+---+
2  | B | B | B | B | B | B | B | B |
   +---+---+---+---+---+---+---+---+
1  | T | S | L | D | K | L | S | T |
   +---+---+---+---+---+---+---+---+
     a   b   c   d   e   f   g   h
  *)
  function  PosToText(Stellung : TSchachbrett; L_Engl : Boolean) : string;
  (*
  Z must contain all information of position and the from- and to-square of the move to be checked.
  Returns true if the move is legal
  *)
  function  IsMovePossible(Astllg : PSchachbrett; Z : PZug; fi : PFischerInfo) : Boolean;
  (*
  Writes all legal moves to sl: the string is the number of from square (on TSchachbrett),
  the object is the integer of to-square
  *)
  function  GetAllMoves(Astllg : PSchachbrett; Z : PZug; sl : TStringlist) : integer;
  (*
  Checks whether the position is Mate
  *)
  function  IsMate(Astllg : PSchachbrett; Z : PZug) : Boolean;
  (*
  Checks for "game runs", mate, stalemate or draw by insufficient material
  (gsRun, gsMate, gsPatt, gsDraw)
  *)
  function  GetGameState(Astllg : PSchachbrett; Z : PZug) : TGameState;
  (*
  true = the specified side (White) has enough material to attain a mate
  *)
  function  hasEnoughMaterial(Astllg : PSchachbrett; White : Boolean) : Boolean;
  (*
  Movepart is the from-square OR the to-square.
  The function returns a possible to-square or from-square, respectively
  *)
  function  GetPropose(Astllg : PSchachbrett; Z : PZug; movepart : integer; fi : PFischerInfo) : integer;
  (*
  Converts a FEN or a Position (TSchachbrett) to a string consisting of the characters printed as chess pieces with a special font
  You can define such a font and the characters of pieces in "schachfont.pas" (fontfile = '' if fontname is defined),
  or you define a fontfile from which the characters can be read (see the function "readfontfile" for the format)
  *)
  function  FENtoSFDia(AFEN : string; sFontName : string; fontfile : String) : string;
  function  PostoSFDia(stllg : TSchachbrett; sFontName : string; fontfile : String) : string;
  function  readFontFile(section, filename : String; var piecesChars : TChessFontChars) : Boolean;
  (*
  true if TSchachbrett may be a starting position of a FRC game
  *)
  function  maybeFischerPos(b : TSchachbrett) : Boolean;
  (*
  true if TSchachbrett is the normal starting position
  *)
  function  isFirstPosition(b : TSchachbrett) : Boolean;
  (*
  true if both are identical
  *)
  function  isEqualPosition(b1, b2 : TSchachbrett) : Boolean;
  (*
  helping functions for TChessDatabase
  *)
  procedure AssignAndresetFile(var f : File; recSize : Integer; Filename : String);
  function  isCCFDatabase(Filename : String) : Boolean;

implementation

uses ChessDatabase;

function  cc(feld : byte) : byte;  {wandelt 21..98 in 0..63 um}
begin
  cc:=((feld div 10)-2)*8+((feld mod 10)-1);
end;

function  c(feld : byte) : byte;  {wandelt 0..63 in 21..98 um}
begin
  c:=(feld div 8)*10 + (((Feld+8) mod 8)) + 21;
end;

function Coord2Number(feldname : String) : Integer;
begin
  Result := -1;
  if length(feldname) < 2 then exit;
  try
    Result := c((ord(upcase(feldname[1])) - 65) + (StrToInt(copy(feldname, 2, length(feldname))) - 1) shl 3);
  except
  end;    
end;

function GetPercent(part, Hundred : integer) : integer;
begin
  Result := 0;
  if Hundred = 0 then exit;
  Result := trunc(part * 100 / Hundred);
end;

procedure ResetEventData(Var e : TEventData);
begin
  with E do begin
    Weiss:='N.N.';Schwarz:='N.N.';Quelle:='';
    Annotator:='';Ergebnis:='*';ort:='';
    Eco := ''; FENString := '';
    Jahr:=0; Tag := 0; Monat := 0;
    Runde := 0; WElo:=0;SElo:=0;
    PlyCount := 0;
    EventDateJahr:=0; EventDateTag := 0; EventDateMonat := 0;
    TimeControlGameSec := 0;
    TimeControlMoveSec := 0;
    MetaDescription := '';MetaKeywords := '';
    WhiteClubID := ''; BlackClubID := '';
    tempdata := '';
    WhitedifferenceToGMTValue := '';
    BlackdifferenceToGMTValue := '';
    DateTimeFormatValue := '';
    GameIdValue := '';
    if assigned(slRestData) then slRestData.Clear;
  end;
end;

procedure SetEventData(Var e : TEventData; s : string);
Var Temp : string;
    fehler, i : integer;

function filtere(zeile : string) : string;
Var b : integer;
    hilf : string;
begin
  b:=1;
  while (b<=length(zeile)) and (zeile[b]<>'"') do inc(b);
  hilf:='';
  inc(b);
  while (b<=length(zeile)) and (not (zeile[b]='"')) do begin
    hilf:=hilf+zeile[b];
    inc(b);
  end;
  Filtere:=hilf;
end;

begin
  with e do begin
    if Uppercase(copy(s, 1, length(PGNPartieStart)))=Uppercase(PGNPartieStart) then begin
      quelle:=Filtere(s);
      if quelle='?' then quelle:='';
    end
    else if Uppercase(copy(s, 1, 6))='[SITE ' then begin
      ort:=Filtere(s);
      if Ort='?' then Ort:='';
    end
    else if Uppercase(copy(s, 1, 7))='[WHITE ' then begin
      weiss:=Filtere(s);
      if Weiss='?' then Weiss:='';
    end
    else if Uppercase(copy(s, 1, 7))='[BLACK ' then begin
      schwarz:=Filtere(s);
      if Schwarz='?' then Schwarz:='';
    end
    else if Uppercase(copy(s, 1, 11))='[ANNOTATOR ' then annotator:=Filtere(s)
    else if Uppercase(copy(s, 1, 8))='[RESULT ' then ergebnis:=Filtere(s)
    else if Uppercase(copy(s, 1, 10))='[WHITEELO ' then begin
      temp:=Filtere(s);
      val(temp, WElo, Fehler);
      if fehler<>0 then WElo:=0;
    end
    else if Uppercase(copy(s, 1, 10))='[BLACKELO ' then begin
      temp:=Filtere(s);
      val(temp, SElo, Fehler);
      if fehler<>0 then SElo:=0;
    end
    else if Uppercase(copy(s, 1, 10))='[PLYCOUNT ' then begin
      temp:=Filtere(s);
      val(temp, Plycount, Fehler);
      if fehler<>0 then Plycount:=0;
      if Plycount <> 0 then Plycount := (Plycount shr 1) + 1;
    end
    else if Uppercase(copy(s, 1, 6))='[DATE ' then begin
      temp:=Filtere(s);
      val(copy(temp, 1, 4), Jahr, Fehler);
      if fehler<>0 then jahr:=0;
      val(copy(temp, 6, 2), Monat, Fehler);
      if fehler<>0 then Monat:=0;
      val(copy(temp, 9, 2), Tag, Fehler);
      if fehler<>0 then Tag:=0;
    end
    else if Uppercase(copy(s, 1, 7))='[ROUND ' then begin
      temp:=Filtere(s);
      val(temp, Runde, Fehler);
      if fehler<>0 then Runde:=0;
    end
    else if Uppercase(copy(s, 1, 5))='[ECO ' then Eco:=Filtere(s)
    else if Uppercase(copy(s, 1, 5))='[FEN ' then
      FENString := Filtere(s)
    else if Uppercase(copy(s, 1, 14))='[METAKEYWORDS ' then
       MetaKeywords := MetaKeywords + Filtere(s)
    else if Uppercase(copy(s, 1, 17))='[METADESCRIPTION ' then
       MetaDescription := MetaDescription + Filtere(s)
    else if Uppercase(copy(s, 1, 13))='[WHITECLUBID ' then
       WhiteClubID := WhiteClubID + Filtere(s)
    else if Uppercase(copy(s, 1, 13))='[BLACKCLUBID ' then
       BlackClubID := BlackClubID + Filtere(s)
    else if Uppercase(copy(s, 1, 7))='[SETUP ' then  {darf nicht unter RestData fallen}

    else if Uppercase(copy(s, 1, 11))='[EVENTDATE ' then begin
      temp:=Filtere(s);
      val(copy(temp, 1, 4), EventDateJahr, Fehler);
      if fehler<>0 then EventDatejahr:=0;
      val(copy(temp, 6, 2), EventDateMonat, Fehler);
      if fehler<>0 then EventDateMonat:=0;
      val(copy(temp, 9, 2), EventDateTag, Fehler);
      if fehler<>0 then EventDateTag:=0;
    end
    else if Uppercase(copy(s, 1, 13))='[TIMECONTROL ' then begin
      temp := Filtere(s);
      i := Pos('+', temp);
      if i <> 0 then begin
        val(trim(copy(temp, 1, i - 1)), TimeControlGameSec, Fehler);
        if fehler<>0 then TimeControlGameSec:=0;
        val(trim(copy(temp, i + 1, length(temp))), TimeControlMoveSec, Fehler);
        if fehler<>0 then TimeControlMoveSec:=0;
      end;
    end
    else if Uppercase(copy(s, 1, length(WhitedifferenceToGMT) + 2))='[' + uppercase(WhitedifferenceToGMT) + ' ' then begin
      WhitedifferenceToGMTValue := Filtere(s);
    end
    else if Uppercase(copy(s, 1, length(BlackdifferenceToGMT) + 2))='[' + uppercase(BlackdifferenceToGMT) + ' ' then begin
      BlackdifferenceToGMTValue := Filtere(s);
    end
    else if Uppercase(copy(s, 1, length(DateTimeFormat) + 2))='[' + uppercase(DateTimeFormat) + ' ' then begin
      DateTimeFormatValue := Filtere(s);
    end
    else if Uppercase(copy(s, 1, length(GameId) + 2))='[' + uppercase(GameId) + ' ' then begin
      GameIdValue := Filtere(s);
    end

    else if assigned(slRestData) then begin
       slRestData.Add(s);
    end;
  end;
end;

function CheckPGN(s : String) : Boolean;
Var f : TextFile;
    p : string;
    i : integer;
    EmptyLines : Boolean;
begin
  Result := false;
  if not FileExists(s) then exit;
  try
    assignFile(f, s);
    reset(f);   { wird auch ausgefhrt bei readonly }
    p:= '';
    i := 0;
    EmptyLines := true;
    while (not eof(f)) and (Pos(PGNPartieStart, p) = 0) and (i < 1000) do begin
      readln(f, p);
      if p <> '' then EmptyLines := false;
      inc(i);
    end;
    CloseFile(f);
    if (Pos(PGNPartieStart, p) <> 0) or (i = 0) or (EmptyLines)
      then Result := true;
  except
  end;
end;

function PosToFEN(Position : TSchachbrett; PosInfo : TZug; nr : integer; FischerInfo : PFischerInfo) : string;
var h : string;
    i, j, k, z : integer;
begin
  j:=56;
  h:='';
  z:=0;
  repeat
    for i:=j to j+7 do begin
      k:=Position[c(i)];
      if k<>LEER then begin
         if z<>0 then begin
           h:=h+IntToStr(z);
           z:=0;
         end;
         h:=h+FENFigurennamen[k];
      end
      else inc(z);
    end;
    if z<>0 then begin
      h:=h+IntToStr(z);
      z:=0;
    end;
    dec(j, 8);
    if not j<0 then h:=h+'/';
  until j<0;
  if PosInfo.amZug=1 then h:=h+' w ' else h:=h+' b ';
  if wkr in PosInfo.Rochademoeglichkeiten then h:=h+'K';
  if wlr in PosInfo.Rochademoeglichkeiten then h:=h+'Q';
  if skr in PosInfo.Rochademoeglichkeiten then h:=h+'k';
  if slr in PosInfo.Rochademoeglichkeiten then h:=h+'q';
  if PosInfo.Rochademoeglichkeiten=[] then h:=h+'-';
  if assigned(Fischerinfo) then h:=h+' '+ FRCKennung + ' 0'
  else begin
    if PosInfo.epFeld<>LEER then h:=h+' '+feldnamen[cc(PosInfo.epFeld)]+' 0'
    else h:=h+' - 0';
  end;
  h:=h+' '+IntToStr(nr);
  Result:=h;
end;

{
Liest Stellung, falls in "FEN" eine angegeben ist
}

function FENToPos(epd : string; Var Stllg : TSchachBrett; Var StllgInfo : TZug; Var znr : word; FischerInfo : PFischerInfo) : Boolean;
Var  q, Reihe, hilf, z, i, fehler : integer;
     h : string;
     tempStllg : TSchachBrett;
const reihen : array[1..8] of shortint=(55,47,39,31,23,15,7,-1);

begin
  Result := false;
  if epd = '' then exit;
  try
    if assigned(FischerInfo) then with Fischerinfo^ do begin
      FischerLeftWhiteRook := 0;  {werden nur in Fischer-Random-Chess-Ausg.stellg. korrekt belegt!}
      FischerRightWhiteRook := 0;
      FischerLeftBlackRook := 0;
      FischerRightBlackRook := 0;
      Enabled := false;
    end;
    tempstllg:=LeereStellung;
    z:=0;
    i:=55;  {es geht von links oben nach rechts unten - anders als Feldnummerierung}
    q:=0;
    Reihe := 0;
    repeat;
      inc(q);
      if not( epd[q]='/') then begin
        if epd[q] in ['1'..'8'] then begin
          Val(epd[q],hilf, fehler);
          i:=i+hilf;
          z:=z+hilf;
          if epd[q]='8' then     {falls in FEN kein Schrgstrich als Trenner ist}
            if epd[q+1]='8' then begin
              dec(i, 16);  {nchste Reihe}
              inc(Reihe);
            end;
        end
        else begin
          inc(z);
          inc(i);
          if (i < 0) or (i > 63)
            then exit;
          case epd[q] of
            'K' : begin
                    tempStllg[c(i)]:=wK;
                    if assigned(Fischerinfo) then FischerInfo.FischerWhiteKing := c(i);
                  end;
            'Q' : tempStllg[c(i)]:=wD;
            'R' : begin
                    tempStllg[c(i)]:=wT;
                    if assigned(Fischerinfo) then begin
                      if FischerInfo.FischerLeftWhiteRook = 0 then FischerInfo.FischerLeftWhiteRook := c(i) { bei FEN lesen kommt immer zuerst der linke Turm}
                      else FischerInfo.FischerRightWhiteRook := c(i);
                    end;
                  end;
            'B' : tempStllg[c(i)]:=wL;
            'N' : tempStllg[c(i)]:=wS;
            'P' : tempStllg[c(i)]:=wB;
            'k' : begin
                    tempStllg[c(i)]:=sK;
                    if assigned(Fischerinfo) then FischerInfo.FischerBlackKing := c(i);
                  end;
            'q' : tempStllg[c(i)]:=sD;
            'r' : begin
                    tempStllg[c(i)]:=sT;
                    if assigned(Fischerinfo) then begin
                      if FischerInfo.FischerLeftBlackRook = 0 then FischerInfo.FischerLeftBlackRook := c(i)  {bei FEN lesen kommt immer zuerst der linke Turm}
                      else FischerInfo.FischerRightBlackRook := c(i);
                    end;
                  end;
            'b' : tempStllg[c(i)]:=sL;
            'n' : tempStllg[c(i)]:=sS;
            'p' : tempStllg[c(i)]:=sB;
          end;
        end;
      end
      else begin
        dec(i, 16);  {nchste Reihe}
        inc(Reihe);
      end;
    until (z>=64) or (q>=length(epd));
    if Reihe <> 7 then exit;  {nur dann war es korrekter FEN}
    Stllg := tempStllg;
    inc(q);
    with StllgInfo do begin
      epFeld:=0;
      amZug:=1;
      Rochademoeglichkeiten:=[];
      epSchlag:=LEER;
      UFigur:=LEER;
      geschlFigur:=LEER;
      RochadeArt:=KEINEROCHADE;
      attr1:=0;
      attr2:=0;
      {... b KQkq e3 0 1"}   {zuletzt steht Zugnummer}
      while q<length(epd) do begin
       if not (epd[q] in [' ','0'..'9']) then begin
        if epd[q+1] in ['0'..'9'] then begin
          if  epd[q] + epd[q+1] = FRCKennung then begin  {dann ist es Fischerausgangsstellung (f1 als ep-Feld)}
            if assigned(Fischerinfo) then Fischerinfo.Enabled := true;
            znr := 1;
          end
          else begin
            if amZug=-1 then epFeld:=c(ord(upcase(epd[q]))-65 +16)
            else epFeld:=c(ord(upcase(epd[q]))-65 +40);
          end;
        end
        else begin
          case epd[q] of
            'w' : amZug:=1;
            'b' : amZug:=-1;
            'K' : Rochademoeglichkeiten:=Rochademoeglichkeiten+[wkr];
            'Q' : Rochademoeglichkeiten:=Rochademoeglichkeiten+[wlr];
            'k' : Rochademoeglichkeiten:=Rochademoeglichkeiten+[skr];
            'q' : Rochademoeglichkeiten:=Rochademoeglichkeiten+[slr];
            end;
          end;
        end;
        inc(q);
      end;
      q:=length(epd);
      h:='';
      while (q>0) and (epd[q]<>' ') do begin
        h:=epd[q]+h;
        dec(q);
      end;
      val(h, i, z);
      if z=0 then znr:=i;
    end;
    Result := true;
  except
  end;
end;

function AppendGameToPGN(FileName, Games : string) : Boolean;
Var f : File;
    cswrite : Pointer;
begin
  Result := false;
  if not CheckPGN(FileName) then exit;
  if Games = '' then exit;
  if Games[1] <> #13 then Games := #13#10#13#10 + Games;  {Leerzeile sollte dazwischen}
  GetMem(cswrite, length(Games) + 1);
  StrCopy(cswrite, PChar(Games));
  assignFile(f, FileName);
  reset(f, 1);
  try
    seek(f, FileSize(f));
    BlockWrite(f, cswrite^, length(Games));
    CloseFile(f);
    Result := true;
  finally
    FreeMem(cswrite, length(Games) + 1);
  end;
end;

procedure SetFischerInfo(FischerInfo : PFischerInfo; Stllg : TSchachBrett);
Var i : integer;
begin
  if assigned(FischerInfo) then with FischerInfo^ do begin
    FischerLeftWhiteRook := 0;
    FischerRightWhiteRook := 0;
    FischerLeftBlackRook := 0;
    FischerRightBlackRook := 0;
    enabled := true;
    for i:=0 to 63 do begin
      case Stllg[c(i)] of
        wk : FischerWhiteKing := c(i);
        sk : FischerBlackKing := c(i);
        wt : if FischerLeftWhiteRook = 0 then FischerLeftWhiteRook := c(i)
             else FischerRightWhiteRook := c(i);
        st : if FischerLeftBlackRook = 0 then FischerLeftBlackRook := c(i)
             else FischerRightBlackRook := c(i);
      end;
    end;
  end;
end;

function GetFischerRandomPosition : TSchachbrett;
Var z : integer;
    stllg : TSchachbrett;

function GetSquare(frei : integer) : integer;
Var i : integer;
begin
  z := random(frei);
  i := 0;
  while (cc(a1 + i) < 63) and
  ((i < z) or (stllg[a1 + i] <> LEER)) do inc(i);
  Result := a1 + i;
end;

begin
  stllg := AusgangsStellung;
  For z := 0 to 7 do stllg[a1 + z] := LEER;
  z := random(4);
  stllg[a1 + (z shl 1)] := wL;  {schwarzfeldriger}
  z := random(4);
  stllg[b1 + (z shl 1)] := wL;  {weifeldriger}
  stllg[GetSquare(6)] := wD;
  stllg[GetSquare(5)] := wS;
  stllg[GetSquare(4)] := wS;
  stllg[GetSquare(1)] := wT;
  stllg[GetSquare(1)] := wK;
  stllg[GetSquare(1)] := wT;
  For z := 0 to 7 do stllg[a8 + z] := -stllg[a1 + z];
  Result := stllg;
end;

function GetPiecesNumber(Stellung : TSchachbrett; Zug : TZug) : string;  // gibt Anzahl zurck SaZ (4 + 3)
var i, w, s : integer;
begin
  if Zug.amZug = WEISS then Result := 'WaZ    '
  else Result := 'SaZ    ';
  w := 0;
  s := 0;
  for i := 0 to 63 do begin
    if Stellung[c(i)] > 0 then inc(w);
    if Stellung[c(i)] < 0 then inc(s);
  end;
  Result := Result + '(' + IntToStr(w) + ' + ' +  IntToStr(s) + ')';
end;

function  PosToText(Stellung : TSchachbrett; L_Engl : Boolean) : string;
Var i, j, fnindex, f : integer;
    s : string;
const fn : array[1..12] of string[1] = ('B','T','L','S','D','K','P','R','B','N','Q','K');
begin
  s := '   +---+---+---+---+---+---+---+---+';
  Result := '';
  if L_Engl then fnindex := 6
  else fnindex := 0;
  j := 56;
  while j >= 0 do begin
    Result := Result + #13#10 + s + #13#10 + InttoStr((j shr 3) + 1) + '  |';
    for i := 0 to 7 do begin
      f := Stellung[c(j + i)];
      if f = LEER then Result := Result + '   |'
      else if f < 0 then Result := Result + ' *' +  Fn[abs(f) + fnindex] + '|'
      else Result := Result + ' ' + Fn[abs(f) + fnindex] + ' |';
    end;
    dec(j, 8);
  end;
  Result := Result + #13#10 + s + #13#10 + '     a   b   c   d   e   f   g   h ';
end;


function sucheDame(wo, az : integer; stllg : PSchachbrett; l : TList) : integer;
Var i, j : integer;
begin
  Result:=-1;
  try
    if l = Nil then exit;
    l.Clear;
    for i:=1 to 8 do begin
      j:=wo;
      repeat
        inc(j, DZ[i]);
        if Stllg^[j]=aZ*wD then l.Add(Pointer(j));
      until Stllg^[j]<>LEER;
    end;
    Result:=0;
  except
  end;
end;

function sucheTurm(wo, az : integer; stllg : PSchachbrett; l : TList) : integer;
Var i, j : integer;
begin
  Result:=-1;
  try
    if l = Nil then exit;
    l.Clear;
    for i:=1 to 4 do begin
      j:=wo;
      repeat
        inc(j, TZ[i]);
        if Stllg^[j]=aZ*wT then l.Add(Pointer(j));
      until Stllg^[j]<>LEER;
    end;
    Result:=0;
  except
  end;
end;

function sucheLaeufer(wo, az : integer; stllg : PSchachbrett; l : TList) : integer;
Var i, j : integer;
begin
  Result:=-1;
  try
    if l = Nil then exit;
    l.Clear;
    for i:=1 to 4 do begin
      j:=wo;
      repeat
        inc(j, LZ[i]);
        if Stllg^[j]=aZ*wL then l.Add(Pointer(j));
      until Stllg^[j]<>LEER;
    end;
    Result:=0;
  except
  end;
end;

function sucheSpringer(wo, az : integer; stllg : PSchachbrett; l : TList) : integer;
Var i, j : integer;
begin
  Result:=-1;
  try
    if l = Nil then exit;
    l.Clear;
    for i:=1 to 8 do begin
      j:=wo;
      inc(j, SZ[i]);
      if Stllg^[j]=aZ*wS then l.Add(Pointer(j));
    end;
    Result:=0;
  except
  end;
end;

function sucheKoenig(wo, az : integer; stllg : PSchachbrett; l : TList) : integer;
Var i, j: integer;
begin
  Result:=-1;
  try
    if l = Nil then exit;
    l.Clear;
    for i:=1 to 8 do begin
      j:=wo;
      inc(j, KZ[i]);
      if Stllg^[j]=aZ*wK then l.Add(Pointer(j));
    end;
    Result:=0;  // Rochade wird hier ignoriert - ist nur fr Zugvorschlag
  except
  end;
end;

function sucheSchlagBauer(wo, az : integer; stllg : PSchachbrett; l : TList) : integer;
begin
  Result:=-1;
  try
    if l = Nil then exit;
    l.Clear;
    if Stllg^[wo+(-aZ*9)]=aZ*wB then l.Add(Pointer(wo+(-aZ*9)));
    if Stllg^[wo+(-aZ*11)]=aZ*wB then l.Add(Pointer(wo+(-aZ*11)));
    Result:=0;
  except
  end;
end;

function sucheBauer(wo, az : integer; stllg : PSchachbrett; l : TList) : integer;
begin
  Result:=-1;
  try
    if l = Nil then exit;
    l.Clear;
    if Stllg^[wo+(-aZ*10)]=aZ*wB then l.Add(Pointer(wo+(-aZ*10)))
    else if Stllg^[wo+(-aZ*10)]=LEER then begin
      if ((az = WEISS) and ((wo > h3) and (wo < a5)))
      or ((az <> WEISS) and ((wo > h4) and (wo < a6))) then
        if Stllg^[wo+(-aZ*20)]=aZ*wB then l.Add(Pointer(wo+(-aZ*20)));
    end;
    Result:=0;
  except
  end;
end;

function bedroht(wo, az : Integer; Stllg : PSchachbrett) : Boolean;
Var l : TList;
begin
  Result := true;
  try
    l := TList.Create;
    try
      sucheDame(wo, -az, stllg, l);
      if l.Count <> 0 then exit;
      sucheTurm(wo, -az, stllg, l);
      if l.Count <> 0 then exit;
      sucheLaeufer(wo, -az, stllg, l);
      if l.Count <> 0 then exit;
      sucheSpringer(wo, -az, stllg, l);
      if l.Count <> 0 then exit;
      sucheSchlagBauer(wo, -az, stllg, l);
      if l.Count <> 0 then exit;
      sucheKoenig(wo, -az, stllg, l);
      if l.Count <> 0 then exit;
      Result := false;
    finally
      l.Free;
    end;
  except
  end;
end;

function GetQueenSquares(wo, az : integer; stllg : PSchachbrett; l : TList) : integer;
Var i, j : integer;
begin
  Result:=-1;
  try
    if l = Nil then exit;
    l.Clear;
    for i:=1 to 8 do begin
      j:=wo;
      repeat
        inc(j, DZ[i]);
        if Stllg^[j]=RAND then break;
        if (az = WEISS) and (Stllg^[j] > 0) then break;
        if (az <> WEISS) and (Stllg^[j] < 0) then break;
        l.Add(Pointer(j));
      until Stllg^[j]<>LEER;
    end;
    Result:=0;
  except
  end;
end;

function GetRookSquares(wo, az : integer; stllg : PSchachbrett; l : TList) : integer;
Var i, j : integer;
begin
  Result:=-1;
  try
    if l = Nil then exit;
    l.Clear;
    for i:=1 to 4 do begin
      j:=wo;
      repeat
        inc(j, TZ[i]);
        if Stllg^[j]=RAND then break;
        if (az = WEISS) and (Stllg^[j] > 0) then break;
        if (az <> WEISS) and (Stllg^[j] < 0) then break;
        l.Add(Pointer(j));
      until Stllg^[j]<>LEER;
    end;
    Result:=0;
  except
  end;
end;

function GetBishopSquares(wo, az : integer; stllg : PSchachbrett; l : TList) : integer;
Var i, j : integer;
begin
  Result:=-1;
  try
    if l = Nil then exit;
    l.Clear;
    for i:=1 to 4 do begin
      j:=wo;
      repeat
        inc(j, LZ[i]);
        if Stllg^[j]=RAND then break;
        if (az = WEISS) and (Stllg^[j] > 0) then break;
        if (az <> WEISS) and (Stllg^[j] < 0) then break;
        l.Add(Pointer(j));
      until Stllg^[j]<>LEER;
    end;
    Result:=0;
  except
  end;
end;

function GetKnightSquares(wo, az : integer; stllg : PSchachbrett; l : TList) : integer;
Var i, j : integer;
begin
  Result:=-1;
  try
    if l = Nil then exit;
    l.Clear;
    for i:=1 to 8 do begin
      j:=wo;
      inc(j, SZ[i]);
      if Stllg^[j]<>RAND then begin
        if (Stllg^[j]=LEER) or
        ((az = WEISS) and (Stllg^[j] < 0)) or
        ((az <> WEISS) and (Stllg^[j] > 0)) then
          l.Add(Pointer(j));
      end;
    end;
    Result:=0;
  except
  end;
end;

function GetKingSquares(wo : integer; Z : PZug; stllg : PSchachbrett; l : TList) : integer;
Var i, j, az: integer;
begin
  Result:=-1;
  try
    if l = Nil then exit;
    l.Clear;
    az := Z^.amZug;
    for i:=1 to 8 do begin
      j:=wo;
      inc(j, KZ[i]);
      if Stllg^[j]<>RAND then begin
        if (Stllg^[j]=LEER) or
        ((az = WEISS) and (Stllg^[j] < 0)) or
        ((az <> WEISS) and (Stllg^[j] > 0)) then
          l.Add(Pointer(j));
      end;
    end;
    if az = WEISS then begin           // !bercksichtigt NICHT Fischerrochaden!
      if wkr in z.Rochademoeglichkeiten then begin
        if  ((Stllg^[e1]=wK) and (Stllg^[h1]=wT))
        and ((Stllg^[f1]=LEER) and (not bedroht(f1, az, Stllg)))
        and ((Stllg^[g1]=LEER) and (not bedroht(g1, az, Stllg)))
        and (not bedroht(e1, az, Stllg))
          then l.Add(Pointer(g1));
      end;
      if wlr in z.Rochademoeglichkeiten then begin
        if  ((Stllg^[e1]=wK) and (Stllg^[a1]=wT))
        and (Stllg^[b1]=LEER)
        and ((Stllg^[d1]=LEER) and (not bedroht(d1, az, Stllg)))
        and ((Stllg^[c1]=LEER) and (not bedroht(c1, az, Stllg)))
        and (not bedroht(e1, az, Stllg))
          then l.Add(Pointer(c1));
      end;
    end
    else begin
      if skr in z.Rochademoeglichkeiten then begin
        if  ((Stllg^[e8]=sK) and (Stllg^[h8]=sT))
        and ((Stllg^[f8]=LEER) and (not bedroht(f8, az, Stllg)))
        and ((Stllg^[g8]=LEER) and (not bedroht(g8, az, Stllg)))
        and (not bedroht(e8, az, Stllg))
          then l.Add(Pointer(g8));
      end;
      if slr in z.Rochademoeglichkeiten then begin
        if  ((Stllg^[e8]=sK) and (Stllg^[a8]=sT))
        and (Stllg^[b8]=LEER)
        and ((Stllg^[d8]=LEER) and (not bedroht(d8, az, Stllg)))
        and ((Stllg^[c8]=LEER) and (not bedroht(c8, az, Stllg)))
        and (not bedroht(e8, az, Stllg))
          then l.Add(Pointer(c8));
      end;
    end;
    Result:=0;
  except
  end;
end;

function GetPawnSquares(wo : integer; z : PZug; stllg : PSchachbrett; l : TList) : integer;
Var az : Integer;
begin
  Result:=-1;
  try
    if l = Nil then exit;
    l.Clear;
    az := Z^.amZug;
    if Stllg^[wo+(aZ*10)]=LEER then begin
      l.Add(Pointer(wo+(aZ*10)));
      if ((az = WEISS) and ((wo > h1) and (wo < a3)))
      or ((az <> WEISS) and ((wo > h6) and (wo < a8))) then
        if Stllg^[wo+(aZ*20)]=LEER then l.Add(Pointer(wo+(aZ*20)));
    end;
    if az = WEISS then begin
      if ((Stllg^[wo+(aZ*9)]<>RAND) and (Stllg^[wo+(aZ*9)]< 0))
      or (wo+(aZ*9) = z.epFeld)
        then l.Add(Pointer(wo+(aZ*9)));
      if ((Stllg^[wo+(aZ*11)]<>RAND) and (Stllg^[wo+(aZ*11)]< 0))
      or (wo+(aZ*11) = z.epFeld)
        then l.Add(Pointer(wo+(aZ*11)));
    end
    else begin
      if ((Stllg^[wo+(aZ*9)]<>RAND) and (Stllg^[wo+(aZ*9)]> 0))
      or (wo+(aZ*9) = z.epFeld)
        then l.Add(Pointer(wo+(aZ*9)));
      if ((Stllg^[wo+(aZ*11)]<>RAND) and (Stllg^[wo+(aZ*11)]> 0))
      or (wo+(aZ*11) = z.epFeld)
        then l.Add(Pointer(wo+(aZ*11)));
    end;
    Result:=0;
  except
  end;
end;

function GetKing(white : Boolean; stllg : PSchachbrett) : integer;
Var i, j : integer;
begin
  Result := -1;
  try
    if white then j := wK else j := sK;
    for i := 0 to 63 do begin
      if Stllg^[c(i)] = j then begin
        Result := c(i);
        exit;
      end;
    end;
  except
  end;
end;

function fiavailable(f : PFischerInfo) : Boolean;
begin
  Result := false;
  if f = Nil then exit;
  if f.enabled then Result := true;
end;

function GetKingSquaresFi(wo : integer; Z : PZug; stllg : PSchachbrett; l : TList; fi : PFischerInfo) : integer;
Var i, j, az: integer;
    moeglich : Boolean;
begin
  Result:=-1;
  try
    if l = Nil then exit;
    l.Clear;
    az := Z^.amZug;
    for i:=1 to 8 do begin
      j:=wo;
      inc(j, KZ[i]);
      if Stllg^[j]<>RAND then begin
        if (Stllg^[j]=LEER) or
        ((az = WEISS) and (Stllg^[j] < 0)) or
        ((az <> WEISS) and (Stllg^[j] > 0)) then
          l.Add(Pointer(j));
      end;
    end;
    if fiavailable(fi) then with fi^ do begin
      if az = WEISS then begin           // !bercksichtigt NICHT Fischerrochaden!
        moeglich := false;
        if (wo = FischerWhiteKing) AND
        (not bedroht(FischerWhiteKing, Z.amZug, Stllg)) then BEGIN
          if wkr in z.Rochademoeglichkeiten then begin
            if (Stllg[FischerRightWhiteRook] = wT) AND
            (
              (Stllg[f1] = LEER) or (f1 = FischerRightWhiteRook) or (f1 = FischerWhiteKing)
            ) AND
            (
              (Stllg[g1] = LEER) or (g1 = FischerRightWhiteRook) or (g1 = FischerWhiteKing)
            )
            then begin
              moeglich := true;
              j := FischerWhiteKing + 1;
              while (j <= FischerRightWhiteRook) and (j < H1) do begin
                if not (
                  ((Stllg[j] = LEER) or (j = FischerRightWhiteRook)) and
                  (not bedroht(j, z.amZug, Stllg))
                )
                then begin
                  moeglich := false;
                  break;
                end;
                inc(j);
              end;
              if moeglich then
                l.Add(Pointer(-g1));
            end;
          end;
          if wlr in z.Rochademoeglichkeiten then begin
            if (Stllg[FischerLeftWhiteRook] = wT) AND
            (
              (Stllg[d1] = LEER) or (d1 = FischerLeftWhiteRook) or (d1 = FischerWhiteKing)
            ) and
            (
              (Stllg[c1] = LEER) or (c1 = FischerLeftWhiteRook) or (c1 = FischerWhiteKing)
            )
            then begin
              moeglich := true;
              if FischerWhiteKing = B1 then j := C1
              else j := FischerWhiteKing - 1;
              while (j >= FischerLeftWhiteRook) and (j > B1) do begin
                if not (
                  ((Stllg[j] = LEER) or (j = FischerLeftWhiteRook)) and
                  (not bedroht(j, z.amZug, Stllg))
                )
                then begin
                  moeglich := false;
                  break;
                end;
                dec(j);
              end;
              if moeglich then
                l.Add(Pointer(-c1));
            end;
          end;
        end;
      end
      else begin
        moeglich := false;
        if (wo = FischerBlackKing) AND
        (not bedroht(FischerBlackKing, Z.amZug, Stllg)) then BEGIN
          if skr in z.Rochademoeglichkeiten then begin
            if (Stllg[FischerRightBlackRook] = sT) AND
            (
              (Stllg[f8] = LEER) or (f8 = FischerRightBlackRook) or (f8 = FischerBlackKing)
            ) AND
            (
              (Stllg[g8] = LEER) or (g8 = FischerRightBlackRook) or (g8 = FischerBlackKing)
            )
            then begin
              moeglich := true;
              j := FischerBlackKing + 1;
              while (j <= FischerRightBlackRook) and (j < H8) do begin
                if not (
                  ((Stllg[j] = LEER) or (j = FischerRightBlackRook)) and
                  (not bedroht(j, z.amZug, Stllg))
                )
                then begin
                  moeglich := false;
                  break;
                end;
                inc(j);
              end;
              if moeglich then
                l.Add(Pointer(-g8));
            end;
          end;
          if slr in z.Rochademoeglichkeiten then begin
            if (Stllg[FischerLeftBlackRook] = sT) AND
            (
              (Stllg[d8] = LEER) or (d8 = FischerLeftBlackRook) or (d8 = FischerBlackKing)
            ) and
            (
              (Stllg[c8] = LEER) or (c8 = FischerLeftBlackRook) or (c8 = FischerBlackKing)
            )
            then begin
              moeglich := true;
              if FischerBlackKing = B8 then j := C8
              else j := FischerBlackKing - 1;
              while (j >= FischerLeftBlackRook) and (j > B8) do begin
                if not (
                  ((Stllg[j] = LEER) or (j = FischerLeftBlackRook)) and
                  (not bedroht(j, z.amZug, Stllg))
                )
                then begin
                  moeglich := false;
                  break;
                end;
                dec(j);
              end;
              if moeglich then
                l.Add(Pointer(-c8));
            end;
          end;
        end;
      end;
    end
    else begin
      if az = WEISS then begin           // !bercksichtigt NICHT Fischerrochaden!
        if wkr in z.Rochademoeglichkeiten then begin
          if  ((Stllg^[e1]=wK) and (Stllg^[h1]=wT))
          and ((Stllg^[f1]=LEER) and (not bedroht(f1, az, Stllg)))
          and ((Stllg^[g1]=LEER) and (not bedroht(g1, az, Stllg)))
          and (not bedroht(e1, az, Stllg))
            then l.Add(Pointer(g1));
        end;
        if wlr in z.Rochademoeglichkeiten then begin
          if  ((Stllg^[e1]=wK) and (Stllg^[a1]=wT))
          and (Stllg^[b1]=LEER)
          and ((Stllg^[d1]=LEER) and (not bedroht(d1, az, Stllg)))
          and ((Stllg^[c1]=LEER) and (not bedroht(c1, az, Stllg)))
          and (not bedroht(e1, az, Stllg))
            then l.Add(Pointer(c1));
        end;
      end
      else begin
        if skr in z.Rochademoeglichkeiten then begin
          if  ((Stllg^[e8]=sK) and (Stllg^[h8]=sT))
          and ((Stllg^[f8]=LEER) and (not bedroht(f8, az, Stllg)))
          and ((Stllg^[g8]=LEER) and (not bedroht(g8, az, Stllg)))
          and (not bedroht(e8, az, Stllg))
            then l.Add(Pointer(g8));
        end;
        if slr in z.Rochademoeglichkeiten then begin
          if  ((Stllg^[e8]=sK) and (Stllg^[a8]=sT))
          and (Stllg^[b8]=LEER)
          and ((Stllg^[d8]=LEER) and (not bedroht(d8, az, Stllg)))
          and ((Stllg^[c8]=LEER) and (not bedroht(c8, az, Stllg)))
          and (not bedroht(e8, az, Stllg))
            then l.Add(Pointer(c8));
        end;
      end;
    end;
    Result:=0;
  except
  end;
end;


function IsMovePossible(Astllg : PSchachbrett; Z : PZug; fi : PFischerInfo) : Boolean; { untersttzt nicht Fischerrochade }
Var i, f : integer;
    l : TList;
    stllg : TSchachbrett;
    zug : TZug;
begin
  Result := false;
  try
    l := TList.Create;
    try
      stllg := AStllg^;
      Zug := Z^;
      f := stllg[Zug.von];
      if (f = LEER) or (f = RAND) then exit;
      if (Zug.amZug = WEISS) and (f < 0) then exit;
      if (Zug.amZug <> WEISS) and (f > 0) then exit;
      case abs(f) of
        wD : GetQueenSquares(Zug.von, Zug.amZug, @Stllg, l);
        wT : GetRookSquares(Zug.von, Zug.amZug, @Stllg, l);
        wL : GetBishopSquares(Zug.von, Zug.amZug, @Stllg, l);
        wS : GetKnightSquares(Zug.von, Zug.amZug, @Stllg, l);
        wK : begin
               if fi = Nil then GetKingSquares(Zug.von, @Zug, @Stllg, l)
               else begin
                 GetKingSquaresFi(Zug.von, @Zug, @Stllg, l, fi);
                 if fiavailable(fi) then
                   for i := 0 to l.Count - 1 do
                     if longint(l[i]) < 0 then
                       l[i] := Pointer(abs(longint(l[i])));
                end;
              end;
        wB : GetPawnSquares(Zug.von, @Zug, @Stllg, l);
      end;
      if l.IndexOf(Pointer(Zug.Nach)) <> -1 then begin
        if Zug.Von <> Zug.nach then begin // kann bei firoch. auftreten
          Stllg[Zug.nach] := Stllg[Zug.Von];
          Stllg[Zug.Von] := LEER;
        end;  
        if (abs(f) = wB) and (Zug.epFeld = Zug.nach) then begin
          Stllg[Zug.nach +(-Zug.amZug*10)] := LEER;
        end;
        i := GetKing(Zug.amZug = WEISS, @Stllg);
        if i <> -1 then Result := not bedroht(i, Zug.amZug, @Stllg);
      end;
    finally
      l.Free;
    end;
  except
  end;
end;

function GetAllMoves(Astllg : PSchachbrett; Z : PZug; sl : TStringlist) : integer; { untersttzt nicht Fischerrochade }
Var i, j, k, v, w, n, f : integer;
    l : TList;
    stllg : TSchachbrett;
    zug : TZug;
begin
  Result := 0;
  try
    l := TList.Create;
    try
      Zug := Z^;
      stllg := AStllg^;
      i := GetKing(Zug.amZug = WEISS, @Stllg);
      if i <> -1 then
      For j := 0 to 63 do begin
        v := c(j);
        f := stllg[v];
        if (not ((f = LEER) or (f = RAND)))
        and (not ((Zug.amZug = WEISS) and (f < 0)))
        and (not ((Zug.amZug <> WEISS) and (f > 0))) then begin
          case abs(f) of
            wD : GetQueenSquares(v, Zug.amZug, @Stllg, l);
            wT : GetRookSquares(v, Zug.amZug, @Stllg, l);
            wL : GetBishopSquares(v, Zug.amZug, @Stllg, l);
            wS : GetKnightSquares(v, Zug.amZug, @Stllg, l);
            wK : GetKingSquares(v, @Zug, @Stllg, l);
            wB : GetPawnSquares(v, @Zug, @Stllg, l);
          end;
          for k := 0 to l.Count - 1 do begin
            n := longint(l[k]);
            Stllg[n] := Stllg[v];
            Stllg[v] := LEER;
            if (abs(f) = wB) and (Zug.epFeld = n) then
              Stllg[n+(-Zug.amZug*10)] := LEER;
            if abs(f) = wK then w := n
            else w := i;
            if not bedroht(w, Zug.amZug, @Stllg) then begin
              if sl <> Nil then sl.AddObject(IntToStr(c(j)), l[k]);
              inc(Result);
              if (abs(f) = wB) and ((n < a2) or (n > h7)) then  // Umwandlung
                inc(Result, 3);
            end;
            stllg := AStllg^;
          end;
        end;
      end;
    finally
      l.Free;
    end;
  except
    Result := -1;
  end;
end;

function IsMate(Astllg : PSchachbrett; Z : PZug) : Boolean; { untersttzt nicht Fischerrochade }
begin
  Result := false;
  try
    Result := GetAllMoves(Astllg, Z, Nil) = 0;
    if Result then
      Result := bedroht(GetKing(Z^.amZug = WEISS, AStllg), Z^.amZug, AStllg);
  except
  end;
end;

function GetGameState(Astllg : PSchachbrett; Z : PZug) : TGameState; { untersttzt nicht Fischerrochade }
begin
  Result := gsRun;
  try
    if GetAllMoves(Astllg, Z, Nil) = 0 then Result := gsMate;
    if Result = gsMate then begin
      if not bedroht(GetKing(Z^.amZug = WEISS, AStllg), Z^.amZug, AStllg)
        then Result := gsPatt;
    end
    else begin
      if (not hasEnoughMaterial(Astllg, true)) and (not hasEnoughMaterial(Astllg, false))
        then Result := gsDraw;
    end;
  except
  end;
end;

function hasEnoughMaterial(Astllg : PSchachbrett; White : Boolean) : Boolean;
Var i, q, r, b, n, p, bq, br, bb, bn, bp : Integer;
    BlackBishopSquareColorIsWhite, WhiteBishopSquareColorIsWhite : Boolean;

  function onlyWhiteKing : Boolean;
  begin
    Result := (q = 0) and (r = 0) and (b = 0) and (n = 0) and (p = 0);
  end;

  function onlyBlackKing : Boolean;
  begin
    Result := (bq = 0) and (br = 0) and (bb = 0) and (bn = 0) and (bp = 0);
  end;

  // genau eine Figur

  function oneWhiteQueen : Boolean;
  begin
    Result := (q = 1) and (r = 0) and (b = 0) and (n = 0) and (p = 0);
  end;

  function oneBlackQueen : Boolean;
  begin
    Result := (bq = 1) and (br = 0) and (bb = 0) and (bn = 0) and (bp = 0);
  end;

  function oneWhiteRook : Boolean;
  begin
    Result := (q = 0) and (r = 1) and (b = 0) and (n = 0) and (p = 0);
  end;

  function oneBlackRook : Boolean;
  begin
    Result := (bq = 0) and (br = 1) and (bb = 0) and (bn = 0) and (bp = 0);
  end;

  function oneWhiteBishop : Boolean;
  begin
    Result := (q = 0) and (r = 0) and (b = 1) and (n = 0) and (p = 0);
  end;

  function oneBlackBishop : Boolean;
  begin
    Result := (bq = 0) and (br = 0) and (bb = 1) and (bn = 0) and (bp = 0);
  end;

  function oneWhiteKnight : Boolean;
  begin
    Result := (q = 0) and (r = 0) and (b = 0) and (n = 1) and (p = 0);
  end;

  function oneBlackKnight : Boolean;
  begin
    Result := (bq = 0) and (br = 0) and (bb = 0) and (bn = 1) and (bp = 0);
  end;

  function oneWhitePawn : Boolean;
  begin
    Result := (q = 0) and (r = 0) and (b = 0) and (n = 0) and (p = 1);
  end;

  function oneBlackPawn : Boolean;
  begin
    Result := (bq = 0) and (br = 0) and (bb = 0) and (bn = 0) and (bp = 1);
  end;

  // mindestens eine Figur

  function WhiteQueen : Boolean;
  begin
    Result := q > 0;
  end;

  function BlackQueen : Boolean;
  begin
    Result := bq > 0;
  end;

  function WhiteRook : Boolean;
  begin
    Result := r > 0;
  end;

  function BlackRook : Boolean;
  begin
    Result := br > 0;
  end;

  function WhiteBishop : Boolean;
  begin
    Result := b > 0;
  end;

  function BlackBishop : Boolean;
  begin
    Result := bb > 0;
  end;

  function WhiteKnight : Boolean;
  begin
    Result := n > 0;
  end;

  function BlackKnight : Boolean;
  begin
    Result := bn > 0;
  end;

  function WhitePawn : Boolean;
  begin
    Result := p > 0;
  end;

  function BlackPawn : Boolean;
  begin
    Result := bp > 0;
  end;

  // mind 2 Bauern

  function atLeastTwoWhitePawns : Boolean;
  begin
    Result := p > 1;
  end;

  function atLeastTwoBlackPawns : Boolean;
  begin
    Result := bp > 1;
  end;

  // 2 Springer

  function twoWhiteKnights : Boolean;
  begin
    Result := (q = 0) and (r = 0) and (b = 0) and (n = 2) and (p = 0);
  end;

  function twoBlackKnights : Boolean;
  begin
    Result := (bq = 0) and (br = 0) and (bb = 0) and (bn = 2) and (bp = 0);
  end;

  // Randbauern

  function whiteAPawn : Boolean;
  begin
    Result := (Astllg^[a2] = wb) or (Astllg^[a3] = wb) or (Astllg^[a4] = wb) or
              (Astllg^[a5] = wb) or (Astllg^[a6] = wb) or (Astllg^[a7] = wb);
  end;

  function blackAPawn : Boolean;
  begin
    Result := (Astllg^[a2] = sb) or (Astllg^[a3] = sb) or (Astllg^[a4] = sb) or
              (Astllg^[a5] = sb) or (Astllg^[a6] = sb) or (Astllg^[a7] = sb);
  end;

  function whiteHPawn : Boolean;
  begin
    Result := (Astllg^[h2] = wb) or (Astllg^[h3] = wb) or (Astllg^[h4] = wb) or
              (Astllg^[h5] = wb) or (Astllg^[h6] = wb) or (Astllg^[h7] = wb);
  end;

  function blackHPawn : Boolean;
  begin
    Result := (Astllg^[h2] = sb) or (Astllg^[h3] = sb) or (Astllg^[h4] = sb) or
              (Astllg^[h5] = sb) or (Astllg^[h6] = sb) or (Astllg^[h7] = sb);
  end;

begin
  Result := true;
  try
    q := 0; r := 0; b := 0; n := 0; p := 0;
    bq := 0; br := 0; bb := 0; bn := 0; bp := 0;
    For i := 0 to 63 do begin
      case Astllg^[c(i)] of
        wd : inc(q);
        wt : inc(r);
        wl : begin
               inc(b);
               WhiteBishopSquareColorIsWhite := FeldFarbe[i] = 0;
             end;
        ws : inc(n);
        wb : inc(p);
        sd : inc(bq);
        st : inc(br);
        sl : begin
               inc(bb);
               BlackBishopSquareColorIsWhite := FeldFarbe[i] = 0;
             end;
        ss : inc(bn);
        sb : inc(bp);
      end;
    end;
    if White then begin
      if onlyWhiteKing then
        Result := false
      else if oneWhiteKnight then begin
        if not (blackAPawn or blackHPawn) then
          Result := false;
      end
      else if oneWhiteBishop then begin
        if not (
        atLeasttwoBlackPawns or
        (blackPawn and blackBishop)// nicht alle Flle bercksichtigt - Matt in 2 nur gegen 2 L auf versch. Farbe
        ) then
          Result := false;
      end
      else if twoWhiteKnights then begin
        if onlyBlackKing or
        ((not BlackPawn) and (not BlackBishop) and (not BlackKnight)) then
          Result := false
      end;
    end
    else begin
      if onlyBlackKing then
        Result := false
      else if oneBlackKnight then begin
        if not (whiteAPawn or whiteHPawn) then
          Result := false;
      end
      else if oneBlackBishop then begin
        if not (
        atLeasttwoWhitePawns or
        (whitePawn and whiteBishop)// nicht alle Flle bercksichtigt - Matt in 2 nur gegen 2 L auf versch. Farbe
        ) then
          Result := false;
      end
      else if twoBlackKnights then begin
        if onlyWhiteKing or
        ((not whitePawn) and (not whiteBishop) and (not whiteKnight)) then
          Result := false
      end;
    end;
  except
  end;
end;

function  GetPropose(Astllg : PSchachbrett; Z : PZug; movepart : integer; fi : PFischerInfo) : integer;
Var f : integer;
    l : TList;
    stllg : TSchachbrett;
    zug : TZug;

function moeglich(firstlast : Boolean) : integer;
Var i : integer;
begin
  Result := 0;
  try
    for i := l.Count - 1 downto 0 do begin
      if firstlast then begin
        Zug.von := movepart;  { kann verndert werden (nur epfeld und Rochadem. nicht)}
        Zug.nach := integer(l[i]);
      end
      else begin
        Zug.von := integer(l[i]);
        Zug.nach := movepart;
      end;
      if IsMovepossible(AStllg, @zug, fi) then begin
        Result := integer(l[i]);
        exit;
      end;
    end;
  except
  end;
end;

begin
  Result := 0;
  try
    l := TList.Create;
    try
      stllg := AStllg^;
      Zug := Z^;
      f := stllg[movepart];
      if f = RAND then exit;
      if ((Zug.amZug = WEISS) and (f > 0))
      or ((Zug.amZug <> WEISS) and (f < 0)) then begin
        case abs(f) of
          wD : GetQueenSquares(movepart, Zug.amZug, @Stllg, l);
          wT : GetRookSquares(movepart, Zug.amZug, @Stllg, l);
          wL : GetBishopSquares(movepart, Zug.amZug, @Stllg, l);
          wS : GetKnightSquares(movepart, Zug.amZug, @Stllg, l);
          wK : GetKingSquares(movepart, @Zug, @Stllg, l);
          wB : GetPawnSquares(movepart, @Zug, @Stllg, l);
        end;
        Result := moeglich(true);
      end
      else begin
        sucheDame(movepart, Zug.amZug, @stllg, l);
        Result := moeglich(false);
        if Result = 0 then begin
          sucheTurm(movepart, Zug.amZug, @stllg, l);
          Result := moeglich(false);
          if Result = 0 then begin
            sucheLaeufer(movepart, Zug.amZug, @stllg, l);
            Result := moeglich(false);
            if Result = 0 then begin
              sucheSpringer(movepart, Zug.amZug, @stllg, l);
              Result := moeglich(false);
              if Result = 0 then begin
                sucheSchlagBauer(movepart, Zug.amZug, @stllg, l);
                Result := moeglich(false);
                if Result = 0 then begin
                  sucheBauer(movepart, Zug.amZug, @stllg, l);
                  Result := moeglich(false);
                  if Result = 0 then begin
                    sucheKoenig(movepart, Zug.amZug, @stllg, l);
                    Result := moeglich(false);
                  end;
                end;
              end;
            end;
          end;
        end;
      end;
    finally
      l.Free;
    end;
  except
  end;
end;

function readFontFile(section, filename : String; var piecesChars : TChessFontChars) : Boolean;
const sInvalid : String = 'xxxxxxxxxx';

  function c(var s : String) : boolean;
  begin
    if trim(s) = '' then s := '';
    if s = sInvalid then Result := false
    else Result := true;
  end;

begin
  Result := false;
  try
    if not FileExists(filename) then exit;
    with piecesChars, TIniFile.Create(filename) do try
      sfRo1   := ReadString(section, 'sfRo1'   ,sInvalid);
      sfRo2   := ReadString(section, 'sfRo2'   ,sInvalid);
      sfRo3   := ReadString(section, 'sfRo3'   ,sInvalid);
      sfRo4   := ReadString(section, 'sfRo4'   ,sInvalid);
      sfRo5   := ReadString(section, 'sfRo5'   ,sInvalid);
      sfRo6   := ReadString(section, 'sfRo6'   ,sInvalid);
      sfRo7   := ReadString(section, 'sfRo7'   ,sInvalid);
      sfRo8   := ReadString(section, 'sfRo8'   ,sInvalid);
      sfRl1   := ReadString(section, 'sfRl1'   ,sInvalid);
      sfRl2   := ReadString(section, 'sfRl2'   ,sInvalid);
      sfRl3   := ReadString(section, 'sfRl3'   ,sInvalid);
      sfRl4   := ReadString(section, 'sfRl4'   ,sInvalid);
      sfRl5   := ReadString(section, 'sfRl5'   ,sInvalid);
      sfRl6   := ReadString(section, 'sfRl6'   ,sInvalid);
      sfRl7   := ReadString(section, 'sfRl7'   ,sInvalid);
      sfRl8   := ReadString(section, 'sfRl8'   ,sInvalid);
      sfRr1   := ReadString(section, 'sfRr1'   ,sInvalid);
      sfRr2   := ReadString(section, 'sfRr2'   ,sInvalid);
      sfRr3   := ReadString(section, 'sfRr3'   ,sInvalid);
      sfRr4   := ReadString(section, 'sfRr4'   ,sInvalid);
      sfRr5   := ReadString(section, 'sfRr5'   ,sInvalid);
      sfRr6   := ReadString(section, 'sfRr6'   ,sInvalid);
      sfRr7   := ReadString(section, 'sfRr7'   ,sInvalid);
      sfRr8   := ReadString(section, 'sfRr8'   ,sInvalid);
      sfRu1   := ReadString(section, 'sfRu1'   ,sInvalid);
      sfRu2   := ReadString(section, 'sfRu2'   ,sInvalid);
      sfRu3   := ReadString(section, 'sfRu3'   ,sInvalid);
      sfRu4   := ReadString(section, 'sfRu4'   ,sInvalid);
      sfRu5   := ReadString(section, 'sfRu5'   ,sInvalid);
      sfRu6   := ReadString(section, 'sfRu6'   ,sInvalid);
      sfRu7   := ReadString(section, 'sfRu7'   ,sInvalid);
      sfRu8   := ReadString(section, 'sfRu8'   ,sInvalid);
      sfElo   := ReadString(section, 'sfElo'   ,sInvalid);
      sfEro   := ReadString(section, 'sfEro'   ,sInvalid);
      sfElu   := ReadString(section, 'sfElu'   ,sInvalid);
      sfEru   := ReadString(section, 'sfEru'   ,sInvalid);
      sfwFl   := ReadString(section, 'sfwFl'   ,sInvalid);
      sfsFl   := ReadString(section, 'sfsFl'   ,sInvalid);
      sfwFwK  := ReadString(section, 'sfwFwK'  ,sInvalid);
      sfwFsK  := ReadString(section, 'sfwFsK'  ,sInvalid);
      sfsFwK  := ReadString(section, 'sfsFwK'  ,sInvalid);
      sfsFsK  := ReadString(section, 'sfsFsK'  ,sInvalid);
      sfwFwD  := ReadString(section, 'sfwFwD'  ,sInvalid);
      sfwFsD  := ReadString(section, 'sfwFsD'  ,sInvalid);
      sfsFwD  := ReadString(section, 'sfsFwD'  ,sInvalid);
      sfsFsD  := ReadString(section, 'sfsFsD'  ,sInvalid);
      sfwFwT  := ReadString(section, 'sfwFwT'  ,sInvalid);
      sfwFsT  := ReadString(section, 'sfwFsT'  ,sInvalid);
      sfsFwT  := ReadString(section, 'sfsFwT'  ,sInvalid);
      sfsFsT  := ReadString(section, 'sfsFsT'  ,sInvalid);
      sfwFwL  := ReadString(section, 'sfwFwL'  ,sInvalid);
      sfwFsL  := ReadString(section, 'sfwFsL'  ,sInvalid);
      sfsFwL  := ReadString(section, 'sfsFwL'  ,sInvalid);
      sfsFsL  := ReadString(section, 'sfsFsL'  ,sInvalid);
      sfwFwS  := ReadString(section, 'sfwFwS'  ,sInvalid);
      sfwFsS  := ReadString(section, 'sfwFsS'  ,sInvalid);
      sfsFwS  := ReadString(section, 'sfsFwS'  ,sInvalid);
      sfsFsS  := ReadString(section, 'sfsFsS'  ,sInvalid);
      sfwFwB  := ReadString(section, 'sfwFwB'  ,sInvalid);
      sfwFsB  := ReadString(section, 'sfwFsB'  ,sInvalid);
      sfsFwB  := ReadString(section, 'sfsFwB'  ,sInvalid);
      sfsFsB  := ReadString(section, 'sfsFsB'  ,sInvalid);

      Result :=                                 
      c(sfRo1) and
      c(sfRo2) and
      c(sfRo3) and
      c(sfRo4) and 
      c(sfRo5) and 
      c(sfRo6) and 
      c(sfRo7) and 
      c(sfRo8) and 
      c(sfRl1) and 
      c(sfRl2) and 
      c(sfRl3) and 
      c(sfRl4) and 
      c(sfRl5) and 
      c(sfRl6) and 
      c(sfRl7) and 
      c(sfRl8) and 
      c(sfRr1) and
      c(sfRr2) and
      c(sfRr3) and 
      c(sfRr4) and 
      c(sfRr5) and 
      c(sfRr6) and 
      c(sfRr7) and 
      c(sfRr8) and 
      c(sfRu1) and 
      c(sfRu2) and 
      c(sfRu3) and 
      c(sfRu4) and 
      c(sfRu5) and 
      c(sfRu6) and 
      c(sfRu7) and 
      c(sfRu8) and
      c(sfElo) and
      c(sfEro) and 
      c(sfElu) and 
      c(sfEru) and 
      c(sfwFl) and 
      c(sfsFl) and 
      c(sfwFwK) and
      c(sfwFsK) and
      c(sfsFwK) and
      c(sfsFsK) and
      c(sfwFwD) and
      c(sfwFsD) and
      c(sfsFwD) and
      c(sfsFsD) and
      c(sfwFwT) and
      c(sfwFsT) and
      c(sfsFwT) and
      c(sfsFsT) and
      c(sfwFwL) and
      c(sfwFsL) and
      c(sfsFwL) and
      c(sfsFsL) and
      c(sfwFwS) and
      c(sfwFsS) and
      c(sfsFwS) and
      c(sfsFsS) and
      c(sfwFwB) and
      c(sfwFsB) and
      c(sfsFwB) and
      c(sfsFsB);

    finally
      free;
    end;
  except
  end;
end;

function PostoSFDia(stllg : TSchachbrett; sFontName : string; fontfile : String) : string;
Var sfChars : PChessfontChars;
    k, l : integer;
    userChars : TChessFontChars;

function Choose(sw, ss : string) : string;
begin
  if FeldFarbe[k] = 0 then Result := sw
  else Result := ss;
end;

function GetLeftBorder : string;
begin
  Result := '';
  with sfChars^ do case l of
    56 : Result := sfRl8;
    48 : Result := sfRl7;
    40 : Result := sfRl6;
    32 : Result := sfRl5;
    24 : Result := sfRl4;
    16 : Result := sfRl3;
     8 : Result := sfRl2;
     0 : Result := sfRl1;
  end;
end;

function GetRightBorder : string;
begin
  Result := '';
  with sfChars^ do case l of
    56 : Result := sfRr8;
    48 : Result := sfRr7;
    40 : Result := sfRr6;
    32 : Result := sfRr5;
    24 : Result := sfRr4;
    16 : Result := sfRr3;
     8 : Result := sfRr2;
     0 : Result := sfRr1;
  end;
end;

begin
   Result := '';
   if trim(uppercase(sFontName)) = trim(uppercase(TSchachfonts[0])) then sfChars := @sfChars1Echecs
   else if sFontName = TSchachfonts[1] then sfChars := @sfCharsDiagramTTHabsburg
   else if sFontName = TSchachfonts[2] then sfChars := @sfCharsChess_Berlin
   else if sFontName = TSchachfonts[3] then sfChars := @sfCharsChess_Merida
//   else ...
   else begin
     if readFontFile(trim(sFontName), fontfile, userChars) then
       sfChars := @userChars
     else
       sfChars := @sfChars1Echecs;
   end;
   with sfChars^ do begin
      Result := sfElo + sfRo1 + sfRo2 + sfRo3 + sfRo4 + sfRo5 + sfRo6 + sfRo7 + sfRo8 + sfEro + noReturn;
      l := 56;
      repeat
        Result := Result + GetLeftBorder;
        for k := l to l + 7 do begin
          case Stllg[c(k)] of
            LEER: Result := Result + Choose(sfwFl, sfsFl);
            wk  : Result := Result + Choose(sfwFwK, sfsFwK);
            wd  : Result := Result + Choose(sfwFwD, sfsFwD);
            wt  : Result := Result + Choose(sfwFwT, sfsFwT);
            wl  : Result := Result + Choose(sfwFwL, sfsFwL);
            ws  : Result := Result + Choose(sfwFwS, sfsFwS);
            wb  : Result := Result + Choose(sfwFwB, sfsFwB);
            sk  : Result := Result + Choose(sfwFsK, sfsFsK);
            sd  : Result := Result + Choose(sfwFsD, sfsFsD);
            st  : Result := Result + Choose(sfwFsT, sfsFsT);
            sl  : Result := Result + Choose(sfwFsL, sfsFsL);
            ss  : Result := Result + Choose(sfwFsS, sfsFsS);
            sb  : Result := Result + Choose(sfwFsB, sfsFsB);
          end;     
        end;
        Result := Result + GetRightBorder + noReturn;
        dec(l, 8);
      until l < 0;
      Result := Result + sfElu + sfRu1 + sfRu2 + sfRu3 + sfRu4 + sfRu5 + sfRu6 + sfRu7 + sfRu8 + sfEru;
  end;
end;

function FENtoSFDia(AFEN : string; sFontName : string; fontfile : String) : string;
Var znr : word;
    Stllg : TSchachbrett;
    zug : TZug;
begin
  Result := '';
  if not FENtoPos(AFen, stllg, zug, znr, Nil) then exit;
  Result := PostoSFDia(stllg, sFontName, fontfile);
end;

function maybeFischerPos(b : TSchachbrett) : Boolean;
Var i, j : Integer;
    equal, ktchanged : Boolean;
begin
  Result := false;
  equal := true;
  ktchanged := false;
  for i := 0 to 63 do begin
    j := c(i);
    equal := equal and (b[j] = Ausgangsstellung[j]);
    if j = e1 then
      if b[j] <> wK then
        ktchanged := true;
    if (j = a1) or (j = h1) then
      if b[j] <> wT then
        ktchanged := true;
    if j = e8 then
      if b[j] <> sK then
        ktchanged := true;
    if (j = a8) or (j = h8) then
      if b[j] <> sT then
        ktchanged := true;
    if i < 8 then begin
      if not (b[j] > 0) then
        exit;
    end
    else if (i > 7) and (i < 16) then begin
      if b[j] <> wB then
        exit;
    end
    else if (i > 15) and (i < 48) then begin
      if b[j] <> 0 then
        exit;
    end
    else if (i > 47) and (i < 56) then begin
      if b[j] <> sB then
        exit;
    end
    else begin
      if not (b[j] < 0) then
        exit;
    end;
  end;
  if (equal) or (not ktchanged) then
    Result := false
  else
    Result := true;
end;

function isFirstPosition(b : TSchachbrett) : Boolean;
var i : Integer;
begin
  Result := false;
  for i:=0 to 119 do
    if b[i]<>Ausgangsstellung[i] then exit;
  Result := true;
end;

function isEqualPosition(b1, b2 : TSchachbrett) : Boolean;
var i : Integer;
begin
  Result := false;
  for i:=0 to 119 do
    if b1[i]<>b2[i] then exit;
  Result := true;
end;

procedure AssignAndResetFile(var f : File; recSize : Integer; Filename : String);
var fm : Byte;
begin
  fm := Filemode;
  try
    AssignFile(f, Filename);
    if FileGetAttr(Filename) and faReadOnly <> 0 then
      Filemode := fmOpenRead;
    if recSize > 0 then
      Reset(f, 1)
    else
      Reset(f);
  finally
    Filemode := fm;
  end;
end;

function isCCFDatabase(Filename : String) : Boolean;
var ccfFile : TChessDatabase;
begin
  Result := false;
  try
    ccfFile := TChessDatabase.Create(Nil); //ccf
    try
      Result := ccfFile.isChessDatabase(Filename);
    finally
      ccfFile.Free;
    end;
  except
  end;
end;

end.

