unit TimeLabel;

(* very simple component of a chess clock *)

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, extctrls;

type

  TTimeLabel = class(TLabel)
  constructor Create(AOwner : TComponent); override;
  destructor Destroy; override;
  private
    { Private-Deklarationen }
    t : TTimer;
    lastTime : TDateTime;
    rColor : TColor;
    FSecTotal : integer;
    FSecPerMove : integer;
    FMilliSecActual : integer;
    FMoveActual : integer;
    FTimeControl1afterMove : integer;
    FTimeControl1PlusSec : integer;
    FTimeControl2afterMove : integer;
    FTimeControl2PlusSec : integer;
    FColorWarning : TColor;
    FshowCentiSecs : Boolean;
    FSecWarning : integer;
    FMoveRecommended : TNotifyEvent;
    FMoveNecessary : TNotifyEvent;
    FZue : TNotifyEvent;
    function getRunning : boolean;
    function showTime : string;
    procedure setSecTotal(i : Integer);
    procedure setMilliSecActual(i : Integer);
    procedure setMoveActual(i : Integer);
    procedure setshowCentiSecs(b : Boolean);

    procedure setCaption(s : String);
    procedure resetTimeHelpvalues;
    procedure adjustTime;
    function  getCentiSec(millis : Integer) : integer;
    function  getSec(millis : integer) : Integer;
  protected
    { Protected-Deklarationen }
    procedure handleTime(Sender: TObject); virtual;
    function TriggerMoveNecessary : Boolean; virtual;
    function TriggerMoveRecommended : Boolean; virtual;
  public
    (*
    stops or starts the clock, respectively (while starting an increment will be added)
    *)
    procedure punch;
    (*
    starts the clock
    *)
    procedure start;
    (*
    stops the clock
    *)
    procedure stop;
    (*
    repaints the clock
    *)
    procedure refreshDisplay;
    (*
    adjusts the time (rarely used)
    *)
    procedure doHandleTime;
    (*
    true if the clock is not stopped
    *)
    property Running : Boolean read getRunning;
    (*
    the current milliseconds
    *)
    property MilliSecActual : Integer read FMilliSecActual write setMilliSecActual;
  published
    (*
    when the specified value is reached, the color will be changed
    *)
    property SecWarning : Integer read FSecWarning write FSecWarning;
    (*
    Color, when SecWarning is reached
    *)
    property ColorWarning : TColor read FColorWarning write FColorWarning;
    (*
    can be adjusted - if a determined time control is reached, the necessary seconds will be added
    *)
    property MoveActual : Integer read FMoveActual write setMoveActual;
    (*
    first time control
    *)
    property TimeControl1afterMove : Integer read FTimeControl1afterMove write FTimeControl1afterMove;
    (*
    second time control
    *)
    property TimeControl2afterMove : Integer read FTimeControl2afterMove write FTimeControl2afterMove;
    (*
    seconds to add after the first time control
    *)
    property TimeControl1PlusSec : Integer read FTimeControl1PlusSec write FTimeControl1PlusSec;
    (*
    seconds to add after the second time control
    *)
    property TimeControl2PlusSec : Integer read FTimeControl2PlusSec write FTimeControl2PlusSec;
    (*
    displays tenth seconds
    *)
    property showCentiSecs : Boolean read FshowCentiSecs write setshowCentiSecs;
    (*
    the time limit
    *)
    property SecTotal : integer read FSecTotal write setSecTotal;
    (*
    time to add after each move
    *)
    property SecPerMove : integer read FSecPerMove write FSecPerMove default 5;
    (*
    not yet implemented
    *)
    property MoveRecommended : TNotifyEvent read FMoveRecommended write FMoveRecommended;
    (*
    not yet implemented
    *)
    property MoveNecessary : TNotifyEvent read FMoveNecessary write FMoveNecessary;
    (*
    Time is over
    *)
    property Zue : TNotifyEvent read FZue write FZue;
  end;

implementation

constructor TTimeLabel.Create(AOwner : TComponent);
begin
  inherited Create(AOwner);
  t := TTimer.Create(self);
  t.Enabled := false;
  t.Interval := 1000;
  t.OnTimer := handleTime;
  FSecTotal := 5 * 60;
  FSecPerMove := 5;
  FSecWarning := 30;
  ColorWarning := clYellow;
  setCaption(showTime);
  rColor := -1;
  resetTimeHelpvalues;
end;

destructor TTimeLabel.Destroy;
begin
  t.Free;
  inherited destroy;
end;

procedure TTimeLabel.setCaption(s : String);
begin
  if Caption <> s then begin
    Caption := s;
    Invalidate;
  end;  
end;

procedure TTimeLabel.adjustTime;
Var dt : TDateTime;

begin
  dt := Now;
  if lastTime <> 0 then
    FMilliSecActual := trunc(FMilliSecActual - 3600 * 1000 * 24 * (abs(dt - lastTime)));
  lastTime := dt;
end;

procedure TTimeLabel.handleTime(Sender: TObject);
begin
  adjustTime;
  if FMilliSecActual < 0 then FMilliSecActual := 0;
  if (FMilliSecActual div 1000 <= SecWarning) and (FMilliSecActual <> 0) then begin
    if Color <> ColorWarning then
      rColor := Color;
    Color := ColorWarning;
  end
  else
    if rColor <> -1 then
      Color := rColor;
  setCaption(showTime);
  if assigned(MoveNecessary) and TriggerMoveNecessary then
    FMoveNecessary(self)
  else
  if assigned(MoveRecommended) and TriggerMoveRecommended then
    FMoveRecommended(self);
  if FMilliSecActual = 0 then begin
    if assigned(Zue) then
      FZue(self);
  end;
end;

function TTimeLabel.getCentiSec(millis : integer) : Integer;
begin
  Result := round(millis / 100);
end;

function TTimeLabel.getSec(millis : integer) : Integer;
begin
  Result := round(millis / 1000);
end;

function TTimeLabel.showTime : string;
var iRest, i : integer;
    sStd, sMin, sSec, sCentiSec : string;
    faktor : Integer;

  function formatTime(i : integer) : string;
  begin
    Result := IntToStr(i);
    if length(Result) = 1 then Result := '0' + Result; 
  end;

begin
  if FshowCentiSecs then
    faktor := 10
  else
    faktor := 1;
  sStd := ''; sMin := ''; sSec := '';
  if FshowCentiSecs then
    i := getCentiSec(FMilliSecActual)
  else
    i := getSec(FMilliSecActual);
  iRest := i mod (3600 * faktor);
  i := i div (3600 * faktor);
  if i > 0 then sStd := formatTime(i);
  i := iRest;
  iRest := i mod (60 * faktor);
  i := i div (60 * faktor);
  if i > 0 then sMin := formatTime(i);
  i := iRest;
  if FshowCentiSecs then begin
    iRest := i mod 10;
    i := i div 10;
  end;
  if i > 0 then sSec := formatTime(i);
  if FshowCentiSecs then begin
    i := iRest;
    if i > 0 then sCentiSec := IntToStr(i);
  end;
  Result := '';
  if sStd <> '' then Result := sStd + ':';
  if sMin <> '' then Result := Result + sMin + ':'
  else Result := Result + '00:';
  if sSec <> '' then begin
    if FshowCentiSecs then
      Result := Result + sSec + ':'
    else
      Result := Result + sSec;
  end    
  else begin
    if FshowCentiSecs then
      Result := Result + '00:'
    else
      Result := Result + '00';
  end;
  if FshowCentiSecs then begin
    if sCentiSec <> '' then Result := Result + sCentiSec
    else Result := Result + '0';
  end;
end;

function TTimeLabel.getRunning : boolean;
begin
  Result := t.Enabled;
end;

procedure TTimeLabel.punch;   // addiert im Unterschied zu start einen Zeitaufschlag
begin
  if t.Enabled then begin
    adjustTime;
    FMilliSecActual := FMilliSecActual + FSecPerMove * 1000;
    setCaption(showTime);
  end;
  resetTimeHelpvalues;
  t.Enabled := not t.Enabled;
end;

procedure TTimeLabel.resetTimeHelpvalues;
begin
  lastTime := Now;
end;

procedure TTimeLabel.start;
begin
  resetTimeHelpvalues;
  t.Enabled := true;
end;

procedure TTimeLabel.stop;
begin
  if t.Enabled then begin
    adjustTime;
    t.Enabled := false;
  end;  
end;

procedure TTimeLabel.setSecTotal(i : Integer);
begin
  FSecTotal := i;
  FMilliSecActual := i * 1000;
  setCaption(showTime);
end;

procedure TTimeLabel.setMilliSecActual(i : integer);
begin
  FMilliSecActual := i;
  resetTimeHelpvalues;
  setCaption(showTime);
end;

procedure TTimeLabel.setMoveActual(i : Integer);
begin
  FMoveActual := i;
  adjustTime;
  if i = FTimeControl1afterMove then
    setSecTotal(getSec(FMilliSecActual) + TimeControl1PlusSec)
  else if i = FTimeControl2afterMove then
    setSecTotal(getSec(FMilliSecActual) + TimeControl2PlusSec);
end;

procedure TTimeLabel.setshowCentiSecs(b : Boolean);
begin
  if FshowCentiSecs = b then exit;
  FshowCentiSecs := b;
  if b then
    t.Interval := 100
  else
    t.Interval := 1000;
  refreshDisplay;  
end;

procedure TTimeLabel.refreshDisplay;
begin
  setCaption(showTime);
end;

procedure TTimeLabel.doHandleTime;
begin
  HandleTime(nil);
end;

(*
Diese beiden Methoden sollten berschrieben werden:
In Abhngigkeit von SecActual, SecTotal und SecPerMove sollte
TriggerMoveRecommended TRUE zurckgeben,
  wenn es Zeit ist, den Zug auszufhren - die Anwendung kann die Aufforderung aber ignorieren,
  falls sie auf Grund der Bewertung der Meinung ist, da noch etwas Rechenzeit erforderlich ist;
TriggerMoveNecessary  TRUE zurckgeben,
  wenn der Zug wegen drohender Zeitnot unbedingt ausgefhrt werden mu
*)

function TTimeLabel.TriggerMoveNecessary : Boolean;
begin
  Result := true;
end;

function TTimeLabel.TriggerMoveRecommended : Boolean;
begin
  Result := true;
end;

end.
