unit UCustomDatabase;

interface

uses Classes, SysUtils, Forms, Controls, Windows, Messages;

{$H+}  // ansistrings verwenden

const
  idxID = 1234567890;

  idxHeaderLength = 8;
  idxKennung      = 0;
  idxTrashSize    = 1;
  idxCount        = 2;
  idxRes          = 3;
  idxUser1        = 4;
  idxUser2        = 5;
  idxUser3        = 6;
  idxUser4        = 7;


  esc       = #255;
  term      = #0;
  aliasEsc  = #5;  // 1-4 kommt schon in mailtexten vor
  aliasTerm = #6;

  delchar   = #7;  // damit werden gelschte Datenstze berschrieben

  errmesscount = 10;
  errmess : array[0..errmesscount - 1] of String =
    (
    'Error while creating index file.',
    'Write protection could not be deleted.',
    'Error.',
    'Index out of range.',
    'Data could not be removed.',
    'Data could not be read.',
    'Data could not be replaced.',
    'File could not be compressed.',
    'Data could not be added.',
    'Wrong file extensions.'
    );

type

// Strings mit 0 am Ende
// im int64-indexfile nach header (Kennung, trashsize) startposition des nchsten eintrags (nach letzter 0)
// ist um 1 erhht, da null ausgeschlossen werden mu (gelschte eintrge werden mit Minus markiert)

ByteFile = File of Byte;
Int64File = File of Int64;

TDatabaseCheckEvent = procedure(Sender : TObject; var checkstep : int64; var cancel : Boolean) of Object;
TDatabaseTraceEvent = procedure(Sender : TObject; var cancel : Boolean; percent : Integer) of Object;

TCustomDatabase = class(TCustomControl)
    destructor destroy; override;
  private
    FTag : Integer;
    HandleExt : String;
    MainFilename : String;
    mainFile : ByteFile;
    mainFileTime : TDateTime;
    indexFilename : String;
    indexFile : int64File;
    indexFileTime : TDateTime;
    FFilesOpenedExternally : Boolean;
    stateFileName : String;
    stateFile : ByteFile;
    StateFileExt : String;
    err, compresscount : int64;
    FDatabaseCheckTrace : TDatabaseCheckEvent;
    FDatabaseCompressTrace : TDatabaseTraceEvent;
    FCreateNewIndexfileTrace : TDatabaseTraceEvent;
    FOverwriteDeletedData, FBackupBeforeCompress : Boolean;
    FFilesPrepared : Boolean;
    FOnCreateNewIndexFile, FOnDatabaseError : TNotifyEvent;
    stateFileOpen, Cancel : Boolean;
    FmaxCheckSec : Integer;
    User1,User2,User3,User4 : Int64;
    FWorking : Boolean;
    FWriteProtected : Boolean;
    FFlushAfterSaving : Boolean;
    function  createNewIndexFile : Boolean;
    function  isMainFileOk : Boolean;
    function  docheck(beforeCompress : Boolean) : Boolean;
    function  check(beforeCompress : Boolean) : Boolean;
    function  getIndex(idx : int64; var deleted : Boolean) : int64;
    function  doGet(idx : int64; var Data : String;
      maxToRead : int64; onlyUndeleted : Boolean; mainidx : int64; var recordIsDeleted : Boolean) : Integer;
    function  dodelete(idx : int64; var wasAlreadyDeleted : Boolean) : Boolean;
    function  doadd(Data : String; replace : Integer; noEncode : Boolean) : Boolean;
    function  DelWriteProtect(s : string) : Boolean;
    function  handleToIndex(handle : String) : int64;
    function  isHandleOk(handle : String) : Boolean;
    function  getHandle(idx : Integer) : String;
    function  getStringLength : Integer;
    function  getString(var Data : String; maxToRead : Integer) : Boolean;
    function  doCount : int64;
    function  doTrashedSize : int64;
    procedure setFileAges;
    procedure SetFilesOpened(b : Boolean);
    function  closeFiles : Boolean;
    function  openFiles : Boolean;
    function  getRecord(var handle, Data : String; maxToRead : int64; noDecode, first : Boolean) : Boolean;
    procedure incCount;
    procedure triggerDatabaseError;
    function  checkTerm(mainFileIdx : int64) : Boolean;
    procedure openIndexFile;
    procedure closeAll;
    procedure SetWriteProtect(s : string);
    function  isWriteProtected(aFilename : String) : Boolean;
  protected
    procedure closeStateFile;
    function  openStateFile : Boolean;
    procedure prepareFiles(Filename, IndexFileExtension, StateFileExtension : String);
    procedure prepareFilesEx(Filename, IndexFileExtension, StateFileExtension : String;
      iUser1, iUser2, iUser3, iUser4 : Int64);
    function  toggleFlag(handle : String; position : Integer) : Boolean;
  public
    property FlushAfterSaving : Boolean read FFlushAfterSaving write FFlushAfterSaving;
    (*
    Since the files are closed after each access,
    you may open them with this property before processing a lot of accesses.
    Finally it must be set to false.
    *)
    property  FilesOpen : Boolean read FFilesOpenedExternally write SetFilesOpened;
    (*
    Is true after "Loadfile" or "CheckandLoadfile" of TChessdatabase was invoked.
    *)
    property  FilesPrepared : Boolean read FFilesPrepared;
    property  Tag : Integer read FTag write FTag;
    (*
    The name of the main file (with the extension "mainFileExt" -
    two other files belong to this database, also)
    *)
    property  Filename : String read MainFilename;
    (*
    The database is checked for errors while loading.
    You may set this value to limit the time of this check.
    *)
    property  maxCheckSec : Integer read FmaxCheckSec write FmaxCheckSec;
    (*
    In this event you can determine the number of records which will be checked for errors.
    TPGN using TChessdatabase sets this parameter due to the size of the database.
    Set this parameter to a high value to avoid the check.
    *)
    property  OnDatabaseCheckTrace : TDatabaseCheckEvent read FDatabaseCheckTrace write FDatabaseCheckTrace;
    (*
    Invoked while executing "Compress" which irrecoverably removes records marked as deleted.
    If you cancel the process in this event, NO game will be removed.
    *)
    property  OnDatabaseCompressTrace : TDatabaseTraceEvent read FDatabaseCompressTrace write FDatabaseCompressTrace;
    (*
    If true, the delete function marks the record as deleted
    AND overwrites the data with any stuff.
    *)
    property  OverwriteDeletedData : Boolean read FOverwriteDeletedData write FOverwriteDeletedData;
    (*
    Backups the main file and the index file (extension ".tmp") before compressing
    *)
    property  BackupBeforeCompress : Boolean read FBackupBeforeCompress write FBackupBeforeCompress;
    (*
    If a database check while loading failed, it is tried to create a new index file.
    Please note, it could take a longer while!
    *)
    property  OnCreateNewIndexFile : TNotifyEvent read FOnCreateNewIndexFile write FOnCreateNewIndexFile;
    (*
    You can cancel the process in this event.
    *)
    property  OnCreateNewIndexfileTrace : TDatabaseTraceEvent read FCreateNewIndexfileTrace write FCreateNewIndexfileTrace;
    (*
    Serious error in the database
    *)
    property  OnDatabaseError : TNotifyEvent read FOnDatabaseError write FOnDatabaseError;
    (*
    True if the database is in process
    *)
    property  Working : Boolean read FWorking;
    (*
    If a file is write protected,
    the component removes the write protection temporarily to grant successful access
    but sets this property, and it's your choice to write any changes
    *)
    property  WriteProtected : Boolean read FWriteProtected;
    (*
    If a function returns false,
    you can uses the following both functions to get information
    *)
    function  getLastError : int64;
    function  getLastErrorMess : String;
    (*
    Returns a record ("Data") by a handle delivered by the add method.
    It can be marked as deleted.
    In addition, you can limit the max. number of characters which shall be read
    (if the data contain a header, for instance, and you only need that header).
    All #0 characters are replaced while saving a record, and thus it must be decoded when it is read.
    If it's sure that the data do not contain any #0, you can set noDecode to true to save time.
    *)
    function  get(handle : String; var Data : String; maxToRead : int64;
      noDecode : Boolean; var recordIsDeleted : Boolean) : Boolean;
    (*
    Gets a record by its index ("idx") in the database from 0..x.
    If "idx" is -1, you MUST set "mainidx". This is the index of the starting byte in the data (main) file
    (rarely used - it only avoids to access the index file to get that starting byte)
    Otherwise "mainidx" is not needed.
    *)
    function  getbyIndex(idx : int64; var handle : String; var Data : String;
      maxToRead : int64; onlyUndeleted{, noOpenFiles}, noDecode : Boolean;
      mainidx : int64; var recordIsDeleted : Boolean) : Integer;
    (*
    Returns a list of all "mainidx" belonging to the records.
    See above how "mainidx" can be used.
    *)
    procedure getAllIndexe(l : TStringList);
    (*
    Returns all handles of the records. Can be used with "get" then.
    *)
    procedure getAllHandles(l : TStringList);
    (*
    Replaces records.
    Do only set "NoEncode" to true if you are really sure that the data does NOT contain any #0!
    *)
    function  replace(var handle : String; Data : String; noEncode : Boolean) : Boolean;
    function  replacebyIndex(idx : int64; Data : String; noEncode : Boolean) : Boolean;
    (*
    Deletes a record.
    *)
    function  delete(handle : String) : Boolean;
    function  deletebyIndex(idx : int64) : Boolean;
    (*
    Since the delete functions only mark the records as deleted,
    you could restore them.
    Only after invoking "Compress" the records will be removed irrecoverably.
    !!! Both functions are NOT yet implemented !!!
    *)
    function  undelete(handle : String) : Boolean;
    function  undeletebyIndex(idx : int64) : Boolean;
    (*
    See the "replace" function for the "noEncode" parameter or set it always to false.
    Remember the handle to change the added record later.
    *)
    function  add(var handle : String; Data : String; noEncode : Boolean) : Boolean;
    (*
    Size of deleted data (records marked as deleted
    AND the stuff increasing after replacing records
    *)
    function  TrashedSize : int64;
    (*
    Return the records marked as deleted.
    *)
    function  TrashedCount : int64;
    (*
    Number of records (without deleted redords)
    *)
    function  Count : int64;
    (*
    Number of all records.
    *)
    function  CountWithDeletedRecords : int64;
    (*
    Returns the size of the main file
    *)
    function  Size : int64;
    (*
    It must be invoked to remove deleted records irrecoverably
    or after many replaced records.
    Check "Trashsize" to see if it's nessecary
    *)
    function  compress : Boolean;
    (*
    Deletes all files belonging to this database
    *)
    function  DeleteDatabase : Boolean;
    (*
    Returns the first and last record.
    *)
    function  getFirstRecord(var handle, Data : String; maxToRead : int64; noDecode : Boolean) : Boolean;
    function  getLastRecord(var handle, Data : String; maxToRead : int64; noDecode : Boolean) : Boolean;
    (*
    If you want to move the database, you must set the new path.
    It's better if you free the component, copy the files and recreate it.
    *)
    procedure changePath(newFolder : String; doNotUseStateFile : Boolean);
    function indexFileExists : Boolean;
  end;

  function  DecodeData(Data : String) : String;
  function  EncodeData(Data : String) : String;
  procedure AssignAndResetByteFile(var f : ByteFile; Filename : String);
  procedure AssignAndResetInt64File(var f : int64File; Filename : String);

implementation

uses StdCtrls;

function doEncodeData(Data : String) : String;
var i : Integer;
begin
  Result := '';
  i := 1;
  while i <= length(Data) do begin
    if (Data[i] = esc) or (Data[i] = term) then begin
      Result := Result + esc;
      if Data[i] = esc then
        Result := Result + aliasEsc
      else
        Result := Result + aliasTerm;
    end
    else
      Result := Result + Data[i];
    inc(i);
  end;
end;

function doDecodeData(Data : String) : String;
var i : Integer;
begin
  Result := '';
  i := 1;
  while i <= length(Data) do begin
    if (Data[i] = esc) and (i < length(Data)) then begin
      inc(i);
      if Data[i] = aliasEsc then
        Result := Result + esc
      else
        Result := Result + Term;
    end
    else
      Result := Result + Data[i];
    inc(i);
  end;
end;

function codeData(Data : String; encode : Boolean) : String;
var i, j : int64;
    step : Integer;
begin
  Result := '';
  step := 10000;
  i := 1;
  while i <= length(Data) do begin
    if encode then
      Result := Result + doEncodeData(copy(Data, i, step))
    else begin
      j := i + step - 1;
      if (j <= length(Data)) and (Data[j] = Esc) then
        inc(step);
      Result := Result + doDecodeData(copy(Data, i, step));
    end;  
    inc(i, step);
  end;
end;

function DecodeData(Data : String) : String;
begin
  Result := codeData(Data, false);
end;

function EncodeData(Data : String) : String;
begin
  Result := codeData(Data, true);
end;

function TCustomDatabase.indexFileExists : Boolean;
begin
  Result := FileExists(indexFileName);
end;

procedure TCustomDatabase.prepareFilesEx(Filename, IndexFileExtension, StateFileExtension : String;
      iUser1, iUser2, iUser3, iUser4 : Int64);
begin
  if FWorking then exit;
  User1 := iUser1;
  User2 := iUser2;
  User3 := iUser3;
  User4 := iUser4;
  prepareFiles(Filename, IndexFileExtension, StateFileExtension);
end;

procedure TCustomDatabase.closeAll;
begin
  closeStateFile;
  MainFilename := '';
end;

function TCustomDatabase.isWriteProtected(aFilename : String) : Boolean;
var f : File;
begin
  Result := false;
  try
    if FileExists(Filename) then begin
      Result := FileGetAttr(Filename) and faReadOnly <> 0;
    end;
  except
  end;
end;

procedure TCustomDatabase.prepareFiles(Filename, IndexFileExtension, StateFileExtension : String);
var b : Byte;
    e1, e2, e3 : String;

  procedure createDatabase(aFilename : String);
  var f : File;
  begin
    if not FileExists(aFilename) then begin
      assignFile(f, aFilename);
      rewrite(f);
      closeFile(f);
    end
  end;

begin
  if FWorking then exit;
  FFilesPrepared := true;
  try
    e1 := uppercase(IndexFileExtension);
    e2 := uppercase(StateFileExtension);
    e3 := uppercase(ExtractFileExt(Filename));
    if (e1 = e2) or (e1 = e3) or (e2 = e3) then begin
      err := 10;
      exit;
    end;

    closeAll;

    StateFileExt := StateFileExtension;
    MainFilename := Filename;
    IndexFilename := ChangeFileExt(Filename, IndexFileExtension);

    FWriteProtected := (isWriteProtected(MainFilename))
    or (isWriteProtected(IndexFilename));

    if not openStateFile then begin
      err := 3;
      MainFilename := '';
      exit;
    end;

    createDatabase(indexFilename);
    //if not
    DelWriteProtect(IndexFilename); //then begin
      (*err := 2;
      MainFilename := ''
      exit;
    end;*)

    createDatabase(MainFilename);
    DelWriteProtect(MainFilename);  // das wird dann noch bei jedem Speichern geprft

    HandleExt := ':' + MainFilename;//ExtractFileName(MainFilename);

    mainFileTime  := -1;
    indexFileTime := -1;
    if not check(false) then
      if not createNewIndexFile then begin
        err := 1;
        MainFilename := '';
      end;
    setFileAges;
  except
    err := 3;
    MainFilename := '';
  end;
end;

destructor TCustomDatabase.destroy;
begin
  closeStateFile;
  if (FWriteProtected) and (Mainfilename <> '') then begin
    setWriteProtect(Mainfilename);
    setWriteProtect(Indexfilename);
  end;
  inherited destroy;
end;

function TCustomDatabase.openStateFile : Boolean;
var b : Byte;
begin
  Result := false;
  try
    stateFileName := ChangeFileExt(Filename, StateFileExt);
    DelWriteProtect(stateFileName);
    if not isWriteProtected(stateFileName) then begin
      if (FileExists(stateFileName)) and
      (not DeleteFile(PChar(stateFileName))) then
        exit;
      assignFile(stateFile, stateFileName);
      rewrite(stateFile); // nur stateFile geffnet halten - Exception wenn von anderer Datenbank geffnet --> Filename wird nicht gesetzt
      write(stateFile, b);
      stateFileOpen := true;
      Result := true;
    end
    else
      if WriteProtected then  // dann sollten hauptdateien auch schreibgeschtzt gewesen sein, wenn schon das ffnen der datei nicht geht
        Result := true;
  except
  end;
end;

procedure TCustomDatabase.closeStateFile;
begin
  if not stateFileOpen then exit;
  try          // wenn mb vorher gelscht wurde, kommt hier exception - ist normal
    if FileExists(stateFileName) then
      closeFile(stateFile);
  except
  end;
  stateFileOpen := false;
end;

function TCustomDatabase.DeleteDatabase : Boolean;
begin
  if FWorking then begin
    Result := false;
    exit;
  end;
  closeStateFile;
  Result := DeleteFile(PChar(stateFilename)) and
    DeleteFile(PChar(indexFilename)) and
    DeleteFile(PChar(mainFilename));
  if mainFilename = '' then
    Result := true;     
  mainFilename := '';
end;

procedure TCustomDatabase.setFileAges;
begin
  mainFileTime  := FileDateToDateTime(FileAge(mainFilename));
  indexFileTime := FileDateToDateTime(FileAge(indexFilename));
end;

function TCustomDatabase.DelWriteProtect(s : string) : Boolean;
Var i : integer;
begin
  try
    i := FileGetAttr(s);
    if i and faReadOnly <> 0 then
      FileSetAttr(s, i xor faReadOnly);
    i := FileGetAttr(s);
    Result := i and faReadOnly = 0;
  except
    Result := false;
  end;
end;

procedure TCustomDatabase.SetWriteProtect(s : string);
begin
  try
    if FileExists(s) then
      FileSetAttr(s, FileGetAttr(s) or faReadOnly);
  except
  end;
end;

function TCustomDatabase.isMainFileOk : Boolean;
var b1, b2 : Boolean;
begin
  err := 0;
  Result := MainFilename <> '';
  if Result then begin
    if (not FileExists(MainFilename))
    or (not FileExists(IndexFilename)) then
      Result := false;
  end;
  if Result then begin
    b1 := DelWriteProtect(MainFilename);
    b2 := DelWriteProtect(IndexFilename);
    if b1 <> b2 then   // entweder mssen beide schreibgeschtzt (dann kommt beim add fehler) oder beide beschreibbar sein
      Result := false;
  end;
  if Result then
  if (
  (mainFileTime <> -1) and
  (mainFileTime  <>  FileDateToDateTime(FileAge(mainFilename)))
  )
      or
  (
  (indexFileTime <> -1) and
  (indexFileTime <> FileDateToDateTime(FileAge(indexFilename)))
  )
    then inc(compresscount);    // dann werden vorherige handles ungltig und es geht nur noch add
end;

procedure TCustomDatabase.triggerDatabaseError;
begin
  if assigned(OnDatabaseError) then
    OnDatabaseError(self);
end;

function TCustomDatabase.openFiles : Boolean;
begin
  if FFilesOpenedExternally then begin
    Result := true;
    exit;
  end;
  Result := false;
  if FWorking then exit;
  try
    if not isMainFileOk then begin
      triggerDatabaseError;
      exit;
    end;
    openIndexFile;
    try
      AssignAndResetByteFile(mainFile, mainFilename);
      Result := true;
    except
      try
        closeFile(indexFile);
      except
      end;  
    end;
  except
  end;
end;

function TCustomDatabase.closeFiles : Boolean;
begin
  Result := false;
  if (FFilesOpenedExternally) or (FWorking) then
    exit;
  Result := true;
  try
    if FFlushAfterSaving then
      FlushFileBuffers(TFileRec(indexFile).Handle);
    closeFile(indexFile);
  except
    Result := false;
  end;
  try
    if FFlushAfterSaving then
      FlushFileBuffers(TFileRec(mainFile).Handle);
    closeFile(mainFile);
  except
    Result := false;
  end;
  setFileAges;
end;

procedure TCustomDatabase.SetFilesOpened(b : Boolean);
begin
  if FFilesOpenedExternally = b then exit;
  if FWorking then exit;
  if b then begin
    openFiles;
    FFilesOpenedExternally := true;
  end
  else begin
    FFilesOpenedExternally := false;
    closeFiles;
  end;
end;

function TCustomDatabase.checkTerm(mainFileIdx : int64) : Boolean;
var j : Byte;
begin
  Result := false;
  if mainFileIdx <> 0 then begin
    seek(mainFile, mainFileIdx - 1);
    read(mainFile, j);
    if j = Byte(term) then
      Result := true;
  end
  else Result := true;
end;

function TCustomDatabase.docheck(beforeCompress : Boolean) : Boolean;
var number, i, k, isize, step : int64;
    j : Byte;
    t, maxt : TTime;
begin
{$I-}
  try
    Result := false;
    FWorking := true;
    Screen.Cursor := crHourglass;
    try
      t := Now;
      maxt := maxCheckSec / 24 / 3600;
      if Filesize(indexFile) < idxHeaderLength then exit;
      step := 0;
      seek(indexFile, idxHeaderLength);
      isize := FileSize(indexFile);
      k := idxHeaderLength;
      number := 0;
      while not eoF(indexFile) do begin
        inc(k);
        read(indexFile, i);
        if (number <> - 1) and (i > 0) then
          inc(number);
        i := abs(i);
        if i > 0 then
          dec(i);
        if i > FileSize(mainFile) then
          exit;
        if not checkTerm(i) then
          exit;
        if step > 1 then begin
          if k + step <= isize then begin
            seek(indexFile, k + step);
            inc(k, step);
          end
          else seek(indexFile, isize);
        end;
        Cancel := false;
        if beforeCompress then begin
          if assigned(FDatabaseCompressTrace) then
            FDatabaseCompressTrace(self, Cancel, trunc(k * 100 / isize));
        end
        else begin
          if assigned(FDatabaseCheckTrace) then
            FDatabaseCheckTrace(self, step, cancel);
          if step <> 0 then
            number := -1;  //ungltig machen, da anzahl-check bei bersprungenen datenstzen nicht mehr funktioniert
        end;
        if cancel then begin
          if not beforeCompress then
            Result := true;
          exit;
        end;
        if maxCheckSec > 0 then
          if Now - t > maxt then
            exit;
        if IOResult <> 0 then exit;
      end;
      i := Filesize(mainFile) - 1;
      if i > 0 then begin
        seek(mainFile, i);
        read(mainFile, j);
        if j <> Byte(term) then exit;
      end;
      if (number = -1) or (beforeCompress) or (doCount = number) then
        Result := true
      else
        messagebeep(0);  
    finally
      Screen.Cursor := crDefault;
      FWorking := false;
    end;
  except
  end;
{$I+}
end;

function TCustomDatabase.check(beforeCompress : Boolean) : Boolean;
begin
  Result := true;
  if not openFiles then exit;
  try
    Result := docheck(beforeCompress);
  finally
    closeFiles;
  end;
end;

function TCustomDatabase.createNewIndexFile : Boolean;
var i, j, iCount, iGes : int64;
    b : Byte;
    ii : Integer;
    cancel, deleted : Boolean;
begin
  Result := false;
  if FWorking then exit;
  if assigned(FOnCreateNewIndexFile) then
    FOnCreateNewIndexFile(self);
  try
    Screen.Cursor := crHourglass;
    openFiles;
    FWorking := true;
    try
      reset(indexFile);
      iGes := FileSize(IndexFile) - idxHeaderlength;
      if iGes <= 0 then iGes := 1;
      rewrite(indexFile);
      j := idxID;
      write(indexFile, j);
      j := 0;
      for ii := 1 to idxHeaderLength - 1 do begin
        case ii of
          idxUser1 : write(indexFile, User1);
          idxUser2 : write(indexFile, User2);
          idxUser3 : write(indexFile, User3);
          idxUser4 : write(indexFile, User4);
          else
            write(indexFile, j);
        end;
      end;
      j := 0;
      i := 0;
      iCount := 0;
      deleted := true;
      while not eoF(mainFile) do begin
        read(mainFile, b);
        inc(i);
        if (b = Byte(term)) then begin
          inc(j);

          if deleted then
            j := -j;

          write(indexFile, j);
          j := i;
          if not deleted then
            inc(iCount);

          if assigned(FCreateNewIndexfileTrace) then begin
            Cancel := false;
            FCreateNewIndexfileTrace(self, cancel, trunc(iCount * 100 / iGes));
            if Cancel then begin
              if assigned(FOnDatabaseError) then
                FOnDatabaseError(self);
              exit;
            end;
          end;

          deleted := true;
        end
        else
          if b <> Byte(delchar) then
            Deleted := false;
      end;
      truncate(indexFile);
      seek(indexFile, idxCount);
      write(indexFile, iCount);
      indexFileTime := FileDateToDateTime(FileAge(indexFilename));

      i := Filesize(mainFile) - 1;
      if i > 0 then begin   // neues indexfile wird manchmal erstellt, wenn letzter eintrag nicht mit 0 abgeschlossen ist (irgendwelche daten angehngt), was hier nicht behoben wird - deshalb das prfen u. ev. angngen
        seek(mainFile, i);
        read(mainFile, b);
        if b <> Byte(term) then begin
          seek(mainFile, i + 1);
          b := 0;
          write(mainFile, b);
        end;
      end;

      Result := true;
    finally
      FWorking := false;
      if assigned(FCreateNewIndexfileTrace) then
        FCreateNewIndexfileTrace(self, cancel, 0);
      Screen.Cursor := crDefault;
      closeFiles;
    end;
  except
    if assigned(FOnDatabaseError) then
      FOnDatabaseError(self);
  end;
end;

function TCustomDatabase.compress : Boolean;
var i, j, k : int64;
    b : Byte;
    newF : File;
    s, newFilename, newIndexFileName : String;
    newIndexFile : File of int64;
    NumRead, NumWritten : integer;
    Buf : PChar;
    Bufsize : int64;
    finished : Boolean;
    ii, fileGes, fileCurr : Integer;
    f : TForm;
begin
  try
    Result := false;
    if FWorking then exit;
    Cancel := false;

    if FBackupBeforeCompress then begin
      CopyFile(PChar(MainFilename), PChar(MainFilename + '.tmp'), false);
      CopyFile(PChar(indexFilename), PChar(indexFilename + '.tmp'), false);
    end;

    if not openFiles then exit;
    FWorking := true;
    try
      if doTrashedSize = 0 then exit;

      newFilename := ChangeFileExt(MainFilename, '.maintmp');
      newIndexFilename := ChangeFileExt(MainFilename, '.indextmp');

      assignFile(newf, newFilename);
      rewrite(newf, 1);
      assignFile(newIndexFile, newIndexFilename);
      rewrite(newIndexFile);

(*      f := Nil;
      if (Filesize(Mainfile) > 1000000) and (not assigned(FDatabaseCompressTrace))
      then begin
        f := TForm.Create(self);
        f.Caption := 'Compressing ' + ExtractFilename(MainFilename);
        f.Show;
      end;*)
      Screen.Cursor := crHourglass;
      try
        if not docheck(true) then begin
          if Cancel then
            exit;
          if not createNewIndexFile then
            exit;
        end;
        reset(indexFile);
        fileGes := FileSize(indexFile) - idxHeaderLength;
        if fileGes <= 0 then fileGes := 1;
        fileCurr := 0;
        for ii := 1 to idxHeaderLength do begin  // header kopieren
          read(indexFile, j);
          write(newindexFile, j);
        end;
        seek(indexFile, idxHeaderLength);
        k := 1;  // index in maifile - eins mehr wegen negativ setzen beim lschen
        while (not Cancel) and (not eoF(indexFile)) do begin
          read(indexFile, i);
          inc(fileCurr);
          if i > 0 then begin
            dec(i);
            seek(mainFile, i);
            bufsize := getStringLength + 1;
            seek(mainFile, i);
            if Bufsize > 1 then begin  // < 1 sollte nie auftreten
              GetMem(Buf, Bufsize);
              try
                BlockRead(mainFile, Buf^, Bufsize, NumRead);
                if NumRead <> 0 then begin
                  BlockWrite(newF, Buf^, NumRead, NumWritten);
                  write(NewIndexFile, k);
                  inc(k, NumWritten);
                end
                else
                  Cancel := true;  // dann immmer abbrechen weil das fehler sein sollte
              finally
                FreeMem(Buf, Bufsize);
              end;
            end;
          end;
          if assigned(FDatabaseCompressTrace) then begin
            FDatabaseCompressTrace(self, Cancel, trunc(fileCurr * 100 / fileGes));
          end;
        end;
        i := 0;  // Trash aktualisieren
        seek(newindexFile, idxTrashSize);
        write(newindexFile, i);
      finally
        Screen.Cursor := crDefault;
        closeFile(newIndexFile);
        closeFile(newF);
  (*      if not assigned(FDatabaseCompressTrace) then
          if f <> Nil then begin
            PostMessage(f.Handle, WM_CLOSE, 0, 0);
          end;*)
          // geht nicht, seltsamer fehler*)
      end;
    finally
      FWorking := false;
      closeFiles;
    end;
    if not cancel then begin  // ansonsten gecancelled
      DeleteFile(PChar(MainFilename));
      if RenameFile(newFilename, MainFilename) then begin
        DeleteFile(PChar(indexFilename));
        if RenameFile(newindexFilename, indexFilename) then begin
          inc(compresscount);  // danach werden alle vorherigen handles ungltig - Daten werden nur noch hinzugefgt (compress sollte nicht bei geladenen Daten aufgerufen werden)
          Result := true;
        end;
      end;
    end
    else begin
      DeleteFile(PChar(newFilename));
      DeleteFile(PChar(newIndexFilename));
    end;
  except
  end;
  if not Result then if err = 0 then err := 7;
end;

function TCustomDatabase.getString(var Data : String; maxToRead : Integer) : Boolean;
const Bufsize : integer = 2049;
var NumRead: integer;
    Buf: PChar;
    s : String;
begin
  Result := false;
  if FWorking then exit;
  try
    Data := '';
    Getmem(Buf, Bufsize);
    try
      FillChar(Buf^, Bufsize, #0);
      while not eoF(mainFile) do begin
        blockread(mainFile, Buf^, Bufsize - 1, NumRead);
        s := StrPas(Buf);
        Data := Data + s;
        if (maxtoread > 0) and (length(Data) > maxtoRead) then
          break;
        if length(s) < Bufsize - 1 then
          break;
      end;
      Result := true;
    finally
      Freemem(Buf, Bufsize);
    end;
  except
  end;
end;

function TCustomDatabase.getStringLength : Integer;
var Data : String;
begin
  Result := 0;
  if getString(Data, 0) then
    Result := length(Data);
end;

function TCustomDatabase.doGet(idx : int64; var Data : String;
   maxToRead : int64; onlyUndeleted : Boolean; mainidx : int64; var recordIsDeleted : Boolean) : Integer;
var i : int64;
    deleted : Boolean;
  //  b : Byte;
begin
  Result := -1;  // -1 bei fehler, ansonsten 0 bzw. 1 wenn kein fehler aber gelschte daten bei gesetzten OnlyUndeleted
  recordIsDeleted := false;
  i := mainidx;  // -1 wenn index in mainfile nicht bekannt, dann wird index anhand der nummer des datensatzes ermittelt
  Deleted := false;
  if i = -1 then
    i := getIndex(idx, deleted);
  if i <> -1 then begin
    if onlyUndeleted and deleted then begin
      Result := 1;
      exit;
    end;
    if not checkTerm(i) then
      exit;
    seek(mainFile, i);
    if getString(Data, maxToRead) then begin
      Result := 0;
      recordIsDeleted := Deleted;
    end;
  end;
end;


(*function TCustomDatabase.isMainfileIndexOk(idx : int64) : Boolean;
var i : int64;
    j : Byte;
begin
  Result := false;
  try
    i := abs(idx) - 1;   //!!! hier wird angepat --> nur aufrufen, wenn noch gemacht
    if i = 0 then Result := true  // erster Datensatz
    else begin
      seek(mainFile, i - 1);
      read(mainFile, j);
      if j = Byte(term) then
        Result := true;
    end;
  except
  end;
end;*)

// in indexfile index fr die hauptdatei lesen
function TCustomDatabase.getIndex(idx : int64; var deleted : Boolean) : int64;
var i, j : int64;
begin
  Result := -1;
  if FWorking then exit;
  deleted := false;
  err := 4;
  if idx < -1 then exit;
  i := idxHeaderLength + idx;     // filesize kommt nicht mit int64 zurecht!
  if i >= FileSize(indexFile) then exit;
  err := 0;
  seek(indexFile, i);
  read(indexFile, j);
(*  if not isMainfileIndexOk(j) then begin
    triggerDatabaseError;
    exit;
  end;*)
  deleted := j < 0;
  Result := abs(j);
  if Result > 0 then
    dec(Result);
end;

function TCustomDatabase.toggleFlag(handle : String; position : Integer) : Boolean;
var c : Char;
    b : Byte;
    idx, i : int64;
    deleted : Boolean;
begin
  Result := false;
  if FWorking then exit;
  try
    if not isHandleOk(handle) then exit;
    idx := handleToIndex(handle);
    openFiles;
    try
      i := getIndex(idx, deleted);
      if i <> -1 then begin
        if not checkTerm(i) then
          exit;
        seek(mainFile, i + position);
        read(mainFile, b);
        c := Char(b);
        if (c = '0') or (c = '1') then begin
          if c = '0' then c := '1'
          else c := '0';
          seek(mainFile, i + position);
          write(mainFile, Byte(c));
          Result := true;
        end;
      end;
    finally
      closeFiles;
    end;
  except
  end;
  if not Result then if err = 0 then err := 2;
end;


function TCustomDatabase.dodelete(idx : int64; var wasAlreadyDeleted : Boolean) : Boolean;
var //b : Byte;
    Bufsize : integer;
    i, j, k, size : int64;
    Buf : PChar;
    NumRead, NumWritten : integer;
    deleted : Boolean;
begin
  Result := false;
  if FWorking then exit;
  wasAlreadyDeleted := false;
  Bufsize := 2049;
  NumWritten := -1;
  i := getIndex(idx, deleted);
  if i <> -1 then begin
    if deleted then begin
      wasAlreadyDeleted := true;
      exit;
    end;
    if not checkTerm(i) then
      exit;
    seek(mainFile, i);
    Bufsize := getStringLength;
    seek(mainFile, i);
    if (FOverwriteDeletedData) and (Bufsize > 0) then begin
      GetMem(Buf, Bufsize);
      try
        seek(mainFile, i);
        FillChar(Buf^, Bufsize, delchar);  
        BlockWrite(mainFile, Buf^, Bufsize, NumWritten);
      finally
        FreeMem(Buf, Bufsize);
      end;
    end;
    seek(indexFile, idxHeaderLength + idx);  // minus vor index des gelschten satzes
    k := -(i + 1);
    write(indexFile, k);
    k := doTrashedSize + Bufsize + 1;  // Trash aktualisieren
    seek(indexFile, idxTrashSize);
    write(indexFile, k);
    if NumWritten <> 0 then
      Result := true;
  end;
end;

function TCustomDatabase.doadd(Data : String; replace : Integer; noEncode : Boolean) : Boolean;
var i : Byte;
    a, l : int64;
    NumWritten: integer;
    Buf : PChar;
begin
  Result := false;
  if FWorking then exit;
  if not noEncode then
    Data := EncodeData(Data);
    if Pos(#0, Data) <> 0 then begin
      Result := false;
      exit;
    end;
  l := length(Data) + 1;  // strpcopy schreibt dann abschlieende 0
  GetMem(Buf, l);
  a := FileSize(mainFile);
  try
    seek(mainFile, a);
    StrPCopy(Buf, Data);
    BlockWrite(mainFile, Buf^, l, NumWritten);
  finally
    FreeMem(Buf, l);
  end;
  if (replace = -1) then
    seek(indexFile, FileSize(indexFile))
  else
    seek(indexFile, replace + idxHeaderLength); // wert von replace ist schon in replacebyindex geprft
  inc(a);  // eins mehr weil erster eintrag nicht 0 sein darf (gelschte werden mit minus gekennzeichnet)  
  write(indexFile, a);
  Result := true;
end;

function TCustomDatabase.getLastError : int64;
begin
  Result := err;
end;

function TCustomDatabase.getLastErrorMess : String;
var i : int64;
begin
  Result := '';
  i := getLastError - 1;
  if (i > -1) and (i < errmesscount) then
    Result := errmess[i];
end;

function TCustomDatabase.doTrashedSize : int64;
begin
  seek(indexFile, idxTrashSize);
  read(indexFile, Result);
end;

function TCustomDatabase.TrashedSize : int64;
begin
  Result := 0;
  if FWorking then exit;
  try
    if MainFilename = '' then exit;
    if not FileExists(indexFileName) then exit;
    if not FFilesOpenedExternally then
      openIndexFile;
    try
      Result := doTrashedSize;
    finally
      if not FFilesOpenedExternally then
        closeFile(indexFile);
    end;
  except
  end;
end;

procedure TCustomDatabase.openIndexFile;
begin
  if FWorking then exit;
  AssignAndResetInt64File(indexFile, indexFileName);
end;

function TCustomDatabase.Size : int64;
begin
  Result := 0;
  if FWorking then exit;
  try
    if MainFilename = '' then exit;
    if not FFilesOpenedExternally then
      AssignAndResetByteFile(mainFile, mainFileName);
    try
      Result := FileSize(mainFile);
    finally
      if not FFilesOpenedExternally then
        closeFile(mainFile);
    end;
  except
  end;
end;

function TCustomDatabase.TrashedCount : int64;
begin
  Result := 0;
  if FWorking then exit;
  if MainFilename = '' then exit;
  if not FFilesOpenedExternally then
    openIndexFile;
  try
    Result := FileSize(indexFile) - idxHeaderLength - doCount;
  finally
    if not FFilesOpenedExternally then
      closeFile(indexFile);
  end;
end;

function TCustomDatabase.CountWithDeletedRecords : int64;
begin
  Result := 0;
  if FWorking then exit;
  if MainFilename = '' then exit;
  if not FFilesOpenedExternally then
    openIndexFile;
  try
    Result := FileSize(indexFile) - idxHeaderLength;
  finally
    if not FFilesOpenedExternally then
      closeFile(indexFile);
  end;
end;

function TCustomDatabase.doCount : int64;
begin
  seek(indexFile, idxCount);
  read(indexFile, Result);
end;

function TCustomDatabase.Count : int64;
begin
  Result := 0;
  if FWorking then exit;
  try
    if MainFilename = '' then exit;
    if not FFilesOpenedExternally then
      openIndexFile;
    try
      Result := doCount;
    finally
      if not FFilesOpenedExternally then
        closeFile(indexFile);
    end;
  except
  end;
end;

function TCustomDatabase.handleToIndex(handle : String) : int64;
var i : Integer;
begin
  Result := -1;
  try
    i := Pos('-', handle);
    if i = 0 then exit;
    inc(i);
    Result := StrToInt(copy(handle, i, Pos(':', handle) - i));
  except
  end;
end;

function TCustomDatabase.isHandleOk(handle : String) : Boolean;
var i, j : int64;          // mu form 0-12 haben (compresscount - index)
begin
  Result := false;
  try
    i := Pos('-', handle);
    if i = 0 then exit;
    j := Pos(':', handle);
    StrToInt(copy(handle, i + 1, j - (i + 1)));
    i := StrToInt(copy(handle, 1, i - 1));
    if i <> compresscount then exit;
    if HandleExt <> copy(handle, j, length(handle)) then exit;
    Result := true;
  except
  end;
end;

function TCustomDatabase.getHandle(idx : Integer) : String;
var i : Integer;
begin
  if idx = -1 then  // neues handle
    i := Filesize(indexFile) - idxHeaderlength - 1
  else
    i := idx;
  Result := IntToStr(compresscount) + '-' + IntToStr(i) + HandleExt;
end;

procedure TCustomDatabase.incCount;
var i : int64;
begin
  if FWorking then exit;
  i := doCount + 1;
  seek(indexFile, idxCount);
  write(indexFile, i);
end;

function TCustomDatabase.add(var handle : String; Data : String; noEncode : Boolean) : Boolean;
var i : int64;
begin
  Result := false;
  handle := '';
  if FWorking then exit;
  try
    openFiles;
    try
      Result := doadd(Data, -1, noEncode);
      if Result then begin
        handle := getHandle(-1);
        incCount;
      end;
    finally
      closeFiles;
    end;
  except
  end;
  if not Result then if err = 0 then err := 8;
end;

function TCustomDatabase.replace(var handle : String; Data : String; noEncode : Boolean) : Boolean;
begin
  Result := false;
  if FWorking then exit;
  if not isHandleOk(handle) then
    Result := add(handle, Data, noEncode)
  else begin
    Result := replacebyIndex(handleToIndex(handle), Data, noEncode);
  end;
end;

function TCustomDatabase.replacebyIndex(idx : int64; Data : String; noEncode : Boolean) : Boolean;
var b : Boolean;
    i : Integer;
    wasAlreadyDeleted : Boolean;
begin
  Result := false;
  if FWorking then exit;
  try
    openFiles;
    try
      i := getIndex(idx, b);
      if i <> -1 then begin
        Result := dodelete(idx, wasAlreadyDeleted);   // wegen trash aktual.
        b := doadd(Data, idx, noEncode);
        if ((not Result) or (wasAlreadyDeleted))  // bei wasAlreadyDeleted wurde eintrag von doadd auf ungelscht gesetzt und count mu inc. werden
        and (b) then
          incCount;
        Result := b and (Result or wasAlreadyDeleted);
      end;
    finally
      closeFiles;
    end;
  except
  end;
  if not Result then if err = 0 then err := 7;
end;

procedure TCustomDatabase.getAllHandles(l : TStringList);
var i, j : int64;
begin
  l.Clear;
  if FWorking then exit;
  if not OpenFiles then exit;
  try
    i := 0;
    seek(IndexFile, idxHeaderLength);
    while not eof(IndexFile) do begin
      read(IndexFile, j);
      if j > 0 then
        l.Add(getHandle(i));
      inc(i);  
    end;
  finally
    CloseFiles;
  end;      
end;


procedure TCustomDatabase.getAllIndexe(l : TStringList);
var j : int64;
    i, f : Integer;
    b : Boolean;
begin
  l.Clear;
  if FWorking then exit;
  if not OpenFiles then exit;
  try
    f := 0;
    seek(IndexFile, idxHeaderLength);
    while not eof(IndexFile) do begin
      read(IndexFile, j);
 //     if isMainfileIndexOk(j) then     // danach wird meist doget gemacht - dort wird geprft auf term
      l.Add(IntToStr(j))
   //   else
   //     inc(f);
    end;
(*    if f <> 0 then
      triggerDatabaseError;*)
  finally
    CloseFiles;
  end;
end;

function TCustomDatabase.getFirstRecord(var handle, Data : String; maxToRead : int64; noDecode : Boolean) : Boolean;
begin
  Result := getRecord(handle, Data, maxToRead, noDecode, true);
end;

function TCustomDatabase.getLastRecord(var handle, Data : String; maxToRead : int64; noDecode : Boolean) : Boolean;
begin
  Result := getRecord(handle, Data, maxToRead, noDecode, false);
end;

function TCustomDatabase.getRecord(var handle, Data : String; maxToRead : int64; noDecode,
  first : Boolean) : Boolean;
var i, j, k : int64;
    recordIsDeletedDummy : Boolean;  // wenn das gebraucht werden sollte, mu das auch nach getlastrecord etc. weitergereicht werden
begin
  Result := false;
  if FWorking then exit;
  try
    if MainFilename = '' then exit;
    if not FileExists(indexFileName) then exit;
    if not FFilesOpenedExternally then
      openIndexFile;
    try
      j := -1;
      if first then begin  // ansonsten ist letzte satz verlangt
        seek(indexFile, idxHeaderLength);
        while not eoF(indexFile) do begin
          read(indexFile, i);
          if i > 0 then begin
            j := i;
            break;
          end;
        end;
      end
      else begin
        k := FileSize(indexFile) - 1;
        if k >= idxHeaderLength then
        while k > -1 do begin
          seek(indexFile, k);
          read(indexFile, i);
          if i > 0 then begin
            j := i;
            break;
          end;
          dec(k);
        end;
      end;
    finally
      if not FFilesOpenedExternally then
        closeFile(indexFile);
    end;
    if j <> -1 then         // erste datensatznummer ist nur dummy - j ist schon index in mainfile
      Result := getbyIndex(-1, handle, Data, maxToRead, true, {false,} noDecode, j - 1, recordIsDeletedDummy) >= 0;
  except
  end;
end;


function TCustomDatabase.get(handle : String; var Data : String; maxToRead : int64;
  noDecode : Boolean; var recordIsDeleted : Boolean) : Boolean;
var dummy : String;
begin
  Result := false;
  if FWorking then exit;
  if not isHandleOk(handle) then exit;
  Result := getbyIndex(handleToIndex(handle), dummy, Data, maxToRead, false, {false,} noDecode, -1, recordIsDeleted) >= 0;
end;

function TCustomDatabase.getbyIndex(idx : int64; var handle : String; var Data : String;
  maxToRead : int64; onlyUndeleted{, noOpenFiles}, noDecode : Boolean;
  mainidx : int64; var recordIsDeleted : Boolean) : integer;

  procedure dogetbyIndex;
  begin
    Result := doGet(idx, Data, maxToRead, onlyUndeleted, mainidx, recordIsDeleted);
    if Result = 0 then begin
      if not noDecode then
        Data := DecodeData(Data);
      handle := getHandle(idx);
    end;
  end;

begin
  Result := -1;
  handle := '';
  if FWorking then exit;
  try
  {  if noOpenFiles then
      dogetbyIndex
    else begin}
      openFiles;
      try
        dogetbyIndex;
      finally
        closeFiles;
      end;
 //   end;
  except
  end;
  if Result <> -1 then if err = 0 then err := 6;
end;

function TCustomDatabase.delete(handle : String) : Boolean;
var k : int64;
begin
  Result := false;
  if FWorking then exit;
  if not isHandleOk(handle) then exit;
  Result := deletebyIndex(handleToIndex(handle));
end;

function TCustomDatabase.deletebyIndex(idx : int64) : Boolean;
var i : int64;
    wasAlreadyDeleted : Boolean;
begin
  Result := false;
  if FWorking then exit;
  try
    openFiles;
    try
      Result := dodelete(idx, wasAlreadyDeleted);
      if Result then begin
        i := doCount - 1;
        if i < 0 then i := 0;
        seek(indexFile, idxCount);
        write(indexFile, i);
      end;
    finally
      closeFiles;
    end;
  except
  end;
  if not Result then if err = 0 then err := 5;
end;

procedure TCustomDatabase.changePath(newFolder : String; doNotUseStateFile : Boolean);
var s : String;
begin
  try
    if FWorking then exit;
    if MainFilename = '' then exit;
    s := newFolder;
    if (s <> '') and (s[length(s)] <> '\') then s := s + '\';
    try closeStateFile; except end;
    MainFilename := s + ExtractFilename(MainFilename);  // hauptdateien sind nicht geffnet
    IndexFilename := s + ExtractFilename(IndexFilename);
    if not doNotUseStateFile then
      self.openStateFile;
  except
  end;
end;

procedure AssignAndResetInt64File(var f : int64File; Filename : String);
var fm : Byte;
begin
  fm := Filemode;
  try
    AssignFile(f, Filename);
    if FileGetAttr(Filename) and faReadOnly <> 0 then
      Filemode := fmOpenRead;
    Reset(f);
  finally
    Filemode := fm;
  end;
end;

procedure AssignAndResetByteFile(var f : ByteFile; Filename : String);
var fm : Byte;
begin
  fm := Filemode;
  try
    AssignFile(f, Filename);
    if FileGetAttr(Filename) and faReadOnly <> 0 then
      Filemode := fmOpenRead;
    Reset(f);
  finally
    Filemode := fm;
  end;
end;

function TCustomDatabase.undelete(handle : String) : Boolean;
begin
end;

function TCustomDatabase.undeletebyIndex(idx : int64) : Boolean;
begin
end;

end.
