unit udgtboard;

interface

uses Windows, SysUtils, Classes, Controls, Forms,
     stdctrls, ExtCtrls, Graphics, Dialogs;

const Namedll = 'dgtb100.dll';

      LNG_ENGLISH = 0;
      LNG_OTHER   = 1;

      ERR_DLL_NOT_FOUND = 0;
      ERR_DLL_NOT_INIT  = 1;
      ERR_COM_NOT_OPEN  = 2;

type

  TNormBoard = array[0..63] of Char;
  TProcessBoardEvents = procedure; stdcall;

  TComInit = function (ComPortName : PChar; BoardComMode : Integer; ProcessBoardEvents : TProcessBoardEvents) : Integer; stdcall;
  TComClose = function : Boolean; stdcall;
  TserveDGTBoard = procedure; stdcall;

  TOnErrorEvent = procedure (Sender: TObject;
    ErrorNumber : Integer; ErrorMessage : String) of object;

  TDGTBoard = class(TCustomControl)
  private
    Flib : THandle;
    FExeFolder : String;
    FPollthread : TThread;
    FLng : Integer;
    FComPort : Integer;
    FComPortName : array[0..6] of char;
    FPollMillis : Integer;

    FOnError : TOnErrorEvent;
    FOnBoardPositionChanged,
    FOnStableBoardPosition,
    FOnRotationChanged : TNotifyEvent;

    procedure setComPort(i : Integer);
    function  getPolling : Boolean;
    procedure setPollMillis(i : Integer);

    procedure FreeLib;

  protected
    procedure Paint; override;
    procedure Loaded; override;
    procedure doProcessBoardEvents(eventID : Integer);
    procedure doError(err : Integer);

  public
    Board : array[0..63] of integer;

    constructor Create(AComponent : TComponent);override;
    destructor  Destroy; override;
    procedure SetBounds(ALeft, ATop, AWidth, AHeight: Integer); override;

    procedure poll;
    function  Start : Integer;
    procedure Stop;
    function  BoardInit : Integer;
    procedure BoardClose;

    property Lng : Integer read FLng write FLng;
    property Polling : Boolean read getPolling;

  published
    property ComPort : Integer read FComPort write setComPort;
    property PollMillis : Integer read FPollMillis write setPollMillis;

    property OnError : TOnErrorEvent read FOnError write FOnError;
    property OnBoardPositionChanged : TNotifyEvent read FOnBoardPositionChanged write FOnBoardPositionChanged;
    property OnStableBoardPosition : TNotifyEvent read FOnStableBoardPosition write FOnStableBoardPosition;
    property OnRotationChanged : TNotifyEvent read FOnRotationChanged write FOnRotationChanged;
  end;

var
  BoardEventID   : ^Char;
  NormBoard      : ^TNormBoard;
  ComInit        : TComInit;
  ComClose       : TComClose;
  serveDGTBoard  : TserveDGTBoard;

  procedure ProcessBoardEvents; stdcall; export;

  procedure Register;

implementation

uses dgtb100;

var DGTBoard : TDGTBoard;

procedure Register;
begin
  RegisterComponents('Chess', [TDGTBoard]);
end;

//------------------
// Messages
//------------------

const MAX_LNG   = 2;
      MAX_MESS  = 5;

      Mess : array[0..MAX_MESS - 1, 0..MAX_LNG - 1] of String = (
      (('The dll ' + NameDll + ' was not found.')      , ('')),
      (('The dll ' + NameDll + ' could not be loaded.'), ('')),
      (('COM could not be opened.')                    , ('')),
      ((''), ('')),
      ((''), (''))
      );

function getMess(iMess, iLng : Integer) : String;
begin
  Result := 'Error';
  if ((iMess < 0) or (iMess > MAX_MESS - 1))
  or ((iLng < 0) or (iLng > MAX_LNG - 1)) then
    exit;
  Result := Mess[iMess, iLng];
end;

//--------------
// Callback
//--------------

procedure ProcessBoardEvents; stdcall; export;
begin
  if (BoardEventID = Nil) or (DGTBoard = Nil)
    then exit;
  DGTBoard.doProcessBoardEvents(ord(BoardEventID^));
end;

//------------------
// TPollthread
//------------------

type

  TPollThread = class(TThread)
  private
    FOwner : TDGTBoard;
  protected
    procedure Execute; override;
  public
    constructor Create(AOwner : TDGTBoard);
  end;

constructor TPollthread.Create(AOwner : TDGTBoard);
begin
  inherited Create(true);
  FOwner := AOwner;
  FreeOnTerminate := true;
  Resume;
end;

procedure TPollthread.Execute;
begin
  while not Terminated do begin
    synchronize(FOwner.Poll);
    sleep(FOwner.PollMillis);
  end;
end;

//----------------
// TDGTBoard
//----------------

constructor TDGTBoard.Create(AComponent : TComponent);
begin
  inherited Create(AComponent);
  DGTBoard := self;
  FExeFolder := ExtractFilePath(Application.Exename);
  ComPort := 3;
  PollMillis := 200;
end;

destructor TDGTBoard.Destroy;
begin
  DGTBoard := Nil;
  if Polling then
    Stop
  else
    BoardClose;
  inherited Destroy;
end;

procedure TDGTBoard.FreeLib;
begin
  FreeLibrary(FLib);
  FLib := 0;
end;

// Public functions

procedure TDGTBoard.poll;
begin
  if @serveDGTBoard <> Nil then
    serveDGTBoard;
end;

function TDGTBoard.BoardInit : Integer;
begin
  Result := 0;
  Flib := LoadLibrary(PChar(Namedll));
  if FLib <> 0 then begin

    BoardEventID  := GetProcAddress(FLib, '_BoardEventID');
    NormBoard     := GetProcAddress(FLib, '_NormBoard');
    @ComInit      := GetProcAddress(FLib, 'ComInit');
    @ComClose     := GetProcAddress(FLib, 'ComClose');
    @serveDGTBoard:= GetProcAddress(FLib, 'serveDGTBoard');

    if (@BoardEventID <> Nil) and (@ComInit <> Nil)
    and (@ComClose <> Nil)  and (@serveDGTBoard <> Nil)
    then begin
      if ComInit(@FComPortName, INITUPDATESAFE, ProcessBoardEvents) <> 0 then begin
        FreeLib;
        doError(ERR_COM_NOT_OPEN);
        Result := ERR_COM_NOT_OPEN;
      end;
    end
    else begin
      FreeLib;
      doError(ERR_DLL_NOT_INIT);
      Result := ERR_DLL_NOT_INIT;
    end;
  end
  else begin
    if not FileExists(FExeFolder + NameDLL) then begin
      doError(ERR_DLL_NOT_FOUND);
      Result := ERR_DLL_NOT_FOUND;
    end
    else begin
      doError(ERR_DLL_NOT_INIT);
      Result := ERR_DLL_NOT_INIT;
    end;
  end;
end;

procedure TDGTBoard.BoardClose;
begin
  if FLib <> 0 then begin
    if @ComClose <> Nil then
      ComClose
    else
      doError(ERR_DLL_NOT_INIT);
    FreeLib;
  end;
end;

procedure TDGTBoard.Stop;
begin
  if Assigned(FPollThread) then begin
    FPollthread.Terminate;
    FPollthread := Nil;
    BoardClose;
  end;
end;

function TDGTBoard.Start : Integer;
begin
  if not Assigned(FPollThread) then begin
    Result := BoardInit;
    if Result = 0 then
     FPollThread := TPollThread.Create(Self);
  end
  else Result := 0;
end;

// Properties

function TDGTBoard.getPolling : Boolean;
begin
  Result := Assigned(FPollThread);
end;

procedure TDGTBoard.setPollMillis(i : Integer);
begin
  if i < 1 then
    exit;
  FPollMillis := i;
end;

procedure TDGTBoard.setComPort(i : Integer);
var sCom : String;
begin
  if i < 1 then
    exit;
  FComPort := i;
  sCom := 'COM' + IntToStr(i);
  StrPCopy(@FComPortName, sCom);
end;

// Events

procedure TDGTBoard.doProcessBoardEvents(eventID : Integer);
Var i : Integer;
begin
  if Normboard <> Nil then
    for i := 0 to 63 do
      Board[i] := ord(NormBoard[i]);
  case eventID of
    BOARDPOSITIONCHANGED:
      if assigned(FOnBoardPositionChanged) then FOnBoardPositionChanged(self);
    STABLEBOARDPOSITION :
      if assigned(FOnStableBoardPosition) then FOnStableBoardPosition(self);
    ROTATIONCHANGED     :
      if assigned(FOnRotationChanged) then FOnRotationChanged(self);
  end;
end;

procedure TDGTBoard.doError(err : Integer);
begin
  if Assigned(FOnError) then
    FOnError(self, err, getMess(err, FLng));
end;

// component functions

var
  aBitmap: TBitmap;
  Loaded: Boolean;

procedure LoadBitmap;
begin
  if Loaded then exit;
  Loaded := True;
  if aBitmap = nil then aBitmap := TBitmap.Create;
  try
    aBitmap.LoadFromResourceName(HInstance, 'TDGTBOARD');
  except
 //   on E:Exception do ShowMessage(E.Message);
  end;
end;

procedure TDGTBoard.Paint;
begin
  if csDesigning in ComponentState then begin
    Canvas.TextOut(0, 0, 'DGT');
    Canvas.StretchDraw(Canvas.ClipRect, aBitmap);
  end;
  inherited Paint;
end;


procedure TDGTBoard.Loaded;
begin
  if csDesigning in ComponentState then LoadBitmap
  else SetBounds(0,0,0,0);
  inherited Loaded;
end;

procedure TDGTBoard.SetBounds(ALeft, ATop, AWidth, AHeight: Integer);
begin
  if csDesigning in ComponentState then
  begin
    AWidth := 25;
    AHeight := 25;
  end;
  inherited SetBounds(ALeft, ATop, AWidth, AHeight);
end;


end.
