unit Notation;

(*
The moves of the notation are saved in the data type "TZug".
Please see in the unit "SchachProcs" the various fields of this data type.
Also, see there the type "TSchachbrett" (especially the constant "AusgangsStellung")
and the values of the various pieces ("wk", "wd", "wt", etc.)
The numbers of squares can be converted from 0..63 (a1..h8) by the function "c()" to the corresponding number on TSchachbrett".
*)

{$B-}

interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
  Forms, Dialogs, ExtCtrls, StdCtrls, Buttons, Menus, clipbrd,
  SchachProcs;

const
  {Attr2 - can save only one value (!, ?, !! etc)}
  nAusruf = 1;
  n2Ausruf = 2;
  nFrage = 3;
  n2Frage = 4;
  nAusrufFrage = 5;
  nFrageAusruf = 6;

  {infAttr - you must "FinfCharzugzwang", etc. initialize with the corresponding characters to adjust it according to a special chess font}
  infzugzwang                 =22;
  infeinzigerzug              = 8;
  infplusminus                =18;
  infplusleichtminus          =16;
  infplusgleich               =14;
  infgleich                   =11;
  infunklar                   =13;
  infgleichplus               =15;
  infleichtminusplus          =17;
  infminusplus                =19;
  infkompensation             =44;
  infgegenspiel               =132;
  infinitiative               =36;
  infangriff                  =40;
  infzeitnot                  =138;
  infvorsprung                =32;
  infneuerung                 =146;
  infmitderidee               =140;
  infBesserIst                =142;

  // the following is only necessary to save more than one symbol per move to ccf databases
  infgegenspielEx             = 63;
  infzeitnotEx                = 62;
  infneuerungEx               = 61;

  // the following default character must NOT be contained in the normal notation displayed

  defaultinfCharzugzwang       : Char = #105;
  defaultinfChareinzigerzug    : Char = #106;
  defaultinfCharplusminus      : Char = #107;
  defaultinfCharplusleichtminus: Char = #108;
  defaultinfCharplusgleich     : Char = #109;
  defaultinfChargleich         : Char = #110;
  defaultinfCharunklar         : Char = #111;
  defaultinfChargleichplus     : Char = #112;
  defaultinfCharleichtminusplus: Char = #113;
  defaultinfCharminusplus      : Char = #114;
  defaultinfCharkompensation   : Char = #115;
  defaultinfChargegenspiel     : Char = #116;
  defaultinfCharinitiative     : Char = #117;
  defaultinfCharangriff        : Char = #118;
  defaultinfCharzeitnot        : Char = #119;
  defaultinfCharvorsprung      : Char = #121;
  defaultinfCharneuerung       : Char = #122;
  defaultinfCharmitderidee     : Char = #124;
  defaultinfCharBesserIst      : Char = #126;



type

     TPartiedaten = record
       EventData : TEventData;
       Stellung : TSchachbrett;
       Zug : TZug;       {it contains info about the position (castling possibilities) as well as info about the move (from/to square)}
       Zugnummer : word;
       Kommentar : string; {if a comment is written before the first move}
     end;

     PBaumList = ^TBaumList;
     PBaum = ^TBaum;
     TBaumlist = record B : PBaum;  BL, rBL : PBaumList; end;
     PPBaumList = ^Pointer;

     TBaum = record
       B, rB : PBaum;
       VarWurzel : PBaumList;
       Zug : TZug;
       Kommentar : String;
       ss, sl : longint;
       Zugnummer : integer; {Partiedaten.Zugnummer + current move number}
       hidden : Boolean;
     end;

     PRettBrett = ^TRettBrett;
     TRettBrett = record
       stellung : TSchachbrett;
       Zug : TZug;
       Wurzel : PBaumList;
       Baum : PBaum;
     end;

     TNotatScrollbar = class(TScrollbar)
       procedure Change; override;
     end;

     TNotatPaintbox = class(TPaintbox)
       procedure paint; override;
       procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y:Integer); override;
     end;

     TPiecesNumber = array[-6..6] of integer;
     TPiecesFilter = class(TObject)
     private
       function isItemEmpty(i : Integer) : Boolean;
     public
       total, min, max : TPiecesNumber;
       checkPiece : array[-6..6] of Boolean;
       procedure Clear;
       function isEqual : Boolean;
       function isEmpty : Boolean;
     end;

     TMoveChangeEvent = procedure (Sender: TObject; Zug : TZug; Stellung : TSchachbrett; NextMove : Boolean) of object;
     TErrorEvent  = procedure (Sender: TObject; Fehlermeldung : string) of object;
     TFischerInfoEvent = procedure (Sender: TObject; FischerInfo : TFischerInfo) of object;

     TNodeInfo = class
     private
       Baumindex : PBaum;
       FLChild, FRChild: Boolean;
       function getMove : TZug;
       function getComment : String;
     public
       property LChild : Boolean read FLChild;
       property RChild : Boolean read FRChild;
       property Move : TZug read getMove;
       property Comment : String read getComment;
     end;


     TNotation = class(TCustomPanel)
     constructor create(AOwner : TComponent);override;
     destructor destroy;override;
     private
       Fueberschrift : boolean;
       FEnglish : Byte;
       FOnMoveChange : TMoveChangeEvent;
       FOnFischerInfo : TFischerInfoEvent;
       FOnError : TErrorEvent;
       FGeaendert : Boolean;
       FOn50MovesRule : TNotifyEvent;
       FOn3Repetition : TNotifyEvent;
       FOnGameChanged : TNotifyEvent;
       FPartiedaten : TPartiedaten;

       FFontHeadLine, FFont0Var, FFontComment : TFont;
       FAutoBold : Boolean;
       FFont1Var : TFont;
       FFont2Var : TFont;
       FIndent0Var, FIndent1Var, FIndent2Var : word;
       FZeilenAbstand : word;
       FBevelwidthNotat, FBorderwidthNotat : word;
       FFRC : PFischerInfo;

       FLockNotatUpdate : Boolean;
       Fcapture : Boolean;

       VarDialog : TForm;
       Einfuegen, Ueberschreiben, Abbruch, Replace : TButton;
       lZuege : TListBox;

       (*zugnummer, *)ebene : integer;
       klammerzu, klammerauf : Boolean; {fr klettere}
       notat, ret : string;

       NoStellung : PSchachbrett;
       NoZug : PZug;
       inBetracht : array[1..10] of integer; {in Betracht knnen hchstens 10 gleiche Figuren kommen}
       inBetrachtindex : byte;
       Wurzelindex : PBaumList;
       Wurzel : PBaumList;
       Wurzelliste : PBaumList; {fr Verzweigungen}
       Baum : PBaum;
       Baumindex : PBaum;

       ZBaumlist : PPBaumList;   {Zeiger auf Zeiger auf Wurzelliste - zum Var. lschen}
       inHauptvariante : Boolean;  {in dem Fall darf nicht aufgewertet, aber nach vorn geschnitten werden}
       uebergabe : integer;  {fr Variantenwahl}
       gemeldet : Boolean;
       udBi : PBaum;         {fr UnDo}
       udwurzel : PBaumlist;
       ud : integer; {0 = kein UnDo mglich, -1 = Schneiden hinten, -2 = Schneiden vorn, 1.. = Lschen (Var.Nr)}
       udPartiedaten : TPartiedaten; {bei Schneiden nach vorn ndern die sich}

       pb : TNotatPaintbox;
       SLeiste : TNotatScrollbar;
       wrap, commwrap : TStringlist;
       bf : TBitmap;
       ZugRect : TRect;
       rNotatBaumindex : PBaum;
       rFont : TFont;   (**)
       Texthoehe : integer;
       FirstLine, LastLine : integer;
       UeberschriftEnde : integer;
       FontNoVar : TFont;

       FUserDefinedPieces : Boolean;
       PiecesNames : array[0..5] of string;
       Saving : Boolean;  { bei Speichern nicht ev. benutzerdef. Figurenbez. benutzen}

       checkPos : Boolean;  { fr containsPos}
       checkPosBoard : TSchachBrett;

       FErrorMoves : String;
       FcheckMoves : Boolean;

       FMarkedMoveColor : TColor;
       FcorrChessMode, FcorrChessMode1 : Boolean;

       FinfCharzugzwang        ,
       FinfChareinzigerzug     ,
       FinfCharplusminus       ,
       FinfCharplusleichtminus ,
       FinfCharplusgleich      ,
       FinfChargleich          ,
       FinfCharunklar          ,
       FinfChargleichplus      ,
       FinfCharleichtminusplus ,
       FinfCharminusplus       ,
       FinfCharkompensation    ,
       FinfChargegenspiel      ,
       FinfCharinitiative      ,
       FinfCharangriff         ,
       FinfCharzeitnot         ,
       FinfCharvorsprung       ,
       FinfCharneuerung        ,
       FinfCharmitderidee      ,
       FinfCharBesserIst       : Char;

       pf : TPiecesFilter;
       checkStellung : Boolean;
       FCtrlPressed, FAltPressed : Boolean;

       niList : TList;

       hideMoves, showLastMoves, hideLastMoves, showMoves : Boolean;
       NextToSolve : PBaum;
       allSolved : Boolean;

       procedure ZugVor;
       procedure ZugZurueck(z : TZug);
       procedure VarEinfuegen;
       procedure BaumEinfuegen(z : TZug);
       function  sucheKoenig(wo : byte; schachpruefung : boolean) : boolean;
       function  sucheDame(wo : byte; schachpruefung : boolean) : boolean;
       function  sucheTurm(wo : byte; schachpruefung : boolean) : boolean;
       function  sucheLaeufer(wo : byte; schachpruefung : boolean) : boolean;
       function  sucheSpringer(wo : byte; schachpruefung : boolean) : boolean;
       function  sucheBauer(wo : byte; schachpruefung : boolean) : boolean;
       procedure loescheBaum(Pbl1 : PPBaumList);
       procedure klettere(bl1 : PBaumlist; neux : Boolean);
       procedure klettereEx;
       function  retract(NotatSelstart : longint) : Boolean;
       function  notiereZug(z : TZug) : string;
       procedure Dialog_ini;
       procedure Varianten_anzeigen(keineZuege : Boolean);
       procedure loesche_UD;
       function  SchreibePartie(ToClipBoard : Boolean; Var PGNString : string; onlyHeader : Boolean) : integer;

       procedure SetUberschrift(w : Boolean);
       procedure SetGeaendert(w : Boolean);
       function  GetEnglish : Boolean;
       procedure SetEnglish(w : Boolean);
       procedure SetFontHeadLine(f : TFont);
       procedure SetFontComment(f : TFont);
       procedure SetFont0Var(f : TFont);
       procedure SetFont1Var(f : TFont);
       procedure SetFont2Var(f : TFont);
       procedure SetIndent0Var(i : word);
       procedure SetIndent1Var(i : word);
       procedure SetIndent2Var(i : word);
       procedure SetZeilenabstand(i : word);
       procedure SetBevelwidthNotat(i : word);
       procedure SetBorderwidthNotat(i : word);

       procedure LoescheNotatImFenster;
       procedure KD(Sender: TObject; var Key: Word; Shift: TShiftState);
       procedure MD(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);

       procedure DoIniPartie;
       procedure SetPbCaret;
       procedure WMSize(var Msg: TMsg); message WM_SIZE;
       procedure DoMouseDown(x, y : integer);
       function  GetNumber(index : integer) : longint;
       function  GetTextWidth(ALine, a, l : integer) : integer;
       procedure SetFontsInLine(sl : TStringlist; a, l : integer; f, fc : TFont);
       function  DeleteCommKlammer(szeile : string) : string;
       function  NumberToString(nr : longint) : string;
       function  GetIndent(Font : TFont) : Integer;
       procedure PaintPb;
       procedure SetNotatBounds(pos : longint);
       function  GetLineNumber(b : PBaum; Var Linestart : longint) : integer; {Zeilennummer von der ersten sichtbaren an}
       function  IsNewScrollbarPos(b : PBaum) : integer;
       procedure AdjustPb(b : PBaum);
       procedure checkFischerPos;
       function doschneide(wie : Boolean; dummy : TObject) : Boolean;
       function  doAddZug(aZug : TZug) : Integer;

       procedure clearniList;
       procedure refreshniList;
       procedure createAndRefreshniList;

     protected
       procedure KeyDown(Var Key: Word; Shift: TShiftState); override;
       procedure loaded; override;
     public
       (*
       The color uses if a move is marked with "nErster/nLetzter" (appears with ">e4<" in the notation)
       Must be set in "Attr1" of TZug in the method "AddZug(TZug)"
       This is NOT the currently active move!
       *)
       property MarkedMoveColor : TColor read FMarkedMoveColor write FMarkedMoveColor default clYellow;
       (*
       Should be false
       *)
       property corrChessMode : Boolean read FcorrChessMode write FcorrChessMode;
       property corrChessMode1 : Boolean read FcorrChessMode1 write FcorrChessMode1;

       (*
       when an error while reading a pgn game occured,
       this string contains the moves which could not be read
       *)
       property ErrorMoves : String read FErrorMoves;
       (*
       if true, each move will be checked for legality while loading a pgn game
       *)
       property checkMoves : Boolean read FcheckMoves write FcheckMoves;

       (*
       the data of pgn header
       *)
       property Partiedaten : TPartiedaten read FPartiedaten write FPartiedaten;
       (*
       if true, the notation will not be repainted
       *)
       property LockNotatUpdate : Boolean read FLockNotatUpdate write FLockNotatUpdate;
       (*
       tells the notation a new starting position
       Please note that here following information of TZug must be defined:
       TRochademoeglichkeiten; epFeld, amZug (the info about position).
       The remaining fields (the info about move) may be undefined here
       *)
       procedure NeueStellung(Stellung : TSchachbrett; Zug : TZug; z : integer);
       (*
       the information for Fischer-Random-Chess (positions of kings and rooks)
       Please see the demo program for the use of FRC
       *)
       property  FRC : PFischerInfo read FFRC write FFRC;
       (*
       a new move can be added by the following function.
       Here ALL fields must be defined (unlike in "NeueStellung")
       *)
       function  AddZug(aZug : TZug) : Boolean;
       (*
       Only for corr chess mode
       *)
       function  AddZugEx(aZug : TZug) : Integer;
       (*
       Cuts a variation: true = after the marked move
                         false= before the marked move
       *)
       function  schneide(wie : Boolean) : Boolean;
       (*
       Only for corr chess mode
       *)
       function  schneideEx(wie : Boolean; dummy : TObject) : Boolean;
       (*
       promotes a variation
       *)
       procedure aufwerten;
       (*
       deletes a variation
       *)
       procedure loeschen;
       (*
       undo the last deletion
       *)
       procedure UnDo;
       (*
       writes # (the notation cannot the chess rules)
       *)
       procedure Matt;
       (*
       a move back
       *)
       function  Goback : Boolean;
       (*
       next move
       *)
       function  Goahead : Boolean;
       (*
       next move without displaying the variation choice
       *)
       function  Goahead_inHauptVariante(Var von, nach, UFigur, RochadeArt : shortint) : Boolean; {ohne Variantendialog}
       (*
       jumps to a previous variation
       *)
       function  GoPreviousVar : Boolean;
       (*
       jumps to a next variation
       *)
       function  GoNextVar : Boolean;
       (*
       jumps to the start of game
       *)
       function  GoToStart : Boolean;
       (*
       jumps to the end of game
       *)
       function  GoToEnd : Boolean;
       (*
       returns true if the last move is currently marked
       *)
       function  isLastPos : Boolean;

       (*
       loads the game in PGNString
       *)
       function  SetGame(PGNString : String) : integer;
       (*
       delivers the game as pgn
       *)
       function  GetGame : string;
       (*
       delivers only the header in pgn notation
       *)
       function  GetHeader : string;
       (*
       only for corr chess mode
       *)
       function  getGameExx(whiteCondMoves : Boolean; slMainVar : TStringlist;
         exPGNHeader : Boolean; bias, timeformat : String; limitPGNLength : Boolean;
         writeGameID : String; ICCF, pgnInICCFTable : Boolean) : string;

       (*
       gets the comment of current move
       *)
       function  GetKommentar : string;
       (*
       sets the comment of current move
       *)
       procedure SetKommentar(k : string);
       (*
       initializes a new game
       *)
       procedure iniPartie;
       (*
       copies the game to the clipboard as pgn
       *)
       procedure CopyGameToClipboard;
       (*
       pastes a pgn game from the clipboard
       *)
       function  PasteGameFromClipBoard : Boolean;
       (*
       sets !, ?, !? for a move
       *)
       procedure SetZugBewertung(s : string);
       (*
       repaints the notation
       *)
       procedure NotationRefresh;
       (*
       gets the notation as displayed (no pgn)
       *)
       function  GetNotation : string;
       (*
       gets the position at a index of the notation string (rarely used)
       *)
       function  GetPositionAt(Stelle : integer) : TSchachbrett;  {Stelle ist Index im Notations-String (wie Mausklick darauf)}
       (*
       gets a record of the data for the starting position, such as who moves at first and e.p. and castling possibilities
       *)
       function  GetFirstPosInfo : PZug;
       (*
       he current move
       *)
       function  GetMove : PZug;
       (*
       the current position
       *)
       function  GetPosition : PSchachbrett;
       (*
       sets new names of pieces
       *)
       procedure SetPiecesNames(sK, sQ, sR, sB, sN, sP : String);
       (*
       gets the current names of pieces
       *)
       procedure GetPiecesNames(Var sK, sQ, sR, sB, sN, sP : String);
       (*
       reset the names to normal pgn notation
       *)
       procedure DelPiecesNames;
       (*
       gets the current move as displayed in the notation (not pgn!)
       *)
       function  getCurrentMoveFromNotation : string;
       (*
       gets the current move as e2e4, e7e8Q, etc.
       *)
       function  getCurrentMoveAse2e4 : string;
       (*
       the root of the move tree (rarely used)
       *)
       function  getRoot : PBaumList;
       (*
       a special function for a program; cannot be explained quickly
       *)
       function  ImportNextMove(OtherNotationRoot : PBaumList) : TZug;  {vergleicht OtherNottaion mit eigener und gibt nchsten Zug zurck, falls eigene mind. einen weiteren enthlt}
       (*
       removes all moves
       *)
       procedure Clear;
       (*
       checks whether any moves are added/available
       *)
       function  Empty : Boolean;
       (*
       returns the number of plies
       *)
       function  PlyCount : Integer;
       (*
       returns the Date header
       *)
       function  getHeaderDate : TDate;
       (*
       returns true if white must move next
       *)
       function  WhitesTurn : Boolean;
       (*
       true if the last move is black
       *)
       function  WhitesTurnOnEnd : Boolean;
       (*
       returns the next move number (not then ply number, in the starting position 1)
       *)
       function  GetNextMoveNumber : Integer;
       (*
       returns the next move number as string (if black has to move: '1...', for instance)
       *)
       function  GetNextMoveNumberAsString : string;
       (*
       returns true if the game contains the position saved with "setCheckPosBoard"
       *)
       function  containsPos(apf : TPiecesFilter; chckpos : Boolean) : Boolean;
       (*
       setups a search position
       *)
       procedure setCheckPosBoard(stllg : TSchachbrett);
       (*
       the currently marked move belongs to a sub variation
       *)
       function  inVariation : Boolean;
       (*
       gets the main variation in the form "3552 6457 ...".
       The number are the square numbers of TSchachbrett.
       No promotion info is provided!
       *)
       function  getOnlyMainVariation : String;
       (*
       Only for corr chess mode
       *)
       procedure createWurzel;
       (*
       Returns the index in the notation string ("getNotation) of the marked move
       *)
       function getIdxOfMarkedMove : Integer;
       (*
       Delivers the move without number or comments
       *)
       function  notiereRawZug(z : TZug) : string;
       (*
       Can be used to access the move properties directly; Nil if no move is selected
       *)
       function  getBaumindex : Pbaum;
       (*
       true if the current move belongs to a variation starting from the main variation
       *)
       function  IsVarErsterOrdnung : Boolean;
       (*
       true if there are moves marked with "nErster (>) or nLetzter (<) in "TZug.Attr1"
       (this is NOT the currently active move)
       *)
       function  AreMovesMarked : integer;
       (*
       Deletes all markers of marked moves (">e4<" to "e4")
       *)
       procedure deleteMarkers;
       (*
       Removes all comments
       *)
       procedure removeComments;
       (*
       Removes all variations
       *)
       procedure removeVariations;
       (*
       The paramter is a string representing the values of the symbols of Chess Informant in a special chess font
       Please see the implementation of this method to get the order of the characters
       *)
       procedure SetInformantSigns(InformantSigns : String);

       (*
       This ia called by TChessDatabase to compress the game before saving
       *)
       function getCompressedGame : string;
       (*
       Not yet implemented
       *)
       procedure setCompressedGame(aGame : String);
       (*
       Delivers TZug.Attr2 as string ("!", "!?", etc.)
       *)
       function Attr2toString(i : shortint) : String;
       (*
       Delivers the character of the symbol of Chess Informant as defined with "SetInformantSigns".
       In "TZug.infAttr" it is saved as a special code,
       and with this method you can get the corresponding character
       *)
       function infAttrtoString(i : byte) : String;
       (*
       See Attr2toString, but this method returns the string as written to the pgn notation ("$12", etc.)
       *)
       function Attr2toPGNString(i : shortint) : String;
       (*
       See infAttrtoString, but this method returns the string as written to the pgn notation ("$12", etc.)
       *)
       function infAttrtoPGNString(i : byte) : String;
       (*
       Sets a symbol of the Chess Informant.
       Here you must use one of the constants "infzugzwang", etc.
       *)
       procedure SetInfAttr(a : byte);
       (*
       Scrolls the notation to the end
       ("GoToEnd" only displays the last move - some variations could follow)
       *)
       procedure ScrollToEnd;

       function  getMoveIndex : Integer;
       function setMoveAtIndex(idx : Integer) : Boolean;
       function setMoveAtBaumIndex(baumidx : PBaum) : PBaum;
       function setMoveAtBaumIndexEx(baumidx : PBaum) : PBaum;

       (*
       functions for solving mode
       *)
       procedure hideNextMoves;
       procedure showNextMoves;
       procedure hideAllMoves;
       procedure showAllMoves;
       function  Goahead_inHauptVariante_and_showMove(var SolvingComplete : Boolean; var NextHiddenMove : PBaum;
         Var von, nach, UFigur, RochadeArt : shortint) : Boolean;

       (*
       the notation was modified
       *)
       property geaendert : Boolean read FGeaendert write SetGeaendert;

       (*
       true if the last added move is a capture
       *)
       property capture : Boolean read Fcapture;
       (*
       You can set it before using "AddZug()" to force a new variation
       *)
       property CtrlPressed : Boolean write FCtrlPressed;
       (*
       You can set it before using "AddZug()" to force that the remeining variation is overwritten
       *)
       property AltPressed : Boolean write FAltPressed;

       (*
       Some functions to access the variation tree
       *)
       function getNodeAt(index : Integer) : TNodeInfo;
       function setPosOfNodeAt(index : Integer) : Boolean;
       function NodesCount : Integer;

       (*
       This must be adjusted for the font used for the notation
       *)
       property infCharzugzwang       : Char read FinfCharzugzwang        write FinfCharzugzwang        ;
       property infChareinzigerzug    : Char read FinfChareinzigerzug     write FinfChareinzigerzug     ;
       property infCharplusminus      : Char read FinfCharplusminus       write FinfCharplusminus       ;
       property infCharplusleichtminus: Char read FinfCharplusleichtminus write FinfCharplusleichtminus ;
       property infCharplusgleich     : Char read FinfCharplusgleich      write FinfCharplusgleich      ;
       property infChargleich         : Char read FinfChargleich          write FinfChargleich          ;
       property infCharunklar         : Char read FinfCharunklar          write FinfCharunklar          ;
       property infChargleichplus     : Char read FinfChargleichplus      write FinfChargleichplus      ;
       property infCharleichtminusplus: Char read FinfCharleichtminusplus write FinfCharleichtminusplus ;
       property infCharminusplus      : Char read FinfCharminusplus       write FinfCharminusplus       ;
       property infCharkompensation   : Char read FinfCharkompensation    write FinfCharkompensation    ;
       property infChargegenspiel     : Char read FinfChargegenspiel      write FinfChargegenspiel      ;
       property infCharinitiative     : Char read FinfCharinitiative      write FinfCharinitiative      ;
       property infCharangriff        : Char read FinfCharangriff         write FinfCharangriff         ;
       property infCharzeitnot        : Char read FinfCharzeitnot         write FinfCharzeitnot         ;
       property infCharvorsprung      : Char read FinfCharvorsprung       write FinfCharvorsprung       ;
       property infCharneuerung       : Char read FinfCharneuerung        write FinfCharneuerung        ;
       property infCharmitderidee     : Char read FinfCharmitderidee      write FinfCharmitderidee          ;
       property infCharBesserIst      : Char read FinfCharBesserIst       write FinfCharBesserIst           ;

     published
       (*
       displays the game header
       *)
       property Ueberschrift : Boolean read FUeberschrift write SetUberschrift;
       (*
       displays pgn notation (otherwise German names of pieces are used)
       *)
       property English : Boolean read GetEnglish write SetEnglish;
       (*
       font for the header (please note, there are some problems with various fonts yet)
       *)
       property FontHeadLine : TFont read FFontHeadLine write SetFontHeadLine;
       (*
       font for the comments
       *)
       property FontComment : TFont read FFontComment write SetFontComment;
       (*
       font for main variation
       *)
       property Font0Var : TFont read FFont0Var write SetFont0Var;
       (*
       font for the first variation
       *)
       property Font1Var : TFont read FFont1Var write SetFont1Var;
       (*
       font for the second variation
       *)
       property Font2Var : TFont read FFont2Var write SetFont2Var;
       (*
       Indent for the main variation
       *)
       property Indent0Var : word read FIndent0Var write SetIndent0Var;
       (*
       Indent for the first variation
       *)
       property Indent1Var : word read FIndent1Var write SetIndent1Var;
       (*
       Indent for the second variation
       *)
       property Indent2Var : word read FIndent2Var write SetIndent2Var;
       (*
       writes the main variation boldly only if sub variations exist
       *)
       property AutoBold : Boolean read FAutoBold write FAutoBold;
       (*
       space between the lines
       *)
       property Zeilenabstand : word read FZeilenabstand write SetZeilenabstand;


       property Color;
       property BevelOuter;
       property BevelInner;
       property BevelWidthNotat : word read FBevelwidthNotat write SetBevelwidthNotat;
       property BorderWidthNotat : word read FBorderwidthNotat write SetBorderwidthNotat;
       property Align;
       property Ctl3D;
       property Cursor;
       property DragCursor;
       property DragMode;
       property Enabled;
       property FullRepaint;
       property HelpContext;
       property Hint;
       property ShowHint;
       property PopupMenu;
       property Visible;

      { property OnClick;  nur auf Panel bezogen - nicht auf Paintbox
       property OnDblClick;
       property OnDragDrop;
       property OnDragOver;
       property OnEndDrag;
       property OnMouseDown;
       property OnMouseMove;
       property OnMouseUp;
       property OnStartDrag; }
       property OnResize;

       (*
       a new move was selected (by the user)
       *)
       property OnMoveChange : TMoveChangeEvent read FOnMoveChange write FOnMoveChange;
       (*
       an error while reading the pgn file occurred
       *)
       property OnError : TErrorEvent read FOnError write FOnError;
       (*
       the fifty-moves-rule happened
       *)
       property On50MovesRule : TNotifyEvent read FOn50MovesRule write FOn50MovesRule;
       (*
       the same position occurred three times
       *)
       property On3Repetition : TNotifyEvent read FOn3Repetition write FOn3Repetition;
       (*
       the notation was modified
       *)
       property OnGameChanged : TNotifyEvent read FOnGameChanged write FOnGameChanged;
       (*
       a Fischer-Random-Chess game was read.
       Set the FRC property then
       *)
       property OnFischerInfo : TFischerInfoEvent read FOnFischerInfo write FOnFischerInfo;

     end;

const
  {or-Werte fr Attr1}

  nErster  = $1;
  nLetzter = $2;
  nWarVorherLetzter = $4;  // nur fr UnDo - hat sonst keine Bedeutung

  PGNErster  = '$998';
  PGNLetzter = '$999';

{or-Werte fr Attr1}

  nLetter = $10;  {in Notation mu ein Buchstabe frs Ausgangsfeld eingefgt werden - 2 Figuren konnten aufs Zielfeld}
  nDigit = $20;
  nSchach = $40;
  nMatt = $80;

  nFischerrochade = $8;


implementation

uses ChessDatabase;

const
 {fr sucheDame, sucheTurm usw.}
  LZ : array[1..4] of integer = (-9,-11,9,11);  { Laeufer Richtungen }
  TZ : array[1..4] of integer = (-1,10,1,-10);  { Turm               }
  SZ : array[1..8] of integer = (19,21,12,-8,-19,-21,-12,8);   { Springer           }
  DZ : array[1..8] of integer = (-9,-11,9,11,-1,10,1,-10);
  KZ : array[1..8] of integer = (-9,-11,9,11,-1,10,1,-10);
  figurnamen : array[1..12] of string[1] = ('','T','L','S','D','K','','R','B','N','Q','K');
  sonderzeichen  = ['(',')','[',']','{','}',#$A,#$D,';',' '];

procedure TPiecesFilter.Clear;
begin
  FillChar(total, SizeOf(total), 0);
end;

function TPiecesFilter.isEqual : Boolean;
var i : Integer;
begin
  Result := false;
  For i := -5 to 5 do
  if i <> 0 then begin
    if checkPiece[i] then
      if (total[i] < min[i]) or (total[i] > max[i]) then
        exit;
  end;
  Result := true;
end;

function TPiecesFilter.isItemEmpty(i : Integer) : Boolean;
begin
  Result := not checkPiece[i];
end;

function TPiecesFilter.isEmpty : Boolean;
var i : Integer;
begin
  Result := false;
  For i := -5 to 5 do if i <> 0 then
    if checkPiece[i] then
      exit;
  Result := true;
end;

function TNodeInfo.getMove : TZug;
begin
  Result := Baumindex^.Zug;
end;

function TNodeInfo.getComment : String;
begin
  Result := Baumindex^.Kommentar;
end;


(*procedure TNotation.WndProc(var MsgRec: TMessage);
begin
  case MsgRec.Msg of
  end;
  inherited WndProc(MsgRec);
end;*)

procedure TNotation.WMSize(var Msg: TMsg);
Var i : integer;
begin
  i := Borderwidth + Bevelwidth;
  pb.left := i;
  pb.Top := i;
  pb.width := self.width - SLeiste.Width - (i shl 1);
  pb.height := self.height - (i shl 1);
  SLeiste.left := pb.left + pb.width;
  SLeiste.Top := pb.Top;
  SLeiste.height := pb.height;
  SLeiste.LargeChange := (SLeiste.height div Texthoehe) - 1;
  Klettere(wurzel, true);
end;

procedure TNotation.NotationRefresh;
begin
  klettere(wurzel, true);
end;

function TNotation.GetNotation : string;
begin
  Result := Notat;
end;

function  TNotation.GetPositionAt(Stelle : integer) : TSchachbrett;
begin
  Result := Partiedaten.Stellung;
  if Retract(Stelle) then begin
    Result := NoStellung^;
    if assigned(FOnMoveChange) then FOnMoveChange(self, NoZug^, NoStellung^, false);
  end;  
end;

procedure TNotation.loaded;
begin
  inherited loaded;
  pb.Canvas.Font.Assign(Font0Var);
  Texthoehe := pb.canvas.TextHeight('Mg') + ZeilenAbstand;
end;

constructor TNotation.Create(AOwner : TComponent);
begin
  inherited create(AOwner);
//  theOwner := AOwner;
  FinfCharzugzwang         := defaultinfCharzugzwang;
  FinfChareinzigerzug      := defaultinfChareinzigerzug     ;
  FinfCharplusminus        := defaultinfCharplusminus       ;
  FinfCharplusleichtminus  := defaultinfCharplusleichtminus ;
  FinfCharplusgleich       := defaultinfCharplusgleich      ;
  FinfChargleich           := defaultinfChargleich          ;
  FinfCharunklar           := defaultinfCharunklar          ;
  FinfChargleichplus       := defaultinfChargleichplus      ;
  FinfCharleichtminusplus  := defaultinfCharleichtminusplus ;
  FinfCharminusplus        := defaultinfCharminusplus       ;
  FinfCharkompensation     := defaultinfCharkompensation    ;
  FinfChargegenspiel       := defaultinfChargegenspiel      ;
  FinfCharinitiative       := defaultinfCharinitiative      ;
  FinfCharangriff          := defaultinfCharangriff         ;
  FinfCharzeitnot          := defaultinfCharzeitnot         ;
  FinfCharvorsprung        := defaultinfCharvorsprung       ;
  FinfCharneuerung         := defaultinfCharneuerung        ;
  infCharmitderidee            := defaultinfCharmitderidee          ;
  infCharBesserIst             := defaultinfCharBesserIst           ;

  Caption := '';
  New(NoStellung);
  New(NoZug);
  ret:=#$D+#$A;
  MarkedMoveColor := clRed;

  SLeiste := TNotatScrollbar.Create(self);
  SLeiste.Kind := sbVertical;
  SLeiste.parent := self;
  pb := TNotatPaintbox.Create(self);
  pb.Canvas.pen.color := Color;
  pb.Canvas.brush.color := Color;
  pb.Canvas.brush.style := bsSolid;
  pb.parent := self;

  pb.width := self.width - SLeiste.Width;
  pb.height := self.height;
  SLeiste.left := pb.width;
  SLeiste.height := self.height;

  FAutoBold := true;
  FFontHeadLine := TFont.Create;
  FFontComment := TFont.Create;
  FFont0Var := TFont.Create;
  FFont1Var := TFont.Create;
  FFont2Var := TFont.Create;
  FontNoVar := TFont.Create;
  rFont := TFont.Create;
  wrap := TStringlist.Create;
  bf := TBitmap.Create;
  commwrap := TStringlist.Create;
  wrap.sorted := true;
  FIndent0Var := 2;
  FIndent1Var := 10;
  FIndent2Var := 18;
  FZeilenAbstand := 3;
  FBevelWidthNotat := 1;
  BevelOuter := bvLowered;
  FPartiedaten.EventData.slRestData := TStringList.Create;
  FPartiedaten.EventData.slRestData.Sorted := true;
  FPartiedaten.EventData.slRestData.Duplicates := dupIgnore;

  DoIniPartie;
  with TBitmap.Create do try
    Canvas.Font.Assign(FFont0Var);
    Texthoehe := Canvas.TextHeight('Mg') + ZeilenAbstand;
  finally
    free;
  end;
end;

destructor TNotation.destroy;
begin
  try
    Dispose(NoStellung);
    Dispose(NoZug);
    if wurzel<>Nil then begin
      ZBaumlist:=addr(wurzel);
      loescheBaum(ZBaumlist);
    end;
    if UDWurzel<>Nil then begin
      ZBaumlist:=addr(UDWurzel);
      loescheBaum(ZBaumlist);
    end;
    SLeiste.free;
    pb.free;
    wrap.free;
    bf.free;
    commwrap.free;
    FFontHeadLine.free;
    FFontComment.free;
    FFont0Var.free;
    FFont1Var.free;
    FFont2Var.free;
    FontNoVar.Free;
    rFont.free;
    FPartiedaten.EventData.slRestData.Free;

    inherited destroy;
  except
  end;
end;

{--------------------------------------------------------------------------------------------
Fgt eine Verzweigung im Baum ein - Wurzelindex mu auf VarWurzel eines Baumeintrages zeigen
---------------------------------------------------------------------------------------------}
procedure TNotation.VarEinfuegen;
begin
  if Wurzelindex<>Nil then begin
    while Wurzelindex^.BL<>Nil do begin
      Wurzelindex:=Wurzelindex^.BL;  {Index auf nchste Verzweigung}
    end;
    New(Wurzelliste);  {neues Ende der Liste};
    Wurzelindex^.BL:=Wurzelliste;
    Wurzelliste^.BL:=Nil;
    Wurzelliste^.rBL:=Wurzelindex;
    Wurzelliste^.B:=Nil;
    Wurzelindex:=Wurzelliste;
  end
  else begin
    New(Baumindex^.VarWurzel);
    Baumindex^.VarWurzel^.B:=Nil;
    Baumindex^.VarWurzel^.BL:=Nil;
    Baumindex^.VarWurzel^.rBL:=Nil;
    Wurzelindex:=Baumindex^.VarWurzel;
  end;
end;

{------------------------------------------------------
Fgt eine neuen Baum - der Aufrufer mu noch Verketten
-------------------------------------------------------}
procedure TNotation.BaumEinfuegen(z : TZug);
Begin
  New(Baum);
  Baum^.Zug:=z;
  Baum^.B:=Nil;
  Baum^.rB:=Nil;
  Baum^.VarWurzel:=Nil;
  Baum^.Kommentar:='';
  Baum.hidden := false;
end;

{----------------------------------------------------------------------
Aktualisiert NoStellung
----------------------------------------------------------------------}
procedure TNotation.ZugVor;
begin
  NoStellung^[NoZug^.nach]:=NoStellung^[NoZug^.von];
  NoStellung^[NoZug^.von]:=0;
  if NoZug^.epSchlag<>LEER then NoStellung^[NoZug^.epSchlag]:=0;
  if NoZug^.RochadeArt<>KEINEROCHADE then begin
    if NoZug^.RochadeArt=KURZEROCHADE then begin
      if NoZug^.amZug=SCHWARZ then begin
        NoStellung^[h1]:=Leer;
        NoStellung^[f1]:=wT;
      end
      else begin
        NoStellung^[h8]:=Leer;
        NoStellung^[f8]:=sT;
      end;
    end
    else if NoZug^.RochadeArt=LANGEROCHADE then begin
      if NoZug^.amZug=SCHWARZ then begin
        NoStellung^[a1]:=Leer;
        NoStellung^[d1]:=wT;
      end
      else begin
        NoStellung^[a8]:=Leer;
        NoStellung^[d8]:=sT;
      end;
    end
    else if NoZug^.RochadeArt=KURZEFISCHERROCHADE then begin
      if not assigned(FRC) then exit;
      if NoZug^.amZug = SCHWARZ then NoStellung^[FRC^.FischerRightWhiteRook]:=LEER
      else NoStellung^[FRC^.FischerRightBlackRook]:=LEER;
      NoStellung^[NoZug^.nach-1]:=-NoZug^.amZug*W_T;
      NoStellung^[NoZug^.nach]:=-NoZug^.amZug*W_K;
    end
    else if NoZug^.RochadeArt=LANGEFISCHERROCHADE then begin
      if not assigned(FRC) then exit;
      if NoZug^.amZug = SCHWARZ then NoStellung^[FRC^.FischerLeftWhiteRook]:=LEER
      else NoStellung^[FRC^.FischerLeftBlackRook]:=LEER;
      NoStellung^[NoZug^.nach+1]:=-NoZug^.amZug*W_T;
      NoStellung^[NoZug^.nach]:=-NoZug^.amZug*W_K;
    end;
  end;
  if NoZug^.UFigur<>LEER then begin
    NoStellung^[NoZug^.nach]:=NoZug^.UFigur;
  end;
end;

procedure TNotation.ZugZurueck(z : TZug);
begin
  NoStellung^[z.von]:=NoStellung^[z.Nach];
  if z.Ufigur <> Leer then NoStellung^[z.von] := -z.amZug;
  if NoZug^.epSchlag<>LEER then begin
    NoStellung^[NoZug^.epSchlag]:=Z.amZug;
    NoStellung^[Z.Nach]:=LEER;
  end
  else NoStellung^[Z.Nach]:=Z.geschlFigur;
  if NoZug^.RochadeArt<>KEINEROCHADE then begin
    if NoZug^.RochadeArt=KURZEROCHADE then begin
      if NoZug^.amZug=SCHWARZ then begin
        NoStellung^[f1]:=Leer;
        NoStellung^[h1]:=wT;
      end
      else begin
        NoStellung^[f8]:=Leer;
        NoStellung^[h8]:=sT;
      end;
    end
    else if NoZug^.RochadeArt=LANGEROCHADE then begin
      if NoZug^.amZug=SCHWARZ then begin
        NoStellung^[d1]:=Leer;
        NoStellung^[a1]:=wT;
      end
      else begin
        NoStellung^[d8]:=Leer;
        NoStellung^[a8]:=sT;
      end;
    end
    else if NoZug^.RochadeArt=KURZEFISCHERROCHADE then begin
      if not assigned(FRC) then exit;
      NoStellung^[z.nach-1]:=LEER;
      if NoZug^.amZug = SCHWARZ then NoStellung^[FRC^.FischerRightWhiteRook]:=wT
      else NoStellung^[FRC^.FischerRightBlackRook]:=sT;
      NoStellung^[z.von]:=-NoZug^.amZug*wK;
    end
    else if NoZug^.RochadeArt=LANGEFISCHERROCHADE then begin
      if not assigned(FRC) then exit;
      NoStellung^[z.nach+1]:=LEER;
      if NoZug^.amZug = SCHWARZ then NoStellung^[FRC^.FischerLeftWhiteRook]:=wT
      else NoStellung^[FRC^.FischerLeftBlackRook]:=sT;
      NoStellung^[z.von]:=-NoZug^.amZug*wK;
    end;
  end;
end;

function TNotation.sucheDame(wo : byte; schachpruefung : boolean) : boolean;
Var i, j, aZ : integer;
begin
  sucheDame:=false;
  if Schachpruefung then aZ:=-NoZug^.amZug else aZ:=NoZug^.amZug;
  for i:=1 to 8 do begin
    j:=wo;
    repeat
      inc(j, DZ[i]);
      if NoStellung^[j]=aZ*wD then begin
        if schachpruefung=false then begin   {bei Schachpruefung darf inBetracht nicht verndert werden}
          inc(inBetrachtindex);
          inBetracht[inBetrachtindex]:=j;
        end;
        sucheDame:=true;
      end;
    until NoStellung^[j]<>LEER;
  end;
end;

function TNotation.sucheTurm(wo : byte; schachpruefung : boolean) : boolean;
Var i, j, aZ : integer;
begin
  sucheTurm:=false;
  if Schachpruefung then aZ:=-NoZug^.amZug else aZ:=NoZug^.amZug;
  for i:=1 to 4 do begin
    j:=wo;
    repeat
      inc(j, TZ[i]);
      if NoStellung^[j]=aZ*wT then begin
        if schachpruefung=false then begin   {bei Schachpruefung darf inBetracht nicht verndert werden}
          inc(inBetrachtindex);
          inBetracht[inBetrachtindex]:=j;
        end;
        sucheTurm:=true;
      end;
    until NoStellung^[j]<>LEER;
  end;
end;

function TNotation.sucheLaeufer(wo : byte; schachpruefung : boolean) : boolean;
Var i, j, aZ : integer;
begin
  sucheLaeufer:=false;
  if Schachpruefung then aZ:=-NoZug^.amZug else aZ:=NoZug^.amZug;
  for i:=1 to 4 do begin
    j:=wo;
    repeat
      inc(j, LZ[i]);
      if NoStellung^[j]=aZ*wL then begin
        if schachpruefung=false then begin   {bei Schachpruefung darf inBetracht nicht verndert werden}
          inc(inBetrachtindex);
          inBetracht[inBetrachtindex]:=j;
        end;
        sucheLaeufer:=true;
      end;
    until NoStellung^[j]<>LEER;
  end;
end;

function TNotation.sucheSpringer(wo : byte; schachpruefung : boolean) : boolean;
Var i, j, aZ : integer;
begin
  sucheSpringer:=false;
  if Schachpruefung then aZ:=-NoZug^.amZug else aZ:=NoZug^.amZug;
  for i:=1 to 8 do begin
    j:=wo;
    inc(j, SZ[i]);
    if NoStellung^[j]=aZ*wS then begin
      if schachpruefung=false then begin   {bei Schachpruefung darf inBetracht nicht verndert werden}
        inc(inBetrachtindex);
        inBetracht[inBetrachtindex]:=j;
      end;
      sucheSpringer:=true;
    end;
  end;
end;

function TNotation.sucheKoenig(wo : byte; schachpruefung : boolean) : boolean;
Var i, j, aZ : integer;
begin
  sucheKoenig:=false;
  if Schachpruefung then aZ:=-NoZug^.amZug else aZ:=NoZug^.amZug;
  for i:=1 to 8 do begin
    j:=wo;
    inc(j, KZ[i]);
    if NoStellung^[j]=aZ*wK then begin
      if schachpruefung=false then begin   {bei Schachpruefung darf inBetracht nicht verndert werden}
        inc(inBetrachtindex);
        inBetracht[inBetrachtindex]:=j;
      end;
      sucheKoenig:=true;
    end;
  end;
end;

function TNotation.sucheBauer(wo : byte; schachpruefung : boolean) : boolean;
Var aZ : integer;
begin
  sucheBauer:=false;
  if Schachpruefung then aZ:=-NoZug^.amZug else aZ:=NoZug^.amZug;
  if (NoStellung^[wo+(-aZ*9)]=aZ*wB)
  or (NoStellung^[wo+(-aZ*11)]=aZ*wB) then begin
    if schachpruefung=false then begin   {bei Schachpruefung darf inBetracht nicht verndert werden}
      inc(inBetrachtindex);
{      inBetracht[inBetrachtindex]:=j;  Bei Bauern ist das unntig}
    end;
    sucheBauer:=true;
  end;
end;

procedure TNotation.DoIniPartie;
begin
  Geaendert:=false;
  NoStellung^:=Ausgangsstellung;
  with Partiedaten do begin
    ResetEventData(FPartiedaten.EventData);
    kommentar:='';
    stellung:=Ausgangsstellung;
    zugnummer:=1;
  end;
  with Partiedaten.Zug do begin
    Rochademoeglichkeiten:=[wkr, wlr, skr, slr];
    amZug:=1;
    epFeld:=LEER;
    epSchlag:=LEER;
    UFigur:=LEER;
    geschlFigur:=LEER;
    RochadeArt:=KEINEROCHADE;
    attr1:=0;
    attr2:=0;
    infattr:=0;
    infAttrBefore := 0;
  end;
  FPartiedaten.EventData.slRestData.Clear;
  if assigned(FOnMoveChange) then FOnMoveChange(self, Partiedaten.Zug, Partiedaten.Stellung, false);
end;

procedure TNotation.iniPartie;(**)
begin
  DoIniPartie;
  LoescheNotatImFenster;
end;

procedure TNotation.Clear;
begin
  LoescheNotatImFenster;
end;

function TNotation.getIdxOfMarkedMove : Integer;
begin
  Result := -1;
  if BaumIndex = Nil then exit;
  Result := Baumindex^.ss;
end;

{
Lesen der Zge und erstellen des Variantenbaumes
}

function TNotation.SetGame(PGNString : String) : integer;
Var
       tempbrett : TRettbrett;
       Kommentar : string;
       rettbrett : PRettbrett;
       RettBrettstack : TList;

procedure SetAttr2(s : string);
begin
  if not assigned(Baumindex) then exit;

  if s = '1' then Baumindex^.Zug.attr2 := nAusruf
  else if s = '3' then Baumindex^.Zug.attr2 := n2Ausruf
  else if s = '2' then Baumindex^.Zug.attr2 := nFrage
  else if s = '4' then Baumindex^.Zug.attr2 := n2Frage
  else if s = '5' then Baumindex^.Zug.attr2 := nAusrufFrage
  else if s = '6' then Baumindex^.Zug.attr2 := nFrageAusruf

  else if s = '22' then Baumindex^.Zug.infattr := infzugzwang
  else if s = '8'  then Baumindex^.Zug.infattr := infeinzigerzug
  else if s = '18' then Baumindex^.Zug.infattr := infplusminus
  else if s = '16' then Baumindex^.Zug.infattr := infplusleichtminus
  else if s = '14' then Baumindex^.Zug.infattr := infplusgleich
  else if s = '11' then Baumindex^.Zug.infattr := infgleich
  else if s = '13' then Baumindex^.Zug.infattr := infunklar
  else if s = '15' then Baumindex^.Zug.infattr := infgleichplus
  else if s = '17' then Baumindex^.Zug.infattr := infleichtminusplus
  else if s = '19' then Baumindex^.Zug.infattr := infminusplus
  else if s = '44' then Baumindex^.Zug.infattr := infkompensation
  else if s = '132'then Baumindex^.Zug.infattr := infgegenspiel
  else if s = '36' then Baumindex^.Zug.infattr := infinitiative
  else if s = '40' then Baumindex^.Zug.infattr := infangriff
  else if s = '138'then Baumindex^.Zug.infattr := infzeitnot
  else if s = '32' then Baumindex^.Zug.infattr := infvorsprung
  else if s = '146'then Baumindex^.Zug.infattr := infneuerung

  else if s = '140'then Baumindex^.Zug.infattrBefore := infmitderidee
  else if s = '142'then Baumindex^.Zug.infattrBefore := infBesserIst    ;
end;


function analyse : boolean;
var i, j, k : longint;
    zug, feld : string;
    hilfswort : string;
    hilf, figur : shortint;
    linie, reihe : byte;
    fehler : integer;
    kl, keineZuege : boolean;

function Figurenart(b : char) : shortint;
begin
  case b of
    'K' : Figurenart:=wK*NoZug^.amZug;
    'Q', 'D' : Figurenart:=wD*NoZug^.amZug;
    'R', 'T' : Figurenart:=wT*NoZug^.amZug;
    'B', 'L' : Figurenart:=wL*NoZug^.amZug;
    'N', 'S' : Figurenart:=wS*NoZug^.amZug;
    'P' : Figurenart:=wB*NoZug^.amZug;
    else begin
      if b in ['a'..'h'] then
        Figurenart := wB*NoZug^.amZug
      else Figurenart := -100;
    end;
  end;
end;

function CheckLetter(c : char) : Boolean;
begin
  if not (c in ['a'..'h']) then Result := true
  else Result := false;
end;

function CheckDigit(c : char) : Boolean;
begin
  if not (c in ['1'..'8']) then Result := true
  else Result := false;
end;

function CheckLong(ss : string) : string;
Var i : integer;
    bauer : Boolean;

begin
  Result := ss;
  try
    i := 1;
    while i <= length(Result) do begin
      if (Result[i] = 'x') or (Result[i] = '-') then begin
        bauer := false;
        if (i > 2) and (Result[i - 1] in ['1'..'8'])  // Rochade wre O
        and (Result[i - 2] in ['a'..'h']) then begin
          if (i - 3 = 0) or
          ((i - 3 > 0) and (not (Result[i - 3] in ['A'..'Z']))) then begin // dann ist es Bauer
            dec(i);    // auf Zahl und lschen
            Delete(Result, i , 1);  // steht jetzt auf x oder - (letzteres auch weg
            bauer := true;
          end;
        end;
        if (Result[i] = '-') and (upcase(Result[i - 1]) <> 'O') and (Result[i - 1] <> '0') then begin
          if Bauer then begin
            dec(i);  // Zahl ist schon weg - Ausgangsreihe noch weg (ist ja kein Schlagfall)
            Delete(Result, i , 1);
          end;
          Delete(Result, i , 1);
        end;
      end;
      inc(i);
    end;
  except
  end;
end;

var hilf1 : shortint;
    lastMove : TZug;
    eebene : Integer;
    onlySqVon : String;
    onlySqI, onlySqJ : Integer;

begin
  analyse:=false; {wird true bei Fehler}
  keineZuege:=true; {falls das so bleibt, wird am Schlu Wurzel freigegeben}
  RettBrettstack:=TList.Create;
  RettBrettstack.Capacity:=10;
  UD:=0;
  if UDWurzel<>Nil then begin
    ZBaumlist:=addr(UDWurzel);
    loescheBaum(ZBaumlist);
  end;
  if wurzel<>nil then begin
    ZBaumlist:=addr(wurzel);
    loescheBaum(ZBaumlist);
  end;
  New(Wurzel);
  Wurzel^.B:=Nil;
  Wurzel^.BL:=Nil;
  Wurzel^.rBL:=Nil;
  Wurzelindex:=Wurzel;
  Baumindex:=Nil;
  i:=0;
  kl:=true;
  kommentar:='';
  NoStellung^:=Partiedaten.Stellung; {}

  checkFischerPos;

  repeat
    inc(i);
    if i>=length(notat) then begin
      if kommentar<>'' then FPartiedaten.kommentar:=kommentar+' ';
      exit; {keine Zge eingegeben}
    end;
    if notat[i]='{' then kl:=false;
    if kl=false then kommentar:=kommentar+notat[i];
    if notat[i]='}' then kl:=true;
  until (notat[i]='.' ) and (kl);  {am Anfang knnte auch ein Kommentar stehen}
  {erste Zugnummer wird in FenToPos ermittelt!!!}
  if length(kommentar)>1 then begin  {bei Zeilensprung in PGN wird im Komm. ein Leerzeichen eingefgt - mu raus}
    if kommentar[2]=' ' then delete(kommentar, 2, 1);
    if kommentar[length(kommentar)-1]=' ' then delete(kommentar, length(kommentar)-1, 1);
  end;
  if kommentar<>'' then FPartiedaten.kommentar:=kommentar+' ';
  NoZug^:=Partiedaten.Zug;

  lastMove := NoZug^;
  eebene := 0;
  FErrorMoves := '';

  dec(i);
  repeat
    inc(i);
    if notat[i]='{' then begin
      kommentar:='{';
      repeat
        inc(i);
        if i>length(notat) then  begin analyse:=true; exit end;
        kommentar:=kommentar+notat[i];
      until notat[i]='}';
      if kommentar[2]=' ' then delete(kommentar, 2, 1);
      if kommentar[length(kommentar)-1]=' ' then delete(kommentar, length(kommentar)-1, 1);
      if assigned(Baumindex) then begin
        if trim(Baumindex^.kommentar) <> '' then
          Baumindex^.kommentar:= Baumindex^.kommentar + ' | ' + kommentar
        else
          Baumindex^.kommentar:=kommentar;
      end;
    end;
    IF notat[i]='$' then begin  {Ausrufezeichen etc}
      hilfswort:='';
      while (not (notat[i]=' ')) and (not (notat[i]=')')) and (i <= length(notat))
      do begin
        if notat[i] in ['0'..'9'] then hilfswort:=hilfswort+notat[i];
        delete(notat, i, 1); {nach $12 o.. mu Leerzeichen stehen; i braucht nicht inc zu werde, da String "nachrutscht"}
      end;
      SetAttr2(Hilfswort);

      if assigned(Baumindex) // falls doch schon vor ersten Zug so etwas stehen sollte - unwahrscheinlich
      then begin // neueM. siehe Parameterbergabe
        Val(hilfswort, j, k);
        case j of
          998 : Baumindex^.Zug.attr1:=Baumindex^.Zug.attr1 or nErster;
          999 : Baumindex^.Zug.attr1:=Baumindex^.Zug.attr1 or nLetzter;
        end;
      end;


    end;
    if notat[i]='(' then begin  {Variantenbeginn}
      inc(eebene);

      New(RettBrett);
      RettBrett^.Stellung:=tempbrett.Stellung;
      RettBrett^.Zug:=tempbrett.Zug;
      RettBrett^.Wurzel:=Wurzelindex;
      RettBrett^.Baum:=Baumindex;
      RettBrettstack.add(rettbrett);
      NoStellung^:=tempbrett.Stellung;  {Zug zurck vvv}
      NoZug^:=tempbrett.Zug;
      Wurzelindex:=Baumindex^.VarWurzel;
      VarEinfuegen;
    end;
    if (notat[i]=')') and (Rettbrettstack.Count <> 0) then begin  {Variantenende}
      dec(eebene);

      RettBrett:=RettBrettstack.Items[RettBrettstack.count-1];
      NoStellung^:=RettBrett^.Stellung; {Stellung vor Variantenbeginn wieder als aktuelle restaurieren}
      NoZug^:= RettBrett^.Zug;
      Wurzelindex:= RettBrett^.Wurzel; {steht jetzt auf letzter Variante}
      Baumindex:=RettBrett^.Baum;
      dispose(Rettbrett);
      RettBrettstack.remove(RettBrettstack.Items[RettBrettstack.count-1]);  {Letzte Stellung entfernen}
      tempbrett.stellung:=NoStellung^;
      tempBrett.Zug:=NoZug^;
      NoZug^.amZug:=-NoZug^.amZug;
      zugvor;  {letzten Zug von Hauptvariante wieder vorspielen}
    end;
    if (notat[i] in ['A'..'Z','a'..'z','O', 'o'{Fischerroch.}]) or
    ((i < length(notat) - 1) and (notat[i] = '0') and (notat[i + 1] = '-') and (notat[i + 2] = '0'))
    then begin  {Hier mu Zug losgehen}
      with NoZug^ do begin
        epFeld:=LEER;
        epSchlag:=LEER;
        UFigur:=LEER;
        geschlFigur:=LEER;
        RochadeArt:=KEINEROCHADE;
        attr1:=0;
        attr2:=0;
        infattr := 0;
        infAttrBefore := 0;
      end;
      zug:='';
      repeat
        if i>length(notat) then  begin analyse:=true; exit end;
        if (notat[i]<>'=') and (notat[i]<>'#') and (notat[i]<>'+') then zug:=zug+notat[i]; {filtern}
        if notat[i]='+' then NoZug^.attr1:=NoZug^.attr1 or nSchach;
        if notat[i]='#' then NoZug^.attr1:=NoZug^.attr1 or nMatt;
        inc(i);
      until (i>length(notat)) or (notat[i]=' ') or (notat[i]=')') or (notat[i]='*');
      {auch wenn kein Erg. am Ende steht, mu Leerzeichen und "*" kommen}
      if notat[i]=')' then dec(i);
      {damit im nchsten Durchlauf nochmal die Klammer entdeckt wird und Variantenende registriert wird}

      if (Zug[1]='O') or (Zug[1]='o') or (Zug[1]='0') then begin {Rochade; Fischerrochade (eine Null statt ein O !)}
        if (
        {(}(Zug[1]='O') or (Zug[1]='0'){)}
        or {(}(Zug[1]='o')
        ) and (not assigned(FRC)){)} then begin
          if NoZug^.amZug=WEISS then begin
            NoZug^.von:=e1;
            if length(zug)>3 then begin
              NoZug^.RochadeArt:=LANGEROCHADE;
              NoZug^.nach:=c1  {groe}
            end
            else begin
              NoZug^.RochadeArt:=KURZEROCHADE;
              NoZug^.nach:=g1;
            end;            {kleine}
          end
          else begin
            NoZug^.von:=e8;
            if length(zug)>3 then begin
              NoZug^.RochadeArt:=LANGEROCHADE;
              NoZug^.nach:=c8  {groe}
            end
            else begin
              NoZug^.RochadeArt:=KURZEROCHADE;
              NoZug^.nach:=g8;
            end;     {kleine}
          end;
        end
        else begin      {Fischer-Rochade}
          if NoZug^.amZug=WEISS then begin
            NoZug^.von:=FRC^.FischerWhiteKing;
            if length(zug)>3 then begin
              NoZug^.RochadeArt:=LANGEFISCHERROCHADE;
              NoZug^.nach:=c1;  {von u. nach-Feld wird bei R. eigentlich nicht gebraucht, nimmt hier die Ausgangsfelder auf (Zielfelder bleiben ja gleich)}
            end
            else begin
              NoZug^.RochadeArt:=KURZEFISCHERROCHADE;
              NoZug^.nach:=g1;
            end;            {kleine}
          end
          else begin
            NoZug^.von:=FRC^.FischerBlackKing;
            if length(zug)>3 then begin
              NoZug^.RochadeArt:=LANGEFISCHERROCHADE;
              NoZug^.nach:=c8;  {groe}
            end
            else begin
              NoZug^.RochadeArt:=KURZEFISCHERROCHADE;
              NoZug^.nach:=g8;
            end;     {kleine}
          end;
        end;
      end
      else begin
        zug := Checklong(zug);

        if not (length(zug)>=2) then
          begin analyse:=true; exit end; {Zug mu mind. 2 Zeichen lang sein}
        if length(zug)=2 then Figur:=NoZug^.amZug*wB
        else figur:=Figurenart(Zug[1]);
        if figur = -100 then
          begin analyse:=true; exit end;
        if zug[length(zug)] in ['A'..'Z'] then NoZug^.UFigur:=Figurenart(zug[length(zug)]); {Umwandlung}
        if NoZug^.UFigur = -100 then
          begin analyse:=true; exit end;
        feld:='';
        if zug[1] in ['A'..'Z'] then j:=1 else j:=0;  {bei Bauernzug steht kein Grobuchstabe davor}
        repeat
          inc(j);
          if (zug[j]<>'x') and (zug[j]<>':') then feld:=feld+Zug[j];
        until (j>=length(zug)) or ((zug[j] in ['1'..'9'])
        {es wird davon ausgegangen, da Zug mit Zahl abgeschlossen wird - "exd5" nicht "exd"}
              and (not (zug[j-1] in ['A'..'Z'])) {bei "e4" ist Schlu - aber nicht bei "N3d4"}
              and (not (zug[j+1] in ['a'..'z']))); {fr Flle wo z.B. Sd2e4 notiert werden muss}

        feld := lowercase(feld);
        if length(feld) < 2 then
          begin analyse:=true; exit end;
        if CheckLetter(feld[length(feld)-1]) then
          begin analyse:=true; exit end;
        j:=ord(feld[length(feld)-1])-97; {vorletztes Zeichen ist Linie - "a" (ASCII=97) mu dabei 0 werden}
        if CheckDigit(feld[length(feld)]) then
          begin analyse:=true; exit end;
        val(feld[length(feld)], k, fehler);  {zuletzt mu Zahl stehen , die Reihe bezeichnet}
        if fehler<>0 then
          begin analyse:=true; exit end;
        dec(k);  {erste Reihe mu 0 werden}
        NoZug^.nach:=c((k*8)+j);
        NoZug^.geschlFigur:=NoStellung^[NoZug^.nach];

        if (abs(Figur)=wB) and (length(feld)=4) then begin
          onlySqVon := lowercase(copy(feld, 1, 2));
          if (not CheckLetter(onlySqVon[1])) and
          (not CheckDigit(onlySqVon[2])) then begin
            onlySqI:=ord(onlySqVon[1])-97;
            val(onlySqVon[2], onlySqJ, fehler);
            if fehler=0 then begin
              dec(onlySqJ);  {erste Reihe mu 0 werden}
              onlySqI:=c((onlySqJ*8)+onlySqI);
              if (onlySqI >= 0) and (onlySqI < 120) then begin
                if (NoStellung^[onlySqI] <> LEER) and
                (abs(NoStellung^[onlySqI]) <> wB) then
                  Figur := abs(NoStellung^[onlySqI]);
              end;
            end;
          end;
        end;

        if abs(Figur)=wB then begin
          if length(feld)=4 then begin
            Delete(feld, 2, 1);  {dann war es Langnotation e2e4}
          end;

          if length(feld)=3 then begin {Schlagzug (ed5 o..)}
            if CheckLetter(feld[1]) then
              begin analyse:=true; exit end;
            NoZug^.von:= c((ord(feld[1])-97)+((k-NoZug^.amZug)*8));
            {Ausgangsfeld ist eine Reihe nach oben (abh. von NoZug^.amZug) und die Linie die im ersten Zeichen steht}
            if NoStellung^[NoZug^.nach]=LEER then begin
              NoZug^.epSchlag:=NoZug^.nach-(NoZug^.amZug*10);  {e.p,Schlag -  geschl. Bauern entfernen}
              NoZug^.geschlFigur:=-NoZug^.amZug;
            end;
          end
          else begin
            if NoStellung^[NoZug^.nach-(NoZug^.amZug*10)]<>LEER then NoZug^.von:=NoZug^.nach-(NoZug^.amZug*10)
            else begin
              NoZug^.von:=NoZug^.nach-(NoZug^.amZug*20); {Doppelschritt}
              NoZug^.epFeld:=NoZug^.nach-(NoZug^.amZug*10);
            end;
          end;
        end
        else begin
          inBetrachtindex:=0;
          {Index fr das Array, in dem alle Felder gespeichert werden, wo Figuren stehen, die aufs Zielfeld knnen}
          case abs(figur) of
            wD : sucheDame(NoZug^.nach, false);
            wT : sucheTurm(NoZug^.nach, false);
            wL : sucheLaeufer(NoZug^.nach, false);
            wS : sucheSpringer(NoZug^.nach, false);
            wK : sucheKoenig(NoZug^.nach, false);
          end;
          if inBetrachtindex<1 then
            begin analyse:=true; exit end;
          if inBetrachtindex>1 then begin  {mehr als eine Figur kann aufs Zielfeld}
            if length(Feld)>=3 then begin
            {entsprechende Figur durch Linien oder Reihenangabe bezeichnet ->zus. Zeichen an 1.Stelle (fg3 o.)}
              if length(Feld)=3 then begin
                for j:=1 to inBetrachtindex do begin
                  linie:=cc(inBetracht[j]) mod 8;
                  reihe:=cc(inBetracht[j]) div 8;
                  if feld[1] in ['1'..'8'] then begin
                    val(feld[1], hilf, fehler);
                    if Reihe=hilf-1 then begin
                      NoZug^.von:=inBetracht[j];  {hilf ergibt von 1-8, reihe aber von 0-7}
                      NoZug^.attr1:=NoZug^.attr1 or nDigit;
                      {zweites Bit von links bezeichnet, da in PGN-notation Ausgangsfeld der jeweiligen
                      Figur angegeben werden muss}
                    end;
                  end
                  else if ord(feld[1])-97=linie then begin
                    NoZug^.von:=inBetracht[j];
                    NoZug^.attr1:=NoZug^.attr1 or nLetter;
                  end;
                end;
              end
              else begin  {falls 3 Figuren aufs Zielfeld knnen Sd2e4 z.B.}
                if (CheckDigit(feld[2])) or (CheckLetter(feld[1])) then
                   begin analyse:=true; exit end;
                val(feld[2], hilf, fehler);
                NoZug^.von:=c(((hilf-1)*8) + (ord(feld[1])-97));
                NoZug^.attr1:=NoZug^.attr1 or nDigit or nLetter;
              end;
            end
            else begin
              hilf := e1;
              for j:=0 to 63 do if NoStellung^[c(j)]=NoZug^.amZug*wK then hilf:=c(j); {Knigsfeld suchen}
              hilf1 := NoStellung^[NoZug^.nach];
              NoStellung^[NoZug^.nach] := figur;
              for j:=1 to inBetrachtindex do begin
                NoStellung^[inBetracht[j]]:=LEER; {erste mgliche Figur entfernen}
                if (not(sucheDame(hilf, true))) and (not(sucheTurm(hilf, true)))
                and (not(sucheLaeufer(hilf, true)))
                  then NoZug^.von:=inBetracht[j]; {kein Langschrittler kann Schach geben ->Figur war nicht gefesselt}
                NoStellung^[inBetracht[j]]:=figur;
              end;
              NoStellung^[NoZug^.nach] := hilf1;
            end;
          end
          else NoZug^.von:=inBetracht[1];
        end;
      end;
      //!!! kleiner bug: rochaderecht wird erst danach gesetzt
      // beim Restaur. von tempbRett wird ein ev. vorher vorhandenes wiederbelebt
      // --> wenn bei K/T Zug Variante losgeht, und bei Rckkehr von K/T auch wieder,
      // ist in der HVar. ungerechtfertigt eine Rochade noch mglich
      tempbRett.Stellung:=NoStellung^;
      tempbRett.Zug:=NoZug^;          
               
      // das behebt oben erwhnten bug, auer wenn K/T unmittelbar im nchsten zug wieder zurckziehen (und dort auch var. losgeht)
      if FRC = Nil then begin     
        if NoStellung^[e1] = 0 then begin
          NoZug^.Rochademoeglichkeiten:=NoZug^.Rochademoeglichkeiten-[wkr];
          NoZug^.Rochademoeglichkeiten:=NoZug^.Rochademoeglichkeiten-[wlr];
        end;
        if NoStellung^[e8] = 0 then begin
          NoZug^.Rochademoeglichkeiten:=NoZug^.Rochademoeglichkeiten-[skr];
          NoZug^.Rochademoeglichkeiten:=NoZug^.Rochademoeglichkeiten-[slr];
        end;
        if NoStellung^[h1] <= 0 then
          NoZug^.Rochademoeglichkeiten:=NoZug^.Rochademoeglichkeiten-[wkr];
        if NoStellung^[a1] <= 0 then
          NoZug^.Rochademoeglichkeiten:=NoZug^.Rochademoeglichkeiten-[wlr];

        if NoStellung^[h8] >= 0 then
          NoZug^.Rochademoeglichkeiten:=NoZug^.Rochademoeglichkeiten-[skr];
        if NoStellung^[a8] >= 0 then
          NoZug^.Rochademoeglichkeiten:=NoZug^.Rochademoeglichkeiten-[slr];
      end;


      if checkMoves then begin
        if eebene = 0 then begin  // in var. funktioniert prfung noch nicht - frs fernschach reicht das erst mal
          lastMove.von := NoZug^.von;
          lastMove.nach := NoZug^.nach;
          if not IsMovePossible(NoStellung, @lastMove, FRC) then
            FErrorMoves := FErrorMoves + ' ' + zug;
        end;
      end;


      case NoStellung^[NoZug^.von] of
        wT :  begin
                if NoZug^.von=h1 then NoZug^.Rochademoeglichkeiten:=NoZug^.Rochademoeglichkeiten-[wkr];
                if NoZug^.von=a1 then NoZug^.Rochademoeglichkeiten:=NoZug^.Rochademoeglichkeiten-[wlr];
              end;
        sT :  begin
                if NoZug^.von=h8 then NoZug^.Rochademoeglichkeiten:=NoZug^.Rochademoeglichkeiten-[skr];
                if NoZug^.von=a8 then NoZug^.Rochademoeglichkeiten:=NoZug^.Rochademoeglichkeiten-[slr];
              end;
        wK :  NoZug^.Rochademoeglichkeiten:=NoZug^.Rochademoeglichkeiten-[wkr, wlr];
        sK :  NoZug^.Rochademoeglichkeiten:=NoZug^.Rochademoeglichkeiten-[skr, slr];
      end;

      case NoZug^.nach of
        h1 : NoZug^.Rochademoeglichkeiten:=NoZug^.Rochademoeglichkeiten-[wkr];
        a1 : NoZug^.Rochademoeglichkeiten:=NoZug^.Rochademoeglichkeiten-[wlr];
        h8 : NoZug^.Rochademoeglichkeiten:=NoZug^.Rochademoeglichkeiten-[skr];
        a8 : NoZug^.Rochademoeglichkeiten:=NoZug^.Rochademoeglichkeiten-[slr];
      end;

      NoZug^.amZug:=-NoZug^.amZug;

      if eebene = 0 then
        lastMove := NoZug^;

      BaumEinfuegen(NoZug^);
      if Wurzelindex^.B=Nil then begin
        Wurzelindex^.B:=Baum;  {Anfang der Liste - wird mit Baum.B fortgesetzt}
        if assigned(Baumindex) then Baum^.rB:=Baumindex;
      end
      else begin
        Baum^.rB:=Baumindex;
        Baumindex^.B:=Baum;  {Verketten}
      end;
      Baumindex:=Baum;
      ZugVor;
      keineZuege:=false;
    end;
  until (i>=length(notat)) or (notat[i]='*') or (notat[i]='-');
  For i:=0 to RettBrettstack.Count-1 do begin
    rettbrett := RettBrettstack.Items[i];
    dispose(rettbrett);
  end;
  RettBrettstack.free;
  if keineZuege then begin dispose(wurzel);wurzel:=Nil;end;
end;

var slTemp : TStringlist;
    s : string;
    j : integer;
    p : TMoveChangeEvent;
    ZuegeStart : integer;
 //   FischerInfo : PFischerInfo;

begin
  Result:=0;
  try
  
  p := FOnMoveChange;
  FOnMoveChange := Nil;  {Ereignis in DoIniPartie nicht aufrufen, da es anschlieend auch nach unvollst. Einlesen gemacht wird}
  IniPartie;
  FOnMoveChange := p;
  slTemp := TStringList.Create;
 // New(FischerInfo);
  try
    if trim(PGNString) = '' then
      exit;

    j := 0;
    ZuegeStart := 0;
    slTemp.Text := PGNString;
    while j < slTemp.Count do begin
      s := trim(slTemp[j]);
      if (s <> '') and (s[1] = '[') then break;
      inc(j);
    end;
    while j < slTemp.Count do begin
      s := trim(slTemp[j]);
      if (s <> '') and (s[1] <> '[') then break;
      if s = '' then break;
      SetEventData(FPartiedaten.EventData, s);
      inc(j);
    end;
    notat := '';
    while j < slTemp.Count do begin
      if (slTemp[j] = '') or (slTemp[j][1] <> '%') then
        notat := notat + slTemp[j] + ' ';
      inc(j);
    end;
 //   FischerInfo.Enabled := false;
    if Partiedaten.EventData.FENString <> '' then
       FENToPos(Partiedaten.EventData.FENString,
       FPartiedaten.Stellung, FPartiedaten.Zug, FPartiedaten.Zugnummer, Nil);//FischerInfo);
//    if assigned(OnFischerinfo) then OnFischerInfo(self, FischerInfo^);  {an enabled kann Hauptprogramm erkennen, ob auf FRC umgestellt werden muss}
    notat := trim(notat);
    if notat = '' then exit;
    if analyse then { gibt bei Fehler(!) true}
    begin
      Result:=-1;
      if assigned(OnError) then OnError(self, mess[7+FEnglish]);   {Partie wurde zwar eingelesen, aber nicht vollstndig}
      gemeldet:=true;  {"klettere" wird noch ausgefhrt u. setzt das wieder auf false}
    end;
  finally
    slTemp.Free;
//    Dispose(FischerInfo);
    Baumindex:=Nil;  {nach Stellung laden ist noch kein Zug indiziert}
    klettere(wurzel, true);
    if assigned(FOnMoveChange) then FOnMoveChange(self, Partiedaten.Zug, Partiedaten.Stellung, false)
  end;
  except
    Result := -1;
  end;  
end;

function TNotation.GetNextMoveNumberAsString : string;
begin
  if not assigned(Baumindex) then begin
    if Partiedaten.Zug.amZug = WEISS then Result := IntToStr(PartieDaten.ZugNummer) + '...'
    else Result := IntToStr(PartieDaten.ZugNummer) + '.';
  end
  else begin
    if Baumindex^.Zug.amZug = WEISS then Result := IntToStr(Baumindex^.ZugNummer + 1) + '.'
    else Result := IntToStr(Baumindex^.ZugNummer) + '...';
  end;
end;

function TNotation.GetNextMoveNumber : Integer;
begin
  if not assigned(Baumindex) then begin
    if Partiedaten.Zug.amZug = WEISS then Result := PartieDaten.ZugNummer
    else Result := PartieDaten.ZugNummer;
  end
  else begin
    if Baumindex^.Zug.amZug = WEISS then Result := Baumindex^.ZugNummer + 1
    else Result := Baumindex^.ZugNummer;
  end;
end;

function TNotation.inVariation : Boolean;
var  rBaum, b : PBaum;
begin
  Result := false;
  b := Baumindex;
  while assigned(b) do begin
    rBaum := b;
    b := b^.rB;
    if (assigned(b)) and (rBaum<>b^.B) {springt in bergeordnete Var. und mu deshalb noch einen Zug zurck}
    then begin
      Result := true;
      exit;
    end;
  end;
end;

function TNotation.WhitesTurnOnEnd : Boolean;
var b : PBaum;
begin
  b := Nil;
  if (wurzelindex <> Nil) and (wurzelindex.B <> Nil) then begin
    b := wurzelindex.B;
    while b^.B <> Nil do
      b := b^.B;
  end;    
  if not assigned(B) then begin
    Result := Partiedaten.Zug.amZug = WEISS;
  end
  else begin
    Result := B^.Zug.amZug = WEISS;
  end;
end;

function TNotation.WhitesTurn : Boolean;
begin
  if not assigned(Baumindex) then begin
    Result := Partiedaten.Zug.amZug = WEISS;
  end
  else begin
    Result := Baumindex^.Zug.amZug = WEISS;
  end;
end;

function TNotation.getCurrentMoveAse2e4 : string;
begin
  if not assigned(Baumindex) then Result := ''
  else begin
    Result := feldnamen[cc(Baumindex^.Zug.von)] + feldnamen[cc(Baumindex^.Zug.nach)];
    case abs(Baumindex^.Zug.UFigur) of
      wD : Result := Result + 'q';
      wT : Result := Result + 'r';
      wL : Result := Result + 'b';
      wS : Result := Result + 'n';
    end;
  end;
end;

function TNotation.getCurrentMoveFromNotation : string;

  procedure toEnglish(var S : String);
  begin
    while Pos('D', S) > 0 do  S[Pos('D', S)] := 'Q';
    while Pos('T', S) > 0 do  S[Pos('T', S)] := 'R';
    while Pos('L', S) > 0 do  S[Pos('L', S)] := 'B';
    while Pos('S', S) > 0 do  S[Pos('S', S)] := 'N';
  end;

begin
  if not assigned(Baumindex) then Result := ''
  else begin
    Result := Result + copy(notat,Baumindex^.ss, Baumindex^.sl);
    toEnglish(Result);
  end;
end;

function TNotation.notiereRawZug(z : TZug) : string;
var z1 : TZug;
    iEngl : Integer;
    eError : TErrorEvent;
    upieces : Boolean;
begin
  Result := '';
  iEngl := FEnglish;
  eError := OnError;
  upieces := FUserDefinedPieces;
  try
    FEnglish := 1;
    FUserDefinedPieces := false;
    z1 := z;
    z1.attr2 := 0;
    z1.infattr := 0;
    z1.infAttrBefore := 0;
    Result := notiereZug(z1);
  finally
    OnError := eError;
    FEnglish := iEngl;
    FUserDefinedPieces := upieces;
  end;
end;

function TNotation.Attr2toString(i : shortint) : String;
begin
  case i of
    nAusruf : Result := '!';
    n2Ausruf : Result := '!!';
    nFrage : Result := '?';
    n2Frage : Result := '??';
    nAusrufFrage : Result := '!?';
    nFrageAusruf : Result := '?!';
    else Result := '';
  end;
end;

function TNotation.Attr2toPGNString(i : shortint) : String;
begin
  case i of
    nAusruf : Result := '$1';
    n2Ausruf : Result := '$3';
    nFrage : Result := '$2';
    n2Frage : Result := '$4';
    nAusrufFrage : Result := '$5';
    nFrageAusruf : Result := '$6';
    else Result := '';
  end;
end;

function TNotation.infAttrtoString(i : byte) : String;
begin
  Result := '';
  if not saving then begin
    case i of
    infzugzwang         : Result :=  infCharzugzwang;
    infeinzigerzug      : Result :=  infChareinzigerzug;
    infplusminus        : Result :=  infCharplusminus;
    infplusleichtminus  : Result :=  infCharplusleichtminus;
    infplusgleich       : Result :=  infCharplusgleich;
    infgleich           : Result :=  infChargleich;
    infunklar           : Result :=  infCharunklar;
    infgleichplus       : Result :=  infChargleichplus;
    infleichtminusplus  : Result :=  infCharleichtminusplus;
    infminusplus        : Result :=  infCharminusplus;
    infkompensation     : Result :=  infCharkompensation;
    infgegenspiel       : Result :=  infChargegenspiel;
    infinitiative       : Result :=  infCharinitiative;
    infangriff          : Result :=  infCharangriff;
    infzeitnot          : Result :=  infCharzeitnot;
    infvorsprung        : Result :=  infCharvorsprung;
    infneuerung         : Result :=  infCharneuerung;
    infmitderidee       : Result :=  infCharmitderidee;
    infBesserIst        : Result :=  infCharBesserIst ;
    end;
  end
  else begin
    case i of
    infzugzwang         : Result :=  defaultinfCharzugzwang;
    infeinzigerzug      : Result :=  defaultinfChareinzigerzug;
    infplusminus        : Result :=  defaultinfCharplusminus;
    infplusleichtminus  : Result :=  defaultinfCharplusleichtminus;
    infplusgleich       : Result :=  defaultinfCharplusgleich;
    infgleich           : Result :=  defaultinfChargleich;
    infunklar           : Result :=  defaultinfCharunklar;
    infgleichplus       : Result :=  defaultinfChargleichplus;
    infleichtminusplus  : Result :=  defaultinfCharleichtminusplus;
    infminusplus        : Result :=  defaultinfCharminusplus;
    infkompensation     : Result :=  defaultinfCharkompensation;
    infgegenspiel       : Result :=  defaultinfChargegenspiel;
    infinitiative       : Result :=  defaultinfCharinitiative;
    infangriff          : Result :=  defaultinfCharangriff;
    infzeitnot          : Result :=  defaultinfCharzeitnot;
    infvorsprung        : Result :=  defaultinfCharvorsprung;
    infneuerung         : Result :=  defaultinfCharneuerung;
    infmitderidee       : Result :=  defaultinfCharmitderidee;
    infBesserIst        : Result :=  defaultinfCharBesserIst ;
    end;
  end;
end;

function TNotation.infAttrtoPGNString(i : byte) : String;
begin
  Result :=  '$' + IntToStr(i);
end;


function TNotation.notiereZug(z : TZug) : string;
Var s, sf : string;
    fnindex : integer; {index in Figurennamen (engl. u. Deutsche}

procedure SetAttrZwei;
begin
  s := s + Attr2toString(z.Attr2);
  s := s + infAttrtoString(z.infattr);
end;

function GetFigurnamen(index : integer) : string;
Var i : integer;
    s : String;
begin
  Result := '';
  i := index;
  try
    if (Saving) or (not FUserDefinedPieces) then Result := Figurnamen[i]
    else begin
      s := Figurnamen[i];
      if i > 6 then i := i - 6;
      Result := PiecesNames[i - 1];
      if Result = '' then Result := s;
    end;
  except
  end;
end;

begin
  fnindex:=FEnglish*6;
  s:='';
  if z.Rochadeart<>LEER then begin
    case z.Rochadeart of
      KURZEROCHADE : s:='O-O';
      LANGEROCHADE : s:='O-O-O';
      KURZEFISCHERROCHADE : s:='O-O';//'o-o';
      LANGEFISCHERROCHADE : s:='O-O-O';//'o-o-o';
    end;
    if z.attr1 and nSchach <> 0 then s:=s+'+';
    if z.attr1 and nMatt <> 0 then s:=s+'#';
    SetAttrZwei;
    Result:=s;
    exit;
  end;
  if (abs(NoStellung^[z.von])=wB) and (NoStellung^[z.nach]<>LEER)
  or (z.epschlag<>LEER) then begin
    s:=s+char((cc(z.von) mod 8)+97);
  end
  else if abs(NoStellung^[z.von])<>0 then
    s:=s+GetFigurnamen(abs(NoStellung^[z.von])+fnindex);
  if z.attr1 and nLetter <> 0 then s:=s+char((cc(z.von) mod 8)+97);
  if z.attr1 and nDigit <> 0 then s:=s+IntTostr((cc(z.von) div 8)+1);
  if (NoStellung^[z.nach]<>LEER) or (z.epschlag<>LEER) then s:=s+'x';
  s:=s+feldnamen[cc(z.nach)];
  if z.UFigur<>LEER then s:=s+'='+GetFigurnamen(abs(z.UFigur)+fnindex);
  if z.attr1 and nSchach <> 0 then s:=s+'+';
  if z.attr1 and nMatt <> 0 then s:=s+'#';

  SetAttrZwei;

  {!!!test
  if z.von = z.nach then s := 'Empty'
  else    }

  if abs(NoStellung^[z.von])=0 then begin
    if not gemeldet then begin
      if FEnglish = 1 then sf := 'Error while reading the moves.' else
      sf := 'Beim Einlesen der Zge ist ein Fehler aufgetreten.';
      if assigned(OnError) then OnError(self, sf);
      gemeldet:=true;
    end;
  end;
  Result:=s;
end;

function TNotation.GetIndent(Font : TFont) : Integer;
begin
  if font = FFont1Var then Result := Indent1Var
  else if font = FFont2Var then Result := Indent2Var
  else Result := Indent0Var;
end;

function TNotation.GetNumber(index : integer) : longint;
begin
  Result := -1;
  if (index < wrap.count) and (wrap.count <> 0) then
    Result := StrToInt(trim(wrap[index]));
end;

function TNotation.NumberToString(nr : longint) : string;
begin
  Result := Format('%' + IntToStr(15) + 'd', [nr]);
end;

function  TNotation.getOnlyMainVariation : String;
var b : PBaum;
begin
  Result := '';
  if wurzel = Nil then exit;
  b := wurzel^.B;
  while assigned(b) do begin
    Result := Result + IntToStr(b^.Zug.von) + IntToStr(b^.Zug.nach) + ' ';
    b := b^.b;
  end;
  Result := trim(Result);
end;

// nur fr fernschachprogramm
procedure TNotation.deleteMarkers;
var i : shortint;
    l : PBaumlist;

procedure rek(b : PBaum);
Var bl : PBaumlist;
begin
  while assigned(b) do begin
    if b^.Zug.Attr1 and i <> 0 then
      if not geaendert then
        geaendert:=true;
    b^.Zug.Attr1:=b^.Zug.Attr1 or i xor i;
    bl:=b^.VarWurzel;
    while assigned(bl) do begin
      rek(bl^.B);
      bl:=bl^.BL;
    end;
    b:=b^.B;
  end;
end;

begin
  i := 0 or nErster or nLetzter or nWarVorherLetzter;
  l := Wurzel;
  if l <> Nil then
    rek(l^.B);
  //klettere(wurzel, true) wird in dem Fall extra gemacht vom FS-Prog.
end;
//

procedure TNotation.removeComments;
var l : PBaumlist;

procedure doremove(b : PBaum);
Var bl : PBaumlist;
begin
  while assigned(b) do begin
    if b^.Kommentar <> '' then begin
      if not geaendert then
        geaendert:=true;
      b^.Kommentar := '';
    end;
    bl:=b^.VarWurzel;
    while assigned(bl) do begin
      doremove(bl^.B);
      bl:=bl^.BL;
    end;
    b:=b^.B;
  end;
end;

begin
  l := Wurzel;
  if l <> Nil then begin
    doremove(l^.B);
    if geaendert then
      klettere(wurzel, true);
  end;
end;

procedure TNotation.removeVariations;
var l : PBaumlist;
    li : TList;

procedure doremove(b : PBaum);
Var bl : PBaumlist;
    i : Integer;
begin
  while assigned(b) do begin
    if assigned(b^.VarWurzel) then begin
      li.Clear;
      li.Add(addr(b^.VarWurzel));
      bl:=b^.VarWurzel;
      while assigned(bl^.bl) do begin
        li.Add(addr(bl^.BL^));
        bl:=bl^.BL;
      end;
      for i := 0 to li.Count - 1 do begin
        ZBaumlist := li[i];
        if ZBaumlist <> Nil then begin
          loescheBaum(ZBaumlist);
          if not geaendert then
            geaendert:=true;
        end;    
      end;
    end;
    b:=b^.B;
  end;
end;

begin
  l := Wurzel;
  if l <> Nil then begin
    li := TList.Create;
    try
      doremove(l^.B);
      if geaendert then
        klettere(wurzel, true);
    finally
      li.Free;
    end;
  end;
end;

function  TNotation.getGameExx(whiteCondMoves : Boolean; slMainVar : TStringlist;
  exPGNHeader : Boolean; bias, timeformat : String; limitPGNLength : Boolean;
  writeGameID : String; ICCF, pgnInICCFTable : Boolean) : string;
var ersterZug : boolean;
    rStellung : TSchachbrett;
    rZug : TZug;
    i, Zugnummer : integer;
    b1, bopp, bown : PBaum;
    s, snotat : String;
    klammeraufx, klammerzux : Boolean;
    ebenex : Integer;
    lastLen : Integer;
    writeCondMoves, NotationsRefreshNec : Boolean;
    writingCondMoves : Boolean;

    function SetICCF(i : integer) : String;
    begin
      Result := IntToStr((cc(i) mod 8) + 1) + IntTostr((cc(i) div 8) + 1);
    end;

    function GetICCFMove(AZug : TZug) : string;
    begin
    if pgnInICCFTable then begin
      Result := notiereRawZug(AZug);
      while length(Result) < 6 do
        Result := Result + ' ';
      Result := Result + ' ';  
      exit;
    end;
    Result := SetICCF(Azug.von) + SetICCF(Azug.nach);
    if Azug.UFigur <> LEER then begin
      case abs(Azug.UFigur) of
        wD : Result := Result + '1 ';
        wT : Result := Result + '2 ';
        wL : Result := Result + '3 ';
        wS : Result := Result + '4 ';
      end;
    end
    else Result := Result + '  ';
    end;

    function isMoveErster(b : PBaum) : Boolean;
    begin
      Result := b^.Zug.attr1 and nErster <> 0;
    end;

    function isMoveLetzter(b : PBaum) : Boolean;
    begin
      Result := b^.Zug.attr1 and nLetzter <> 0;
    end;

    function isOwnMove(zug : TZug) : Boolean;
    begin
      Result :=
      ((whiteCondMoves) and (Zug.amZug = SCHWARZ))
      or ((not whiteCondMoves) and (Zug.amZug = WEISS))
    end;

    function isLastMoveMarked(b : PBaum; i : shortint) : Boolean;
  begin
    Result :=
    (b^.rB <> Nil) and
    (b^.rB.Zug.attr1 and i <> 0)       // markierung vorher
  end;

  function isLastMoveLetzter(b : PBaum) : Boolean;
  begin
    Result := isLastMoveMarked(b, nLetzter);
  end;

  function isLastMoveErster(b : PBaum) : Boolean;
  begin
    Result := isLastMoveMarked(b, nErster);
  end;


procedure rek(b : PBaum);
Var rettbrett : TRettbrett;
    rettZugNummer : word;
    l : PBaumList;
    ii, jj : Integer;
    snumber : String;
    forceNumber : Boolean;
    durchlauf : Integer;
    sMove, sMove1, sNr : String;
    p : Pointer;

  function endsWithRightMove(pb : PBaum) : Boolean;
  begin
    Result := false;
    if pb = Nil then exit;
    while pb^.B <> Nil do
      pb := pb^.B;
    if ((whiteCondMoves) and (pb.Zug.amZug = SCHWARZ))
    or ((not whiteCondMoves) and (pb.Zug.amZug = WEISS)) then
      Result := true;
  end;

begin
  if b=Nil then exit;  {drfte nicht auftreten}
  durchlauf := 0;
  repeat
   (* if (ebenex = 1) and (durchlauf = 0) then begin
      if b^.Kommentar = '' then begin
        b^.Kommentar := '{'+condmoves+'}';
      end;
    end;*)

    forceNumber := false;
    if ebenex = 0 then begin
      if writeCondMoves then begin
        writeCondMoves := false;
        snotat := snotat + ret + '{' + acceptedcondmoves + '} ';
        forceNumber := true;
      end;
      if b = bopp  then begin  // vorletzter zug
        if b^.B <> bown then
          writeCondMoves := true;
        snotat := snotat + ret;
      end;
      if b = bown then  // letzter zug
          snotat := snotat + ret;
      if (snotat <> '') and (snotat[length(snotat)] = #10) then
        forceNumber := true;
    end;

    inc(zugnummer);
    if odd(Zugnummer) then begin
      sNr := inttostr((Zugnummer+1) shr 1)+'.';
      snotat:=snotat + sNr + ' '
    end
    else begin
      sNr := inttostr((Zugnummer+1) shr 1)+'...';
      if (klammerzux) or (klammeraufx) or (ersterZug) or (forceNumber) then
        snotat := snotat + sNr + ' ';
    end;
    ersterZug:=false;
    klammeraufx:=false;
    klammerzux:=false;
    NoZug^:=b^.Zug;
    sMove := notiereRawZug(NoZug^);
    if (writingCondMoves) and (ICCF) then
      snotat := snotat + GetICCFMove(NoZug^)
    else
      snotat := snotat + sMove;

    if (slMainVar <> Nil) and (ebenex = 0) then begin
      p := Nil;
      if isOwnMove(NoZug^) then
        p := Pointer(1);
      jj := Pos('...', sNr);
      if jj <> 0 then
        sNr := '...';
      if ICCF then
        sMove1 := GetICCFMove(NoZug^)
      else begin
        sMove1 := sNr + sMove;
        if NoZug^.attr1 and nErster <> 0 then
          sMove1 := '>' + sMove1;
        if NoZug^.attr1 and nLetzter <> 0 then
          sMove1 := sMove1 + '<';
      end;
      slMainVar.AddObject(sMove1, p);
    end;

    if (ExPGNHeader) and
    ((b = bown) or (b = bopp)) then
      snotat := snotat + TimeLineStart + '}';  // danach wird beim zeit einfgen gesucht '  { (2 spaces)

    if (b^.VarWurzel<>Nil)
    and
    (  // schreibt nur variante, die im letzten zug losgeht (Ev.zge)
      (ebenex > 0)    // in evzugfolge allen verzweigugen nachgehen
      or (b = bown)
    )
    then begin
      inc(ebenex);
      l:=b^.Varwurzel;
      repeat
        if (
          (ebenex > 1) and
          (l^.B <> Nil) and (endsWithRightMove(l^.B^.B))
        )
        or
        (
          (l^.B <> Nil) and
          (b^.Zug.von = l^.B^.Zug.von)  and (b^.Zug.nach = l^.B^.Zug.nach)
          and (endsWithRightMove(l^.B^.B))
        )
        then begin
        if ebenex = 1 then begin
          snotat := snotat + ret + '{' + condmoves + '}';
          writingCondMoves := true;
          lastLen := length(snotat);
          if l^.B <> Nil then
            if l^.B^.Kommentar = '' then begin
              l^.B^.Kommentar := '{c.m.}';// '{'+ condmoves +'}';
              NotationsRefreshNec := True;
            end;
        end;
        Rettbrett.Stellung:=NoStellung^;
        Rettbrett.Zug:=NoZug^;
        RettZugNummer := ZugNummer;
        dec(Zugnummer);
        klammeraufx:=true;
        case ebenex of
          1 : begin  {vorher Var. beendet ')'? - dann kein ret, weil es beim Beenden eingefgt wurde}
                snotat:=snotat+' (';
              end;
          2 : begin
                snotat:=snotat+' (';
              end;
          else snotat:=snotat+' (';
        end;
        if l^.B<>Nil then rek(l^.B);
        while snotat[length(snotat)]=' ' do delete(snotat,length(snotat),1);
        case ebenex of
          1 : begin
                snotat:=snotat+')';
              end;
          2 : begin
                snotat:=snotat+')';
              end;
          else snotat:=snotat+')';
        end;
        klammerzux:=true;
        NoStellung^:=Rettbrett.Stellung;
        NoZug^:=Rettbrett.Zug;
        ZugNummer:=RettZugnummer;
        end;
        l:=l^.BL;
        writingCondMoves := false;
      until l=Nil;
      dec(ebenex);
      if ebenex = 0 then begin
        snotat:=snotat+' ';
        exit;
      end;
    end;
    zugvor;

    if (limitPGNLength) and (length(snotat) - lastLen > 70) then begin
      snotat:=snotat + ret;
      lastLen := length(snotat);
    end
    else
      snotat:=snotat+' ';

    if b = bown then
      exit;

    inc(durchlauf);
    b:=b^.B;
  until b=Nil;
end;

var sladj : TStringlist;
    k, l : Integer;

begin
  rStellung:=NoStellung^;
  rZug:=NoZug^;
  try
    writeCondMoves := false;
    lastLen := 0;
    zugnummer:=(Partiedaten.zugnummer-1) shl 1;
    if Partiedaten.Zug.amZug=SCHWARZ then inc(zugnummer);
    NoStellung^:=Partiedaten.Stellung;
    snotat:='';
    ebenex:=0;
    ersterZug:=true;
    bopp := Nil;
    bown := Nil;
    NotationsRefreshNec := false;
    if assigned(wurzel) then begin
      b1:=wurzel^.B;
      // gegner u. eigenen zug suchen
      while b1 <> Nil do begin
        if (isMoveErster(b1))
        or (
          (b1^.B <> Nil) and (b1^.B^.B = Nil)
          and (not (isOwnMove(b1^.Zug)))
        )
        or (
          (b1^.B = Nil)  // eigener zug noch nicht eingetragen
          and (not (isOwnMove(b1^.Zug)))
        )
        then if bopp = Nil then
            bopp := b1;

        if (
          (isLastMoveLetzter(b1))
        )
        or (
          (b1^.B = Nil) and
          (isOwnMove(b1^.Zug))
        )
        then if bown = Nil then
            bown := b1;

        b1 := b1^.B;
      end;

      rek(wurzel^.B);
    end;
    if ExPGNHeader then with Partiedaten.EventData do begin
        // nur die angabe fr den eigenen zug schreiben
      if not whiteCondMoves then
        BlackdifferenceToGMTValue := bias
      else
        WhitedifferenceToGMTValue := bias;
      DateTimeFormatValue := timeformat;
      if (whiteCondMoves)  // wei erstellt gameid
      and (writeGameID <> '') then
        GameIDValue := writeGameID;

      Result := getHeader;
      //nach result geschrieben ist, wieder resetten 
      WhitedifferenceToGMTValue := '';
      BlackdifferenceToGMTValue := '';
      DateTimeFormatValue := '';
      GameIDValue := '';

      Result := Result + #13#10 + trim(snotat) + #13#10 + Partiedaten.EventData.Ergebnis;
    end
    else begin
      Result := getHeader;  // die letzten 2 Zge werden separat auf eigene Zeilen geschrieben - ndern, wenn kein expgn
      sladj := TStringlist.Create;
      try
        sladj.Text := snotat;
        if sladj.Count > 1 then begin  // wenn nur 1 zeile, ist das nicht ntig
          k := Pos('...', sladj[sladj.count - 1]);
          if k <> 0 then begin
            l := k - 1;
            while (l > 0) and (sladj[sladj.count - 1][l] in ['0'..'9']) do
              dec(l);
            if l = 0 then
              sladj[sladj.count - 1] := copy(sladj[sladj.count - 1], 6, length(sladj[sladj.count - 1]));
          end;
          snotat := '';
          for k := 0 to sladj.count - 1 do
            snotat := snotat + trim(sladj[k]) + ' ';
        end;
      finally
        sladj.Free;
      end;
//      hier ergebnis auf gleicher zeile wegen idiot. pgn reader
      Result := Result + #13#10 + trim(snotat) + ' ' + Partiedaten.EventData.Ergebnis;
    end;

  finally
    NoStellung^:=rStellung;
    NoZug^:=rZug;
  end;
  if NotationsRefreshNec then
    NotationRefresh;
end;

procedure TNotation.setCompressedGame(aGame : String);
begin
  // statt ber den PGN-Umweg mit DataToPGN und SetGame wrde das hier viel schneller gehen (fr Suchfunktionen, etc.)
end;

function  TNotation.getCompressedGame : string;
var rStellung : TSchachbrett;
    rZug : TZug;
    bits, bits1, bits2 : TBits;
    ebene : Integer;
    i : Integer;
    VarStart, first : Boolean;

  procedure Bits2Result;
  var i, j : Integer;
      b, b1 : Byte;
  begin
    i := bits.Size mod 8;
    if i <> 0 then begin
      bits.Size := bits.Size + 8 - i;
      bits[bits.Size - 1] := true;  // letzte bits werden nicht gelesen - Wert egal - null vermeiden
    end;
    i := 0;
    while true do begin
      b := 0;
      b1 := $80;
      for j := 0 to 7 do begin
        if i + j >= bits.Size then begin
          bits.size := 0;
          exit;
        end;
        if bits[i + j] then
          b := b or b1;
        b1 := b1 shr 1;  
      end;
      Result := Result + Char(b);
      inc(i, 8);
    end;
  end;

  procedure AddBitField(theBits : TBits);
  var i, j : Integer;
  begin
    j := bits.Size;
    bits.Size := bits.Size + theBits.Size;
    for i := 0 to theBits.Size - 1 do
      bits[i + j] := theBits[i];
  end;

  procedure addBits(b : Byte; number : Integer; theBits : TBits);
  var i, l : Integer;
      k : Byte;
  begin
    i := theBits.Size;
    theBits.Size := theBits.Size + number;
    k := $80;
    k := k shr (8 - number);
    for l := 1 to number do begin
      if b and k <> 0 then
        theBits[i] := true
      else
        theBits[i] := false;
      k := k shr 1;
      inc(i);
    end;
  end;

procedure rek(b : PBaum);
Var rettbrett : TRettbrett;
    l : PBaumList;
    b1, b2 : Byte;
    i : Integer;
begin
  if b=Nil then exit;  {drfte nicht auftreten}
  repeat
    NoZug^ := b^.Zug;

    if not first then
      addBits(1, 1, bits);
    first := false;

    if Varstart then
      addBits(0, 1, bits)
    else
      addBits(1, 1, bits);
    Varstart := false;

    addBits(cc(b^.Zug.von), 6, bits);
    addBits(cc(b^.Zug.nach), 6, bits);
    if (b^.Zug.UFigur <> 0)
    or (b^.Zug.attr1 <> 0) or (b^.Zug.attr2 <> 0)
    or (b^.Zug.infattr <> 0) or (b^.Zug.infattrBefore <> 0)
    or (b^.Kommentar <> '')
    then begin

      addBits(0, 1, bits);

      bits1.Size := 0;
      bits1.Size := 5;
      for i := 0 to bits1.Size - 1 do bits1[i] := true;
      bits2.Size := 0;
      if b^.Zug.UFigur <> 0 then begin
        bits1[0] := false;
        case abs(b^.Zug.UFigur) of
          wS   : b1 := cdPromotionKnight ;
          wL   : b1 := cdPromotionBishop ;
          wT   : b1 := cdPromotionRook   ;
          else   b1 := cdPromotionQueen  ;
        end;
        addBits(b1, 2, bits2);
      end;
      if b^.Zug.attr1 <> 0 then begin
        bits1[1] := false;
        addBits(b^.Zug.attr1 shr 4, 4{8}, bits2);
        //nLetter = $10;  nDigit = $20;  nSchach = $40;  nMatt = $80; nErster  = $1;  nLetzter = $2;  nWarVorherLetzter = $4;
      end;
      if b^.Zug.attr2 <> 0 then begin
        bits1[2] := false;    // !,?, !?, etc.
        addBits(b^.Zug.attr2, 8, bits2);
      end;
      if (b^.Zug.infattr <> 0) or (b^.Zug.infattrBefore <> 0) then begin
        b2 := b^.Zug.infattr;
        case b2 of
          infgegenspiel : b2 := infgegenspielEx;
          infzeitnot    : b2 := infzeitnotEx;
          infneuerung   : b2 := infneuerungEx;
        end;
        if b^.Zug.infattrBefore = infmitderidee then b2 := b2 or 128;
        if b^.Zug.infattrBefore = infBesserIst  then b2 := b2 or  64;

        bits1[3] := false;     // informatorzeichen
        addBits(b2, 8, bits2);
      end;
      if b^.Kommentar <> '' then begin
        bits1[4] := false;
      end;
      AddBitField(bits1);
      AddBitField(bits2);
      if b^.Kommentar <> '' then begin
        Bits2Result;
        Result := Result + b^.Kommentar + cdStringLimiter;
      end;
    end
    else
      addBits(1, 1, bits);

    if (b^.VarWurzel<>Nil) then begin
      l:=b^.Varwurzel;
      repeat
      // das folgende kann weg (wahrscheinlich)
        VarStart := true;
        Rettbrett.Stellung:=NoStellung^;
        Rettbrett.Zug:=NoZug^;
     // END das folgende kann weg (wahrscheinlich)

        Varstart := true;

     // das folgende kann weg (wahrscheinlich)
        inc(ebene);
    // END das folgende kann weg (wahrscheinlich)

        if l^.B<>Nil then rek(l^.B);

      // das folgende kann weg (wahrscheinlich)
        dec(ebene);
      // END das folgende kann weg (wahrscheinlich)

        addBits(0, 1, bits);

    // das folgende kann weg (wahrscheinlich)
        NoStellung^:=Rettbrett.Stellung;
        NoZug^:=Rettbrett.Zug;
    // END das folgende kann weg (wahrscheinlich)

        l:=l^.BL;
      until l=Nil;
    end;

 // das folgende kann weg (wahrscheinlich)
    zugvor;
// END das folgende kann weg (wahrscheinlich)

    b:=b^.B;
  until b=Nil;
end;

begin
  Result := '';
  // Fischer games must be saved as pgn string
  if FFRC <> Nil then begin
    Result := GetGame;
    exit;
  end;

  rStellung:=NoStellung^;
  rZug:=NoZug^;
  bits := TBits.Create;
  bits1 := TBits.Create;
  bits2 := TBits.Create;
  with Partiedaten.EventData do try
    if Jahr <> 0 then
      Result := Result + evDateJahr + Char(hi(Jahr) + 1) + Char(lo(Jahr) + 1);
    if Monat <> 0 then
      Result := Result + evDateMonat + Char(hi(Monat) + 1) + Char(lo(Monat) + 1);
    if Partiedaten.EventData.Tag <> 0 then
      Result := Result + evDateTag + Char(hi(Partiedaten.EventData.Tag) + 1) + Char(lo(Partiedaten.EventData.Tag) + 1);
    if Runde <> 0 then
      Result := Result + evRound + Char(hi(Runde) + 1) + Char(lo(Runde) + 1);
    if WElo <> 0 then
      Result := Result + evWElo + Char(hi(WElo) + 1) + Char(lo(WElo) + 1);
    if SElo <> 0 then
      Result := Result + evSElo + Char(hi(SElo) + 1) + Char(lo(SElo) + 1);
    if PlyCount <> 0 then
      Result := Result + evPlyCount + Char(hi(PlyCount) + 1) + Char(lo(PlyCount) + 1);
    if EventDateJahr <> 0 then
      Result := Result + evEventDateJahr + Char(hi(EventDateJahr) + 1) + Char(lo(EventDateJahr) + 1);
    if EventDateMonat <> 0 then
      Result := Result + evEventDateMonat + Char(hi(EventDateMonat) + 1) + Char(lo(EventDateMonat) + 1);
    if EventDateTag <> 0 then
      Result := Result + evEventDateTag + Char(hi(EventDateTag) + 1) + Char(lo(EventDateTag) + 1);
    if TimeControlGameSec <> 0 then
      Result := Result + evGameSec + Char(hi(TimeControlGameSec) + 1) + Char(lo(TimeControlGameSec) + 1);
    if TimeControlMoveSec <> 0 then
      Result := Result + evMoveSec + Char(hi(TimeControlMoveSec) + 1) + Char(lo(TimeControlMoveSec) + 1);

    if Partiedaten.EventData.Weiss <> '' then
      Result := Result + evWeiss + Partiedaten.EventData.Weiss + cdStringLimiter;
    if Partiedaten.EventData.Schwarz <> '' then
      Result := Result + evSchwarz + Partiedaten.EventData.Schwarz + cdStringLimiter;
    if quelle <> '' then
      Result := Result + evquelle + quelle + cdStringLimiter;
    if Annotator <> '' then
      Result := Result + evAnnotator + Annotator + cdStringLimiter;
    if ort <> '' then
      Result := Result + evort + ort + cdStringLimiter;
    if ergebnis <> '' then
      Result := Result + evergebnis + ergebnis + cdStringLimiter;
    if Eco <> '' then
      Result := Result + evEco + Eco + cdStringLimiter;
    if WhitedifferenceToGMTValue <> '' then
      Result := Result + evWhitediffToGMT + WhitedifferenceToGMTValue + cdStringLimiter;
    if BlackdifferenceToGMTValue <> '' then
      Result := Result + evBlackdiffToGMT + BlackdifferenceToGMTValue + cdStringLimiter;
    if DateTimeFormatValue <> '' then
      Result := Result + evDateTimeFormat + DateTimeFormatValue + cdStringLimiter;
    if GameIdValue <> '' then
      Result := Result + evGameId + GameIdValue + cdStringLimiter;
    if MetaDescription <> '' then
      Result := Result + evMetaDescription + MetaDescription + cdStringLimiter;
    if MetaKeywords <> '' then
      Result := Result + evMetaKeywords + MetaKeywords + cdStringLimiter;

    if not isFirstPosition(Partiedaten.Stellung) then
      Result := Result + evPosition + PosToFen(Partiedaten.Stellung, Partiedaten.Zug, Partiedaten.Zugnummer, FRC) + cdStringLimiter;
    if Partiedaten.Kommentar <> '' then
      Result := Result + evFirstComment + Partiedaten.Kommentar + cdStringLimiter;

    Result := Result + cdHeaderLimiter;

    NoStellung^:=Partiedaten.Stellung;
    if wurzel = Nil then exit;
    ebene := 0;
    VarStart := false;
    first := true;
    rek(wurzel^.B);

    addBits(1, 1, bits);

    Bits2Result;
  finally
    NoStellung^:=rStellung;
    NoZug^:=rZug;
    bits.free;
    bits1.free;
    bits2.free;
  end;
end;

procedure TNotation.klettere(bl1 : PBaumlist; neux : Boolean);
var ersterZug : boolean;
    rStellung : TSchachbrett;
    rZug : TZug;
    i, Zugnummer : integer;
    inHauptvar : Boolean;
    b1 : PBaum;
    FontHauptVar : TFont;
    checkingPos : Boolean;

    hide, show : Boolean;
    currEbene : Integer;

procedure AddRet(f : TFont);
begin
  notat:=notat+' ';
  wrap.AddObject(NumberToString(length(notat) + 1), F);
end;

procedure formatiere;
Var hilf : String; {auch um FontBreite zu bestimmen}
    i, j, k, l, m : longint;
    Indent : integer;
    font : TFont;
    tw : integer;

procedure SetTw;
begin
  if i = 1 then begin
    if Partiedaten.Kommentar = '' then bf.canvas.Font.assign(Font)
    else bf.canvas.Font.assign(FontComment);
  end
  else if Ueberschrift then begin
    if i = UeberschriftEnde + 1 then
      bf.canvas.Font.assign(Font)
  end;
  if (notat[i] <> '}') and (notat[i] <> '{') then
  tw := tw + bf.canvas.textwidth(notat[i])
  else begin
    if (notat[i] = '}') then bf.canvas.Font.assign(Font)
    else bf.canvas.Font.assign(FontComment);
  end;
end;

begin
  try
    Indent := 0;

    hilf:='';
    tw:=0;
    i:=0;
    l := 0;
    m := 0;
    while i<length(notat) do begin
      inc(i);
      if i=GetNumber(l) then begin  {im ersten Element von wrap steht immer 1 (was Funkt. zurckgibt), so da die Font u. Indent immer zuerst init. werden}
        font := wrap.objects[l] as TFont;
        Indent := GetIndent(Font);
        SetTw;
        if bf.canvas.Font.Style <> Font.Style then
          bf.canvas.Font.Style := Font.Style;  // wegen wechsel - falls hvar fett
        tw := bf.canvas.textwidth(notat[i]);
        inc(l);
      end
      else begin
        SetTw;
        if tw>=pb.width - Indent then begin
          hilf:='';
          j:=i;
          repeat
            hilf:=notat[i]+hilf; {ntig, falls ein langes Kommentarwort ohne Leerzeichen drinnen steht}
            dec(i);
          until (notat[i]=' ') or (i=0);
          k:=j-3;
          while (notat[j]<>' ') and (j<length(notat)) do begin
            inc(j);
            hilf:=hilf+notat[j];
          end;
          if bf.canvas.textwidth(hilf{+'    '})>=pb.width - Indent then {dann wars langes Wort - trennen}
            if k>2 then i:=k;
          if i<=m then exit; {dann war TNotat so eng , da RET am Anfang eingefgt werden sollte - nicht zulassen}
          m := i;

          wrap.AddObject(NumberToString(i + 1), Font);
          hilf:='';
          tw := 0;
        end;
      end;
    end;
  except
  end;
end;

procedure docheckPos;
var i, j : Integer;
    ok : Boolean;
begin
  ok := true;
  if pf <> Nil then
    pf.Clear;
  For i := 0 to 63 do begin
    j := c(i);  {Die Funktion c wandelt 0..63 in die Indexe der entsprechenden TSchachbrett-Felder um}
    if pf <> Nil then
      if NoStellung^[j] <> LEER then
        inc(pf.total[NoStellung^[j]]);
    if checkStellung then
      if checkPosBoard[j] <> LEER then begin
        if checkPosBoard[j] <> NoStellung^[j] then begin
          ok := false;
          Break;
        end;
      end;
  end;
  if (ok) and (pf <> Nil) then
    ok := pf.isEqual;
  checkPos := not ok;
end;

procedure rek(b : PBaum);
Var rettbrett : TRettbrett;
    rettZugNummer : word;
    l : PBaumList;
    ii : Integer;
    snumber : String;
begin
  if b=Nil then exit; 
  repeat

    if b = Baumindex then begin
      if hideLastMoves then begin
        hide := true;
        currEbene := Ebene;
      end;
      if showLastMoves then begin
        show := true;
        currEbene := Ebene;
      end;
    end;

    inc(zugnummer);
    b^.ss := length(notat) + 1;  {selstart merken}
    b^.Zugnummer := (Zugnummer+1) shr 1;

    if b^.Zug.attr1 and nErster <> 0 then notat:=notat+'>';

    if not saving then
      notat := notat + infAttrToString(b^.Zug.infattrBefore);

    if odd(Zugnummer) then begin
      notat:=notat+inttostr((Zugnummer+1) shr 1)+'.';
    end
    else if (klammerzu) or (klammerauf) or (ersterZug) {S in Ausgangsstllg. am Zug}
      then begin
        notat:=notat+inttostr((Zugnummer+1) shr 1)+'...';
      end;

    ersterZug:=false;
    klammerauf:=false;
    klammerzu:=false;
    NoZug^:=b^.Zug;

    if hide then
      b^.hidden := true;
    if show then
      b^.hidden := false;

    if (not saving) and (b^.hidden) then
      notat:=notat + '---'
    else begin
      notat:=notat+ notiereZug(NoZug^);
      if saving then
        notat := notat + infAttrToString(b^.Zug.infattrBefore);
    end;

    if b^.hidden then begin
      allSolved := false;
      NextToSolve := b;
    end;

    if b^.Zug.attr1 and nLetzter <> 0 then notat:=notat+'<';

    B^.sl:=length(notat) - (b^.ss - 1);
    if length(b^.kommentar)>2 then begin
      if b^.kommentar[2]=' ' then delete(b^.kommentar, 2,1);
      {bei Zeilenumruch in Pgn-Datei kommt Leerzeichen vor Klammer - das mu raus}
      if b^.kommentar[length(b^.kommentar)-1]=' ' then delete(b^.kommentar, length(b^.kommentar)-1,1);
      if b^.kommentar<>'' then begin
        commwrap.AddObject(IntToStr(length(notat) + 2), Pointer(length(b^.Kommentar)));
        notat:=notat+' '+b^.kommentar;
      end;  
    end;
    if b^.VarWurzel<>Nil  then begin

      inc(ebene);
      l:=b^.Varwurzel;
      repeat
        inHauptvar:=false;
        Rettbrett.Stellung:=NoStellung^;
        Rettbrett.Zug:=NoZug^;
        RettZugNummer := ZugNummer;
        dec(Zugnummer);
        klammerauf:=true;
        case ebene of
          1 : begin  {vorher Var. beendet ')'? - dann kein ret, weil es beim Beenden eingefgt wurde}
                if GetNumber(wrap.count - 1) + 1 <> length(notat) then begin
                  snumber := NumberToString(length(notat) + 2);
                  ii := wrap.IndexOf(snumber);
                  if ii <> -1 then
                    wrap.Objects[ii] := Pointer(FFont1Var)
                  else
                    wrap.AddObject(snumber, FFont1Var);
                end;
                notat:=notat+' [';
              end;
          2 : begin
                if GetNumber(wrap.count - 1) + 1 <> length(notat) then begin
                  snumber := NumberToString(length(notat) + 2);
                  ii := wrap.IndexOf(snumber);
                  if ii <> -1 then
                    wrap.Objects[ii] := Pointer(FFont2Var)
                  else
                  wrap.AddObject(snumber, FFont2Var);
                end;
                notat:=notat+' (';
              end;
          else notat:=notat+' (';
        end;

        if l^.B<>Nil then rek(l^.B);
        if GetNumber(wrap.count - 1) - 1 = length(notat)  {Zeilenumbruch vor Klammer?}
          then wrap.delete(wrap.count - 1);{raus}
        while notat[length(notat)]=' ' do delete(notat,length(notat),1);
        case ebene of
          1 : begin
                notat:=notat+']';
                wrap.AddObject(NumberToString(length(notat) + 2), FontHauptVar);
                inHauptvar:=true;
              end;
          2 : begin
                notat:=notat+')';
                wrap.AddObject(NumberToString(length(notat) + 2), FFont1Var);
              end;
          else notat:=notat+')';
        end;
        klammerzu:=true;
        NoStellung^:=Rettbrett.Stellung;
        NoZug^:=Rettbrett.Zug;
        ZugNummer:=RettZugnummer;
        l:=l^.BL;
      until l=Nil;
      dec(ebene);
    end;

    zugvor;

    if checkPos then begin
      docheckPos;
    end;

    if b=Baumindex then begin
      rStellung:=NoStellung^;
      rZug:=NoZug^;
      inHauptvariante:=inHauptvar;
    end;
    notat:=notat+' ';

    b:=b^.B;
  until b=Nil;
  gemeldet:=false;

  if (hideLastMoves) and (currEbene = Ebene) then
      hide := false;
  if (showLastMoves) and (currEbene = Ebene) then
      show := false;

end;

procedure OrtEintragen;
begin
  with Partiedaten.EventData do begin
    if (ort<>'') and (ort<>'?') then begin
       notat := notat + ort;
       if jahr<>0 then notat:=notat+', '
       else AddRet(FFontHeadLine);
    end;
    if jahr<>0 then begin
      notat:=notat+inttostr(jahr);
      AddRet(FFontHeadLine);
    end;
  end;
end;

function CheckVar : Boolean;
Var b : PBaum;
begin
  Result := false;
  if not assigned(bl1) then exit;
  try
    b := bl1^.B;
    while assigned(b) do begin
      if assigned(b^.VarWurzel) then begin
        Result := true;
        break;
      end;
      b := b^.b;
    end;
  except
  end;
end;

begin
  hide := false;
  show := false;
  currEbene := -1;
  if hideMoves then
    hide := true;
  if showMoves then
    show := true;

  Wrap.Clear;
  CommWrap.Clear;
  inHauptvariante:=true;
  inHauptvar:=true;
  rStellung:=Partiedaten.Stellung;
  rZug:=Partiedaten.Zug;
  zugnummer:=(Partiedaten.zugnummer-1) shl 1;
  if Partiedaten.Zug.amZug=SCHWARZ then inc(zugnummer);
  NoStellung^:=Partiedaten.Stellung;

  CheckingPos := checkPos;
  if checkPos then begin
    docheckPos;
  end;

  if (not CheckVar) or (not FAutoBold) then FontHauptVar := FFont0Var
  else begin
    FontNoVar.assign(FFont0Var);
    FontNoVar.Style := FontNoVar.Style + [fsBold];
    FontHauptVar := FontNoVar;
  end;
  notat:='';
  UeberschriftEnde := 0;
  if ueberschrift then begin
    wrap.AddObject(NumberToString(1), FFontHeadLine);
    with Partiedaten.EventData do begin
      notat:=weiss;
      if (weiss <> '') and (schwarz <> '') then notat := notat +' - ';
      notat := notat + schwarz;
      AddRet(FFontHeadLine);
      if (Quelle<>'') and (Quelle<>'?') then begin
        notat:=notat + Quelle;
        if ((ort<>'') and (ort<>'?')) or (jahr <> 0) then begin
          notat:=notat + ', ';
          OrtEintragen;
        end
        else AddRet(FFontHeadLine);
      end
      else OrtEintragen;
      AddRet(FontHauptVar);
    end;
    UeberschriftEnde := length(notat);

  //  wrap.AddObject(NumberToString(UeberschriftEnde), FontHauptVar);

  end
  else wrap.AddObject(NumberToString(1), FontHauptVar);
  if Partiedaten.kommentar<>'' then begin
    commwrap.AddObject(IntToStr(length(notat) + 1), Pointer(length(Partiedaten.Kommentar)));
    notat:=notat+Partiedaten.kommentar+' ';
  end;
  ebene:=0;  {daran werden Einrckungen gemessen}
  ersterZug:=true;
  if assigned(bl1) then begin
    b1:=bl1^.B;
    rek(b1);
  end;

  if (not Saving) and (not CheckingPos) then
    formatiere;

  if GetNumber(wrap.count - 1) - 1 <> length(notat)  {kein Zeilenumbruch?}
    then AddRet(FFont0Var);
  AddRet(FFontHeadLine);
  notat := notat+Partiedaten.EventData.Ergebnis;

  if not Saving then begin

    if wrap.Count - pb.height div TextHoehe > 0
      then SLeiste.Max := wrap.Count - pb.height div TextHoehe
    else SLeiste.Max := 0; {!}

    i := IsNewScrollbarPos(Baumindex);
    if (i <> -1) and (SLeiste.Position <> i) then SLeiste.Position := i
    else SLeiste.Change; {immer neu zeichnen - Change wird auch nach ndern von Position aufger.}
  end;

  NoStellung^:=rStellung;
  NoZug^:=rZug;

  if niList <> Nil then
    refreshniList;

end;

function TNotation.getNodeAt(index : Integer) : TNodeInfo;
var idx : Integer;
begin
  Result := Nil;
  idx := index - 1;
  createAndRefreshniList;
  if (idx > -1) and (idx < niList.Count) then
    Result := niList[idx];
end;

function TNotation.setPosOfNodeAt(index : Integer) : Boolean;
var idx : Integer;
begin
  Result := false;
  idx := index - 1;
  createAndRefreshniList;
  if (idx > -2) and (idx < niList.Count) then begin
    if idx = -1 then
      Baumindex := Nil
    else
      Baumindex := TNodeInfo(niList[idx]).Baumindex;
    klettere(wurzel, true);
    if assigned(FOnMoveChange) then FOnMoveChange(self, NoZug^, NoStellung^, false)
  end;
end;

function TNotation.NodesCount : Integer;
begin
  createAndRefreshniList;
  Result := niList.Count;
end;

procedure TNotation.clearniList;
begin
  if niList = Nil then exit;
  while niList.Count <> 0 do begin
    TNodeInfo(niList[0]).Free;
    niList.Delete(0);
  end;
end;

procedure TNotation.refreshniList;

  procedure rek(l : PBaumlist);
  var b, rb: PBaum;
      bl : PBaumlist;

    procedure addNode(b : PBaum; lc : Boolean);
    var ni : TNodeInfo;
    begin
      ni := TNodeInfo.Create;
      ni.FLChild := b^.B <> Nil;
      ni.FRChild := lc;
      ni.Baumindex := b;
      niList.Add(ni);
    end;

  begin
    if l = Nil then  exit;
    b := l^.B;
    rb := b;
    while b <> Nil do begin   // erst aufs ende
      addNode(b, (b^.VarWurzel <> Nil) or ((rb = b) and (l^.bl <> Nil)));
      if b^.B = Nil then
        break
      else
        b := b^.B;
    end;
    while b <> Nil do begin
      bl := b^.varWurzel;
      while bl <> Nil do begin
        rek(bl);
        bl := bl^.BL;
      end;
      if b = rb then break;
      b := b^.rB;
    end;
  end;

begin
  if niList = Nil then exit;
  clearniList;
  rek(Wurzel);
end;

procedure TNotation.createAndRefreshniList;
begin
  if niList = Nil then begin
    niList := TList.Create;
    refreshniList;
  end;
end;

procedure TNotation.AdjustPb(b : PBaum);
Var i : integer;
begin
  i := IsNewScrollbarPos(Baumindex);
  if (i <> -1) and (SLeiste.Position <> i) then SLeiste.Position := i
  else SetPbCaret;
end;

function TNotation.IsNewScrollbarPos(b : PBaum) : integer;
Var Linestart, i : longint;
Begin
  Result := -1;
  i := GetLineNumber(b, Linestart);
  if i <> -1 then begin
    if i >= LastLine then Result := i - ((LastLine - FirstLine) shr 1)
    else if i + 1 < FirstLine then begin
      Result := i - ((LastLine - FirstLine) shr 1);
      if Result < 0 then Result := 0;
    end;
  end
  else if FirstLine <> 1 then Result := 0
end;

function TNotation.GetLineNumber(b : PBaum; Var Linestart : longint) : integer;
begin
  Result := -1;
  if assigned(B) then begin
    wrap.Find(NumberToString(B^.ss), Result);
    if (wrap.Count > Result) and (Result >= 0) then begin
      LineStart := StrToInt(trim(wrap.strings[Result]));
      if Linestart <> B^.ss then begin
        dec(Result);
        if (wrap.Count > Result)  and (Result >= 0)  then Linestart := StrToInt(trim(wrap.strings[Result]))
        else Result := -1;
      end;
    end
    else Result := -1;
  end;
end;

function TNotation.GetTextWidth(ALine, a, l : integer) : integer;
Var k : integer;
    szeile : string;
    slf : TStringlist;
begin
  Result := 0;
  try
    slf := TStringlist.Create;
    try
      SetFontsInLine(slf, a, l, wrap.objects[ALine] as TFont, FontComment);
      k := 0;
      while k < slf.count - 1 do begin
        bf.Canvas.Font.Assign(slf.Objects[k] as TFont);
        szeile := copy(notat, StrToInt(slf[k]), StrToInt(slf[k + 1]) - StrToInt(slf[k]));
        szeile := DeleteCommKlammer(szeile);
        Result := Result + bf.canvas.TextWidth(szeile);
        inc(k);
      end;
    finally
      slf.free;
    end;
  except
  end;
end;

procedure TNotation.SetPbCaret;
Var i, LineStart : longint;
    pbIndent : integer;
begin
  if assigned(rNotatBaumindex) then begin
    if IsNewScrollbarPos(rNotatBaumindex) = -1 then begin
      pb.canvas.Font.assign(rFont);
      pb.canvas.TextOut(ZugRect.Left, ZugRect.Top,
      copy(notat, rNotatBaumindex^.ss, rNotatBaumindex^.sl));
    end;
  end;
  rNotatBaumindex := Baumindex;

  i := GetLineNumber(Baumindex, Linestart);
  if i <> -1 then begin
    pb.canvas.Font.assign(wrap.objects[i] as TFont);
    pbIndent := GetIndent(wrap.objects[i] as TFont);
    rFont.assign(pb.canvas.Font);

    ZugRect.Left := pbIndent +   {!!!bei [Italic] + not TrueType klappt das nicht}
      GetTextWidth(i, LineStart, Baumindex^.ss - Linestart);
    ZugRect.Right := ZugRect.Left +
      pb.canvas.TextWidth(copy(notat, Baumindex^.ss, Baumindex^.sl));
    ZugRect.Top := TextHoehe * (i + 1 - FirstLine);
    ZugRect.Bottom := ZugRect.Top + TextHoehe;

    if (ZugRect.Top >= 0) and  {sonst wird herausgescrollte Zeile z.T. angezeicgt}
    (Zugrect.Top < (Lastline - FirstLine + 1) * Texthoehe) then begin
      pb.canvas.Brush.Color := clBlack;
      pb.canvas.Font.Color := clWhite;
      pb.canvas.TextOut(ZugRect.Left, ZugRect.Top,
      copy(notat, Baumindex^.ss, Baumindex^.sl));
      pb.canvas.Brush.Color := Color;
      pb.canvas.Font.Color := clBlack;
    end;
  end;
end;

procedure TNotation.SetFontsInLine(sl : TStringlist; a, l : integer; f, fc : TFont);
Var ca, cl, commi : integer;

function SetComm : Boolean;
begin
  inc(commi);
  if commi < commwrap.count then begin
    ca := StrToInt(commwrap[commi]);
    cl := longint(commwrap.Objects[commi]);
    Result := true;
  end
  else begin
    ca := -1;
    cl := -1;
    Result := false;
  end;
end;

begin
  commi := -1;
  sl.Clear;
  sl.AddObject(IntTostr(a), f);
  while SetComm do begin
    if (ca <= a) and (ca + cl > a) then begin  { Zeile geht mit Komm. los }
      sl.Objects[0] := fc;
      if ca + cl >= a + l then break;  { ausschlielich Komm. }
      sl.AddObject(IntTostr(ca + cl), f);
    end;
    if (ca > a) and (ca < a + l) then begin  { Komm. geht in Zeile los}
      if (ca + cl < a + l) then begin { Komm. dazwischen }
        sl.AddObject(IntTostr(ca), fc);
        sl.AddObject(IntTostr(ca + cl), f);
      end
      else begin      { Komm. geht dann weiter }
        sl.AddObject(IntTostr(ca), fc);
        break;
      end;
    end;
    if ca > a + l then break;
  end;
  sl.AddObject(IntTostr(a + l), Nil);
end;

function TNotation.DeleteCommKlammer(szeile : string) : string;
Var m : integer;
begin
  Result := szeile;
  m := Pos('{', Result);
  if m <> 0 then delete(Result, m, 1);
  m := Pos('}', Result);
  if m <> 0 then delete(Result, m, 1);

end;

procedure TNotation.PaintPb;
Var szeile : string;
    i, j, k, l, m, hdiff, diff, Indent : integer;
    inotata, inotate : integer;
    slf : TStringList;

begin
if FLockNotatUpdate then exit;
try
  SetNotatBounds(SLeiste.Position);
  slf := TStringList.Create;
  try
    pb.canvas.brush.Color := color;
    pb.canvas.pen.Color := color;
    pb.canvas.rectangle(0, 0, pb.width, pb.height);
    j := 0;
    i := 0;
    while i < wrap.count do begin
      if (i >= FirstLine - 1) and (i <= LastLine - 1) then begin
        pb.canvas.Font := FontComment;
        hdiff := pb.canvas.TextHeight('Wg');
        pb.canvas.Font.assign(wrap.objects[i] as TFont);
        hdiff := (pb.canvas.TextHeight('Wg') - hdiff) shr 1;
        inotata := GetNumber(i);
        inotate := GetNumber(i + 1);
        if inotate = -1 then inotate := length(notat) + 1;
        if i < wrap.count - 1 then l := inotate - inotata
        else l := length(notat);
        Indent := GetIndent(wrap.objects[i] as TFont);

        SetFontsInLine(slf, inotata, l, wrap.objects[i] as TFont, FontComment);
        k := 0;
        while k < slf.count - 1 do begin                                   
          if slf.Objects[k] as TFont <> FontComment then diff := 0
          else diff := hdiff;
          pb.Canvas.Font.Assign(slf.Objects[k] as TFont);
          szeile := copy(notat, StrToInt(slf[k]), StrToInt(slf[k + 1]) - StrToInt(slf[k]));
          szeile := DeleteCommKlammer(szeile);
          pb.canvas.TextOut(Indent, TextHoehe * j + diff, szeile);

          if (corrChessMode) and (slf.Objects[k] as TFont <> FontComment) then begin
            m := Pos('>', szeile);
            if m <> 0 then begin
              pb.Canvas.Font.Color := MarkedMoveColor;
              pb.canvas.TextOut(Indent + pb.canvas.TextWidth(copy(szeile, 1, m - 1)), TextHoehe * j + diff, '>');
            end;
            m := Pos('<', szeile);
            if m <> 0 then begin
              pb.Canvas.Font.Color := MarkedMoveColor;
              pb.canvas.TextOut(Indent + pb.canvas.TextWidth(copy(szeile, 1, m - 1)), TextHoehe * j + diff, '<');
            end;
          end;


          Indent := Indent + pb.canvas.TextWidth(szeile);
          inc(k);
        end;
        inc(j);
      end;
      inc(i);
    end;
    rNotatBaumindex := Nil;
    SetPbCaret;
  finally
    slf.free;
  end;
except
end;
end;

procedure TNotation.SetNotatBounds(pos : longint);
Var i : integer;
begin
  FirstLine := pos + 1;
  LastLine := FirstLine + (pb.height div TextHoehe) - 1;

  if LastLine > wrap.Count + 1 then begin  {!}
    i := LastLine - (wrap.Count + 1);
    LastLine := LastLine - i;
    FirstLine := FirstLine - i;
    if FirstLine <= 0 then FirstLine := 1;
  end;
end;

procedure TNotation.DoMouseDown(x, y : integer);
Var z, i : longint;
    s : string;
    Indent, inotata, inotate, l, tw, islf, j, diff : integer;
    slf : TStringList;
begin
  slf := TStringList.Create;
  try
    if (Lastline - Firstline + 1) * Texthoehe < Y then exit;
    z := FirstLine + (y div TextHoehe) - 1;
    if z < wrap.count then begin
      i := StrToInt(trim(wrap.strings[z]));
      pb.canvas.Font.assign(wrap.Objects[z] as TFont);
      Indent := GetIndent(wrap.Objects[z] as TFont);

      inotata := GetNumber(z);
      inotate := GetNumber(z + 1);
      if inotate = -1 then inotate := length(notat) + 1;
      if z < wrap.count - 1 then l := inotate - inotata
      else l := length(notat);

      SetFontsInLine(slf, inotata, l, wrap.objects[z] as TFont, FontComment);
      islf := 0;
      tw := Indent;
      diff := 0;
      while islf < slf.count - 1 do begin
        bf.canvas.font.assign(slf.Objects[islf] as TFont);
        s := copy(notat, StrToInt(slf[islf]), StrToInt(slf[islf + 1]) - StrToInt(slf[islf]));
        j := length(s);
        s := DeleteCommKlammer(s);
        diff := diff + (j - length(s));
        j := 1;
        while j <= length(s) do begin
          tw := tw + bf.canvas.TextWidth(s[j]);
          if tw > x then break;          
          inc(i);
          inc(j);
        end;
        if tw > x then break;
        inc(islf);
      end;
      i := i + diff;
      if retract(i) then begin    {gibt nur true zurck, wenn auch ein Zug angeklickt wurde}
        if assigned(FOnMoveChange) then FOnMoveChange(self, NoZug^, NoStellung^, false)
      end
      else if Ueberschrift then begin
        Baumindex := Nil;
        klettere(wurzel, true);
        if assigned(FOnMoveChange) then FOnMoveChange(self, NoZug^, NoStellung^, false)
      end;
    end;
  finally
    slf.free;
  end;
  try SLeiste.SetFocus; except end;
end;

{------------------------
 Methoden Pb und SLeiste
------------------------}

procedure TNotatPaintbox.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y:Integer);
begin
  if not enabled then exit;
  inherited MouseDown(Button, Shift, X, Y);
  if Button = mbLeft then (Parent as TNotation).DoMouseDown(X, Y);
end;

procedure TNotatPaintbox.paint;
begin
  inherited paint;
  (parent as TNotation).paintpb;
end;

procedure TNotatScrollbar.Change;
begin
  inherited Change;
  (parent as TNotation).paintpb;
end;

{--------------------------------}


procedure TNotation.loescheBaum(Pbl1 : PPBaumList);
var  bl1 : PBaumlist;

procedure rek(bl : PBaumlist);
var l : PBaumList;
    lr : TBaumlist;
    br, b : PBaum;
begin
 try
  b:=bl^.B;
  if b=Nil then exit;
  repeat
    if b^.VarWurzel<>Nil then begin
      l:=b^.Varwurzel;
      repeat
        lr:=l^;
        if l^.BL<>Nil then rek(l^.BL);
        l:=lr.BL;
      until l=Nil;
    end;
    b^.kommentar:='';
    br:=b;
    b:=b^.B;
    dispose(br);
  until b=Nil;
 except
 end;
end;

begin
 bl1:=Pbl1^;
 if bl1=Nil then exit;  {drfte nicht auftreten}
 rek(bl1);
 try
  if bl1^.BL=Nil then begin
    if bl1^.rBL=Nil then begin  {dann mu bl Wurzel oder VarWurzel sein}
      dispose(bl1);
      Pbl1^:=Nil;
    end
    else begin
      bl1^.rBL^.BL:=Nil;  {neues Ende im vorhergehenden Eintrag}
      dispose(bl1);
    end;
  end
  else begin
    if bl1^.rBL=Nil then begin  {nchster Eintrag mu dann in Wurzel o. VarWurzel kopiert werden}
      Pbl1^:=bl1^.BL;
      bl1^.Bl^.rBL:=Nil; 
      dispose(bl1);
    end
    else begin
      bl1^.BL^.rBL:=bl1^.rBL;
      bl1^.rBL^.BL:=bl1^.BL;
      dispose(bl1);
    end;
  end;
 except
 end;
end;

function TNotation.retract(NotatSelstart  : longint) : Boolean;
var i, j, k : longint;
    sZug, fehler, zugzahl, zahl : integer;
    space, punkt : boolean;
    zahlwort, fahrplan : string;

procedure suche;
var iFahrplan : integer;
    e : boolean;
    z : integer;
    rBaumindex : PBaum;
begin
  NoStellung^:=Partiedaten.Stellung;
  wurzelindex:=wurzel;
  iFahrplan:=1;
  z:=0;
  if wurzelindex^.B=Nil then exit; 
  Baumindex:=Wurzelindex^.B;
  rBaumindex:=Baumindex;
  while (Baumindex<>Nil) and (z<zugzahl) do begin
    if Baumindex^.VarWurzel<>Nil then begin
      e:=false;
      Wurzelliste:=Baumindex^.VarWurzel;
      while (not(iFahrplan>length(Fahrplan))) and (not e) do begin
        if Fahrplan[iFahrplan]='0' then begin
          if Wurzelliste^.BL=Nil then e:=true
          else Wurzelliste:=Wurzelliste^.BL;
        end
        else begin
          if Wurzelliste^.B=Nil then exit; {drfte nicht auftreten}
          inHauptvariante:=false;
          Baumindex:=Wurzelliste^.B;
          e:=true;
        end;
        inc(iFahrplan);
      end;
    end;
    NoZug^:=Baumindex^.Zug;
    zugvor;
    rBaumindex:=Baumindex;
    Baumindex:=Baumindex^.B; {Baumindex bleibt auf einem Zug nach dem angeklickten stehen}
    inc(z);
  end;
  Baumindex:=rBaumindex;      {Baumindex auf angeklickten Zug}
  NoZug^:=Baumindex^.Zug;
(*  zugnummer:=z-1;*)
end;

function springe(c : char; j : longint) : longint;
var
    c1 : char;
    e : integer;
begin
  case c of
    ')' : c1:='(';
    '}' : c1:='{';     {in Kommentaren geschweifte Klammern nicht zulassen!}
    ']' : c1:='[';
    else c1:='(';
  end;
  e:=1;
  while (e<>0) and (j>1) do begin
    dec(j);
    if (c1 <> '{') and (notat[j] = '}') then j := springe('}', j);
    if notat[j]=c then inc(e);
    if notat[j]=c1 then dec(e);
  end;
  Result:=j;
end;

begin
  Result:=false;
  i:=1;
  zugzahl:=0;
  space:=false;
  punkt:=false;
  fahrplan:='';
  k:=0;
  if notat='' then exit;
  (**)
  if (ueberschrift) and (NotatSelstart <= UeberschriftEnde)  then exit;  {da stand selstart auf berschrift - die kann auch am Anfang sein}
  if length(notat) <= NotatSelstart then exit;  {dann stehen keine Zge drinnen}

  i:=NotatSelstart;
  if (notat[i]='}') or (notat[i]='{') then exit; {im Kommentar (dort drfen keine { stehen) ? --> raus}
  j:=i;
  k:=0;
  while (j>0) and (notat[j]<>'{') and (k=0) do begin
    if notat[j]='}' then k:=1;
    dec(j);
  end;
  if ((j > 0)(**) and (notat[j]='{')) and (k = 0) then exit;
  k:=0;
  while (i<length(notat)) and (notat[i] in sonderzeichen) do inc(i); {von ev. Sonderzeichen weg}
  if (i > 1) and (notat[i - 1] = '{') then exit; (**)
  while (i<length(notat)) and (not(notat[i] in sonderzeichen)) do begin
    inc(i); {hinter Zug setzen - es gibt nach Komm. auch Zge ohne 1. (Punkt)}
  end;
  if notat[i]='}' then exit;  {dann wars im Kommentar}
  if (i<2) or (i>=length(notat)) then exit;  {deckt auch Klick auf Ergebnis ab}
  j:=i-1;
  dec(i);  {steht sonst auf Leer}
  sZug:=-1;
  {wenn Cursor auf "1.wZug" steht wird Zugnummer verdoppelt und es mu ein Halbzug (kein schw. Zug dazu!) abgezogen werden}
  while (i>UeberschriftEnde) do begin
    if (notat[i]='(') or (notat[i]='[') then fahrplan:='1' + fahrplan;
    if (notat[i]=')') or (notat[i]='}') or (notat[i]=']') then begin
      i:=springe(notat[i], i);
      if (notat[i]='(') or (notat[i]='[')
      then fahrplan:='0' + fahrplan;
    end;
    if (notat[i]=' ') or (notat[i]=#$D) then  {auch Zeilenspung beachten}
    if (punkt=false) and (space=false) then begin
      sZug:=0; {s. Zug - davor steht keine Nummer - gerade Halbzugzahl}
      space:=true;
      k:=i;
    end;
    if notat[i]='.' then if punkt=false then begin
      while (i>0) and (not(notat[i] in ['0'..'9'])) do begin
        dec(i);  {Zugnummer lesen}
      end;
      if i+2<length(notat) then
        if notat[i+2]='.' then sZug:=0;
        {auch vor s. Zug kann 1... (Zugnummer) stehen - in dem Fall gerade Halbzugzahl (Zugnummer shl 1)}
      while (i>0) and (notat[i] in ['0'..'9']) do begin  {wegen 2/3 stelligen Nummern}
        zahlwort:=notat[i]+zahlwort;
        dec(i);
      end;
      if i=1 then dec(i);  {an erster Stelle wurde Schleife auf Zugnummer - nicht davor - abgebrochen}
      if k=0 then k:=i;  {nur wenn kein s. Zug markiert werden muss}
      inc(i);  {weil es am Ende noch einmal dec. wird - wrde sonst Klammern usw. bersehen}
      val(zahlwort , zahl, fehler);
      if fehler<>0 then exit;  {drfte nicht auftreten}
      zahl:=zahl-(Partiedaten.Zugnummer-1);
      zahl:=zahl shl 1;
      if Partiedaten.Zug.amZug=-1 then dec(zahl); {wenn Schwarz am Zug in Ausgangsstellung, mu dec. werden}
      zugzahl:=zahl+sZug;  {Anzahl Halbzge ist gefragt}
      punkt:=true;
      space:=true;
    end;
    dec(i);
  end;
//  if notat[k] = #$D then inc(k);        (**)
 (* wrapselstart:=k;
  wrapsellength:=abs(k-j);*)
  inHauptvariante:=true;  {wird in Suche ev. falsch, falls im Fahrplan 1 enthalten ist u. dadurch Var. eingetaucht wird}
  suche;
  SetPbCaret;     (**)
  Result:=true;
end;

procedure TNotation.KeyDown(Var Key: Word; Shift: TShiftState);

procedure ChangeLine(summand : integer);
Var i, j, Linestart, NextLinestart, Next2Linestart, NotatSelstart : longint;
    e : Boolean;
function Checkmove(nss : longint) : Boolean;
begin
  Result := false;
  if retract(NotatSelstart) then begin
    i := IsNewScrollbarPos(Baumindex);
    if (i <> -1) and (SLeiste.Position <> i)
      then SLeiste.Position := i;
    if assigned(FOnMoveChange) then FOnMoveChange(self, NoZug^, NoStellung^, false);
    Result := true;
  end
end;

begin
  if not enabled then exit;
  i := GetLineNumber(Baumindex, Linestart);
  if i = -1 then exit;
  if (wrap.Count > i + summand) and (i + summand >= 0) then begin
    j := (Baumindex^.sl shr 1) + Baumindex^.ss - Linestart;
    NextLinestart := StrToInt(trim(wrap.strings[i + summand]));
    if summand = -1 then begin
      if NextLineStart + j < LineStart
        then NotatSelstart := NextLineStart + j
      else NotatSelstart := Linestart - 3;
    end
    else begin
      if wrap.Count > i + summand + 1 then
        Next2Linestart := StrToInt(trim(wrap.strings[i + summand + 1]))
      else exit;
      if NextLineStart + (Baumindex^.ss - Linestart) < Next2LineStart
        then NotatSelstart := NextLineStart + j
      else NotatSelstart := Next2Linestart - 3;
    end;
    if NotatSelstart > length(notat) then NotatSelstart := length(notat) - 3;
    if NotatSelstart < 1 then NotatSelstart := 1;
    while(NotatSelstart > 1) and (notat[NotatSelstart] = ' ') do dec(NotatSelstart);
    if not CheckMove(NotatSelstart) then begin
      e := false;
      if summand = -1 then begin
        i := 0;
        if wrap.count > FirstLine - 1 then i := StrToInt(trim(wrap.strings[FirstLine - 1]));
        while (NotatSelstart > i) and (length(notat) >= NotatSelstart) do begin
          if ((notat[Notatselstart] in ['1'..'8']) and (notat[Notatselstart - 1] in ['a'..'h'])) or
          (((upcase(notat[Notatselstart]) = 'O')  or (notat[Notatselstart]='0')) and (notat[Notatselstart - 1] = '-')) then begin
            if CheckMove(NotatSelstart) then begin
              e := true;
              break;
            end;
          end;
          dec(NotatSelstart);
        end;
      end
      else begin
        i := length(notat);
        if wrap.count > LastLine - 1 then i := StrToInt(trim(wrap.strings[LastLine - 1]));
        while (NotatSelstart < i) and (length(notat) > NotatSelstart) do begin
          if ((notat[Notatselstart + 1] in ['1'..'8']) and (notat[Notatselstart] in ['a'..'h'])) or
          (upcase(notat[Notatselstart + 1]) = 'O') and (notat[Notatselstart] = '-') then begin
            if CheckMove(NotatSelstart) then begin
              e := true;
              break;
            end;
          end;
          inc(NotatSelstart);
        end;
      end;
      if not e then SLeiste.Position := SLeiste.Position + summand;
    end;
  end;
end;

begin        (**)
  inherited keyDown(key, shift);
  case key of
    38 : ChangeLine(-1);
    40 : ChangeLine(1);
    33 : GoPreviousVar;
    34 : GoNextVar;
    36 : GotoStart;
    35 : GotoEnd;
    39 : GoAhead;
    37 : GoBack;
  end;
end;

procedure TNotation.NeueStellung(Stellung : TSchachbrett; Zug : TZug; z : integer);
begin
  geaendert:=true;
  UD:=0;
  if z<>-1 then FPartiedaten.Zugnummer:=z;
  FPartiedaten.Stellung:=Stellung;
  FPartiedaten.Zug:=Zug;
  NoStellung^:=Stellung;
  checkFischerPos;
//.  ResetEventData(FPartiedaten.EventData);
  LoescheNotatImFenster;
end;

procedure TNotation.LoescheNotatImFenster;
begin
  UD:=0;
  if UDWurzel<>Nil then begin
    ZBaumlist:=addr(UDWurzel);
    loescheBaum(ZBaumlist);
  end;
  if wurzel<>Nil then begin
    ZBaumlist:=addr(wurzel);
    loescheBaum(ZBaumlist);
  end;
  Baumindex:=Nil;
  WurzelIndex := Nil;
  Klettere(wurzel, true);
end;


procedure TNotation.Dialog_ini;
Var P : TPoint;
begin
  VarDialog:=TForm.Create(self);
  with VarDialog do begin
    Font.Name:='MS Sans Serif';
    Font.Size:=8;
    width:=255;
    height:=147;
    BorderIcons:=[biSystemMenu];
  end;
  P.X:=0;
  P.Y:=Height;
  P:=ClientToScreen(P);
  if P.X+VarDialog.Width>Screen.Width then P.X:=P.X-(P.X+VarDialog.Width-Screen.Width);
  if P.Y+VarDialog.Height>Screen.Height then P.Y:=P.Y-(P.Y+VarDialog.Height-Screen.Height);
  VarDialog.Top:=P.Y;
  VarDialog.Left:=P.X;
end;

procedure TNotation.Varianten_anzeigen(keineZuege : Boolean);
Var hs : string;
    j : integer;
begin
  with lZuege do begin
    left:=2;
    top:=2;
    width:=121;
    Height := 113;
    Font.Style:=Font.Style+[fsBold];
    hs:='';
    if keineZuege then begin
      if Partiedaten.Zug.amZug=WEISS then hs:='.'
      else hs:='...';
    end
    else begin
      if baumindex^.Zug.amZug=WEISS then begin (**)
        hs:='.';
      end
      else hs:='...';
    end;
    j:=65;
    if keineZuege then begin
      wurzelindex:=Baumindex^.VarWurzel;
      Items.add('A) '+IntTostr(Baumindex^.zugnummer)+hs+notiereZug(Baumindex^.Zug));
    end
    else begin
      wurzelindex:=Baumindex^.B^.VarWurzel;
      Items.add('A) '+IntTostr(Baumindex^.B.zugnummer)+hs+notiereZug(Baumindex^.B.Zug));
    end;
    while assigned(wurzelindex) do begin
      inc(j);
      if assigned(wurzelindex^.B) then
        Items.add(chr(j)+') '+IntTostr(wurzelindex^.B.zugnummer)+hs+notiereZug(wurzelindex^.B^.Zug));
      wurzelindex:=wurzelindex^.BL;
    end;
    Itemindex:=0;
  end;
end;

function TNotation.AddZug(aZug : TZug) : Boolean;
begin
  Result := doAddZug(aZug) = 0;
end;

function TNotation.AddZugEx(aZug : TZug) : Integer;
begin
  Result := doAddZug(aZug);
end;

function TNotation.doAddZug(aZug : TZug) : Integer;
var linie, reihe, hr, hl, j : byte;
    keineZuege : Boolean;
    rS : TSchachBrett;
    z3, z50, rnach, rFigur, ramZug, rep : integer;
    rRochademoeglichkeiten : TRochademoeglichkeiten;
    e, first, gleich : Boolean;
    rB : PBaum;
    Zug : TZug;
    aendern : integer;
    cpressed, apressed : Boolean;
    mr : TModalResult;
    rZug : TZug;
    moveOk : Boolean;
    rBaumidx : PBaum;
    rStllg : TSchachbrett;
    rZg, lastMove : TZug;
    l : TList;
    i : Integer;

  procedure rek(b : PBaum);
  Var bl : PBaumlist;
      EmptyvonFeld_handled : Boolean;
      rPos : TSchachbrett;
      rMove : TZug;
  begin
    EmptyvonFeld_handled := false;
    while assigned(b) do begin
      bl:=b^.VarWurzel;
      while assigned(bl) do begin
        rPos := NoStellung^;
        rMove := NoZug^;
        rek(bl^.B);
        bl:=bl^.BL;
        NoStellung^ := rPos;
        NoZug^ := rMove;
        lastMove := NoZug^;
      end;

      NoZug^ := b^.Zug;

      if (not EmptyvonFeld_handled) and (NoStellung[NoZug^.von] = 0) then begin
        EmptyvonFeld_handled := true;
        NoZug^.von := Zug.nach;
        l.Add(b);
      end;
      lastMove.von := NoZug^.von;  // letzer Zug mu wegen Stellungs-Infos (Rochadem., etc.) genommen werden
      lastMove.nach := NoZug^.nach;
      if not IsMovePossible(NoStellung, @lastMove, FRC) then
        FErrorMoves := FErrorMoves + ' ' + IntToStr(NoZug^.von) + '-' + IntToStr(NoZug^.Nach);

      lastMove := NoZug^;
      zugvor;

      b:=b^.B;
    end;
  end;

var  bl : PBaumList;
 ba : PBaum;
 rTempVarIndex:PBaumList;
   rSameNode:Boolean;


begin


(*
// You may activate the following if you do not want the variation dialog
// if the same move already exists as variation

if ((Baumindex <> Nil) and (Baumindex^.B <> Nil))
or ((Baumindex = Nil) and (Wurzel <> Nil) and (Wurzel^.B <> Nil)) then begin
  if Baumindex = Nil then begin
    bl := Wurzel^.B^.VarWurzel;
    ba := Wurzel^.B;
  end
  else begin
    bl := Baumindex^.B^.VarWurzel;
    ba := Baumindex^.B;
  end;
  if (Ba.Zug.von = aZug.von) and (Ba.Zug.nach = aZug.nach) and
  (Ba.Zug.UFigur = aZug.UFigur) then begin
    Baumindex := Ba;
    klettere(wurzel, true);
    exit;
  end;
  while assigned(bl) do begin
    if (bl.B.Zug.von = aZug.von) and (bl.B.Zug.nach = aZug.nach) and
    (bl.B.Zug.UFigur = aZug.UFigur) then begin
      Baumindex := bl.B;
      klettere(wurzel, true);
      exit;
    end;
    bl := bl^.BL;
  end;
end;
          *)


  Result:=0;
  cpressed := FCtrlPressed;
  apressed := FAltPressed;
  FCtrlPressed := false;
  FAltPressed := false;
  if corrChessMode then begin
    aendern:=AreMovesMarked;
    if (aendern=2) or (aendern=3) then begin
      if assigned(FOnMoveChange) then FOnMoveChange(self, NoZug^, NoStellung^, false);
      Result:=1;
      exit;
    end;
  end;  

  Zug := aZug;
  UD:=0;
  Zug.attr1:=0;  {ntig, weil es von TBrett u. TMiniMax nicht behandelt u. undefiniert bergeben wird}
  Zug.attr2:=0;
  Zug.infattr:=0;
  Zug.infattrBefore:=0;
  Fcapture := Zug.geschlFigur <> 0;
  rBaumidx := Baumindex;
  if not assigned(wurzel) then begin  {bei Hauptvar. - also Partie lschen -, wir Wurzel freigegeben}
    New(wurzel);
    wurzel^.BL:=Nil;
    Wurzel^.rBL:=Nil;
    wurzel^.B:=Nil;
  end;
  if not assigned(Baumindex) then begin
    Baumindex:=Wurzel^.B; {Sonderfall - kein Zug markiert}
    keineZuege:=true;
  end else keineZuege:=false;
  if not assigned(Baumindex) then begin
    BaumEinfuegen(Zug);
    Wurzel^.B:=Baum;
    Baumindex:=Baum;
  end
  else begin
    if (not assigned(Baumindex^.B)) and (keineZuege=false) then begin
      BaumEinfuegen(Zug);
      Baum^.rB:=Baumindex;
      Baumindex^.B:=Baum;
      Baumindex:=Baum;
    end
    else begin
      Dialog_ini;
      if English then VarDialog.Caption:='Enter new move' else VarDialog.Caption:='Neuen Zug eingeben';

      Einfuegen:=TButton.Create(self);
      with Einfuegen do begin
        ModalResult:=mrOk;
        left:=140;
        top:=2;
        width:=100;
        Height := 25;
        if English then Caption:='New variation' else Caption:='Neue Variante';
      end;
      VarDialog.insertControl(Einfuegen);

      Ueberschreiben:=TButton.Create(self);
      with Ueberschreiben do begin
        ModalResult:=mrIgnore;
        left:=140;
        top:=30;
        width:=100;
        Height := 25;
        if English then Caption:='Overwrite' else Caption:='berschreiben';

        if corrChessMode then
          if aendern=1 then begin
            Enabled:=False;
            aPressed := false;
          end;
      end;
      VarDialog.insertControl(Ueberschreiben);

      Replace:=TButton.Create(self);
      with Replace do begin
        ModalResult:=mrRetry;
        left:=140;
        top:=58;
        width:=100;
        Height := 25;
        if English then Caption:='Replace move' else Caption:='Zug ersetzen';

        if corrChessMode then
          Enabled := false;
      end;
      VarDialog.insertControl(Replace);


      Abbruch:=TButton.Create(self);
      with Abbruch do begin
        ModalResult:=mrCancel;
        left:=140;
        top:=90;
        width:=100;
        Height := 25;
        if English then Caption:='Cancel' else Caption:='Abbruch';
      end;
      VarDialog.insertControl(Abbruch);

      lZuege:=TListBox.Create(self);
      VarDialog.insertControl(lZuege);

      try
      Varianten_anzeigen(keineZuege);

      lZuege.Enabled := false;
      lZuege.Itemindex := -1;

      if (not cpressed) and (not apressed) then
        mr := VarDialog.Showmodal
      else begin
        if cpressed then mr := mrOk
        else mr := mrIgnore;
      end;
      if (mr = mrRetry) and
      (
        (Baumindex = Nil) or
        ((Baumindex^.B = Nil) and (keineZuege=false))
      ) then
        mr := mrCancel;
      case mr of
        mrOK     : begin
                     if keineZuege=false then Baumindex:=Baumindex^.B; {Sonderfall - kein Zug markiert}
                     wurzelindex:=Baumindex^.VarWurzel;
                     VarEinfuegen;  {wurzelindex zeigt danach auf neue Variantenwurzel}
                     BaumEinfuegen(Zug);
                     Baum^.rB:=Baumindex;
                     wurzelindex^.B:=Baum;
                     Baumindex:=Baum;
                   end;
        mrIgnore : begin  {berschreiben}
                     New(Wurzelindex);
                     Wurzelindex^.BL:=Nil;
                     Wurzelindex^.rBL:=Nil;
                     if keineZuege then Wurzelindex^.B:=Baumindex
                     else Wurzelindex^.B:=Baumindex.B;
                     ZBaumlist:=addr(Wurzelindex);
                     loescheBaum(ZBaumlist);  {dabei wird auch wurzelindex wieder frei}
                     BaumEinfuegen(Zug);
                     if keineZuege then begin
                       Baum^.rB:=Nil;
                       Wurzel^.B:=Baum;
                     end
                     else begin
                       Baum^.rB:=Baumindex;
                       Baumindex^.B:=Baum;
                     end;
                     Baumindex:=Baum;
                   end;
        mrCancel : begin
                     if keineZuege then Baumindex:=Nil;
                     if assigned(FOnMoveChange) then FOnMoveChange(self, NoZug^, NoStellung^, false);
                     Result:=-1;
                     exit;
                   end;
        mrRetry : begin
                    moveOk := false;
                    if keineZuege=false then begin
                      Baumindex := Baumindex^.B;
                      lastMove := NoZug^;
                    end
                    else lastMove := Partiedaten.Zug;
                    rZug := Baumindex^.Zug;
                    Baumindex^.Zug := Zug;
                    FErrorMoves := '';
                    rStllg := NoStellung^;
                    rZg := NoZug^;
                    l := TList.Create;
                    try
                      rek(Baumindex);
                      if FErrorMoves = '' then begin
                        moveOk := true;
                        for i := 0 to l.Count - 1 do
                          PBaum(l[i])^.Zug.von := Zug.nach;
                      end;
                  //    else
                    //    ShowMessage(FErrorMoves);
                    finally
                      NoStellung^ := rStllg;
                      NoZug^ := rZg;
                      FErrorMoves := '';
                      l.Free;
                    end;

                    if not moveOk then begin
                      Baumindex^.Zug := rZug;
                      Baumindex := rBaumidx;
                      klettere(wurzel, true);
                      if assigned(FOnMoveChange) then FOnMoveChange(self, NoZug^, NoStellung^, false);
                      Result:=-1;
                      exit;
                    end;
                  end;
      end;
      finally
        Ueberschreiben.Free;
        Einfuegen.Free;
        Abbruch.Free;
        Replace.Free;
        lZuege.Free;
        VarDialog.free;
      end;
    end;
  end;
  NoZug^:=Zug;
  NoZug^.attr1:=0;  {ntig, weil es von TBrett u. TMiniMax nicht behandelt u. undefiniert bergeben wird}
  NoZug^.attr2:=0;
  NoZug^.infattr:=0;
  NoZug^.infattrBefore:=0;
  inBetrachtindex:=0;  {Index fr das Array, in dem alle Felder gespeichert werden, wo Figuren stehen,die auf Zielfeld knnen}
  NoZug^.amZug:=-NoZug^.amZug;
  case abs(NoStellung^[NoZug^.von]) of
    wD : sucheDame(NoZug^.nach, false);
    wT : sucheTurm(NoZug^.nach, false);
    wL : sucheLaeufer(NoZug^.nach, false);
    wS : sucheSpringer(NoZug^.nach, false);
    wK : inBetrachtindex:=1;
    wB : inBetrachtindex:=1;
  end;
  NoZug^.amZug:=-NoZug^.amZug;
  if inBetrachtindex>1 then begin
    hl:=cc(NoZug^.von) mod 8;
    hr:=cc(NoZug^.von) div 8;
    for j:=1 to inBetrachtindex do begin
      if inBetracht[j]<>NoZug^.von then begin
        linie:=cc(inBetracht[j]) mod 8;
        reihe:=cc(inBetracht[j]) div 8;
        {eine andere Figur knnte aufs Zielfeld - Ausganslinie/-Reihe wird in NotiereZug eingefgt bei attr1<>0}
        if hr=reihe then Baumindex^.Zug.attr1:=Baumindex^.Zug.attr1 or nLetter;
        if hl=linie then Baumindex^.Zug.attr1:=Baumindex^.Zug.attr1 or nDigit;
      end;
    end;
    {Falls zweite in Frage kommende Figur weder auf derselben Linie noch Reihe (Springer) steht - Buchstabe}
    if Baumindex^.Zug.attr1 and nDigit = 0 then Baumindex^.Zug.attr1:=Baumindex^.Zug.attr1 or nLetter;
  end;
  zugvor;
  j:=0;  {Schachprfung}
  while (NoStellung^[c(j)]<>NoZug^.amZug*wK) and (j<=63) do inc(j);
  if sucheDame(c(j), true) or sucheTurm(c(j), true) or sucheLaeufer(c(j), true)
  or sucheSpringer(c(j), true) or sucheBauer(c(j), true) then
    Baumindex^.Zug.attr1:=Baumindex^.Zug.attr1 or nSchach;
  klettere(wurzel, true);
  geaendert:=true;
                  {Prfen auf 50 Zge-Regel u. 3-fache Stellungswiederholung}
  rS:=NoStellung^;
  z3:=1;
  rnach:=Baumindex^.Zug.nach;
  rFigur:=NoStellung^[Baumindex^.Zug.nach];
  rRochademoeglichkeiten:=Baumindex^.Zug.Rochademoeglichkeiten;
  rep:=Baumindex^.Zug.epFeld;
  ramZug:=Baumindex^.Zug.amZug;
  z50:=0;
  e:=false;
  rB:=Baumindex;
  first:=true;        {rckwrts prfen}
  while (assigned(rB)) and (z50<100) and (not e) and (z3<3) do begin
    if (rB^.Zug.Rochadeart<>LEER) or (rB^.Zug.epSchlag<>LEER)
    or (rB^.Zug.geschlFigur<>LEER) or (abs(rS[rB^.Zug.nach])=wB)
    or (rB^.Zug.UFigur<>LEER) then e:=true;
    if not e then begin
      if not first then begin
        if rS[rnach]=rFigur then begin
          gleich:=true;
          For j:=0 to 63 do if NoStellung^[c(j)]<>rS[c(j)] then gleich:=false;
          if (rRochademoeglichkeiten=rB^.Zug.Rochademoeglichkeiten)
          and (rep=rB^.Zug.epFeld) and (ramZug=rB^.Zug.amZug) and (gleich) then inc(z3);
        end;
      end else first:=false;
      rS[rB^.Zug.von]:=rS[rB^.Zug.nach];
      rS[rB^.Zug.nach]:=LEER;  {Zug zurck}
      if assigned(rB^.rB) and (rB^.Zug.amZug=rB^.rB^.Zug.amZug) then {an Verzweigung gekommen}
        rb:=rB^.rB^.rB
      else rB:=rB^.rB;
      inc(z50);
    end;
  end;
  if (not assigned(rB)) and (not e) then begin
    gleich:=true;
    For j:=0 to 63 do if NoStellung^[c(j)]<>Partiedaten.Stellung[c(j)] then gleich:=false;
    if (rRochademoeglichkeiten=Partiedaten.Zug.Rochademoeglichkeiten)
    and (rep=Partiedaten.Zug.epFeld) and (ramZug=Partiedaten.Zug.amZug) and (gleich)
      then inc(z3);
  end;
  if z3=3 then if assigned(On3Repetition) then On3Repetition(self);
  if z50=100 then if assigned(On50MovesRule) then On50MovesRule(self);
end;

procedure TNotation.loesche_UD;
begin
  if assigned(UDWurzel) then begin
    ZBaumlist:=addr(UDWurzel);
    loescheBaum(ZBaumlist);
  end;
  New(UDWurzel);
  UDWurzel^.rBL:=Nil;
  UDWurzel^.BL:=Nil;
  UD:=0;
end;

function TNotation.schneideEx(wie : Boolean; dummy : TObject) : Boolean;
begin
  Result := doschneide(wie, dummy);
end;

function TNotation.schneide(wie : Boolean) : Boolean;
begin
  Result := doschneide(wie, Nil);
end;


function TNotation.doschneide(wie : Boolean; dummy : TObject) : Boolean;
Var nLetzterSetzen : Boolean;
begin
  Result := false;
  if not assigned(Baumindex) then exit;  {kein Zug gewhlt - Ausgangsstellung}
  if wie then begin
    if assigned(Baumindex^.B) then begin

      if corrChessmode then
      case AreMovesMarked of
        1,3 : exit;  // vorher darf nicht geschnitten werden, EvZugfolge kann dagegen schon verkrzt werden
        2   : begin
                if dummy <> Nil then begin
                  if ((Integer(dummy) = 1) and (Baumindex^.Zug.amZug=WEISS)) or
                  ((Integer(dummy)=-1) and (Baumindex^.Zug.amZug=SCHWARZ))
                    then exit; // in EvZugfolge Zge nur paarweisse schneiden
                end;
                nLetzterSetzen:=true;
              end;
        0   : nLetzterSetzen:=true;   //!!!l347 letzterSetzen mu auch gemacht werden, wenn nur ein Zug markiert ist!
        4   : nLetzterSetzen:=false;
      end;

      geaendert:=true;
      New(Wurzelindex);
      Wurzelindex^.BL:=Nil;
      Wurzelindex^.rBL:=Nil;
      Wurzelindex^.B:=Baumindex.B;

      loesche_UD;
      UD:=-1;
      UDWurzel^.B:=Baumindex^.B;
      UDBi:=Baumindex;

      Baumindex.B:=Nil;

      if corrChessmode then
      if nLetzterSetzen then begin
        if Baumindex^.Zug.Attr1 and nLetzter <> 0
          then Baumindex^.Zug.Attr1:=Baumindex^.Zug.Attr1 or nWarVorherLetzter
        else Baumindex^.Zug.Attr1:=Baumindex^.Zug.Attr1 or nLetzter;
      end;

      klettere(wurzel, true);
      Result := true;
    end;
  end
  else begin
    if not inHauptvariante then exit;
    geaendert:=true;
    UDPartieDaten:=Partiedaten;

    FPartiedaten.Zugnummer:=Baumindex^.Zugnummer;
    if Baumindex^.Zug.amZug = WEISS then inc(FPartiedaten.Zugnummer);(**)

    FPartiedaten.zug:=NoZug^;
    FPartiedaten.Stellung:=NoStellung^;
    New(Wurzelindex);
    Wurzelindex^.BL:=Nil;
    Wurzelindex^.rBL:=Nil;
    Wurzelindex^.B:=Baumindex^.B;
    if assigned(Baumindex^.B) then Baumindex^.B^.rB:=Nil;
    Baumindex^.B:=Nil;

    loesche_UD;
    UD:=-2;
    UDBi:=Baumindex;
    UDWurzel^.B:=Wurzel^.B;
    if assigned(wurzel) then dispose(wurzel);
    wurzel:=Nil;

    Baumindex:=Nil;
    klettere(wurzelindex, true);
    {wenn wurzelindex^.B Nil, dann wurde vom letzten Zug an geschnitten u. Wurzel mu Nil bleiben}
    if assigned(wurzelindex^.B) then wurzel:=wurzelindex
    else dispose(wurzelindex);
    Result := true;
  end;
end;

procedure TNotation.aufwerten;
var rb, rb1 : PBaum;
    rBaumindex : PBaum;
    rZug : TZug;
    s : string;
begin
  if (inHauptvariante) or (not assigned(Baumindex)) then exit;

  if corrChessMode then
    if IsVarErsterOrdnung then exit;

  geaendert:=true;
  UD:=0;
  rBaumindex:=Baumindex;
  rb1:=Baumindex;
  if assigned(Baumindex^.rB) then Baumindex:=Baumindex^.rB;  {es knnte Var. losgehen - Varwurzel bereits <> NIL}
  while (assigned(Baumindex^.rB)) and
  (not ((assigned(Baumindex^.varwurzel)) and (rBaumindex^.Zug.amZug=Baumindex^.Zug.amZug))) do begin
    rBaumindex:=Baumindex;
    Baumindex:=Baumindex^.rB; {zur Wurzel zurck}
  end;
  rb:=Baumindex^.B;
  Baumindex^.B:=rBaumindex^.B;
  rBaumindex^.B:=rb;
  if assigned(rBaumindex^.B) then rBaumindex^.B^.rB:=rBaumindex;
  if assigned(Baumindex^.B) then Baumindex^.B^.rB:=Baumindex;
  rzug:=Baumindex^.zug;
  s := Baumindex^.Kommentar;
  Baumindex^.zug:=rBaumindex^.zug;
  Baumindex^.Kommentar:=rBaumindex^.Kommentar;
  rBaumindex^.zug:=rzug;
  rBaumindex^.Kommentar:=s;
  if rb1<>rBaumindex
    then Baumindex:=rb1;  {nur Tauschen, wenn nicht erster Variantenzug markiert war}
  klettere(wurzel, true);
end;

procedure TNotation.loeschen;
var vi : PBaumlist;
    rBaumindex : PBaum;
begin
  if (inHauptvariante) or (not assigned(Baumindex)) then exit;
  geaendert:=true;
  rBaumindex:=Baumindex;
  if assigned(Baumindex^.rB) then Baumindex:=Baumindex^.rB;  {es knnte Var. losgehen - Varwurzel bereits <> NIL}
  while (assigned(Baumindex^.rB)) and
  (not ((assigned(Baumindex^.varwurzel)) and (rBaumindex^.Zug.amZug=Baumindex^.Zug.amZug))) do begin
    rBaumindex:=Baumindex;
    Baumindex:=Baumindex^.rB; {zur Wurzel zurck}
  end;
  {!!!}
  if not assigned(Baumindex^.rB) then vi:=wurzel^.B^.Varwurzel
  else vi:=Baumindex^.varwurzel;

  loesche_UD;   {ev. vorherige undo-Var. lschen}
  UD:=1;
  while (vi^.B<>rBaumindex) and (assigned(vi^.Bl)) do begin
    vi:=vi^.Bl;
    inc(UD);
  end;
  UDBi:=Baumindex;
  UDWurzel^.B:=vi^.B;  {Variante auskoppeln und an UDWurzel bergeben}
  vi^.B:=Nil;

  if vi=Baumindex^.varwurzel then  ZBaumlist:=addr(Baumindex^.varwurzel)
  else ZBaumlist:=addr(vi^.rBl^.Bl);
  loescheBaum(ZBaumlist);
  Baumindex:=Baumindex^.rB;
  klettere(wurzel, true);
  if assigned(FOnMoveChange) then FOnMoveChange(self, NoZug^, NoStellung^, false);
end;

procedure TNotation.UnDo;
var i : integer;
    wi, rwi : PBaumlist;                     {  wurzel^.b^.b^.varwurzel^.bl^}
begin
  if (not assigned(UDWurzel)) or (UD=0) then exit;
  if UD>0 then begin
    i:=1;
    wi:=UDBi^.VarWurzel;
    rwi:=wi;
    while (i<UD) and (assigned(wi)) do begin
      rwi:=wi;
      wi:=wi^.BL;
      inc(i);
    end;
    if not (assigned(wi)) then begin  {letzter Eintrag}
      if wi=UDBi^.VarWurzel then UDBi^.VarWurzel:=UDWurzel
      else begin
        rwi^.BL:=UDWurzel;
        UDWurzel^.rBL:=rwi;
      end;
    end
    else begin
      if wi=UDBi^.VarWurzel then begin
        UDBi^.VarWurzel^.rBL:=UDWurzel;
        UDWurzel^.BL:=UDBi^.VarWurzel;
        UDBi^.VarWurzel:=UDWurzel;
      end
      else begin
        UDWurzel^.BL:=wi;
        wi^.rBL:=UDWurzel;
        UDWurzel^.rBL:=rwi;
        rwi^.BL:=UDWurzel;
      end;
    end;
  end
  else begin
    if UD=-1 then begin

      if corrChessMode then
        if UDBi^.Zug.Attr1 and nWarVorherLetzter = 0 then
          UDBi^.Zug.Attr1:=(UDBi^.Zug.Attr1 or nLetzter) xor nLetzter;  

      UDBi^.B:=UDWurzel^.B;
      dispose(UDWurzel);
    end
    else begin
      if not assigned(wurzel) then UDBi^.B:=Nil
      else begin
        UDBi^.B:=wurzel^.B;
        if assigned(wurzel^.B) then wurzel^.B^.rB:=UDBi;
      end;
      Partiedaten:=UDPartiedaten;
      if not assigned(Baumindex) then
        if assigned(FOnMoveChange) then FOnMoveChange(self, Partiedaten.Zug, Partiedaten.Stellung, false);
      Wurzel:=UDWurzel;
    end;
  end;
  UDWurzel:=Nil;
  UD:=0;
  klettere(wurzel, true);
end;

procedure TNotation.Matt;
var k : longint;
begin
 try                 (**)
   if assigned(Baumindex) then begin
     k:=Baumindex^.ss + Baumindex^.sl - 1;
     if notat[k]='+' then begin
       notat[k]:='#';
       Baumindex^.Zug.attr1:=Baumindex^.Zug.attr1 or nMatt xor nSchach;
       Klettere(wurzel, true);
     end;
  end;
 except
 end;
end;

procedure TNotation.KD(Sender: TObject; var Key: Word;  Shift: TShiftState);
begin
  case key of
    27, VK_LEFT : uebergabe:=-1;
    VK_RIGHT, 13, VK_SPACE :  uebergabe:=lZuege.ItemIndex;
  else exit;
  end;
  VarDialog.Close;
end;

procedure TNotation.MD(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  uebergabe:=lZuege.ItemIndex;
  VarDialog.Close;
end;

function TNotation.GoBack : Boolean;
Var rBaum : PBaum;
begin
  Result := false;
  if not assigned(Baumindex) then exit;
  rBaum:=Baumindex;
  Baumindex:=Baumindex^.rB;
  if (assigned(Baumindex)) and (rBaum<>Baumindex^.B) {springt in bergeordnete Var. und mu deshalb noch einen Zug zurck}
    then Baumindex:=Baumindex^.rB;
  zugzurueck(rBaum^.Zug);
  if assigned(Baumindex) then NoZug^ := Baumindex^.Zug
  else NoZug^ := Partiedaten.Zug;
  AdjustPb(Baumindex);
  if assigned(FOnMoveChange) then FOnMoveChange(self, NoZug^, NoStellung^, false);
  Result := true;
end;

function TNotation.Goahead_inHauptVariante_and_showMove(var SolvingComplete : Boolean; var NextHiddenMove : PBaum;
  Var von, nach, UFigur, RochadeArt : shortint) : Boolean;
begin
  Result := Goahead_inHauptVariante(von, nach, UFigur, RochadeArt);
  if Result and (Baumindex <> Nil) then begin
    Baumindex^.hidden := false;
  end;
  klettereEx;
  SolvingComplete := allSolved;
  NextHiddenMove := nextToSolve;
end;

function TNotation.Goahead_inHauptVariante(Var von, nach, UFigur, RochadeArt : shortint) : Boolean; {lst kein OnmoveChangeEreignis aus, gibt aber den gespielten Zug zurck}
Var hBaumindex : PBaum;
begin
  Result := false;
  if not assigned(Baumindex) then begin
    if (not assigned(wurzel)) or (not assigned(wurzel^.B)) then exit;
    hBaumindex:=Wurzel^.B;
  end
  else hBaumindex:=Baumindex^.B;
  if assigned(hBaumindex) then begin
    Baumindex := hBaumindex;
    NoZug^ := Baumindex^.Zug;
    zugvor;
    AdjustPb(Baumindex);
    von := Baumindex^.Zug.von;
    nach := Baumindex^.Zug.nach;
    UFigur := Baumindex^.Zug.UFigur;
    RochadeArt := Baumindex^.Zug.RochadeArt;
    Result := true;
  end;
end;

function TNotation.Goahead : Boolean;
Var i : longint;
    keineZuege : Boolean;
    hBaumindex : PBaum;
begin
  keineZuege:=false;
  Result := false;
  if not assigned(Baumindex) then begin
    if (not assigned(wurzel)) or (not assigned(wurzel^.B)) then exit;
    Baumindex:=Wurzel^.B;
    keineZuege:=true;
  end;
  if not keineZuege then hBaumindex:=Baumindex^.B else hBaumindex:=Baumindex;
  if assigned(hBaumindex) then begin
    uebergabe:=-2;
    if assigned(hBaumindex^.Varwurzel) then begin
      uebergabe:=-1;
      Dialog_ini;
      VarDialog.KeyPreview:=true;
      VarDialog.OnKeydown:=KD;
      if English then VarDialog.Caption:='Select move' else VarDialog.Caption:='Variante whlen';
      lZuege:=TListBox.Create(self);
      VarDialog.insertControl(lZuege);
      lZuege.OnMouseDown:=MD;
      lZuege.Align:=alClient;
      Varianten_anzeigen(keineZuege);
      Vardialog.showmodal;
      lzuege.free;
      VarDialog.free;
      if uebergabe>=0 then begin
        wurzelindex:=hBaumindex^.Varwurzel;
        if uebergabe=0 then begin
          if not keineZuege then Baumindex:=Baumindex^.B;  {Hauptvariante}
      //    inHauptvariante:=true;
        end
        else begin
          inHauptvariante:=false;
          if uebergabe=1 then begin  {VarWurzel - nur eine Verzweigung}
            Baumindex:=Wurzelindex^.B;
          end
          else begin
            i:=1;
            while (i<uebergabe) and (assigned(wurzelindex^.BL)) do begin
              wurzelindex:=wurzelindex^.BL;
              inc(i);
            end;
            Baumindex:=Wurzelindex^.B;
          end;
        end;
      end
      else if keineZuege then Baumindex:=Nil;  {Abbruch gewhlt}
    end
    else if not keineZuege then Baumindex:=Baumindex^.B;  {keine Verzweigungen}
    if uebergabe <> -1 then begin {Abbruch im Var.dialog gewhlt}
      NoZug^ := Baumindex^.Zug;
      zugvor;
      AdjustPb(Baumindex);
      if assigned(FOnMoveChange) then FOnMoveChange(self, NoZug^, NoStellung^, true);
      Result := true;
    end;
  end
  else begin  {Bei Variantenende auf bergeordnete Var. dahinter (falls vorhanden) springen}
    i:=Baumindex^.ss + 1 + Baumindex^.sl;
    while (i<length(notat)) and (notat[i]<>'.') do begin
      inc(i);
      if notat[i]='{' then
        repeat inc(i) until (i>=length(notat)) or (notat[i]='}');
    end;
    if (notat[i]<>'.') or (i>=length(notat)) then exit;
    retract(i);
    MessageBeep(0);
    if assigned(FOnMoveChange) then FOnMoveChange(self, NoZug^, NoStellung^, false);
  end;
end;

function TNotation.GoPreviousVar : Boolean;
begin
  Result := false;
  if not assigned(Baumindex) then exit;
  if assigned(Baumindex^.rB) then Baumindex:=Baumindex^.rB;  {es knnte Var. losgehen - Varwurzel bereits <> NIL}
  while (assigned(Baumindex^.rB)) and
  (not (assigned(Baumindex^.varwurzel))) do begin
    Baumindex:=Baumindex^.rB; {zur Wurzel zurck}
  end;
  Baumindex:=Baumindex^.rB;
  klettere(wurzel, false);
  if assigned(FOnMoveChange) then FOnMoveChange(self, NoZug^, NoStellung^, false);
  Result := true;
end;

function TNotation.GoNextVar : Boolean;
var rBaumindex : PBaum;
begin
  Result := false;
  if (not assigned(wurzel)) or (not assigned(wurzel^.B)) then exit;
  rBaumindex:=Baumindex;
  if assigned(Baumindex) then Baumindex:=Baumindex^.B else Baumindex:=Wurzel^.B;
  if not assigned(Baumindex) then begin
    Baumindex:=rBaumindex;
    exit;
  end;
  while (assigned(Baumindex^.B)) and (not assigned(Baumindex^.B^.VarWurzel)) do
    Baumindex:=Baumindex^.B;
  klettere(wurzel, false);
  if assigned(FOnMoveChange) then FOnMoveChange(self, NoZug^, NoStellung^, false);
  Result := true;
end;

function TNotation.GoToStart : Boolean;
begin
  Result := false;
  if (not assigned(wurzel)) or (not assigned(Baumindex)) or
  (not assigned(wurzel^.B)) then exit;
  Baumindex:=Nil;
  klettere(wurzel, false);
  if assigned(FOnMoveChange) then FOnMoveChange(self, NoZug^, NoStellung^, false);
  Result := true;
end;

function TNotation.GoToEnd : Boolean;
begin
  Result := false;
  if (not assigned(wurzel)) or (not assigned(wurzel^.B)) then exit;
  Baumindex:=wurzel^.B;
  while assigned(Baumindex^.B) do Baumindex:=Baumindex^.B;
  klettere(wurzel, false);
  if assigned(FOnMoveChange) then FOnMoveChange(self, NoZug^, NoStellung^, false);
  Result := true;

  if (corrChessMode) and (Baumindex <> Nil) and (Baumindex^.VarWurzel <> Nil) then
    ScrollToEnd;
end;

procedure TNotation.ScrollToEnd;
begin
  SLeiste.Position := SLeiste.Max;
end;

function TNotation.isLastPos : Boolean;
begin
   if (not assigned(wurzel)) or (not assigned(wurzel^.B)) then begin
     Result := true;
     exit;
   end;
   Result := (assigned(Baumindex)) and (not assigned(Baumindex^.B));
end;

procedure TNotation.setCheckPosBoard(stllg : TSchachbrett);
var i : Integer;
begin
  for i := 0 to 119 do
    checkPosBoard[i] := stllg[i];
end;

function TNotation.containsPos(apf : TPiecesFilter; chckpos : Boolean) : Boolean;
begin
  Result := false;
  checkPos := true;
  pf := apf;
  try
    checkStellung := chckpos;
    klettere(wurzel, false);
    Result := not checkPos;
  finally
    checkPos := false;
    pf := Nil;
  end;
end;

function TNotation.SchreibePartie(ToClipBoard : Boolean; Var PGNString : string; onlyHeader : Boolean) : integer;
var h, h1 : string;
    ri, i, n : longint;
    gleich : Boolean;
    cs : PChar;
    sl : TStringlist;

procedure cwriteln(s : string);
begin
  sl.Add(s)
end;

procedure split(gh, hs : string);
Var i : integer;
begin
  i := 1;
  while i <= length(hs) do begin
    cwriteln(gh + copy(hs, i, 150) + '"]');
    inc(i, 150);
  end;
end;

procedure DoInfInsert(iinf : integer);
Var sinf : string;
begin
  if length(h) > i then begin
    insert(' ', h, i);
    inc(i);
    h[i] := '$';
    inc(i);
    sinf := IntToStr(iinf);
    insert(sinf, h, i);
    inc(i, length(sinf) - 1);
  end;
end;

function infCharsDefined : Boolean;
begin
  Result :=
     (defaultinfCharzugzwang       <>   infCharzugzwang             )
  or (defaultinfChareinzigerzug    <>   infChareinzigerzug          )
  or (defaultinfCharplusminus      <>   infCharplusminus            )
  or (defaultinfCharplusleichtminus<>   infCharplusleichtminus      )
  or (defaultinfCharplusgleich     <>   infCharplusgleich           )
  or (defaultinfChargleich         <>   infChargleich               )
  or (defaultinfCharunklar         <>   infCharunklar               )
  or (defaultinfChargleichplus     <>   infChargleichplus           )
  or (defaultinfCharleichtminusplus<>   infCharleichtminusplus      )
  or (defaultinfCharminusplus      <>   infCharminusplus            )
  or (defaultinfCharkompensation   <>   infCharkompensation         )
  or (defaultinfChargegenspiel     <>   infChargegenspiel           )
  or (defaultinfCharinitiative     <>   infCharinitiative           )
  or (defaultinfCharangriff        <>   infCharangriff              )
  or (defaultinfCharzeitnot        <>   infCharzeitnot              )
  or (defaultinfCharvorsprung      <>   infCharvorsprung            )
  or (defaultinfCharneuerung       <>   infCharneuerung             )
  or (defaultinfCharmitderidee     <>   infCharmitderidee           )
  or (defaultinfCharBesserIst      <>   infCharBesserIst            );
end;

begin
 Result:=-1;
 sl := TStringlist.Create;
 try
  if (not onlyHeader) {and   saving must always be set to true, except for header export
  ((FUserDefinedPieces) or (infCharsDefined))} then begin
    saving := true;
    Klettere(wurzel, true);
  end;

  with Partiedaten.EventData do begin
    if quelle<>'' then h:=quelle else h:='?';
    cwriteln('[Event "'+h+'"]');
    if ort<>'' then h:=ort else h:='?';
    cwriteln('[Site "'+h+'"]');
    if jahr<>0 then h:=IntToStr(jahr) + '.' else h:='????.';
    if Monat<>0 then begin
      if monat < 10 then h := h + '0';
      h:=h + IntToStr(Monat) + '.';
    end
    else h:=h+'??.';
    if Tag<>0 then begin
      if tag < 10 then h := h + '0';
      h:=h + IntToStr(Tag);
    end
    else h:=h+'??';
    //if h <> '????.??.??' then
    cwriteln('[Date "'+h+'"]');
    if runde <> 0 then cwriteln('[Round "'+IntToStr(Runde)+'"]')
    else cwriteln('[Round "-"]');
    if weiss<>'' then h:=weiss else h:='?';
    cwriteln('[White "'+h+'"]');
    if schwarz<>'' then h:=schwarz else h:='?';
    cwriteln('[Black "'+h+'"]');
    cwriteln('[Result "'+ergebnis+'"]');

    if annotator<>'' then cwriteln('[Annotator "'+annotator+'"]');

    if sElo<>0 then cwriteln('[BlackElo "'+IntToStr(sElo)+'"]');
    if eco <> '' then cwriteln('[ECO "'+eco+'"]');

    if EventDatejahr<>0 then h:=IntToStr(EventDatejahr) + '.' else h:='????.';
    if EventDateMonat<>0 then begin
      if EventDatemonat < 10 then h := h + '0';
      h:=h + IntToStr(EventDateMonat) + '.';
    end
    else h:=h+'??.';
    if EventDateTag<>0 then begin
      if tag < 10 then h := h + '0';
      h:=h + IntToStr(EventDateTag);
    end
    else h:=h+'??';
    if h <> '????.??.??' then cwriteln('[EventDate "'+h+'"]');

    gleich:=true;
    for i:=0 to 119 do if Partiedaten.Stellung[i]<>Ausgangsstellung[i] then gleich:=false;
    if gleich then if Partiedaten.Zug.amZug <> 1 then gleich := false;  (**)
    if not gleich then begin
      cwriteln('[FEN "'+ PosToFEN(Partiedaten.Stellung, Partiedaten.Zug, Partiedaten.Zugnummer, FRC) +'"]');
      cwriteln('[SetUp "1"]');
    end;

    if TimeControlGameSec <> 0 then
      cwriteln('[TimeControl "'+IntToStr(TimeControlGameSec) + '+' + IntToStr(TimeControlMoveSec) + '"]');

    if wElo<>0 then cwriteln('[WhiteElo "'+IntToStr(wElo)+'"]');

    if MetaKeywords <> '' then split('[MetaKeywords "', MetaKeywords);
    if MetaDescription <> '' then split('[MetaDescription "', MetaDescription);

    if corrChessMode or corrChessMode1 then begin
      if WhiteClubID <> '' then
        cwriteln('[WhiteClubID "'+WhiteClubID+'"]');
      if BlackClubID <> '' then
        cwriteln('[BlackClubID "'+BlackClubID+'"]');
      if GameIdValue <> '' then
        cwriteln('[' + GameId + ' "'+ GameIdValue +'"]');
    end;
    if corrChessMode then begin
      if WhitedifferenceToGMTValue <> '' then
        cwriteln('[' + WhitedifferenceToGMT + ' "'+WhitedifferenceToGMTValue+'"]');
      if BlackdifferenceToGMTValue <> '' then
        cwriteln('[' + BlackdifferenceToGMT + ' "'+BlackdifferenceToGMTValue+'"]');
      if DateTimeFormatValue <> ''then
        cwriteln('[' + DateTimeFormat + ' "'+DateTimeFormatValue+'"]');
    end;

    For i := 0 to slRestData.Count - 1 do begin
      cwriteln(slRestData[i]);
    end;
  end;

  if onlyHeader then begin
    PGNString := sl.Text;
    exit;
  end;

  cwriteln('');
  h:=notat;
  if ueberschrift then delete(h, 1, UeberschriftEnde);      (**)
  i:=1;
  n:=0;
  while i<=length(h) do begin
    case h[i] of

      '>' : begin
              ri:=i;
              dec(i); // wird am ende increm.
              delete(h, ri, 1);
              while (ri<=length(h)) and (h[ri]<>' ') and (h[ri]<>'<') do inc(ri);
              insert(' '+ PGNErster, h, ri);
            end;
      '<' : begin
              delete(h, i, 1);
              insert(' ' + PGNLetzter, h, i);
            end;


      '[' : h[i]:='(';
      ']' : h[i]:=')';
      'D' : h[i]:='Q';
      'T' : h[i]:='R';
      'L' : h[i]:='B';
      'S' : h[i]:='N';
      #$D : h[i]:=' ';
      #$A : begin
              while (i<=length(h)) and ((h[i]=' ') or (h[i]=#$A)) do delete(h, i, 1);
              dec(i);
            end;
      ' ' : if n>70 then begin
              h[i]:=#1;
              n:=0;
            end;
      '{' : while (i<=length(h)) and (h[i]<>'}') do begin
              case h[i] of
                #$D : h[i]:=' ';
                #$A : begin
                        while (i<=length(h)) and ((h[i]=' ') or (h[i]=#$A)) do delete(h, i, 1);
                        dec(i);
                      end;
                ' ' : begin
                        while (i<length(h)) and (h[i+1]=' ') do delete(h, i+1, 1);
                        if (n>60) and
                        ((i<length(h)) and (h[i+1]<>'[')) then begin
                        { <>'[' - damit nicht mal zufallig ein "[Event" am Zeilenanfang steht}
                          h[i]:=#1;
                          n:=0;
                        end;
                      end;
              end;
              inc(i);
              inc(n);
            end;
      '!' : if length(h) > i then begin
              insert(' ', h, i);
              inc(i);
              h[i] := '$';
              inc(i);
              case h[i] of
                '!' : h[i] := '3';
                '?' : h[i] := '5';
                else insert('1', h, i);
              end;
            end;
      '?' : if length(h) > i then begin
              insert(' ', h, i);
              inc(i);
              h[i] := '$';
              inc(i);
              case h[i] of
                '?' : h[i] := '4';
                '!' : h[i] := '6';
                else insert('2', h, i);
              end;
            end;
      else begin
        if h[i]      = defaultinfCharzugzwang               then  DoInfInsert(     infzugzwang            )
        else if h[i] = defaultinfChareinzigerzug       then  DoInfInsert(     infeinzigerzug         )
        else if h[i] = defaultinfCharplusminus         then  DoInfInsert(     infplusminus           )
        else if h[i] = defaultinfCharplusleichtminus   then  DoInfInsert(     infplusleichtminus     )
        else if h[i] = defaultinfCharplusgleich        then  DoInfInsert(     infplusgleich          )
        else if h[i] = defaultinfChargleich            then  DoInfInsert(     infgleich              )
        else if h[i] = defaultinfCharunklar            then  DoInfInsert(     infunklar              )
        else if h[i] = defaultinfChargleichplus        then  DoInfInsert(     infgleichplus          )
        else if h[i] = defaultinfCharleichtminusplus   then  DoInfInsert(     infleichtminusplus     )
        else if h[i] = defaultinfCharminusplus         then  DoInfInsert(     infminusplus           )
        else if h[i] = defaultinfCharkompensation      then  DoInfInsert(     infkompensation        )
        else if h[i] = defaultinfChargegenspiel        then  DoInfInsert(     infgegenspiel          )
        else if h[i] = defaultinfCharinitiative        then  DoInfInsert(     infinitiative          )
        else if h[i] = defaultinfCharangriff           then  DoInfInsert(     infangriff             )
        else if h[i] = defaultinfCharzeitnot           then  DoInfInsert(     infzeitnot             )
        else if h[i] = defaultinfCharvorsprung         then  DoInfInsert(     infvorsprung           )
        else if h[i] = defaultinfCharneuerung          then  DoInfInsert(     infneuerung            )
        else if h[i] = defaultinfCharmitderidee        then  DoInfInsert(     infmitderidee          )
        else if h[i] = defaultinfCharBesserIst         then  DoInfInsert(     infBesserIst           );
      end;
    end;
    inc(n);
    inc(i);
  end;
  h1:='';
  for i:=1 to length(h) do begin
    if h[i]<>#1 then h1:=h1+h[i]
    else begin
      cwriteln(h1);
      h1:='';
    end;
  end;
  if h1<>'' then cwriteln(h1);
  if ToClipBoard then begin
    with ClipBoard do begin
      open;
      try
        cs := stralloc(length(sl.Text) + 1);
        try
          StrCopy(cs, PChar(sl.Text));
          SetTextBuf(cs);
        finally
          StrDispose(cs);
        end;
      finally
        close;
      end;
    end;
  end
  else PGNString := sl.Text;
  Result:=0;
 finally
   if (not onlyHeader) {and
   ((FUserDefinedPieces) or (infCharsDefined)) }then begin
     saving := false;
     Klettere(wurzel, true);
   end;
   sl.free;
   saving := false;
 end;
end;

procedure TNotation.CopyGameToClipboard;
Var dummy : string;
begin
  schreibePartie(true, dummy, false);
end;

function TNotation.GetGame : string;
begin
  Result := '';
  schreibePartie(false, Result, false);
end;

function TNotation.GetHeader : string;
begin
  Result := '';
  schreibePartie(false, Result, true);
end;

function TNotation.PasteGameFromClipBoard;
begin
  Result := false;
  if not ClipBoard.HasFormat(CF_TEXT) then exit;
  with Clipboard do begin
    open;
    try
      SetGame(Clipboard.AsText);
    finally
      close;
    end;
  end;
  Result := true;
end;

//Fernschach
function TNotation.AreMovesMarked : integer;
Var b : PBaum;
begin
  Result:=4;  //!!!l347
  if not assigned(Baumindex) then begin
    if (not assigned(wurzel)) or (not assigned(wurzel^.B)) then exit
    else b:=wurzel^.B;
  end
  else b:=Baumindex;
  if (assigned(b^.B)) and (b^.B^.Zug.attr1 and nErster <> 0) then begin  // steht unmittelbar vor nErster, dann darf keine Var. eingegeben werden
    Result:=3;
    exit;
  end;
  while assigned(b) do begin
    if b^.Zug.attr1 and nLetzter <> 0 then begin
      Result:=2;
      if Baumindex=b then Result:=0;  // nach letzten EvZug kann natrlich eingegeben werden
      if (assigned(b^.B)) and (Baumindex=b^.B) then Result:=4; // es darf nur ein Antwortzug eingegeben werde - Ev.Varianten mssen als var. eingegeben werden
      exit;
    end;
    b:=b^.B;
    if (assigned(b)) and (b^.Zug.attr1 and nErster <> 0) then begin
      Result:=1;
      exit;  // dann stand b vor dem ersten EvZug (nicht unmittelbar!) - Variante einfgen zulassen
    end;
  end;
end;

function TNotation.IsVarErsterOrdnung : Boolean;
Var b, rb, Muendung : PBaum;
    i : integer;
begin
  Result:=False;
  b:=Baumindex;
  rb:=b;
  i:=0;   //!!! funktioniert noch nicht, wenn Variante im 1. Zug losgeht
  if assigned(b^.rB) then b:=b^.rB;  {es knnte Var. losgehen - Varwurzel bereits <> NIL}
  while (assigned(b^.rB)) and (i<2) do begin        {zur Hauptwurzel zurck}
    if (assigned(b^.varwurzel)) and (rb^.Zug.amZug=b^.Zug.amZug) then begin
      inc(i);  // wenn Zugrecht gleich ist mndet es in bergeordnete Var.
      Muendung:=b;
    end;  
    rb:=b;
    b:=b^.rB;
  end;
  if i=1 then begin  // war Var. 1. Ord. --> prfen, ob Einmndung vor nErster war
    b:=Muendung;
    if b^.Zug.attr1 and nErster = 0 then begin  // wenn Mndung in nErster erfolgte, kann aufgewertet werden
      while (assigned(b)) and (Result=False) do begin
        if b^.Zug.attr1 and nErster <> 0 then Result:=True;
        b:=b^.B;
      end;
    end;
  end;
end;
//

function  TNotation.GetFirstPosInfo : PZug;
begin
  Result := @FPartiedaten.Zug;
end;

function  TNotation.getBaumindex : Pbaum;
begin
  Result := Baumindex;
end;

function  TNotation.GetMove : PZug;
begin
  Result := Nil;
  if Baumindex <> Nil then
    Result := @Baumindex^.Zug;
end;

function  TNotation.GetPosition : PSchachbrett;
begin
  Result := NoStellung;
end;

function  TNotation.Empty : Boolean;
begin
  if (wurzel = Nil) or (wurzel^.B = Nil)
  then Result := true
  else Result := false;
end;

function  TNotation.PlyCount : Integer;
var b: PBaum;
begin
  Result := 0;
  if wurzel = Nil then  exit;
  b := Wurzel^.B;
  while b <> Nil do begin
    inc(Result);
    b := b^.B;
  end;
end;

function TNotation.GetKommentar : string;
Var i : longint;
begin
  if assigned(Baumindex) then Result:=Baumindex^.Kommentar
  else Result:=Partiedaten.Kommentar;
  if Result='' then exit;
  i:=0;
  repeat inc(i) until (Result[i]='{') or (i>=length(Result));
  delete(Result, i, 1);
  i:=length(Result)+1;
  repeat dec(i) until (i<=0) or (Result[i]='}');
  delete(Result, i, 1);
end;

function TNotation.getHeaderDate : TDate;
var year, month, day : Word;
begin
  Result := 0;
  with Partiedaten.EventData do begin
    if (jahr = 0) or (Monat = 0) or (Tag = 0) then
      exit;
    Result := EncodeDate(Jahr, Monat, Tag);
  end;
end;

Procedure TNotation.SetKommentar(k : String);
var i : longint;
begin
  trim(k);
  if k<>'' then begin
    i:=0;
    repeat
      inc(i);
      case k[i] of
        '}' : begin delete(k, i, 1);dec(i);end;
        '{' : begin delete(k, i, 1);dec(i);end;
    (*    '' : begin delete(k, i, 1); insert('Ae', k, i); end;
        '' : begin delete(k, i, 1); insert('Oe', k, i); end;
        '' : begin delete(k, i, 1); insert('Ue', k, i); end;
        '' : begin delete(k, i, 1); insert('ae', k, i); end;
        '' : begin delete(k, i, 1); insert('oe', k, i); end;
        '' : begin delete(k, i, 1); insert('ue', k, i); end;
        '' : begin delete(k, i, 1); insert('ss', k, i); end;  *)
        #$D : begin delete(k, i, 2); insert(' ', k, i); end;   {Zeilensprung wrde Formatierungsroutine stren}
      end;
    until i>=length(k);
  end;
  if k<>'' then k:='{'+k+'}';
  if assigned(Baumindex) then Baumindex^.Kommentar:=k
  else FPartiedaten.Kommentar:=k;
  if not assigned(wurzel) then begin
    New(wurzel);
    wurzel^.rBL:=Nil;
    wurzel^.BL:=Nil;
    wurzel^.B:=Nil;
  end;
  klettere(wurzel, true);
  geaendert:=true;
end;

procedure TNotation.SetZugBewertung(s : string);
begin
  if not assigned(Baumindex) then exit;
  if s = '!' then Baumindex^.Zug.Attr2 := nAusruf
  else if s = '!!' then Baumindex^.Zug.Attr2 := n2Ausruf
  else if s = '?' then Baumindex^.Zug.Attr2 := nFrage
  else if s = '??' then Baumindex^.Zug.Attr2 := n2Frage
  else if s = '!?' then Baumindex^.Zug.Attr2 := nAusrufFrage
  else if s = '?!' then Baumindex^.Zug.Attr2 := nFrageAusruf
  else Baumindex^.Zug.Attr2 := 0;
  klettere(wurzel, true);
  geaendert:=true;
end;

function  TNotation.getRoot : PBaumList;
begin
  Result := Wurzel;
end;

function  TNotation.ImportNextMove(OtherNotationRoot : PBaumList) : TZug;  {vergleicht OtherNottaion mit eigener und gibt nchsten Zug zurck, falls eigene mind. einen weiteren enthlt}
Var b, b1 : PBaum;
begin
  if OtherNotationRoot = Nil then
    b := Nil
  else
    b := OtherNotationRoot.B;
  if Wurzel = Nil then
    b1 := Nil
  else
    b1 := Wurzel.B;
  while (assigned(b)) and (assigned(b1)) do begin
    Result := b1^.Zug;
    if (b^.Zug.von <> b1^.Zug.von) or (b^.Zug.nach <> b1^.Zug.nach) then
      exit;
    b:=b^.B;
    b1:=b1^.B;
  end;
  if assigned(b1) then
    Result := b1^.Zug
  else
    Result.von := 0;  
end;

procedure TNotation.showNextMoves;
begin
  if Baumindex = Nil then begin
    showAllMoves;
    exit;
  end;
  showLastMoves := true;
  try
    klettere(wurzel, true);
  finally
    showLastMoves := false;
  end;
end;

procedure TNotation.hideNextMoves;
begin
  if Baumindex = Nil then begin
    hideAllMoves;
    exit;
  end;
  hideLastMoves := true;
  try
    klettere(wurzel, true);
  finally
    hideLastMoves := false;
  end;
end;

procedure TNotation.hideAllMoves;
begin
  hideMoves := true;
  try
    Baumindex := Nil;
    klettere(wurzel, true);
    if assigned(FOnMoveChange) then FOnMoveChange(self, NoZug^, NoStellung^, false)
  finally
    hideMoves := false;
  end;
end;

procedure TNotation.showAllMoves;
begin
  showMoves := true;
  try
    klettere(wurzel, true);
  finally
    showMoves := false;
  end;
end;

{Property-Routinen}

procedure TNotation.SetUberschrift(w : Boolean);
begin
  if w<>FUeberschrift then FUeberschrift:=w;
  klettere(wurzel, true);
end;

procedure TNotation.SetGeaendert(w : Boolean);
begin
  FGeaendert := w;
  if w then if assigned(OnGameChanged) then OnGameChanged(self);
end;

procedure TNotation.SetFontHeadLine(f : TFont);
begin
  FFontHeadLine.assign(f);
  klettere(wurzel, true);
end;

procedure TNotation.SetFontComment(f : TFont);
Var i : integer;
begin
  FFontComment.assign(f);
  i := FFont0Var.Size;
  if i < FFont1Var.Size then i := FFont1Var.Size;
  if i < FFont2Var.Size then i := FFont2Var.Size;
  if FFontComment.Size > i then FFontComment.Size := i;
  klettere(wurzel, true);
end;

procedure TNotation.SetFont0Var(f : TFont);
begin
  pb.canvas.Font.assign(f); (**)
  FFont0Var.assign(f);
  FFont1Var.Name := f.Name;
  FFont1Var.Size := f.Size;
  FFont2Var.Name := f.Name;
  FFont2Var.Size := f.Size;
  if FFontComment.Size > FFont0Var.Size then FFontComment.Size := FFont0Var.Size;
  Texthoehe := pb.canvas.TextHeight('Mg') + ZeilenAbstand;
  SLeiste.LargeChange := (SLeiste.height div Texthoehe) - 1;
  klettere(wurzel, true);
end;

procedure TNotation.SetFont1Var(f : TFont);
begin
  FFont1Var.Color := f.Color;
  FFont1Var.Style := f.Style;
  klettere(wurzel, true);
end;

procedure TNotation.SetFont2Var(f : TFont);
begin
  FFont2Var.Color := f.Color;
  FFont2Var.Style := f.Style;
  klettere(wurzel, true);
end;

procedure TNotation.SetIndent0Var(i : word);
begin
  if FIndent0Var = i then exit;
  FIndent0Var := i;
  klettere(wurzel, true);
end;

procedure TNotation.SetIndent1Var(i : word);
begin
  if FIndent1Var = i then exit;
  FIndent1Var := i;
  klettere(wurzel, true);
end;

procedure TNotation.SetIndent2Var(i : word);
begin
  if FIndent2Var = i then exit;
  FIndent2Var := i;
  klettere(wurzel, true);
end;

procedure TNotation.SetZeilenabstand(i : word);
begin
  if FZeilenabstand = i then exit;
  FZeilenabstand := i;
  with TBitmap.Create do begin
    try
      Canvas.Font.Assign(FFont0Var);
      Texthoehe := canvas.TextHeight('Mg') + FZeilenAbstand;
      SLeiste.LargeChange := (SLeiste.height div Texthoehe) - 1;
      klettere(wurzel, true);
    finally
      free;
    end;
  end;    
end;

procedure TNotation.SetBorderWidthNotat(i : word);
begin
  Borderwidth := i;
  FBorderWidthNotat := i;
  SendMessage(handle, WM_SIZE, 0, 0);
end;

procedure TNotation.SetBevelWidthNotat(i : word);
begin
  if i < 1 then exit;
  Bevelwidth := i;
  FBevelWidthNotat := i;
  SendMessage(handle, WM_SIZE, 0, 0);
end;

function TNotation.GetEnglish : Boolean;
begin
  if FEnglish=0 then Result:=false else Result:=true;
end;

procedure TNotation.SetEnglish(w : Boolean);
begin
  if ((w) and (FEnglish=1)) or ((not w) and (FEnglish=0)) then exit;
  if w then FEnglish:=1 else FEnglish:=0;
  Klettere(wurzel, true);
end;

procedure TNotation.SetInformantSigns(InformantSigns : String);
begin
  if length(InformantSigns) >  0 then if (InformantSigns[1] <> '#' ) and (ord(InformantSigns[1]) > 127) then infCharzugzwang       := InformantSigns[1]  else infCharzugzwang        := defaultinfCharzugzwang       ;
  if length(InformantSigns) >  1 then if (InformantSigns[2] <> '#' ) and (ord(InformantSigns[2]) > 127) then infChareinzigerzug    := InformantSigns[2]  else infChareinzigerzug     := defaultinfChareinzigerzug    ;
  if length(InformantSigns) >  2 then if (InformantSigns[3] <> '#' ) and (ord(InformantSigns[3]) > 127) then infCharplusminus      := InformantSigns[3]  else infCharplusminus       := defaultinfCharplusminus      ;
  if length(InformantSigns) >  3 then if (InformantSigns[4] <> '#' ) and (ord(InformantSigns[4]) > 127) then infCharplusleichtminus:= InformantSigns[4]  else infCharplusleichtminus := defaultinfCharplusleichtminus;
  if length(InformantSigns) >  4 then if (InformantSigns[5] <> '#' ) and (ord(InformantSigns[5]) > 127) then infCharplusgleich     := InformantSigns[5]  else infCharplusgleich      := defaultinfCharplusgleich     ;
  if length(InformantSigns) >  5 then if (InformantSigns[6] <> '#' ) and (ord(InformantSigns[6]) > 127) then infChargleich         := InformantSigns[6]  else infChargleich          := defaultinfChargleich         ;
  if length(InformantSigns) >  6 then if (InformantSigns[7] <> '#' ) and (ord(InformantSigns[7]) > 127) then infCharunklar         := InformantSigns[7]  else infCharunklar          := defaultinfCharunklar         ;
  if length(InformantSigns) >  7 then if (InformantSigns[8] <> '#' ) and (ord(InformantSigns[8]) > 127) then infChargleichplus     := InformantSigns[8]  else infChargleichplus      := defaultinfChargleichplus     ;
  if length(InformantSigns) >  8 then if (InformantSigns[9] <> '#' ) and (ord(InformantSigns[9]) > 127) then infCharleichtminusplus:= InformantSigns[9]  else infCharleichtminusplus := defaultinfCharleichtminusplus;
  if length(InformantSigns) >  9 then if (InformantSigns[10] <> '#') and (ord(InformantSigns[10]) > 127) then infCharminusplus      := InformantSigns[10] else infCharminusplus       := defaultinfCharminusplus      ;
  if length(InformantSigns) > 10 then if (InformantSigns[11] <> '#') and (ord(InformantSigns[11]) > 127) then infCharkompensation   := InformantSigns[11] else infCharkompensation    := defaultinfCharkompensation   ;
  if length(InformantSigns) > 11 then if (InformantSigns[12] <> '#') and (ord(InformantSigns[12]) > 127) then infChargegenspiel     := InformantSigns[12] else infChargegenspiel      := defaultinfChargegenspiel     ;
  if length(InformantSigns) > 12 then if (InformantSigns[13] <> '#') and (ord(InformantSigns[13]) > 127) then infCharinitiative     := InformantSigns[13] else infCharinitiative      := defaultinfCharinitiative     ;
  if length(InformantSigns) > 13 then if (InformantSigns[14] <> '#') and (ord(InformantSigns[14]) > 127) then infCharangriff        := InformantSigns[14] else infCharangriff         := defaultinfCharangriff        ;
  if length(InformantSigns) > 14 then if (InformantSigns[15] <> '#') and (ord(InformantSigns[15]) > 127) then infCharzeitnot        := InformantSigns[15] else infCharzeitnot         := defaultinfCharzeitnot        ;
  if length(InformantSigns) > 15 then if (InformantSigns[16] <> '#') and (ord(InformantSigns[16]) > 127) then infCharvorsprung      := InformantSigns[16] else infCharvorsprung       := defaultinfCharvorsprung      ;
  if length(InformantSigns) > 16 then if (InformantSigns[17] <> '#') and (ord(InformantSigns[17]) > 127) then infCharneuerung       := InformantSigns[17] else infCharneuerung        := defaultinfCharneuerung       ;
  if length(InformantSigns) > 17 then if (InformantSigns[18] <> '#') and (ord(InformantSigns[18]) > 127) then infCharmitderidee     := InformantSigns[18] else infCharmitderidee      := defaultinfCharmitderidee     ;
  if length(InformantSigns) > 18 then if (InformantSigns[19] <> '#') and (ord(InformantSigns[19]) > 127) then infCharBesserIst      := InformantSigns[19] else infCharBesserIst       := defaultinfCharBesserIst      ;
  Klettere(wurzel, true);
end;

procedure TNotation.SetPiecesNames(sK, sQ, sR, sB, sN, sP : String);
var b : Boolean;
begin
  b := FUserDefinedPieces;
  FUserDefinedPieces := (sK <> '') or (sQ <> '') or(sR <> '') or(sB <> '') or(sN <> '') or (sP <> '');
  PiecesNames[0] := uppercase(sP);
  PiecesNames[1] := uppercase(sR);
  PiecesNames[2] := uppercase(sB);
  PiecesNames[3] := uppercase(sN);
  PiecesNames[4] := uppercase(sQ);
  PiecesNames[5] := uppercase(sK);
  Klettere(wurzel, true);
end;

procedure TNotation.GetPiecesNames(Var sK, sQ, sR, sB, sN, sP : String);
begin
  sP := PiecesNames[0];
  sR := PiecesNames[1];
  sB := PiecesNames[2];
  sN := PiecesNames[3];
  sQ := PiecesNames[4];
  sK := PiecesNames[5];
end;

procedure TNotation.DelPiecesNames;
begin
  PiecesNames[0] := '';
  PiecesNames[1] := 'R';
  PiecesNames[2] := 'B';
  PiecesNames[3] := 'N';
  PiecesNames[4] := 'Q';
  PiecesNames[5] := 'K';
  if FUserDefinedPieces then
    Klettere(wurzel, true);
  FUserDefinedPieces := false;
end;

procedure TNotation.checkFischerPos;
var i : Integer;
    F : PFischerInfo;
begin
  new(F);
  try
  with f^ do begin
      FischerLeftWhiteRook := 0;  {werden nur in Fischer-Random-Chess-Ausg.stellg. korrekt belegt!}
      FischerRightWhiteRook := 0;
      FischerLeftBlackRook := 0;
      FischerRightBlackRook := 0;
      Enabled := false;
  end;
  if maybeFischerpos(NoStellung^) then begin
    FPartiedaten.Zug.Rochademoeglichkeiten := [wkr, wlr, skr, slr];
    f^.enabled := true;
    for i := 0 to 63 do begin
      case NoStellung^[c(i)] of
        wK : f.FischerWhiteKing := c(i);
        wT : begin
               if f.FischerLeftWhiteRook = 0 then f.FischerLeftWhiteRook := c(i) { bei FEN lesen kommt immer zuerst der linke Turm}
               else f.FischerRightWhiteRook := c(i);
            end;
        sK :  f.FischerBlackKing := c(i);

        sT : begin
               if f.FischerLeftBlackRook = 0 then f.FischerLeftBlackRook := c(i)  {bei FEN lesen kommt immer zuerst der linke Turm}
               else f.FischerRightBlackRook := c(i);
            end;
      end;
    end;
  end;
  if assigned(OnFischerinfo) then OnFischerInfo(self, f^);  {an enabled kann Hauptprogramm erkennen, ob auf FRC umgestellt werden muss}
  finally
    dispose(F);
  end;  
end;

procedure TNotation.createWurzel;
begin
  if assigned(wurzel) then exit;
  New(Wurzel);
  Wurzel^.B:=Nil;
  Wurzel^.BL:=Nil;
  Wurzel^.rBL:=Nil;
  Wurzelindex:=Wurzel;
  Baumindex:=Nil;
end;

procedure TNotation.SetInfAttr(a : byte);
begin
  if not assigned(Baumindex) then exit;
  if (a = infmitderidee) or (a = infBesserIst) then
    Baumindex^.Zug.infattrBefore := a
  else
    Baumindex^.Zug.infattr := a;
  klettere(wurzel, true);
  geaendert:=true;
end;

function TNotation.getMoveIndex : Integer;
begin
  if Baumindex = Nil then
    Result := -1
  else
    Result := Baumindex.ss;
end;

function TNotation.setMoveAtIndex(idx : Integer) : Boolean;
begin
  Result := retract(idx);
  if Result then
    if assigned(FOnMoveChange) then FOnMoveChange(self, NoZug^, NoStellung^, false)
end;

function TNotation.setMoveAtBaumIndex(baumidx : PBaum) : PBaum;
var b : PBaum;
begin
  Result := Nil;
  if baumidx = Nil then exit;
  b := baumidx;
  while (b.rB <> Nil) and (b.rB.hidden) do b := b.rB;
  BaumIndex := b;
  BaumIndex^.Hidden := false;
  klettereEx;
  if assigned(FOnMoveChange) then FOnMoveChange(self, NoZug^, NoStellung^, false);
  Result := Baumindex;
end;

procedure TNotation.klettereEx;
begin
  nextToSolve := Nil;
  allSolved := true;
  klettere(wurzel, true);
end;

// for solving mode - ignores variation starting with an "own" move (not to solve)
function TNotation.setMoveAtBaumIndexEx(baumidx : PBaum) : PBaum;
var b, b1 : PBaum;
begin
  b := baumidx;
  b1 := Nil;
  Result := Nil;
  while true do begin
    b := setMoveAtBaumIndex(b);
    Result := b;
    if (b <> Nil) and (b.B = Nil) then begin  // only one move
      b := NextToSolve;
      b1 := b;
    end
    else begin
      if (b = Nil) or (b = b1) then exit;
      if b.Zug.amZug <> Partiedaten.Zug.amZug then begin
        while b <> Nil do begin
          b.hidden := false;
          b1 := b;
          b := b.B;
        end;
        klettereEx;
        b := NextToSolve;
      end
      else exit;
    end;  
  end;
end;


end.
